# Main.pm - The Object Class that provides a Main Object
#$Id: Base.pm,v 1.41 2005/08/16 22:05:21 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.
# Created by James Pattie & Jason A. Pattie, 11/07/2000.

package Portal::Accounting::Base;
use strict;
use HTMLObject::Base;
use HTMLObject::Normal;
use HTMLObject::FrameSet;
use Portal::AppState;
use Portal::Data::Variables;
use Portal::Accounting::Data::Variables;
use Date::Calc qw(Today);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::AppState Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.41 $';

# global variables.

=head1 NAME

Main - Object used to build a Main Menu for accounting package

=head1 SYNOPSIS

  use Portal::Accounting::Main;
  my $obj = Portal::Accounting::Main->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

Main is the Main Menu from which every other part of
the systems will be accessed.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(codeVersion, release, DBVersion)

 Creates a new instance of the Portal::Accounting::Main module.
 See Portal::AppState(3) for a listing of required arguments.

 Instantiates an instance of the Main State. !!Anything passed in here should be
 data that never changes from one run of a command to another such as database handles
 portalSession? input? cookieObj? dont fit that criterion

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( release => "", DBVersion => "", @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }
  $self->{warningString} = "";
  $self->{warning} = 0;
  $self->{debugString} = "";
  $self->{debug} = 0;

  # instantiate anything unique to this module
  my @state = split /::/, ref $self;
  for (my $i = 2; $i < @state; $i++)
  {
    if ($i != 2) { $self->{state} .= "::"; }
    $self->{state} .= $state[$i];
  }
  $self->{release} = $args{release};
  $self->{DBVersion} = $args{DBVersion};
  $self->{codeVersion} = $args{codeVersion};
  $self->{portalVariables} = Portal::Data::Variables->new(langObj => $self->{langObj});
  if ($self->{portalVariables}->error)
  {
    $self->error($self->{portalVariables}->errorMessage);
    return $self;
  }
  $self->{variables} = Portal::Accounting::Data::Variables->new(langObj => $self->{langObj});
  if ($self->{variables}->error)
  {
    $self->error($self->{variables}->errorMessage);
    return $self;
  }
  $self->{formObj} = HTMLObject::Form->new;
  if ($self->{formObj}->error)
  {
    $self->error($self->{formObj}->errorMessage);
    return $self;
  }
  $self->{jsDir}         = $self->{methods}->createBaseURL(type => "App",    linkType => "js",       appConfigObj => $self->{sessionObj}->{store}->{companyAppObj} );
  $self->{imageDir}      = $self->{methods}->createBaseURL(type => "App",    linkType => "image",    appConfigObj => $self->{sessionObj}->{store}->{companyAppObj} );
  $self->{cssDir}        = $self->{methods}->createBaseURL(type => "App",    linkType => "css",      appConfigObj => $self->{sessionObj}->{store}->{companyAppObj} );
  $self->{dbDump}        = $self->{methods}->createBaseURL(type => "App",    linkType => "db_dump",  appConfigObj => $self->{sessionObj}->{store}->{companyAppObj} );
  $self->{portalCGIDir}  = $self->{methods}->createBaseURL(type => "Portal", linkType => "cgibin",      appConfigObj => $self->{sessionObj}->{store}->{companyAppObj} );
  $self->{baseURL}       = $self->{methods}->createBaseURL(type => "App",    linkType => "cgi",      appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  $self->{output} = { app => $self->{app}, state => $self->{state}, command => $self->{input}->{command} };

  #the state is not set until runtime
  $self->{postInfo} .= "<input type=\"hidden\" name=\"app\" value=\"$self->{app}\">\n";
  $self->{postInfo} .= "<input type=\"hidden\" name=\"state\" value=\"$self->{state}\">\n";
  my $windowName = '';
  if (defined $self->{input}->{windowName}) { $windowName = $self->{input}->{windowName}; }
  $self->{getInfo} .= "windowName=$windowName&app=$self->{app}&state=$self->{state}";

  $self->{object} = 'Base'; #name of object to create/edit/delete/save
  $self->{OBJECT} = undef;  #instance of object just created/edited/saved/deleted
  # do validation
  if (!$self->Portal::Accounting::Base::isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  return $self;
}

sub prepDoc
{
  my $self = shift;
  my %args = (doc => undef, @_);
  my %foo;
  my $doc;
  if (defined $args{doc}) { $doc = $args{doc}; }
  else { $doc = $self->SUPER::prepDoc; }

  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->error)
  {
    $self->prefixError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->error("No valid colorScheme returned!<br>\n");
    return undef;
  }
  my %cssEntries = $self->{methods}->generateColorSchemeCSS(colorScheme => $colorScheme, appName => $self->{app});
  if ($self->{methods}->error)
  {
    $self->error($self->{methods}->errorMessage);
    return undef;
  }
  $doc->print(%cssEntries);

  my $title = $self->{langObj}->map($self->{commands}->{$self->{input}->{command}});
  $doc->setTitle( $title);
  $doc->setJavascriptInclude(file => $self->{jsDir}."Accounting.js");
  $foo{body} = "<center><h3>$title</h3></center>";
  $foo{javascript} = "langObj['rightClick'] = \"".$self->langObj->map("rightClick")."\";\n";
  $doc->print( %foo );
  return $doc;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if (length $self->{release} == 0)
  {
    $self->invalid("release", $self->{release});
  }
  if (length $self->{DBVersion} == 0)
  {
    $self->invalid("DBVersion", $self->{DBVersion});
  }
  if (length $self->{codeVersion} == 0)
  {
    $self->invalid("codeVersion", $self->{codeVersion});
  }
  if (not defined $self->{methods})
  { $self->missing("methods");  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item string commify( value)

 requires: value
 returns: commified value
 This routine commifies the specified value according to US standards of money.

=cut
sub commify
{
  my $self = shift;
  my $value = reverse shift;

  $value =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;

  return reverse $value;
}

sub mainButton
{
  my $self = shift;
  my $return;
  my $display = $self->{langObj}->map($self->{commands}->{display});
  $return = << "ENDOF";
  <form name="display">
    <input type="submit" name="submit" value="$display">
    <input type="hidden" name="app" value="$self->{app}">
    <input type="hidden" name="state" value="$self->{state}">
    <input type="hidden" name="command" value="display">
  </form>
ENDOF
  return $return;
}


sub c_new
{
  my $self = shift;
  return $self->c_edit(@_);
}

sub c_save
{
  my $self = shift;
  return $self->c_edit(@_);
}

=item HTMLObject c_edit(arg1 =>, arg2 =>, . . .)

  this routine is to create/edit a new object.  It will present the
  appropriate screen for the correct action.
  arguments such as ledgerCode can be sent and will be passed to objects

=cut
sub c_edit
{
  my $self = shift;
  my $doc = $self->prepDoc;
  my %args = ( command => $self->{input}->{command}, @_ );
  my $command = $args{command};
  my $l = $self->{langObj};
  my $foo;
  eval("use Portal::Accounting::Objects::$self->{object};");
  if ($@) { $self->error($@); return undef;  }
  eval("\$foo = Portal::Accounting::Objects::$self->{object}->new(langObj => \$self->{langObj}, dbHandle => \$self->{dbHandle}, \@_);");
  if ($@) { $self->error($@); return undef;  }
  if ($foo->error) { $doc->print($foo->errorMessage); return $doc;  }
  if ($command eq "save")
  { #validate user input then react
    if ($foo->populate(%{$self->{input}}) && $foo->validateForm(input => $self->{input}, formName => "main"))
    {
      my $r = $foo->write(); #insert/update into database
      $doc->print($foo->debugString);
      if (!$foo->error)
      {
        $doc->setFocus("body");
        $doc->print($l->map("Success"));
        $self->{OBJECT} = $foo;
        return $doc;
      }
    }
  }
  elsif ($command eq "edit")
  {
    my $result = $foo->populate(%{$self->{input}});  #!!populate calls isvalid
    if (! defined $self->{input}->{$foo->{id}}) { $doc->print("<p>To use this object, there must be an id sent in the data!</p>\n");  return $doc; }
#    my $result = $foo->set($foo->{id}, $self->{input}->{$foo->{id}});
    $foo->resetError;
    if (!$foo->error) { $foo->select(); }
    if ($foo->error)
    {
      $self->error("Error".$foo->errorMessage());
    }
  }
  #display form
  my %form = $foo->printForm();
  if ($foo->error)
  {
    $self->error($foo->errorMessage);
  }
  if (defined %form)
  {
    $doc->print(%form);
  }
  else
  {
    $doc->print('error! form is undefined<br>');
    $doc->print("\nformObj:".$foo->{formObj}->error);
  }

  $doc->print($foo->debugString);
  return $doc;
}

=item HTMLObject c_delete()

  presents screens to delete the working object.

=cut
sub c_delete
{
  my $self = shift;
  my %args = (doc => undef, @_);
  my $l = $self->{langObj};
  my $Failed = $l->map("Failed");
  my $Success = $l->map("Success");
  my $confirm = $l->map("confirmDelete");
  my $doc;
  if (defined $args{doc}) { $doc = $args{doc}; }
  else { $doc = $self->prepDoc();  }

  my $foo;
  eval("\$foo = Portal::Accounting::Objects::$self->{object}->new(langObj => \$self->{langObj}, dbHandle => \$self->{dbHandle}, \@_);");
  if (not defined $foo) { $doc->print("<p>Error trying to eval object</p>\n"); return $doc; }
  $foo->populate(%{$self->{input}});
  if ($foo->error)
  {
    $doc->print("Populate failed! $foo->{id}:$self->{input}->{id}<br>\n".$foo->errorMessage."<br>".$foo->debugString);
    return $doc;
  }
  $foo->select();
  if ($foo->error)
  {
    $doc->print("<p>SELECT for delete failed!<br>\n".$foo->errorMessage."<br>".$foo->debugString."FOO</p>");
    return $doc;
  }
  if (defined $self->{input}->{confirmDelete} && $self->{input}->{confirmDelete} eq "1")
  {#do the actual delete
    my $result = $foo->delete();
    if ($foo->error)
    {
      $doc->print($foo->errorMessage);
    }
    else
    {
      my $out = $l->map("Delete") . ": ";
      if ($result == -1)
      {
        $out .= "$Failed<br>".$foo->result;
      }
      elsif ($result >= 0)  {  $out .= "$Success<br>";  }
      else { $out .= "<p>Error in result: '$result'</p>\n";  }
      $doc->print($out."<hr>". $foo->debugString);
    }
    return $doc;
  }
  else
  { #output confirm d
    my $led = "";
    if (defined $args{ledgerCode}) { $led = " <input type=\"hidden\" name=\"ledgerCode\" value=\"$args{ledgerCode}\">\n";  }
    my $yes = $self->{langObj}->map("Yes");
    $doc->print(<<"ENDOF");
<form name="mainForm">
  <p>$confirm<br>
  <input type="submit" name="yes" value="$yes"></p>
  <input type="hidden" name="app" value="$self->{app}">
  <input type="hidden" name="state" value="$self->{state}">
  <input type="hidden" name="command" value="delete">
  <input type="hidden" name="confirmDelete" value="1">
  <input type="hidden" name="id" value="$self->{input}->{id}">
  $led
</form>
ENDOF
  }
  return $doc;
}

sub getDateArray
{
  my $self = shift;
  return Today();
}

sub c_display
{
  my $self = shift;
  my $doc = $self->prepDoc(@_);
  my $confirm = $self->{langObj}->map("Confirm Delete");
  my $foo;
  eval("use Portal::Accounting::Objects::$self->{object};");
  if ($@) { $self->error($@); return undef;  }
  eval("\$foo = Portal::Accounting::Objects::$self->{object}->new(langObj => \$self->{langObj}, dbHandle => \$self->{dbHandle}, \@_);");
  if ($@) { $self->error($@); return undef;  }
  if ($foo->error) { $doc->print($foo->errorMessage); return $doc;  }
  $doc->print($foo->printList);
  $doc->print($foo->printNew);
  $doc->print($foo->debugString);
  $doc->setFocus("javascript");
  $doc->print(<<"ENDOF");
function deleteConfirm(myForm)
{
  if (confirm("$confirm"))
  {
    myForm.confirmDelete.value = 1;
    return true;
  }
  return false;
}
ENDOF
  return $doc;
}

=item boolean warning("string") or boolean warning()

  similar to the error routine this routine will set a warning
  statement and turn on warning if called.  access throught
  $obj->warningString

=cut
sub warning
{
  my $self = shift;
  my @callerArgs = caller(1);
  (my $subName = $callerArgs[3]) =~ s/^(.+)(::)([^:]+)$/$1->$3/;
  my $callerErrStr = "";

  if (scalar @_ > 0)
  {
    # we are setting an warning condition.
    if (scalar @_ == 1)
    {
      $self->{warningString} .= $callerErrStr . @_[0]."<br>";
    }
    else
    {
      # handle the sprintf case.
      if (@_[0] =~ /(?<!\\)%[^%]/)
      {
        # build up the string to eval for the sprintf.
        my $str = "\"@_[0]\"";
        for (my $i=1; $i < scalar @_; $i++)
        {
          $str .= ", \"@_[$i]\"";
        }
        $self->{warningString} .= $callerErrStr;
        eval "\$self->{warningString} .= sprintf($str);";
        if ($@)
        {
          $self->warning($callerErrStr . $@);
          return;
        }
      }
      else
      {
        $self->{warningString} .= $callerErrStr . join(" ", @_);
      }
    }
    $self->{warningString} .= "\n";
    $self->{warning} = 1;
  }

  return $self->{warning};
}

=item boolean debug("string") or boolean debug()

  similar to the error routine this routine will set a debug
  statement and turn on debug if called.  access throught
  $obj->debugString

=cut
sub debug
{
  my $self = shift;
  my @callerArgs = caller(1);
  (my $subName = $callerArgs[3]) =~ s/^(.+)(::)([^:]+)$/$1->$3/;
  my $callerErrStr = "$subName";

  if (scalar @_ > 0)
  {
    # we are setting an error condition.
    if (scalar @_ == 1)
    {
      $self->{debugString} .= $callerErrStr . @_[0]."<br>";
    }
    else
    {
      if (@_[0] eq "debugString")
      {
        my %args = ( @_ );
        if (!exists $args{debugString})  # make sure we get the debugString argument!
        {
          $self->debug($callerErrStr . "<b>debugString</b> is missing!<br>\n");
          return;
        }
        else
        {
          $self->{debugString} .= $callerErrStr . $args{debugString};
        }
      }
      else
      {
        # handle the sprintf case.
        if (@_[0] =~ /(?<!\\)%[^%]/)
        {
          # build up the string to eval for the sprintf.
          my $str = "\"@_[0]\"";
          for (my $i=1; $i < scalar @_; $i++)
          {
            $str .= ", \"@_[$i]\"";
          }
          $self->{debugString} .= $callerErrStr;
          eval "\$self->{debugString} .= sprintf($str);";
          if ($@)
          {
            $self->debug($callerErrStr . $@);
            return;
          }
        }
        else
        {
          $self->{debugString} .= $callerErrStr . join(" ", @_);
        }
      }
    }
    $self->{debugString} .= "\n";
    $self->{debug} = 1;
  }

  return $self->{debug};
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
