# Main.pm - The Object Class that provides a Main Object
#$Id: BaseL.pm,v 1.7 2005/08/11 19:15:01 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.
# Created by James Pattie & Jason A. Pattie, 11/07/2000.

package Portal::Accounting::BaseL;
use strict;
use Portal::Accounting::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);
require Exporter;

@ISA = qw(Portal::Accounting::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.7 $';

# global variables.

=head1 NAME

BaseL - Object used as parent to derive Ledger specific states.

=head1 SYNOPSIS

  use Portal::Accounting::Main;
  my $obj = Portal::Accounting::Main->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

Main is the Main Menu from which every other part of
the systems will be accessed.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(codeVersion, release, DBVersion)

 Creates a new instance of the Portal::Accounting::Main module.
 See Portal::AppState(3) for a listing of required arguments.

 Instantiates an instance of the Main State. !!Anything passed in here should be
 data that never changes from one run of a command to another such as database handles
 portalSession? input? cookieObj? dont fit that criterion

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # do validation
  if (!$self->Portal::Accounting::BaseL::isValid)
  {
    return $self;
  }
  $self->{postInfo} .= "<input type=\"hidden\" name=\"ledgerCode\" value=\"$self->{input}->{ledgerCode}\">\n";
  $self->{getInfo} .= "&ledgerCode=".$self->{input}->{ledgerCode};

  return $self;
}

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }
  # do validation code here.
  if (not defined $self->{input}->{ledgerCode})
  { $self->missing("ledgerCode");  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }
  return 1;
}

sub prepDoc
{
  my $self = shift;
  my %args = (doc => undef, @_);
  my %foo;
  my $doc;
  if (defined $args{doc}) { $doc = $args{doc}; }
  else { $doc = HTMLObject::Normal->new; }
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $title = $self->{langObj}->map($self->{commands}->{$self->{input}->{command}});
  $doc->setTitle($self->{sessionObj}->{store}->{ledgers}->{$ledgerCode}->{name}." $title");
  $doc->setJavascriptInclude(file => $self->{jsDir}."Accounting.js");

  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->error)
  {
    $self->prefixError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->error("No valid colorScheme returned!<br>\n");
    return undef;
  }
  my %cssEntries = $self->{methods}->generateColorSchemeCSS(colorScheme => $colorScheme, appName => $self->{app});
  if ($self->{methods}->error)
  {
    $self->error($self->{methods}->errorMessage);
    return undef;
  }
  $doc->print(%cssEntries);

  my $color = $self->{sessionObj}->{store}->{ledgers}->{$ledgerCode}->{color};
  $foo{body} = "<center><h3><font color=\"$color\">". $self->{sessionObj}->{store}->{ledgers}->{$ledgerCode}->{name}."</font> $title</h3></center>";
  $foo{javascript} = "langObj['rightClick'] = \"".$self->langObj->map("rightClick")."\";\n";
  $doc->print( %foo );
  return $doc;
}

sub mainButton
{
  my $self = shift;
  my $return;
  my $display = $self->{langObj}->map($self->{commands}->{display});
  $return = << "ENDOF";
  <form name="display">
    <input type="submit" name="submit" value="$display">
    <input type="hidden" name="app" value="$self->{app}">
    <input type="hidden" name="state" value="$self->{state}">
    <input type="hidden" name="ledgerCode" value="$self->{input}->{ledgerCode}">
    <input type="hidden" name="command" value="display">
  </form>
ENDOF
  return $return;
}


sub c_new
{
  my $self = shift;
  return $self->SUPER::c_new(ledgerCode => $self->{input}->{ledgerCode}, @_);
}

sub c_save
{
  my $self = shift;
  return $self->SUPER::c_save(ledgerCode => $self->{input}->{ledgerCode}, @_);
}

sub c_edit
{
  my $self = shift;
  return $self->SUPER::c_edit(ledgerCode => $self->{input}->{ledgerCode}, @_);
}

=item HTMLObject c_delete()

  presents screens to delete the working object.

=cut
sub c_delete
{
  my $self = shift;
  return $self->SUPER::c_delete(ledgerCode => $self->{input}->{ledgerCode}, @_);
}

sub c_display
{
  my $self = shift;
  my $doc = $self->prepDoc(@_);
  my $confirm = $self->{langObj}->map("Confirm Delete");
  my $foo;
  my $led = $self->{input}->{ledgerCode};
  eval("use Portal::Accounting::Objects::$self->{object};");
  if ($@) { $self->error($@); return undef;  }
  eval('$foo = Portal::Accounting::Objects::'.$self->{object}.'->new(langObj => $self->{langObj}, dbHandle => $self->{dbHandle}, ledgerCode => $led, @_);');
  if ($@) { $self->error($@); return undef;  }
  if ($foo->error) { $doc->print($foo->errorMessage); return $doc;  }
  $doc->print($foo->printList);
  $doc->print($foo->printNew);
  $doc->print($self->mainButton);
  if ($foo->error)
  {
    $doc->print($foo->errorMessage);
  }
  $doc->print($foo->debugString);
  $doc->setFocus("javascript");
  $doc->print(<<"ENDOF");
function deleteConfirm(myForm)
{
  if (confirm("$confirm"))
  {
    myForm.confirmDelete.value = 1;
    return true;
  }
  return false;
}
ENDOF
  return $doc;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
