# Monitor.pm - The Object Class that provides a Monitor Object
# Copyright (c) 2000-2002, Xperience, Inc. http://www.pcxperience.com/
# All rights reserved.  This program is free software; you can redistribute it
# and/or modify it under the same terms of the GPL

package Portal::Accounting::Config::Monitor;
use strict;
use Portal::Accounting::Base;
use Portal::Data::Variables;
use Portal::Data::Config;
use Portal::Accounting::DB::Main;
use Portal::Accounting::Objects::MonitorObject;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.02';

# global variables.
# modify to represent the commands this state provides.
my %commands = ( display => "Monitor Menu", edit => "Edit", close => "Close Window",
                 updateMailServer => "Update MailServer setting", );

=head1 NAME

Monitor - Object used to build a Monitor Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Monitor;
  my $obj = Portal::Accounting::Monitor->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

Monitor is a Monitor class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::Config::Monitor module.
 See Portal::Accounting::Base(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{accountTypesToMonitor} = [ 0, 1, 111, 2, 3 ];

  $self->{dbObj} = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($self->{dbObj}->didErrorOccur)
  {
    $self->setError($self->{dbObj}->errorMessage);
    return $self;
  }

  # do validation
  if (!$self->isValid())
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item HTMLObject run(command)

 Runs the specified command and returns the resulting HTMLObject or
 undef if an error occured.

=cut

sub run
{
  my $self = shift;
  my %args = ( command => "", @_ );
  my $command = $args{command};
  my $doc;
  my $phrases = $self->{langObj}->{phraseObj}->{phrases};
  my $errorPhrase = $phrases->{error};

  if (!exists $commands{$command})
  {
    $self->setError(sprintf($phrases->{commandNotKnown}, $command));
    return undef;
  }
  my $ledgerCode = $self->{input}->{ledgerCode};

  if ((!exists $self->{input}->{ledgerCode}) || (length $ledgerCode == 0))
  {
    $self->setError("ledgerCode must be defined!<br>\n");
    return undef;
  }
  my $ledger = $self->{input}->{ledger};
  if ((!exists $self->{input}->{ledger}) || (length $ledger == 0))
  {
    $self->setError("ledger must be defined!<br>\n");
    return undef;
  }

  my $windowName = $self->{input}->{windowName};
  if (!exists $self->{input}->{windowName})
  {
    $self->setError("windowName must be defined!<br>\n");
    return undef;
  }
  else
  {
    # see if the window is registered and register if not.
    if (!$self->{sessionObj}->windowExists($windowName))
    {
      $self->{sessionObj}->registerWindow($windowName);
      if ($self->{sessionObj}->error)
      {
        $self->error($self->{sessionObj}->errorMessage);
	return undef;
      }
    }
  }

  $doc = HTMLObject::Normal->new;

  if ($command eq "display")
  {
    # display the Create MonitorAccount main screen.
    $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger);
    if ($self->didErrorOccur)
    {
      $self->setError($self->errorMessage);
      return undef;
    }
  }
  elsif ($command eq "edit")
  {
    my $mode = $self->{input}->{mode};
    my $action = $self->{input}->{action};
    my $id = $self->{input}->{id};

    if ($mode eq "delete")
    {
      my $monitorAccountObj = $self->{dbObj}->getMonitorInfo(id => $id);
      if ($self->{dbObj}->didErrorOccur)
      {
        $self->setError($self->{dbObj}->errorMessage);
        return undef;
      }
      if (!defined $monitorAccountObj)
      {
        $self->setError("id = '$id' does not point to a MonitorAccount entry!<br>\n");
        return undef;
      }

      # call the deletion code.
      $self->{dbObj}->updateMonitorEntry(monitorObj => $monitorAccountObj, action => "delete");
      if ($self->{dbObj}->didErrorOccur)
      {
        # display the Create MonitorAccount entry screen.
        $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger, errorMessage => $self->{dbObj}->errorMessage);
        if ($self->didErrorOccur)
        {
          $self->setError($self->errorMessage);
          return undef;
        }
      }

      # display the Create MonitorAccount main screen.
      $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger, statusMessage => "Deletion of Account Monitor Entry with id = '$id' successfull...");
      if ($self->didErrorOccur)
      {
        $self->setError($self->errorMessage);
        return undef;
      }
    }
    elsif ($mode =~ /^(edit|create)$/)
    {
      if ($action eq $mode)
      {
        # validate and try to create/edit.
        my $monitorAccountObj = $self->validateMonitorAccountInfo(command => $mode, ledgerCode => $ledgerCode, ledger => $ledger, id => $id);
        if ($self->didErrorOccur)
        {
          # display the Create MonitorAccount entry screen.
          $doc = $self->updateMonitorAccount(doc => $doc, command => $mode, ledgerCode => $ledgerCode, ledger => $ledger, id => $id, errorMessage => $self->errorMessage);
          if ($self->didErrorOccur)
          {
            $self->setError($self->errorMessage);
            return undef;
          }
        }
        if ($monitorAccountObj->didErrorOccur)
        {
          # display the Create MonitorAccount entry screen.
          $doc = $self->updateMonitorAccount(doc => $doc, command => $mode, ledgerCode => $ledgerCode, ledger => $ledger, id => $id, errorMessage => $monitorAccountObj->errorMessage);
          if ($self->didErrorOccur)
          {
            $self->setError($self->errorMessage);
            return undef;
          }
        }
        else
        {
          # call the creation code.
          $self->{dbObj}->updateMonitorEntry(monitorObj => $monitorAccountObj);
          if ($self->{dbObj}->didErrorOccur)
          {
            # display the Create MonitorAccount entry screen.
            $doc = $self->updateMonitorAccount(doc => $doc, command => $mode, ledgerCode => $ledgerCode, ledger => $ledger, id => $id, errorMessage => $self->{dbObj}->errorMessage);
            if ($self->didErrorOccur)
            {
              $self->setError($self->errorMessage);
              return undef;
            }
          }

          # display the Create MonitorAccount main screen.
          $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger, statusMessage => "Account Monitor Entry $mode successfull...");
          if ($self->didErrorOccur)
          {
            $self->setError($self->errorMessage);
            return undef;
          }
        }
      }
      else
      {
        # display the Create MonitorAccount entry screen.
        $doc = $self->updateMonitorAccount(doc => $doc, command => $mode, ledgerCode => $ledgerCode, ledger => $ledger, id => $id);
        if ($self->didErrorOccur)
        {
          $self->setError($self->errorMessage);
          return undef;
        }
      }
    }
    else
    {
      # display the Create MonitorAccount main screen.
      $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger, errorMessage => "mode = '$mode' is invalid!");
      if ($self->didErrorOccur)
      {
        $self->setError($self->errorMessage);
        return undef;
      }
    }
  }
  elsif ($command eq "close")
  {
    $doc = $self->close;
  }
  elsif ($command eq "updateMailServer")
  {
    my $mailServer = $self->{input}->{mailServer};
    if ($mailServer !~ /^(.+)$/)
    {
      # display the Create MonitorAccount main screen.
      $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger, errorMessage => "mailServer = '$mailServer' is invalid!");
      if ($self->didErrorOccur)
      {
        $self->setError($self->errorMessage);
        return undef;
      }
    }
    else
    {
      # update.
      my $result = $self->{dbObj}->setConfigEntry(name => "monitorAccountsMailServer", value => $mailServer);
      if ($self->{dbObj}->didErrorOccur)
      {
        $self->setError($self->{dbObj}->errorMessage);
        return undef;
      }
      # display the Create MonitorAccount main screen.
      $doc = $self->displayMonitorAccounts(doc => $doc, ledgerCode => $ledgerCode, ledger => $ledger);
      if ($self->didErrorOccur)
      {
        $self->setError($self->errorMessage);
        return undef;
      }
    }
  }

  return $doc;
}

# HTMLObject displayMonitorAccounts(doc, ledgerCode, ledger, errorMessage, statusMessage)
sub displayMonitorAccounts
{
  my $self = shift;
  my %args = ( doc => undef, ledgerCode => "", ledger => "", errorMessage => "", statusMessage => "", @_ );
  my $doc = $args{doc};
  my $ledgerCode = $args{ledgerCode};
  my $ledger = $args{ledger};
  my $errorMessage = $args{errorMessage};
  my $statusMessage = $args{statusMessage};
  my $phrases = $self->{langObj}->{phraseObj}->{phrases};
  my $errorPhrase = $phrases->{error};
  my $windowName = $self->{input}->{windowName};

  my $refresh = $phrases->{Refresh};
  my $close = $phrases->{Close_Window};

  my $ledgerCodeURI = $doc->encodeString(string => $ledgerCode);
  my $ledgerURI = $doc->encodeString(string => $ledger);

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->didErrorOccur)
  {
    $self->setError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->setError("No valid colorScheme returned!<br>\n");
    return undef;
  }

  # create the needed css entries
  my %cssEntries = $self->{methods}->generateColorSchemeCSS(colorScheme => $colorScheme, appName => "Accounting");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError($self->{methods}->errorMessage);
    return undef;
  }
  $doc->print(%cssEntries);

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError($self->{methods}->errorMessage);
    return $doc;
  }
  my $refreshUrl = $url . "?app=Accounting&state=Config::Monitor&command=display&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName";
  my $closeUrl = $url . "?app=Accounting&state=Config::Monitor&command=close&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName";
  my $createUrl = $url . "?app=Accounting&state=Config::Monitor&command=edit&mode=create&id=-1&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName";
  my $editUrl = $url . "?app=Accounting&state=Config::Monitor&command=edit&mode=edit&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName&id=";
  my $deleteUrl = $url . "?app=Accounting&state=Config::Monitor&command=edit&mode=delete&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName&id=";

  my $noEntriesFoundPhrase = $phrases->{noEntriesFound};
  my $createPhrase = $phrases->{Create} . " " . $phrases->{Monitor} . " " . $phrases->{Account} . " " . $phrases->{Entry};
  my $editPhrase = $phrases->{Edit};
  my $deletePhrase = $phrases->{Delete};

  my $idPhrase = $phrases->{id};
  my $ledgerPhrase = $phrases->{Ledger};
  my $codePhrase = $phrases->{Code};
  my $ledgerCodePhrase = $ledgerPhrase . " " . $codePhrase;
  my $accountsPhrase = $phrases->{Accounts};
  my $emailsPhrase = $phrases->{Emails};
  my $usersPhrase = $phrases->{Users};
  my $operatorPhrase = $phrases->{Operator};
  my $mailTypePhrase = $phrases->{MailType};
  my $forPhrase = $phrases->{for};
  my $monitorPhrase = $phrases->{Monitor};
  my $mailServerPhrase = $self->{langObj}->map("MailServer");
  my $updatePhrase = $self->{langObj}->map("Update");

  my $errorOutput = ($errorMessage ? "<br><b>Error</b>: $errorMessage\n<br>\n" : "");
  my $statusOutput = ($statusMessage ? "<br><b>Status</b>: $statusMessage\n<br>\n" : "");

  $doc->setTitle("$monitorPhrase $accountsPhrase $forPhrase $ledgerPhrase - $ledger");

  $doc->setFocus("body");
  $doc->print(<<"END_OF_CODE");
<center>
<table border="0" width="100%" cellspacing="0" cellpadding="2" class="menu">
  <tr>
    <td align="left" width="100%"><a class="menu" href="$closeUrl" onMouseOver="window.status='$close'; return true;" title="$close">$close</a> | <a class="menu" href="$refreshUrl" onMouseOver="window.status='$refresh'; return true;" title="$refresh">$refresh</a></td>
  </tr>
</table>
<br>
<h2>$monitorPhrase $accountsPhrase $forPhrase $ledgerPhrase - $ledger</h2>
<br>
$errorMessage$statusMessage
<br>
END_OF_CODE

  my $mailServer = $self->{dbObj}->getConfigEntry(name => "monitorAccountsMailServer");
  if ($self->{dbObj}->didErrorOccur)
  {
    $self->setError($self->{dbObj}->errorMessage);
    return undef;
  }
  if (!defined $mailServer)
  {
    $mailServer = "127.0.0.1";  # default to localhost.
  }
  $mailServer = $self->{input}->{mailServer} if (exists $self->{input}->{mailServer});
  $doc->print(<<"END_OF_CODE");
<form name="mailServerF" method="post" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="Config::Monitor">
  <input type="hidden" name="command" value="updateMailServer">
  <input type="hidden" name="ledgerCode" value="$ledgerCode">
  <input type="hidden" name="ledger" value="$ledger">
  <input type="hidden" name="windowName" value="$windowName">
  <table border="0" width="75%" cellspacing="0" cellpadding="2">
    <tr>
      <td align="right">$mailServerPhrase:</td>
      <td align="left"><input type="text" name="mailServer" value="$mailServer"></td>
    </tr>
    <tr>
      <td align="center" colspan="2"><input type="submit" name="update" value="$updatePhrase"></td>
    </tr>
  </table>
</form>
<br>
END_OF_CODE

  my @entries = $self->{dbObj}->getMonitorEntries(ledgerCode => $ledgerCode);
  if ($self->{dbObj}->didErrorOccur)
  {
    $self->setError($self->{dbObj}->errorMessage);
    return undef;
  }
  if (scalar @entries == 0)
  {
    $doc->print("$noEntriesFoundPhrase<br>\n");
  }
  else
  {
    $doc->print(<<"END_OF_CODE");
<table border="1" cellpadding="1" cellspacing="1" width="100%">
  <tr>
    <th>$editPhrase</th><th>$idPhrase</th><th>$ledgerCodePhrase</th><th>$accountsPhrase</th><th>$emailsPhrase</th><th>$usersPhrase</th><th>&lt;</th><th>$operatorPhrase</th><th>&gt;</th><th>$mailTypePhrase</th><th>$deletePhrase</th>
  </tr>
END_OF_CODE
    foreach my $entry (@entries)
    {
      my $tmpUrl = $editUrl . $entry->{id};
      my $tmp2Url = $deleteUrl . $entry->{id};
      my $emails = ($entry->{emails} ? $entry->{emails} : "&nbsp;");
      my $users = ($entry->{users} ? $entry->{users} : "&nbsp;");
      my $lessThan = ($entry->{lessThan} ? $entry->{lessThan} : "&nbsp;");
      my $greaterThan = ($entry->{greaterThan} ? $entry->{greaterThan} : "&nbsp;");
      $doc->print(<<"END_OF_CODE");
  <tr>
    <td align="center"><a href="$tmpUrl" onMouseOver="window.status='$editPhrase'; return true;" title="$editPhrase">e</a></td>
    <td align="center">$entry->{id}</td>
    <td align="center">$entry->{ledgerCode}</td>
    <td>$entry->{accounts}</td>
    <td>$emails</td>
    <td>$users</td>
    <td align="right">$lessThan</td>
    <td align="center">$entry->{operator}</td>
    <td align="right">$greaterThan</td>
    <td align="center">$entry->{mailType}</td>
    <td align="center"><a href="$tmp2Url" onMouseOver="window.status='$deletePhrase'; return true;" title="$deletePhrase">d</a></td>
  </tr>
END_OF_CODE
    }
    $doc->print(<<"END_OF_CODE");
</table>
END_OF_CODE
  }
  $doc->print(<<"END_OF_CODE");
<br>
<a href="$createUrl" onMouseOver="window.status='$createPhrase'; return true;" title="$createPhrase">$createPhrase</a>
END_OF_CODE

  return $doc;
}

# HTMLObject updateMonitorAccount(doc, command, ledgerCode, ledger, id, errorMessage)
sub updateMonitorAccount
{
  my $self = shift;
  my %args = ( doc => undef, command => "edit", ledgerCode => "", ledger => "", id => "-1", errorMessage => "", @_ );
  my $doc = $args{doc};
  my $command = $args{command};
  my $ledgerCode = $args{ledgerCode};
  my $ledger = $args{ledger};
  my $id = $args{id};
  my $errorMessage = $args{errorMessage};
  my $phrases = $self->{langObj}->{phraseObj}->{phrases};
  my $errorPhrase = $phrases->{Error};
  my $windowName = $self->{input}->{windowName};
  my $refresh = $phrases->{Refresh};

  if ($command !~ /^(edit|create)$/)
  {
    $self->setError("command = '$command' is invalid!<br>\n");
    return $doc;
  }
  if ($command eq "create" && $id ne "-1")
  {
    $self->setError("id = '$id' is invalid when command = '$command'!<br>\n");
    return $doc;
  }

  my $ledgerCodeURI = $doc->encodeString(string => $ledgerCode);
  my $ledgerURI = $doc->encodeString(string => $ledger);

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->didErrorOccur)
  {
    $self->setError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->setError("No valid colorScheme returned!<br>\n");
    return undef;
  }

  # create the needed css entries
  my %cssEntries = $self->{methods}->generateColorSchemeCSS(colorScheme => $colorScheme, appName => "Accounting");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError($self->{methods}->errorMessage);
    return undef;
  }
  $doc->print(%cssEntries);

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError($self->{methods}->errorMessage);
    return $doc;
  }
  my $refreshUrl = $url . "?app=Accounting&state=Config::Monitor&command=edit&mode=$command&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName&id=$id";
  my $mainMenuUrl = $url . "?app=Accounting&state=Config::Monitor&command=display&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName";
  my $updateUrl = $url . "?app=Accounting&state=Config::Monitor&command=edit&mode=$command&id=-1&ledgerCode=$ledgerCodeURI&ledger=$ledgerURI&windowName=$windowName";

  my $createPhrase = $phrases->{Create};
  my $editPhrase = $phrases->{Edit};
  my $updatePhrase = ($command eq "edit" ? $editPhrase : $createPhrase);

  my $idPhrase = $phrases->{Id};
  my $ledgerPhrase = $phrases->{Ledger};
  my $codePhrase = $phrases->{Code};
  my $ledgerCodePhrase = $ledgerPhrase . " " . $codePhrase;
  my $accountsPhrase = $phrases->{Accounts};
  my $emailsPhrase = $phrases->{Emails};
  my $usersPhrase = $phrases->{Users};
  my $operatorPhrase = $phrases->{Operator};
  my $mailTypePhrase = $phrases->{MailType};
  my $forPhrase = $phrases->{for};
  my $monitorPhrase = $phrases->{Monitor};

  my $accountsInfo = "";
  my $lessThan = "";
  my $operatorInfo = "";
  my $greaterThan = "";
  my $usersInfo = "";
  my $emails = "";
  my $mailTypeInfo = "";
  my $monitorInfoObj = undef;

  # lookup the entry if id != -1
  if ($id != -1)
  {
    $monitorInfoObj = $self->{dbObj}->getMonitorInfo(id => $id);
    if ($self->{dbObj}->didErrorOccur)
    {
      $self->setError($self->{dbObj}->errorMessage);
      return $doc;
    }
    if (!defined $monitorInfoObj)
    {
      $self->setError("id = '$id' does not point to a MonitorAccount entry!<br>\n");
      return $doc;
    }
  }

  if (defined $monitorInfoObj)
  {
    $lessThan = $monitorInfoObj->{lessThan};
    $greaterThan = $monitorInfoObj->{greaterThan};
    ($emails = $monitorInfoObj->{emails}) =~ s/,/\n/g;
  }
  $lessThan = $self->{input}->{lessThan} if (exists $self->{input}->{lessThan});
  $greaterThan = $self->{input}->{greaterThan} if (exists $self->{input}->{greaterThan});
  $emails = $self->{input}->{emails} if (exists $self->{input}->{emails});

  # gather available Accounts and determine if the user has already selected any.
  my @availableAccountObjs = ();
  foreach my $acctType (@{$self->{accountTypesToMonitor}})
  {
    my @tmpAvailableAccountObjs = $self->{dbObj}->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => $acctType);
    if ($self->{dbObj}->didErrorOccur)
    {
      $self->setError($self->{dbObj}->errorMessage);
      return $doc;
    }
    foreach my $obj (@tmpAvailableAccountObjs)
    {
      push @availableAccountObjs, $obj;
    }
  }
  if (scalar @availableAccountObjs == 0)
  {
    $self->setError("No Accounts assigned to Ledger '$ledger' - '$ledgerCode'.<br>\nFix this error message and make it be a link!<br>\n");
    return $doc;
  }
  my @selectedAccounts = split /\0/, $self->{input}->{accounts};
  my @originalAccounts = ();
  if (defined $monitorInfoObj)
  {
    @originalAccounts = split /,/, $monitorInfoObj->{accounts};
  }

  foreach my $accountObj (@availableAccountObjs)
  {
    my $selected = "";
    if (exists $self->{input}->{accounts})
    {
      $selected = (grep(/$accountObj->{number}/, @selectedAccounts) > 0 ? " selected" : "");
    }
    elsif (defined $monitorInfoObj)
    {
      $selected = (grep(/$accountObj->{number}/, @originalAccounts) > 0 ? " selected" : "");
    }
    $accountsInfo .= "          <option value=\"$accountObj->{number}\"$selected>$accountObj->{name} ($accountObj->{number})</option>\n";
  }

  # gather available Portal users for this company.
  my @availableUserObjs = $self->{authObj}->getListOfUsersForCompany(companyObj => $self->{companyObj});
  if ($self->{authObj}->didErrorOccur)
  {
    $self->setError($self->{authObj}->errorMessage);
    return $doc;
  }
  if (scalar @availableUserObjs == 0)
  {
    $self->setError("No Users created for Company '$self->{companyObj}->{name}' - '$self->{companyObj}->{code}'.<br>\nFix this error message and make it be a link!<br>\n");
    return $doc;
  }
  my @selectedUsers = split /\0/, $self->{input}->{users};
  my @originalUsers = ();
  if (defined $monitorInfoObj)
  {
    @originalUsers = split /,/, $monitorInfoObj->{users};
  }

  foreach my $userObj (@availableUserObjs)
  {
    my $selected = "";
    if (exists $self->{input}->{accounts})
    {
      $selected = (grep(/$userObj->{uname}/, @selectedUsers) > 0 ? " selected" : "");
    }
    elsif (defined $monitorInfoObj)
    {
      $selected = (grep(/$userObj->{uname}/, @originalUsers) > 0 ? " selected" : "");
    }
    $usersInfo .= "          <option value=\"$userObj->{uname}\"$selected>$userObj->{fname} $userObj->{lname} ($userObj->{uname})</option>\n";
  }

  # process the operator (and/or)
  foreach my $operator ("and", "or")
  {
    my $selected = "";
    if (exists $self->{input}->{operator})
    {
      $selected = ($self->{input}->{operator} eq $operator ? " selected" : "");
    }
    elsif (defined $monitorInfoObj)
    {
      $selected = ($monitorInfoObj->{operator} eq $operator ? " selected" : "");
    }
    $operatorInfo .= "          <option value=\"$operator\"$selected>$operator</option>\n";
  }

  # process the mailType value (html/text)
  foreach my $mailType ("html", "text")
  {
    my $selected = "";
    if (exists $self->{input}->{mailType})
    {
      $selected = ($self->{input}->{mailType} eq $mailType ? " selected" : "");
    }
    elsif (defined $monitorInfoObj)
    {
      $selected = ($monitorInfoObj->{mailType} eq $mailType ? " selected" : "");
    }
    $mailTypeInfo .= "          <option value=\"$mailType\"$selected>$mailType</option>\n";
  }

  my $errorOutput = ($errorMessage ? "<br><b>Error</b>: $errorMessage\n<br>\n" : "");

  $doc->setTitle("$monitorPhrase $accountsPhrase $forPhrase $ledgerPhrase - $ledger : $command");
  $doc->setFocus("body");
  $doc->print(<<"END_OF_CODE");
<center>
<table border="0" width="100%" cellspacing="0" cellpadding="2" class="menu">
  <tr>
    <td align="left" width="100%"><a class="menu" href="$mainMenuUrl" onMouseOver="window.status='$monitorPhrase $accountsPhrase'; return true;" title="$monitorPhrase $accountsPhrase">$monitorPhrase $accountsPhrase</a> | <a class="menu" href="$refreshUrl" onMouseOver="window.status='$refresh'; return true;" title="$refresh">$refresh</a></td>
  </tr>
</table>
<br>
<h2>$monitorPhrase $accountsPhrase $forPhrase $ledgerPhrase - $ledger : $command</h2>
<br>
$errorOutput
<form name="monitorAccount" method="post" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="Config::Monitor">
  <input type="hidden" name="command" value="edit">
  <input type="hidden" name="mode" value="$command">
  <input type="hidden" name="ledgerCode" value="$ledgerCode">
  <input type="hidden" name="ledger" value="$ledger">
  <input type="hidden" name="id" value="$id">
  <input type="hidden" name="action" value="$command">
  <input type="hidden" name="windowName" value="$windowName">
  <table border="0" cellpadding="0" cellspacing="5">
    <tr>
      <td align="right">$idPhrase</td>
      <td align="left"><b>$id</b></td>
    </tr>
    <tr>
      <td align="right">$ledgerCodePhrase</td>
      <td align="left"><b>$ledgerCode</b></td>
    </tr>
    <tr>
      <td align="right">$accountsPhrase</td>
      <td align="left">
        <select name="accounts" size="5" multiple>$accountsInfo
        </select>
      </td>
    </tr>
    <tr>
      <td align="right">&lt;</td>
      <td align="left"><input type="text" name="lessThan" value="$lessThan"></td>
    </tr>
    <tr>
      <td align="right">$operatorPhrase</td>
      <td align="left">
        <select name="operator">$operatorInfo
        </select>
      </td>
    </tr>
    <tr>
      <td align="right">&gt;</td>
      <td align="left"><input type="text" name="greaterThan" value="$greaterThan"></td>
    </tr>
    <tr>
      <td align="right">$usersPhrase</td>
      <td align="left">
        <select name="users" size="5" multiple>$usersInfo
        </select>
      </td>
    </tr>
    <tr>
      <td align="right">$emailsPhrase</td>
      <td align="left"><textarea name="emails" rows="5" cols="50" wrap="hard">$emails</textarea></td>
    </tr>
    <tr>
      <td align="right">$mailTypePhrase</td>
      <td align="left">
        <select name="mailType">$mailTypeInfo
        </select>
      </td>
    </tr>
  </table>
  <input type="submit" name="update" value="$updatePhrase">
</form>
END_OF_CODE

  return $doc;
}

# MonitorObject validateMonitorAccountInfo(command, ledgerCode, ledger, id)
sub validateMonitorAccountInfo
{
  my $self = shift;
  my %args = ( command => "edit", ledgerCode => "", ledger => "", id => "-1", @_ );
  my $command = $args{command};
  my $ledgerCode = $args{ledgerCode};
  my $ledger = $args{ledger};
  my $id = $args{id};
  my $phrases = $self->{langObj}->{phraseObj}->{phrases};
  my $errorPhrase = $phrases->{error};
  my $monitorObj = undef;

  if ($command !~ /^(edit|create)$/)
  {
    $self->setError("command = '$command' is invalid!<br>\n");
    return $monitorObj;
  }
  if ($command eq "create" && $id ne "-1")
  {
    $self->setError("id = '$id' is invalid when command = '$command'!<br>\n");
    return $monitorObj;
  }

  my $accounts = "",
  my $lessThan = "";
  my $operator = "";
  my $greaterThan = "";
  my $users = "";
  my $emails = "";
  my $mailType = "";

  $lessThan = $self->{input}->{lessThan};
  $greaterThan = $self->{input}->{greaterThan};
  $emails = $self->{input}->{emails};
  $operator = $self->{input}->{operator};
  $mailType = $self->{input}->{mailType};

  # gather available Accounts and make sure the ones the user selected are valid.
  my @availableAccountObjs = ();
  foreach my $acctType (@{$self->{accountTypesToMonitor}})
  {
    my @tmpAvailableAccountObjs = $self->{dbObj}->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => $acctType);
    if ($self->{dbObj}->didErrorOccur)
    {
      $self->setError($self->{dbObj}->errorMessage);
      return $monitorObj;
    }
    foreach my $obj (@tmpAvailableAccountObjs)
    {
      push @availableAccountObjs, $obj;
    }
  }
  if (scalar @availableAccountObjs == 0)
  {
    $self->setError("No Accounts assigned to Ledger '$ledger' - '$ledgerCode'.<br>\nFix this error message and make it be a link!<br>\n");
    return $monitorObj;
  }
  my @selectedAccounts = split /\0/, $self->{input}->{accounts};
  my $accounts = join(',', @selectedAccounts);

  foreach my $accountNum (@selectedAccounts)
  {
    my $found = 0;
    foreach my $accountObj (@availableAccountObjs)
    {
      if ($accountObj->{number} == $accountNum)
      {
        $found = 1;
        last;
      }
    }
    if (!$found)
    {
      $self->setError("Account with number = '$accountNum' is not assigned to Ledger '$ledger' - '$ledgerCode'!<br>\n");
      return $monitorObj;
    }
  }

  # gather available Portal users for this company.
  my @availableUserObjs = $self->{authObj}->getListOfUsersForCompany(companyObj => $self->{companyObj});
  if ($self->{authObj}->didErrorOccur)
  {
    $self->setError($self->{authObj}->errorMessage);
    return $monitorObj;
  }
  if (scalar @availableUserObjs == 0)
  {
    $self->setError("No Users created for Company '$self->{companyObj}->{name}' - '$self->{companyObj}->{code}'.<br>\nFix this error message and make it be a link!<br>\n");
    return $monitorObj;
  }
  my @selectedUsers = split /\0/, $self->{input}->{users};
  my $users = join(',', @selectedUsers);

  foreach my $uname (@selectedUsers)
  {
    my $found = 0;
    foreach my $userObj (@availableUserObjs)
    {
      if ($userObj->{uname} eq $uname)
      {
        $found = 1;
        last;
      }
    }
    if (!$found)
    {
      $self->setError("User with uname = '$uname' is not assigned to Company '$self->{companyObj}->{name}' - '$self->{companyObj}->{code}'!<br>\n");
      return $monitorObj;
    }
  }

  # fixup the emails.
  $emails =~ s/\s*\n\s*/,/g;
  $emails =~ s/,$//;  # remove the trailing , if it exists.

  # now build up the MonitorObject and return it.
  $monitorObj = Portal::Accounting::Objects::MonitorObject->new(id => $id, ledgerCode => $ledgerCode, accounts => $accounts, emails => $emails, users => $users,
                lessThan => $lessThan, operator => $operator, greaterThan => $greaterThan, mailType => $mailType,
                langObj => $self->{langObj});

  return $monitorObj;
}

sub close
{
  my $self = shift;
  my %args = ( @_ );
  my $command = 'close';
  my $doc;
  my $phrases = $self->{langObj};

  $doc = HTMLObject::Normal->new;
  if (not defined $doc || $doc->didErrorOccur)
  {
    $self->setError("Could not create HTMLObject<br>\n");
    return undef;
  }

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->didErrorOccur)
  {
    $self->setError(errorString => $self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->setError(errorString => "No valid colorScheme returned!<br>\n");
    return undef;
  }

  # create the needed css entries
  my %cssEntries = $self->{methods}->generateColorSchemeCSS(colorScheme => $colorScheme, appName => "Accounting");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }
  $doc->print(%cssEntries);

  my $window = $self->{input}->{windowName};
  if ($self->{sessionObj}->windowExists($window))
  {
    $self->{sessionObj}->unregisterWindow($window);
    if ($self->{sessionObj}->error)
    {
      $self->error($self->{sessionObj}->errorMessage);
      return undef;
    }
  }
  else
  {
    # this is an internal error!  I'm ignoring it for now.
  }

  # generate the document that will close itself.
  $doc->setTitle("XIWA - " . $phrases->map("Close_Window"));

  $doc->setOnload(code => "window.close();");  # output the JavaScript code that
                                              #will be run on the onLoad event of the page.

  return $doc;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::Base(3)

=cut
