# DBUpgrade.pm - The Object Class that provides a DBUpgrade Object
#$Id: DBUpgrade.pm,v 1.35 2005/03/04 23:28:50 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::DBUpgrade;
use strict;
use Portal::Accounting::Main;
use Portal::Accounting::DB::Main;
use Portal::Accounting::DB::Ledger;
use Portal::Accounting::Objects::JournalEntry;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.06';

# global variables.
my $currDBVersion = "1.6";

# modify to represent the commands this state provides.
my %commands = (
  display => "Initialize the Upgrade Process",
  phase1 => "Upgrade to next specified version",
  userinput => "Allow user to interact with Upgrade Process",
  displayTop => "The Top Frame",
  displayPhase => "The Bottom Frame",
  notAdmin => "Upgrade needed, but you are not an admin!",
               );
=head1 NAME

DBUpgrade - Object used to build a DBUpgrade Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::DBUpgrade;
  my $obj = Portal::Accounting::DBUpgrade->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

DBUpgrade is a DBUpgrade class.
This class allows for migration from one version of the database to the
next and allows the user to interact with the migration in a nicer
environment than a command line provides.  This is mainly to allow the
user to specify the stock info or what accounts are to be taxable, etc.
If at all possible, the command line upgrade program should be used as
it is not affected by browser timeout, etc.

This is run from the Accounting module when it detects that a new
version of the database is available (code wise) and the database is not
at the version needed.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::DBUpgrade module.
 See Portal::AppState(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item HTMLObject run(command)

    Runs the specified command and returns the resulting HTMLObject.

=cut

sub run
{
  my $self = shift;
  my %args = ( command => "", @_ );
  my $command = $args{command};
  my $doc;
  my $phrases = $self->langObj->phraseObj->phrases;

  if (!exists $commands{$command})
  {
    $self->setError(errorString => sprintf($phrases->{commandNotKnown}, $command));
    return undef;
  }
  $doc = HTMLObject::Normal->new;

  my $mainDBObj = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return undef;
  }

  my $ledgerDBObj = Portal::Accounting::DB::Ledger->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($ledgerDBObj->didErrorOccur)
  {
    $self->setError(errorString => $ledgerDBObj->errorMessage);
    return undef;
  }

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  # check Database version
  my $dbVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return undef;
  }

  if ($command eq "notAdmin")
  {
    $doc->setTitle("DBUpgrade:  Needed!");
    $url .= "?" . (length $self->{sessionObj}->{store}->{originalDestination} > 0 ? $self->{sessionObj}->{store}->{originalDestination} : "app=Accounting&state=Main&command=display");
    $doc->setFocus("body");
    $doc->print("<center><h1>DB Upgrade Needed!</h1><br>But you are not an admin user!</center><br><br><br>\n");
    $doc->print("Go <a href=\"$url\">here</a> to continue using Xiwa after the database is upgraded.");

    return $doc;
  }

  if ($dbVersion >= $currDBVersion && $command ne "userinput")
  {
    # we don't need to upgrade!  Return to Main->display().
    $doc->setTitle("DBUpgrade:  Not needed!");
    $url .= "?" . (length $self->{sessionObj}->{store}->{originalDestination} > 0 ? $self->{sessionObj}->{store}->{originalDestination} : "app=Accounting&state=Main&command=display");
    $doc->setMetaTag('http-equiv' => "Refresh", content => "1; URL=$url");
    $doc->setFocus("body");
    $doc->print("Continuing...");

    # make sure the app is unlocked!
    if ($self->{applicationObj}->isAppLocked(app => $self->{appObj}->{id}))
    {
      # now unlock the app.
      $self->{applicationObj}->unlockApp(app => $self->{appObj}->{id}, type => "c", company => $self->{companyObj}->{id});
      if ($self->{applicationObj}->didErrorOccur)
      {
        $self->setError(errorString => $self->{applicationObj}->errorMessage);
        return undef;
      }
    }
    return $doc;
  }

  my @currentVersion = ("1.5");
  my %updateFunctions = ("1.4" => "updateToOneFive", "1.5" => "updateToOneSix" );
  my %userInputFunctions = ( "1.2" => "userInputToOneThree" );

  if ($command eq "display")
  {
    # generate frameset
    use HTMLObject::FrameSet;
    $doc = HTMLObject::FrameSet->new;

    $doc->setTitle("DBUpgrade: Database Upgrade");

    my $innerContent;
    $innerContent = $doc->createNoframe("You must have a frames capable browser!");
    $innerContent .= $doc->createFrame(src => "$url?app=Accounting&state=DBUpgrade&command=displayTop", frameborder => 'no', scrolling => 'auto', border => '1', name => 'upgradeTop');
    $innerContent .= $doc->createFrame(src => "$url?app=Accounting&state=DBUpgrade&command=displayPhase", frameborder => 'no', scrolling => 'auto', name => 'upgradeBottom');
    $doc->outerFrameset(rows => '*, 4*', content => $innerContent, border => '1', frameborder => 'no');
  }
  elsif ($command eq "displayTop")
  {
    $doc->setTitle("DBUpgrade: Top Frame");
    $doc->setFocus("body");
    if ($self->{sessionObj}->{store}->{dumpDB})
    {
      $doc->print("<center>Dumping Database at Version <b>$dbVersion</b>...</center>");
    }
    else
    {
      $doc->print("<center>Upgrading from Version <b>$dbVersion</b> to <b>$currDBVersion</b>...</center>");
    }
  }
  elsif ($command eq "displayPhase")
  {
    $doc = $self->displayPhaseScreen(doc => $doc, dbVersion => $dbVersion, mainDBObj => $mainDBObj, ledgerDBObj => $ledgerDBObj, currentVersion => \@currentVersion, updateFunctions => \%updateFunctions);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
  }
  elsif ($command eq "phase1")
  {
    $doc = $self->processPhase(doc => $doc, dbVersion => $dbVersion, mainDBObj => $mainDBObj, ledgerDBObj => $ledgerDBObj, currentVersion => \@currentVersion, updateFunctions => \%updateFunctions);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
  }
  elsif ($command eq "userinput")
  {
    $doc = $self->processUserInput(doc => $doc, dbVersion => $dbVersion, mainDBObj => $mainDBObj, ledgerDBObj => $ledgerDBObj, currentVersion => \@currentVersion, userInputFunctions => \%userInputFunctions);
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
  }

  return $doc;
}

=item HTMLObject displayPhaseScreen(dbVersion, mainDBObj, ledgerDBObj, @currentVersion,
             %updateFunctions, doc, [errorStr, statusStr])

    This method decides what version we need to upgrade to and starts
    the process of calling the correct helper functions, etc.  If the
    helper function decides it needs user interaction, an HTMLObject
    document will be returned with the form, etc. that gets the user
    started on their part of the process.  If no user input is required
    and the upgrade is done for the current version, then undef is
    returned.  When all necessary upgrades are finished, a success
    document is displayed for the user and a link is provided to take
    them back to the main menu.

=cut
sub displayPhaseScreen
{
  my $self = shift;
  my %args = (dbVersion => "", mainDBObj => undef, ledgerDBObj => undef, currentVersion => undef, updateFunctions => undef, doc => undef, errorStr => "", statusStr => "", @_ );
  my $dbVersion = $args{dbVersion};
  my $mainDBObj = $args{mainDBObj};
  my $currentVersion = $args{currentVersion};
  my $updateFunctions = $args{updateFunctions};
  my $doc = $args{doc};
  my $errorStr = $args{errorStr};
  my $statusStr = $args{statusStr};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  # figure out where the current DB Version is in relation to the latest version to upgrade.
  my $index;
  my $found = 0;
  my $count;
  my $nextVersion;
  for ($index=0; $index < scalar @{$currentVersion} && !$found; $index++)
  {
    if ($currentVersion->[$index] == $dbVersion)
    {
      $found = 1;
    }
  }
  if ($found)
  {
    $index--;
    $count = scalar(@{$currentVersion}) - $index;  # This tells me how many there are between it and the end of the array (number of upgrades to perform).
  }
  else
  {
    if ($nextVersion == $currDBVersion)
    {
      $count = 0;
    }
    else
    {
      # unknown version!
      $self->setError(errorString => "Current Database Version = '$dbVersion' is unknown!<br>\nSee the System Administrator!<br>\n");
      return undef;
    }
  }
  $nextVersion = (($index + 1) < scalar @{$currentVersion} ? $currentVersion->[$index+1] : $currDBVersion);

  # now generate the Request for confirmation of upgrade and tell the user how many steps to become current.
  $doc->setTitle("DBUpgrade: Bottom Frame");
  $doc->setFocus("body");
  if ($self->{sessionObj}->{store}->{dumpDB})
  {
    $doc->print(<<"END_OF_CODE");
<center>
<h1>Dumping Database...</h1>
</center>
END_OF_CODE
    if (length $errorStr > 0)
    {
      $doc->print("<font color=\"red\">Error:</font> $errorStr\n<br>\n");
    }
    if (length $statusStr > 0)
    {
      $doc->print("<font color=\"red\">Status:</font> $statusStr\n<br>\n");
    }
    $doc->print(<<"END_OF_CODE");
<center>
<form method="post" name="upgradeForm" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="DBUpgrade">
  <input type="hidden" name="command" value="displayPhase">
END_OF_CODE

    # dump the database
    my $dumpDir = $self->{configObj}->{webRoot} . $self->{configObj}->{siteDir} . "/db_dump/Accounting/$self->{companyObj}->{code}/";

    my $companyAppObj = $self->{sessionObj}->{store}->{companyAppObj};
    my $result = $self->{methods}->dbDump(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                          user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                          dbName => $companyAppObj->{dbName}, dir => $dumpDir, fname => "xiwa_$dbVersion.dump", inserts => 1);
    if ($self->{methods}->didErrorOccur)
    {
      $self->setError(errorString => $self->{methods}->errorMessage);
      return undef;
    }

    $self->{sessionObj}->{store}->{dumpDB} = 0;  # indicate we no longer need to dump the database.

    $doc->print(<<"END_OF_CODE");
  <br>
  <br>
  $result
  <br>
  <br>
  <h1>Database Dumped...</h1>
  <br>
  <br>
  <h2>Do <font color="red">Not</font> reload your browser!</h2>
</form>
END_OF_CODE
  }
  else
  {
    $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade In <font color="red">Progress</font></h1>
</center>
END_OF_CODE
  if (length $errorStr > 0)
  {
    $doc->print("<font color=\"red\">Error:</font> $errorStr\n<br>\n");
  }
  if (length $statusStr > 0)
  {
    $doc->print("<font color=\"red\">Status:</font> $statusStr\n<br>\n");
  }
  $doc->print(<<"END_OF_CODE");
<center>
<form method="post" name="upgradeForm" action="$url">
  <input type="hidden" name="app" value="Accounting">
  <input type="hidden" name="state" value="DBUpgrade">
  <input type="hidden" name="command" value="phase1">
  <input type="hidden" name="nextVersion" value="$dbVersion">
  <br>
  <input type="hidden" name="confirmUpgrade" value="YES">
  <br>
  Current step: Going from Version <b>$dbVersion</b> to <b>$nextVersion</b>.
  <br>
  There is a total of <b>$count</b> individual database upgrades to do. &nbsp;You may have to interact with the upgrade process and then you may not.
  <br>
  The upgrade process may take a while, so please be patient.
  <br>
  <br>
  <h2>Do <font color="red">Not</font> reload your browser!</h2>
</form>
END_OF_CODE
  }

  $doc->setOnload(code => "document.upgradeForm.submit();");

  return $doc;
}

=item HTMLObject processPhase(dbVersion, mainDBObj, ledgerDBObj, @currentVersion,
             %updateFunctions, doc)

    actually validates and then processes the database upgrade for the
    specified version.  Returns HTMLObject if user interaction required,
    else undef if upgrade is complete.

=cut
sub processPhase
{
  my $self = shift;
  my %args = (dbVersion => "", mainDBObj => undef, ledgerDBObj => undef, currentVersion => undef, updateFunctions => undef, doc => undef, @_ );
  my $dbVersion = $args{dbVersion};
  my $mainDBObj = $args{mainDBObj};
  my $ledgerDBObj = $args{ledgerDBObj};
  my $currentVersion = $args{currentVersion};
  my $updateFunctions = $args{updateFunctions};
  my $doc = $args{doc};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  my $confirmUpgrade = $self->{input}->{confirmUpgrade};
  my $nextVersion = $self->{input}->{nextVersion};

  # set dbVersion into the input hash for children functions to get to.
  $self->{input}->{dbVersion} = $dbVersion;

  # make sure we have the input we need.
  if ($confirmUpgrade !~ /^YES|NO$/)
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "confirmUpgrade = '$confirmUpgrade' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }
  if ($nextVersion !~ /^\d+\.\d+$/ || !exists $updateFunctions->{$nextVersion})
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "nextVersion = '$nextVersion' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  if ($confirmUpgrade eq "NO")
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "You must specify <b>Yes</b> before I can Upgrade your database to the next version!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  # call the upgrade procedure for the specified version
  my $function = $updateFunctions->{$nextVersion};
  my @results;  # return results must be [0] = success/failure, [1] = doc, [2] = doc modified, [3] = success message
  eval "\@results = \$self->$function(doc => \$doc, mainDBObj => \$mainDBObj, ledgerDBObj => \$ledgerDBObj);";
  if ($@)
  {
    $self->setError(errorString => "Eval of Upgrade Function = '$function' failed!<br>\nError = '$@'.<br>\n");
    return undef;
  }

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return undef;
  }
  if ($results[2])
  {
    $doc = $results[1];  # return the updated version of $doc.  This is how user interaction is provided for.
  }
  else
  {
    # see if we have another version to upgrade or if we are done.
    # find nextVersion in the array.
    my $index;
    my $count;
    my $found = 0;
    for ($index=0; $index < scalar @{$currentVersion} && !$found; $index++)
    {
      if ($currentVersion->[$index] == $nextVersion)
      {
        $found = 1;
      }
    }
    if ($found)
    {
      $count = scalar(@{$currentVersion}) - $index;  # This tells me how many there are between it and the end of the array (number of upgrades to perform).
    }
    else
    {
      if ($nextVersion == $currDBVersion)
      {
        $count = 0;
      }
      else
      {
        # unknown version!
        $self->setError(errorString => "Current Database Version = '$nextVersion' is unknown!<br>\nSee the System Administrator!<br>\n");
        return undef;
      }
    }

    if ($count > 0)
    {
      $nextVersion = $currentVersion->[$index];  # get the next version to upgrade to.
      $args{dbVersion} = $nextVersion;
      $doc = $self->displayPhaseScreen(%args, statusStr => $results[3]);
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return undef;
      }
      return $doc;
    }
    else
    {
      # generate a page that indicates we are done and provides a link back to the main menu.
      $doc->setTitle("DBUpgrade:  Completed");
      $doc->setFocus("body");
      $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade Completed!</h1>
<br>
<font color="red">Status:</font> $results[3]
<br>
<br>
</center>
<a href="$url?$self->{sessionObj}->{store}->{originalDestination}" onMouseOver="window.status='Original Destination'" target="_top">Continue</a>
END_OF_CODE
      $doc->setLinkDecorations; #/

      # now unlock the app.
      $self->{applicationObj}->unlockApp(app => $self->{appObj}->{id}, type => "c", company => $self->{companyObj}->{id});
      if ($self->{applicationObj}->didErrorOccur)
      {
        $self->setError(errorString => $self->{applicationObj}->errorMessage);
        return undef;
      }
    }
  }

  return $doc;
}

=item HTMLObject processUserInput(dbVersion, mainDBObj, ledgerDBObj, @currentVersion,
             %userInputFunctions, doc)

    actually validates and then processes any user input needed for the
    upgrade for the specified version.  Returns HTMLObject if more user
    input is required, or undef if done with the user input.

=cut
sub processUserInput
{
  my $self = shift;
  my %args = (dbVersion => "", mainDBObj => undef, ledgerDBObj => undef, currentVersion => undef, userInputFunctions => undef, doc => undef, @_ );
  my $dbVersion = $args{dbVersion};
  my $mainDBObj = $args{mainDBObj};
  my $ledgerDBObj = $args{ledgerDBObj};
  my $currentVersion = $args{currentVersion};
  my $userInputFunctions = $args{userInputFunctions};
  my $doc = $args{doc};

  my $url = $self->{methods}->createBaseURL(type => "App", appConfigObj => $self->{sessionObj}->{store}->{companyAppObj});
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  my $confirmUpgrade = $self->{input}->{confirmUpgrade};
  my $nextVersion = $self->{input}->{nextVersion};

  # set dbVersion into the input hash for children functions to get to.
  $self->{input}->{dbVersion} = $dbVersion;

  # make sure we have the input we need.
  if ($confirmUpgrade !~ /^YES|NO$/)
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "confirmUpgrade = '$confirmUpgrade' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }
  if ($nextVersion !~ /^\d+\.\d+$/ || !exists $userInputFunctions->{$nextVersion})
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "nextVersion = '$nextVersion' is invalid!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  if ($confirmUpgrade eq "NO")
  {
    $doc = $self->displayPhaseScreen(%args, errorStr => "You must specify <b>Yes</b> before I can Upgrade your database to the next version!<br>\n");
    if ($self->didErrorOccur)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }

  # call the user input procedure for the specified version
  my $function = $userInputFunctions->{$nextVersion};
  my @results;  # return results must be [0] = success/failure, [1] = doc, [2] = doc modified, [3] = success message
  eval "\@results = \$self->$function(doc => \$doc, mainDBObj => \$mainDBObj, ledgerDBObj => \$ledgerDBObj);";
  if ($@)
  {
    $self->setError(errorString => "Eval of User Input Function = '$function' failed!<br>\nError = '$@'.<br>\n");
    return undef;
  }

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return undef;
  }
  if ($results[2])
  {
    $doc = $results[1];  # return the updated version of $doc.  This is how user interaction is provided for.
  }
  elsif ($results[0])  # we successfully finished whatever we were doing.
  {
    # see if we have another version to upgrade or if we are done.
    # find nextVersion in the array.
    my $index;
    my $count;
    my $found = 0;
    for ($index=0; $index < scalar @{$currentVersion} && !$found; $index++)
    {
      if ($currentVersion->[$index] == $nextVersion)
      {
        $found = 1;
      }
    }
    if ($found)
    {
      $count = scalar(@{$currentVersion}) - $index;  # This tells me how many there are between it and the end of the array (number of upgrades to perform).
    }
    else
    {
      if ($nextVersion == $currDBVersion)
      {
        $count = 0;
      }
      else
      {
        # unknown version!
        $self->setError(errorString => "Current Database Version = '$nextVersion' is unknown!<br>\nSee the System Administrator!<br>\n");
        return undef;
      }
    }

    if ($count > 0)
    {
      $nextVersion = $currentVersion->[$index];  # get the next version to upgrade to.
      $args{dbVersion} = $nextVersion;
      $doc = $self->displayPhaseScreen(%args, statusStr => $results[3]);
      if ($self->didErrorOccur)
      {
        $self->prefixError();
        return undef;
      }
      return $doc;
    }
    else
    {
      # generate a page that indicates we are done and provides a link back to the main menu.
      $doc->setTitle("DBUpgrade:  Completed");
      $doc->setFocus("body");
      $doc->print(<<"END_OF_CODE");
<center>
<h1>Database Upgrade Completed!</h1>
<br>
<font color="red">Status:</font> $results[3]
<br>
<br>
</center>
<a href="$url?$self->{sessionObj}->{store}->{originalDestination}" onMouseOver="window.status='Original Destination'" target="_top">Continue</a>
END_OF_CODE
      $doc->setLinkDecorations;

      # now unlock the app.
      $self->{applicationObj}->unlockApp(app => $self->{appObj}->{id}, type => "c", company => $self->{companyObj}->{id});
      if ($self->{applicationObj}->didErrorOccur)
      {
        $self->setError(errorString => $self->{applicationObj}->errorMessage);
        return undef;
      }
    }
  }

  return $doc;
}

# version specific upgrade subroutines here....

# updateToOneFive
# requires: doc, mainDBObj
# optional:
# returns: array: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1), [3]=status message (if success)
# summary: This updates the database from version 1.4 to 1.5.
sub updateToOneFive
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $success = 0;
  my $modified = 0;
  my $statusStr = "Successfully updated from <b>1.4</b> to <b>1.5</b>.\n";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version 1.4 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.4)
  {
    $self->setError(errorString => "The database is not at version = <b>1.4</b>!<br>\nI can not upgrade to <b>1.5</b>!<br>\n");
    return @result;
  }

  # get the path that our upgrade.psql file is at.
  my $upgradePath = $self->{configObj}->{webRoot} . $self->{configObj}->{siteDir} . "/install/Accounting/db/1.5/";
  my $upgradeFile = $upgradePath . "upgrade.psql";

  # now validate we have an upgrade.psql file
  if (! -r $upgradeFile)
  {
    $self->setError(errorString => "Upgrade file '$upgradeFile' does not exist!  Error = '$!'.<br>\n");
    return @result;
  }

  # now use the Portal provided script to import/run the upgrade.psql file.
  my $companyAppObj = $self->{sessionObj}->{store}->{companyAppObj};
  my $result = $self->{methods}->dbImport(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                        user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                        dbName => $companyAppObj->{dbName}, fname => "$upgradeFile");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return undef;
  }

  # signal success.
  $result[0] = 1;

  return @result;
}

# updateToOneSix
# requires: doc, mainDBObj, ledgerDBObj
# optional:
# returns: array: [0]=success/failure (1/0), [1]=doc, [2]=doc modified (0/1), [3]=status message (if success)
# summary: This updates the database from version 1.5 to 1.6.
sub updateToOneSix
{
  my $self = shift;
  my %args = ( doc => undef, mainDBObj => undef, ledgerDBObj => undef, @_ );
  my $doc = $args{doc};
  my $mainDBObj = $args{mainDBObj};
  my $ledgerDBObj = $args{ledgerDBObj};
  my $success = 0;
  my $modified = 0;
  my $lang = $self->{langObj};
  my $statusStr = "Successfully updated from <b>1.5</b> to <b>1.6</b>.\n";
  my @result = ($success, $doc, $modified, $statusStr);
  my $dbVersion = $self->{input}->{dbVersion};
  my @defaultEquityAccount = ( 3000 );

  if (!defined $doc)
  {
    $self->setError(errorString => "doc is not defined!<br>\n");
    return @result;
  }

  # validate we are still at version 1.5 before we do anything.
  my $currentVersion = $mainDBObj->getDBVersion;
  if ($mainDBObj->didErrorOccur)
  {
    $self->setError(errorString => $mainDBObj->errorMessage);
    return @result;
  }

  if ($currentVersion != 1.5)
  {
    $self->setError(errorString => "The database is not at version = <b>1.5</b>!<br>\nI can not upgrade to <b>1.6</b>!<br>\n");
    return @result;
  }

  # get the path that our upgrade.psql file is at.
  my $upgradePath = $self->{configObj}->{webRoot} . $self->{configObj}->{siteDir} . "/install/Accounting/db/1.6/";
  my $upgradeFile = $upgradePath . "upgrade.psql";

  # now validate we have an upgrade.psql file
  if (! -r $upgradeFile)
  {
    $self->setError(errorString => "Upgrade file '$upgradeFile' does not exist!  Error = '$!'.<br>\n");
    return @result;
  }

  # now use the Portal provided script to import/run the upgrade.psql file.
  my $companyAppObj = $self->{sessionObj}->{store}->{companyAppObj};
  my $result = $self->{methods}->dbImport(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                        user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                        dbName => $companyAppObj->{dbName}, fname => "$upgradeFile");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return @result;
  }

  # Create a database JournalEntry object for use in inserting JournalEntryObject objects into the database
  $self->{transaction} = Portal::Accounting::DB::JournalEntry->new( dbHandle => $self->{dbHandle}, langObj => $self->{langObj} );

  if ($self->{transaction}->didErrorOccur())
  {
    $self->setError(errorString => "Error instantiating JournalEntry object!<br>\n" . $self->{transaction}->errorMessage);
    return @result;
  }

  # Convert beginning_balance_entry_tb table entries into Journal Entry transactions
  my $sql = "SELECT DISTINCT(beginning_balance_entry_tb.ledger_code), MIN(known_years_tb.year),  (SELECT start1 from ledger_tb WHERE code = beginning_balance_entry_tb.ledger_code) AS start FROM beginning_balance_entry_tb GROUP BY ledger_code;";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->setError(errorString => "Error obtaining distinct list of used ledger codes from beginning_balance_tb table!<br>\n" . $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }
  while (my @info = $sth->fetchrow_array)
  {
    # Create journal entries using the books start date less one day
    my @date = ($info[1], split "-", $info[2]);
    @date = Date::Calc::Add_Delta_Days(@date, -1);
    my $fiscalPriorDate = "$date[0]-$date[1]-$date[2]";

    my $transObj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $info[0], journalCode => "G", transDate => $fiscalPriorDate, langObj => $self->{langObj});

    my $sth2 = $self->{dbHandle}->read(sql => "SELECT account_num, sub_account_num, sub_e_code, amount, dorc FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $info[0] ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->setError(errorString => "Error obtaining list of entries from beginning_balance_entry_tb for ledger_code = '" . $info[0] . "'<br>\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    my $debitSum = 0.00;
    my $creditSum = 0.00;
    my $problemString = "ledgerCode = '" . $info[0] . "'<br>\n";
    while (my @info2 = $sth2->fetchrow_array)
    {
      $problemString .= "fiscalPriorDate = '" . $fiscalPriorDate . "', subEntity = '" . $info2[2] . "', accountNum = '" . $info2[0] . "', subAccountNum = '" . $info2[1] . "', dorc = '" . $info2[4] . "', amount = '" . $info2[3] . "'";
      # build up the JournalEntryObject
      $transObj->addRow(rowDate => $fiscalPriorDate, subEntity => $info2[2], accountNum => $info2[0], subAccountNum => $info2[1], dorc => $info2[4], amount => $info2[3]);
      if ($transObj->didErrorOccur)
      {
        $self->setError(errorString => "Error adding row to JournalEntry:" . $transObj->errorMessage . "<br>\n" . $problemString);
        $self->{dbHandle}->rollback;
        return @result;
      }

      if ($info2[4] =~ /^(true|1)$/)
      {
        $debitSum += $info2[3];
      }
      else
      {
        $creditSum += $info2[3];
      }

      $problemString .= ", creditSum = '" . $creditSum . "', debitSum = '" . $debitSum . "'.<br>\n";
    }
    # add equity account amount to offset beginning_balance_entry_tb amounts
    my $sum = 0.00;
    my $dorc = 0;
    my $amount = 0.00;
    $sum = sprintf("%.2f", $debitSum - $creditSum);
    if ($sum < 0.00)
    {
      $dorc = 1;
    }
    else
    {
      $dorc = 0;
    }
    $amount = sprintf("%.2f", abs($sum));
    $problemString .= "equity sum = '" . $sum . "', equity dorc = '" . $dorc . "', equity amount = '" . $amount . "'.<br>\n";
    # hardcode Retained Earnings account as the equity account to offset all the values
    $transObj->addRow(rowDate => $fiscalPriorDate, subEntity => "", accountNum => "3000", subAccountNum => "-1", dorc => $dorc, amount => $amount);
    if ($transObj->didErrorOccur)
    {
      $self->setError(errorString => "Error adding row to JournalEntry:" . $transObj->errorMessage . "<br>\n" . $problemString);
      $self->{dbHandle}->rollback;
      return @result;
    }

    my $refNum = $self->{transaction}->insert(obj => $transObj, commit => 0);
    if (($self->{transaction}->didErrorOccur) || ($refNum == -1))
    {
      $self->setError(errorString => "Error inserting new transaction!\n" . $self->{transaction}->errorMessage . "<br>\n" . $problemString);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # Delete all beginning_balance_entry_tb entries for the current ledger
    my $sth3 = $self->{dbHandle}->write(sql => "DELETE FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $info[0] ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->setError(errorString => "Error deleting entries from beginning_balance_entry_tb table for ledger_code = '" . $info[0] . "'!\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # Create new beginning_balance_entry_tb entry using reference number from journal entry and current ledger code
    $sth3 = $self->{dbHandle}->write(sql => "INSERT INTO beginning_balance_entry_tb (ref_num, ledger_code) VALUES (?, ?)", plug => [ $refNum, $info[0] ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->setError(errorString => "Error inserting updated entry into beginning_balance_entry_tb table for ref_num = '" . $refNum . "', ledger_code = '" . $info[0] . "'!\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # make sure that there are no closeouts after the fiscalPriorDate
    if (not $ledgerDBObj->reopenLedger(ledgerCode => $info[0], reopenDate => $fiscalPriorDate, commit => 0))
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorReOpeningLedger"), $info[0], $fiscalPriorDate) . "<br>\n" . $ledgerDBObj->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }

    # "Closeout" the books for the day prior to the books start date for the current ledger
    if (not $ledgerDBObj->closeoutLedger(ledgerCode => $info[0], closeoutDate => $fiscalPriorDate, equityAccounts => \@defaultEquityAccount, checkDateRange => 0, commit => 0))
    {
      $self->setError(errorString => sprintf($self->{langObj}->map("errorClosingOutLedger"), $info[0], $fiscalPriorDate) . "<br>\n" . $ledgerDBObj->errorMessage);
      $self->{dbHandle}->rollback;
      return @result;
    }
  }
  # commit changes
  $sth = $self->{dbHandle}->commit;
  if ($self->{dbHandle}->didErrorOccur)
  {
    $self->setError(errorString => "Error commiting upgrade of initial balance entries!\n" . $self->{dbHandle}->errorMessage);
    $self->{dbHandle}->rollback;
    return @result;
  }

  # Finish the upgrade by dropping unneeded columns in the beginning_balance_entry_tb table
  $upgradeFile = $upgradePath . "post-upgrade.psql";
  $result = $self->{methods}->dbImport(type => $companyAppObj->{dbType}, host => $companyAppObj->{dbHost}, port => $companyAppObj->{dbPort},
                                        user => $self->{configObj}->{dbUser}, passwd => $self->{configObj}->{dbPasswd},
                                        dbName => $companyAppObj->{dbName}, fname => "$upgradeFile");
  if ($self->{methods}->didErrorOccur)
  {
    $self->setError(errorString => $self->{methods}->errorMessage);
    return @result;
  }

  # signal success.
  $result[0] = 1;
  # use a JavaScript alert message box to indicate to the user that Retained Earnings account was used for all initial balance entries instead of
  # what might have originally been there and that any closeouts were reopened and will need to be closed out again
  $result[3] .= "<script language=\"Javascript1.2\"> alert('" . sprintf($lang->map("initialBalanceUpgrade"), $defaultEquityAccount[0]) . "'); </script>\n";

  return @result;
}

# getNodeTree
# Summary: retrieves a tree of nodes for the given node number from the lot_node_tb table
# Takes: nodeID, mainDBObj
# Returns: two-dimensional array of node ID's for the tree underneath the given node ID with the lot number associated with
#          the node ID
#          undef on failure of some kind
sub getNodeTree
{
  my $self = shift;
  my %args = ( mainDBObj => undef, nodeID => undef, @_ );
  my @treeNodes = [];
  my $mainDBObj = $args{'mainDBObj'};
  my $nodeID = $args{'nodeID'};

  # verify the arguments are valid
  if (not defined $mainDBObj)
  {
    return undef;
  }
  if (not defined $nodeID)
  {
    return undef;
  }

  # look for any children nodes that have the current node as their parent
  my $sth = $mainDBObj->{dbHandle}->read(sql => "SELECT lot, node_id FROM lot_node_tb WHERE parent_id = '$nodeID'");
  if ($mainDBObj->{dbHandle}->didErrorOccur())
  {
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    push @treeNodes, [ ($info[1], $info[0]) ];
    my @tempNodes = $self->getNodeTree(nodeID => $info[1], mainDBObj => $mainDBObj);
    if (not defined $tempNodes[0])
    {
      return undef;
    }
    push @treeNodes, @tempNodes;
  }

  return @treeNodes;
}

=back

=cut
1;
__END__

=head1 NOTE

  See updateTemplate as a template for future DB version updates.

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
