# CGL.pm - The Object Class that provides a CGL Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: CGL.pm,v 1.5 2005/07/22 15:40:48 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Investments::CGL;
use strict;
use HTMLObject::Widgets;
use Portal::Accounting::BaseL;
use Portal::Accounting::Objects::Stock;
use Portal::Accounting::DB::Main;
use Portal::Accounting::DB::Ledger;
use Portal::Accounting::DB::Security;
use Portal::Accounting::DisplayItems;
use Portal::Accounting::Objects::CGLObject;
use Portal::Accounting::Objects::CGLDisplayObject;
use Portal::Accounting::Objects::CGLAssets;
use Portal::Accounting::Objects::CGLAssetObject;
use Portal::Accounting::Objects::CGLStocks;
use Portal::Accounting::Objects::CGLStockObject;
use Portal::Accounting::Objects::CGLStockLotObject;
use Storable;
use POSIX qw(tmpnam);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.5 $';

# global variables.
# modify to represent the commands this state provides.
my %commands = (
  display => "Capital Gains/Losses",
  execute => "Capital Gains/Losses",
  update => "Capital Gains/Losses",
  print => "Capital Gains/Losses" );

=head1 NAME

CGL - Object used to build a CGL Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::CGL;
  my $obj = Portal::Accounting::CGL->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

CGL is a CGL class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::CGL module.
 See Portal::AppState(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{commands} = \%commands;
  $self->{widgets} = HTMLObject::Widgets->new();
  if ($self->{widgets}->error)
  {
    $self->error($self->{widgets}->errorMessage);
    return $self;
  }

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub c_display
{
  my $self = shift;
  my $doc = $self->prepDoc;
  $doc = $self->displayCustomizationScreen(doc => $doc, firstTime => 1);
  if ($self->error)
  {
    $doc->print($self->errorMessage());
  }
  return $doc;
}

sub c_execute
{
  my $self = shift;
  my %args = ( command => "display", @_ );
  my $command = $args{command};
  my $doc = $self->prepDoc;
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
    # validate input
    my $assets = $self->{input}->{assets};
    my $stocks = $self->{input}->{stocks};
    my $invAccount = $self->{input}->{invAccount};
    my $startDate = $self->{input}->{startDate};
    my $endDate = $self->{input}->{endDate};
    my $negativeFormat = $self->{input}->{negativeFormat};
    my $orderBy = $self->{input}->{orderBy};
    my $displayManualEntries = (exists $self->{input}->{displayManualEntries});

    my @assets = split /\0/, $assets;
    my @stocks = split /\0/, $stocks;
    (my $invAccountName = $invAccount) =~ s/^(.+)\|(.+)$/$2/;
    $invAccount =~ s/^(.+)\|(.+)$/$1/;

    my $errorString = "";
    if ($negativeFormat < 0 || $negativeFormat > 2)
    {
      $errorString .= "negativeFormat = '$negativeFormat' is invalid!<br>\n";
    }
    if ($orderBy < 0 || $orderBy > 1)
    {
      $errorString .= "orderBy = '$orderBy' is invalid!<br>\n";
    }

    if (length $errorString > 0)
    {
      $doc = $self->displayCustomizationScreen(doc => $doc, error => $errorString);
      if ($self->error)
      {
        $self->prefixError();
        return undef;
      }
      return $doc;
    }
    else
    {
      # gather the specified data and store in a storable object.

      my $securitiesObj = Portal::Accounting::DB::Security->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
      if ($securitiesObj->error)
      {
        $self->error($securitiesObj->errorMessage);
        return undef;
      }
      my $cglDisplayObj = $securitiesObj->getCGLData(invAccount => $invAccount, assets => \@assets, stocks => \@stocks, startDate => $startDate, endDate => $endDate, ledgerCode => $ledgerCode, orderBy => $orderBy, getManualEntries => $displayManualEntries);
      if (!defined $cglDisplayObj)
      {
        if ($securitiesObj->error)
        {
          $self->error($securitiesObj->errorMessage);
          return undef;
        }
        if ($securitiesObj->{errorCode} == -1)
        {
          $doc = $self->displayCustomizationScreen(doc => $doc, error => "No CGL Data returned for the selected Investment Accounts, Assets and Stocks!");
        }
        else
        {
          $doc = $self->displayCustomizationScreen(doc => $doc, error => "CGL Object not created!");
        }
        if ($self->error)
        {
          $self->prefixError();
          return undef;
        }
        return $doc;
      }
      else
      {
        if ($cglDisplayObj->error)
        {
          $self->error($cglDisplayObj->errorMessage);
          return undef;
        }
      }

      # now generate the Storable Object and save to disk.
      my %data = ( cglDisplayObj => $cglDisplayObj, startDate => $startDate, endDate => $endDate, orderBy => $orderBy, invAccount => ($invAccount ? "$invAccountName ($invAccount)" : ""), displayManualEntries => $displayManualEntries );
      my $filename = tmpnam();
      eval "store (\\\%data, \$filename);";
      if ($@)
      {
        $self->error("Store of file = '$filename' failed!<br>\nError = '$@'.<br>\n");
        return undef;
      }

      # log the name of the temp file for later removal.
      $securitiesObj->logTempFile(filename => $filename);
      if ($securitiesObj->error)
      {
        $self->error($securitiesObj->errorMessage);
        return undef;
      }

      # call the display routine giving it the name of the file to work with.
      $doc = $self->displayCGL(doc => $doc, filename => $filename, negativeFormat => $negativeFormat);
      if ($self->error)
      {
        $self->prefixError();
        return undef;
      }
    }
  return $doc;
}

sub c_update
{
  my $self = shift;
  my %args = ( command => "display", @_ );
  my $command = $args{command};
  my $doc = $self->prepDoc;
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
    my $filename = $self->{input}->{filename};
    my $updateWhat = $self->{input}->{updateWhat};
    my $number = $self->{input}->{number};
    my $negativeFormat = $self->{input}->{negativeFormat};

    if (length $filename == 0 || ! -e $filename)
    {
      $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
      return $doc;
    }

    if ($updateWhat !~ /^(cgl|sub-acct|refresh)$/)
    {
      $self->error("updateWhat = '$updateWhat' is invalid!<br>\n");
      return $doc;
    }
    if (length $number == 0 && $updateWhat ne "refresh")
    {
      $self->error("number = '$number' is invalid!<br>\n");
      return $doc;
    }
    if ($updateWhat eq "sub-acct" && ($number != -1 && $number != -2 && scalar (split /\|/, $number) != 4))
    {
      $self->error("number = '$number' is invalid for Toggling a Sub-Account!<br>\n");
      return $doc;
    }
    if ($updateWhat eq "cgl" && ($number != -1 && $number != -2 && scalar (split /\|/, $number) != 2))
    {
      $self->error("number = '$number' is invalid for Toggling a Sub-Account!<br>\n");
      return $doc;
    }
    if ($negativeFormat < 0 || $negativeFormat > 2)
    {
      $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
      return $doc;
    }

    if ($updateWhat ne "refresh")
    {
      # get the file to work with
      my $args_ref;
      eval "\$args_ref = retrieve(\$filename);";
      if ($@)
      {
        $self->error("Retrieve of file = '$filename' failed!<br>\nError = '$@'.<br>\n");
        return undef;
      }
      my %data = %{$args_ref};

      my $cglDisplayObj = $data{cglDisplayObj};
      if (!defined $cglDisplayObj || !$cglDisplayObj->isValid)
      {
        $self->error("stored cglDisplayObj is invalid or does not exist!<br>\n");
        return $doc;
      }

      # find the account or sub-acct to toggle
      if ($updateWhat eq "cgl")
      {
        if ($number == -1 || $number == -2)
        {
          my @sections = ("Stocks");
          push @sections, "Manual" if ($data{displayManualEntries});
          foreach my $objType (@sections) # Add Assets here
          {
            (my $objSingular = $objType) =~ s/s$//;
            foreach my $cglObj (@{$cglDisplayObj->{"cgl$objType"}})
            {
              if ((scalar @{$cglObj->{cglObjects}->{"cgl$objSingular" . "Objects"}} > 0))
              {
                $cglObj->state($number == -1 ? "o" : "c");
              }
            }
          }
        }
        else  # normal behaviour
        {
          (my $section, my $cgl) = split /\|/, $number;
          foreach my $objType ($section)
          {
            my $found = 0;
            for (my $count=0; $count < scalar @{$cglDisplayObj->{"cgl$objType"}} && !$found; $count++)
            {
              if ($cglDisplayObj->{"cgl$objType"}[$count]->type eq $cgl)
              {
                $cglDisplayObj->{"cgl$objType"}->[$count]->state($cglDisplayObj->{"cgl$objType"}->[$count]->{state} eq "c" ? "o" : "c");
                $found = 1;
              }
            }

            if (!$found)
            {
              $self->error("CGL Entry = '$cgl' not found for $objType!<br>\n");
              return $doc;
            }
          }
        }
      }
      elsif ($updateWhat eq "sub-acct")
      {
        if ($number == -1 || $number == -2)
        {
          my @sections = ("Stocks");
          push @sections, "Manual" if ($data{displayManualEntries});
          foreach my $objType (@sections) # Add Assets here
          {
            (my $objSingular = $objType) =~ s/s$//;
            # loop through all Accounts (open or closed) and toggle the flag of all sub-accts to either be open or closed.
            for (my $count=0; $count < scalar @{$cglDisplayObj->{"cgl$objType"}}; $count++)
            {
              if ((scalar @{$cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}} > 0))
              {
                $cglDisplayObj->{"cgl$objType"}->[$count]->state("o");  # force all Accounts to be open when we deal with all Sub-Accounts.
                for (my $count2=0; $count2 < scalar @{$cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}}; $count2++)
                {
                  $cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}->[$count2]->state($number == -1 ? "o" : "c");
                }
              }
            }
          }
        }
        else  # normal behaviour
        {
          (my $section, my $cgl, my $subAcct, my $type) = split /\|/, $number;
          foreach my $objType ($section)
          {
            (my $objSingular = $objType) =~ s/s$//;
            my $found = 0;
            for (my $count=0; $count < scalar @{$cglDisplayObj->{"cgl$objType"}} && !$found; $count++)
            {
              if ($cglDisplayObj->{"cgl$objType"}->[$count]->{type} eq $cgl)
              {
                # now find the Sub-Account associated with this account.
                my $subAcctFound = 0;
                for (my $count2=0; $count2 < scalar @{$cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}} && !$subAcctFound; $count2++)
                {
                  if ($cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}->[$count2]->{number} == $subAcct &&
                      $cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}->[$count2]->{type} eq $type)
                  {
                    $cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}->[$count2]->state($cglDisplayObj->{"cgl$objType"}->[$count]->{cglObjects}->{"cgl$objSingular" . "Objects"}->[$count2]->{state} eq "c" ? "o" : "c");
                    $subAcctFound = 1;
                  }
                }
                if (!$subAcctFound)
                {
                  $self->error("CGL Entry = '$cgl' found but Sub-Account = '$subAcct', type = '$type' not found for $objType!<br>\n");
                  return $doc;
                }
                $found = 1;
              }
            }

            if (!$found)
            {
              $self->error("CGL Entry = '$cgl' not found for $objType!<br>\n");
              return $doc;
            }
          }
        }
      }

      # write the data back to the file
      $data{cglDisplayObj} = $cglDisplayObj;  # make sure it is back in the data hash.
      eval "store (\\\%data, \$filename);";
      if ($@)
      {
        $self->error("Store of file = '$filename' failed!<br>\nError = '$@'.<br>\n");
        return undef;
      }
    }

    # redisplay the CGL
    $doc = $self->displayCGL(doc => $doc, filename => $filename, negativeFormat => $negativeFormat);
    if ($self->error)
    {
      $self->prefixError();
      return undef;
    }
  return $doc;
}

sub c_print
{
  my $self = shift;
  my %args = ( command => "display", @_ );
  my $command = $args{command};
  my $doc = $self->prepDoc;
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
    my $filename = $self->{input}->{filename};
    my $negativeFormat = $self->{input}->{negativeFormat};

    if (length $filename == 0 || ! -e $filename)
    {
      $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
      return $doc;
    }
    if ($negativeFormat < 0 || $negativeFormat > 2)
    {
      $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
      return $doc;
    }

    $doc = $self->displayCGL(doc => $doc, filename => $filename, negativeFormat => $negativeFormat, printable => "true");
    if ($self->error)
    {
      $self->prefixError();
      return undef;
    }
  return $doc;
}

# Helper functions

# displayCustomizationScreen
# takes: doc, error
# optional: firstTime
# returns: HTMLObject::Normal
sub displayCustomizationScreen
{
  my $self = shift;
  my %args = ( doc => undef, error => "", firstTime => 0, @_ );
  my $doc = $args{doc};
  my $error = $args{error};
  my $firstTime = $args{firstTime};
  my $ledger = $self->{input}->{ledger};
  my $stocks = $self->{input}->{stocks};
  my $assets = $self->{input}->{assets};
  my $invAccount = $self->{input}->{invAccount};
  my $negativeFormat = (exists $self->{input}->{negativeFormat} ? $self->{input}->{negativeFormat} : 0);
  my $orderBy = (exists $self->{input}->{orderBy} ? $self->{input}->{orderBy} : 0);
  my $ledgerCode = $self->{input}->{ledgerCode};
#  my $ledgerCode = $self->{sessionObj}->{store}->{ledgers}->{$ledger}->{code};
  my $phrases = $self->{langObj};
  my $ledgerDBObj = Portal::Accounting::DB::Ledger->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($ledgerDBObj->error)
  {
    $self->error($ledgerDBObj->errorMessage);
    return undef;
  }
  # get date info for current fiscal year to use in date select boxes.
  my $currFiscalYear = $ledgerDBObj->getCurrentFiscalYear($ledgerCode);
  if ($ledgerDBObj->error)
  {
    $self->error($ledgerDBObj->errorMessage);
    return undef;
  }
  my $led = $self->{sessionObj}->{store}->{ledgers}->{$ledgerCode};
  my $tmpStartDate = $led->start(month => 1, fiscalYear => $currFiscalYear);
  my $tmpEndDate = $led->end(month => 12, fiscalYear => $currFiscalYear);

  my $startDate = (exists $self->{input}->{startDate} ? $self->{input}->{startDate} : "$tmpStartDate");
  my $endDate = (exists $self->{input}->{endDate} ? $self->{input}->{endDate} : "$tmpEndDate");

  if (! defined $doc)
  {
    $self->error("doc must be defined!<br>\n");
    return undef;
  }

  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return undef;
  }

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->error)
  {
    $self->prefixError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->error("No valid colorScheme returned!<br>\n");
    return undef;
  }

  # create the needed css entries

  my $ledgerURI = $doc->encodeString(string => $ledger);

  my $execute = $phrases->map("execute");
  my $assetsPhrase = $phrases->map("Assets");
  my $stocksPhrase = $phrases->map("Stocks");
  my $startDatePhrase = $phrases->map("Start")." ".$phrases->map("Date");
  my $endDatePhrase = $phrases->map("End")." ".$phrases->map("Date");
  my $selectAll = $phrases->map("Select All");
  my $unSelectAll = $phrases->map("Un-Select All");
  my $unSelect = $phrases->map("unSelect");
  my $refresh = $phrases->map("refresh");
  my $negativeNumberFormatPhrase = $phrases->map("negativeNumberFormatPhrase");

  my $investmentJournalPhrase = $phrases->map("investmentJournal");
  my $orderByPhrase = $phrases->map("Order By");
  my $investmentAccountsPhrase = $phrases->map("Investment")." ".$phrases->map("Accounts");

  my $type0Checked = ($negativeFormat == 0 ? " checked" : "");
  my $type1Checked = ($negativeFormat == 1 ? " checked" : "");
  my $type2Checked = ($negativeFormat == 2 ? " checked" : "");

  my $order0Checked = ($orderBy == 0 ? " checked" : "");
  my $order1Checked = ($orderBy == 1 ? " checked" : "");

  my $displayManualEntriesChecked = (exists $self->{input}->{displayManualEntries} ? " checked" : ($firstTime ? " checked" : ""));

  my $displayItemObj = Portal::Accounting::DisplayItems->new(langObj => $self->{langObj});

  $doc->setLinkDecorations;

  # gather all Stocks
  my $stock = Portal::Accounting::Objects::SubAccount->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
  $stock->set('type', 70);
  my @Stocks = $stock->selectArrayByType();
  if ($stock->error)
  {
    $self->error($stock->errorMessage);
    return undef;
  }

  # gather all Assets
  $stock->set('type', 55);
  my @Assets = $stock->selectArrayByType();
  if ($stock->error)
  {
    $self->error($stock->errorMessage);
    return undef;
  }

  # gather all available Investment Accounts
  my $account = Portal::Accounting::Objects::Account->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $ledgerCode);
  $account->set('type', 3);
  my @investmentAccounts = $account->selectArrayByType();
  if ($account->error)
  {
    $self->error($account->errorMessage);
    return undef;
  }
  # gather all Capital Gains Accounts (8)
  $account->set('type', 8);
  my @capitalGainsAccounts = $account->selectArrayByType();
  if ($account->error)
  {
    $self->error($account->errorMessage);
    return undef;
  }

  # gather all Capital Loss Accounts (18)
  $account->set('type', 18);
  my @capitalLossAccounts = $account->selectArrayByType();
  if ($account->error)
  {
    $self->error($account->errorMessage);
    return undef;
  }
  my %startDatePicker = $self->{widgets}->generatePickerCode(type => "date", doc => $doc, form => "cgl_customize",
                          itemName => "startDate", itemValue => $startDate, year => $currFiscalYear,
                          phrase => $startDatePhrase, class => "menu", linkClass => "menu", width => 300, height => 250);
  if ($self->{widgets}->error)
  {
    $self->error($self->{widgets}->errorMessage);
    return undef;
  }

  my %endDatePicker = $self->{widgets}->generatePickerCode(type => "date", doc => $doc, form => "cgl_customize",
                          itemName => "endDate", itemValue => $endDate, year => $currFiscalYear,
                          phrase => $endDatePhrase, class => "menu", linkClass => "menu", width => 300, height => 250);
  if ($self->{widgets}->error)
  {
    $self->error($self->{widgets}->errorMessage);
    return undef;
  }

  my $siteDir = $self->{configObj}->{siteDir};

  # display input form with multiselect boxes for Assets and Stocks and a Date Selection Widget.
  $doc->setFocus("body");

  if (scalar @investmentAccounts == 0)
  {
    # give them the link to create the needed account(s) first.
    $doc->print(<<"END_OF_CODE");
You need to create and assign at least 1 investment account!<br>
Go <a href="$siteDir/js/Accounting/accounting/cgi-bin/create_account.cgi?ledger=$ledgerURI">here</a> to create an Investment Account.<br>
END_OF_CODE
    return $doc;
  }

  $doc->setFocus("javascript");
  $doc->print(<<"END_OF_CODE");
var toggleState = 0;  // initially hidden

function toggleAdvancedView()
{
  var d = document.cgl_customize;

  if (toggleState)
  {
    // hide the Advanced Options
    toggleState = 0;
    document.getElementById('invAcctHeader').style.display = 'none';
    document.getElementById('assetHeader').style.display = 'none';
    document.getElementById('stocksHeader').style.display = 'none';
    document.getElementById('invAcctColumn').style.display = 'none';
    document.getElementById('assetColumn').style.display = 'none';
    document.getElementById('stocksColumn').style.display = 'none';
    document.getElementById('displayManualEntries').style.display = 'none';
  }
  else
  {
    // display the Advanced Options
    toggleState = 1;
    document.getElementById('invAcctHeader').style.display = 'table-cell';
    document.getElementById('assetHeader').style.display = 'table-cell';
    document.getElementById('stocksHeader').style.display = 'table-cell';
    document.getElementById('invAcctColumn').style.display = 'table-cell';
    document.getElementById('assetColumn').style.display = 'table-cell';
    document.getElementById('stocksColumn').style.display = 'table-cell';
    document.getElementById('displayManualEntries').style.display = 'table-cell';
  }
}
END_OF_CODE

  $doc->setFocus("body");
  $doc->print(<<"END_OF_CODE");
<br>
<form name="cgl_customize" method="post" action="">
  <input type=hidden name="app" value="$self->{app}">
  <input type=hidden name="state" value="$self->{state}">
  <input type=hidden name="command" value="execute">
  <input type=hidden name="ledger" value="$ledger">
  <input type=hidden name="ledgerCode" value="$ledgerCode">
  <table border=0 cellpadding=1 cellspacing=1 class="menu">
    <tr>
      <th align="center">&nbsp;</th><th align="center" id="invAcctHeader" style="display: none;"><span class="menu">$investmentAccountsPhrase</span></th><th align="center" id="assetHeader" style="display: none;"><span class="menu">$assetsPhrase</span></th><th align="center" id="stocksHeader" style="display: none;"><span class="menu">$stocksPhrase</span></th>
    </tr>
    <tr>
      <td align="right">
END_OF_CODE
  $doc->print(%startDatePicker); $doc->print("<br>");
  $doc->print(%endDatePicker);
  $doc->print(<<"END_OF_CODE");
      </td>
      <td rowspan="5" align="center" id="invAcctColumn" style="display: none;">
        <select name="invAccount" size="10">
END_OF_CODE

  foreach my $accountObj (@investmentAccounts)
  {
    my $selected = ($accountObj->get("id") == $invAccount ? " selected" : "");
    $doc->print("          <option value=\"".$accountObj->get("id")."|".$accountObj->get("name")."\"$selected>".$accountObj->get("name")."</option>\n");
  }

  $doc->print(<<"END_OF_CODE");
        </select>
        <br>
        <input type="button" name="noAccounts" value="$unSelect" onclick="unSelectAll(document.cgl_customize.invAccount)">
      </td>
      <td rowspan="5" align="center" id="assetColumn" style="display: none;">
        <select name="assets" size="10" multiple>
END_OF_CODE

  my @tmpAssets = split /\0/, $assets;
  foreach my $assetObj (@Assets)
  {
    my $selected = (grep(/$assetObj->get("id")/, @tmpAssets) > 0 ? " selected" : ($firstTime ? " selected" : ""));
    $doc->print("          <option value=\"".$assetObj->get("id")."\"$selected>".$assetObj->get("name")."</option>\n");
  }

  $doc->print(<<"END_OF_CODE");
        </select>
        <br>
        <input type="button" name="allAssets" value="$selectAll" onclick="selectAll(document.cgl_customize.assets)">
        <br>
        <input type="button" name="noAssets" value="$unSelectAll" onclick="unSelectAll(document.cgl_customize.assets)">
      </td>
      <td rowspan="5" align="center" id="stocksColumn" style="display: none;">
        <select name="stocks" size="10" multiple>
END_OF_CODE

  my @tmpStocks = split /\0/, $stocks;
  foreach my $stockObj (@Stocks)
  {
    my $selected = (grep(/$stockObj->get("id")/, @tmpStocks) > 0 ? " selected" : "");
    $doc->print("          <option value=\"".$stockObj->get("id")."\"$selected>".$stockObj->get("name")."</option>\n");
  }

  $doc->print(<<"END_OF_CODE");
        </select>
        <br>
        <input type="button" name="allStocks" value="$selectAll" onclick="selectAll(document.cgl_customize.stocks)">
        <br>
        <input type="button" name="noStocks" value="$unSelectAll" onclick="unSelectAll(document.cgl_customize.stocks)">
      </td>
    </tr>
    <tr>
      <th align="center"><span class="menu">$negativeNumberFormatPhrase</span></th>
    </tr>
    <tr>
      <td align="left">
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="radio" name="negativeFormat" value="0"$type0Checked><font color="red">&nbsp;(</font>xxx.xx<font color="red">)</font>
        <br>
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="radio" name="negativeFormat" value="1"$type1Checked>&nbsp;<font color="red">-</font>xxx.xx
        <br>
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="radio" name="negativeFormat" value="2"$type2Checked>&nbsp;xxx.xx<font color="red">-</font>
      </td>
    </tr>
    <tr>
      <th align="center"><span class="menu">$orderByPhrase</span></th>
    </tr>
    <tr>
      <td align="left">
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="radio" name="orderBy" value="0"$order0Checked> ST / LT
        <br>
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="radio" name="orderBy" value="1"$order1Checked> Sub Account
      </td>
    </tr>
    <tr>
      <td align="center">
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="button" name="advancedView" value="Toggle Advanced View" onClick="toggleAdvancedView();">
      </td>
      <td colspan="3" align="right" id="displayManualEntries" style="display: none;">
        &nbsp;&nbsp;&nbsp;&nbsp;<input type="checkbox" name="displayManualEntries" value="1"$displayManualEntriesChecked> Display Manual Entries
      </td>
    </tr>
  </table>
  <input type="submit" name="execute" value="$execute">
</form>
</center>
END_OF_CODE

  if (length $error > 0)
  {
    $doc->print("<br><br><center><font color=\"red\">Error:&nbsp;</font>$error</center>\n");
  }

  return $doc;
}

# displayCGL
# requires: doc, filename, negativeFormat
# optional: printable
# returns: doc or undef if error
# This routine uses the storable object defined by filename to know what data it needs to display.  It then displays the CGL from the
# stored data and provides the user with an interface to selectively browse through the displayed data.
sub displayCGL
{
  my $self = shift;
  my %args = ( doc => undef, filename => "", negativeFormat => 0, printable => "false", @_ );
  my $doc = $args{doc};
  my $filename = $args{filename};
  my $printable = $args{printable};
  my $negativeFormat = $args{negativeFormat};
  my $ledger = $self->{input}->{ledger};
  my $ledgerCode = $self->{input}->{ledgerCode};
  #my $ledgerCode = $self->{sessionObj}->{store}->{ledgers}->{$ledger}->code;
  my $ledgerColor = $self->{sessionObj}->{store}->{ledgers}->{$ledger}->{color};
  my $displayItemObj = Portal::Accounting::DisplayItems->new(langObj => $self->{langObj});
  my $phrases = $self->{langObj};
  my @paddingValue;

  if (! defined $doc)
  {
    $self->error("doc must be defined!<br>\n");
    return $doc;
  }
  if (length $filename == 0 || ! -e $filename)
  {
    $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
    return $doc;
  }

  if ($printable !~ /^(true|false)$/)
  {
    $self->error("printable = '$printable' is invalid!<br>\n");
    return $doc;
  }
  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return $doc;
  }

  # bring into existance the storable object.
  my $args_ref;
  eval "\$args_ref = retrieve(\$filename);";
  if ($@)
  {
    $self->error("Retrieve of file = '$filename' failed!<br>\nError = '$@'.<br>\n");
    return undef;
  }
  my %data = %{$args_ref};

  my $cglDisplayObj = $data{cglDisplayObj};
  if (!defined $cglDisplayObj || !$cglDisplayObj->isValid)
  {
    $self->error("stored cglDisplayObj is invalid or does not exist!<br>\n");
    return $doc;
  }
  my $date = $data{date};
  my $startDate = $data{startDate};
  my $endDate = $data{endDate};
  my $invAccountString = ($data{invAccount} ? "<br>Investment Account: $data{invAccount}" : "");

  # get the users colorScheme to work with.
  my $colorScheme = $self->{applicationObj}->getCurrentColorScheme(userId => $self->{userObj}->{id});
  if ($self->{applicationObj}->error)
  {
    $self->prefixError($self->{applicationObj}->errorMessage);
    return undef;
  }
  if (!defined $colorScheme)
  {
    $self->error("No valid colorScheme returned!<br>\n");
    return undef;
  }

  my $refresh = $phrases->map("Refresh");
  my $printPhrase = $phrases->map("Print");

  my $ledgerURI = $doc->encodeString(string => $ledger);
  my $filenameURI = $doc->encodeString(string => $filename);

  my $url = "?app=$self->{app}&state=$self->{state}&command=\%s&ledger=\%s&filename=\%s&ledgerCode=$ledgerCode&negativeFormat=$negativeFormat";
  my $refreshUrl = sprintf($url, "update", $ledgerURI, $filenameURI);
  $refreshUrl .= "&updateWhat=refresh&number=";
  my $printUrl = sprintf($url, "print", $ledgerURI, $filenameURI);
  my $doAnotherUrl = sprintf($url, "display", $ledgerURI, $filenameURI);

  $doc->setTitle("Capital Gain/Loss: Ledger = '$ledger', '$startDate' to '$endDate'");
  $doc->setLinkDecorations;

  # generate any needed JavaScript functions
  $doc->setFocus("javascript");
  $doc->print(<<"END_OF_CODE");
function toggle(what, number)
{
  var d = document.mainForm;

  d.updateWhat.value = what;
  d.number.value = number;

  d.submit();
}
END_OF_CODE

  # now display the body.
  $doc->setFocus("body");
  if ($printable eq "false")
  {
    $doc->print(<<"END_OF_CODE");
<a href="$refreshUrl" onMouseOver="window.status='$refresh'">$refresh</a>&nbsp;&nbsp;|&nbsp;
<a href="$doAnotherUrl" onMouseOver="window.status='New View'">New View</a>&nbsp;&nbsp;|&nbsp;
<a href="$printUrl" onMouseOver="window.status='$printPhrase'">$printPhrase</a>
<hr>
<form name="mainForm">
<table border="0" cellpadding="0" cellspacing="0" width="100%">
  <tr>
    <td align="left" width="50%">
      <a href="#" onMouseOver="window.status='Open All CGL entries'" onClick="javascript: toggle('cgl', -1); return false;">Open</a>&nbsp;/&nbsp;
      <a href="#" onMouseOver="window.status='Close All CGL entries'" onClick="javascript: toggle('cgl', -2); return false;">Close</a> All Capital Gain/Loss Entries
    </td>
    <td align="right" width="50%">
      <a href="#" onMouseOver="window.status='Open All Sub Accounts'" onClick="javascript: toggle('sub-acct', -1); return false;">Open</a>&nbsp;/&nbsp;
      <a href="#" onMouseOver="window.status='Close All Sub Accounts'" onClick="javascript: toggle('sub-acct', -2); return false;">Close</a> All Sub Accounts
    </td>
  </tr>
</table>
END_OF_CODE
  }

  my $headerString = ($printable eq "true" ? "Return" : "Refresh");
  my $siteDir = $self->{configObj}->{siteDir};
  my $cglDisplayObj = $data{cglDisplayObj};
  my $output = "";

  my %grandTotals = ( cost => "0.00", market => "0.00", gainLoss => "0.00" );
  my %sectionTotals = ( "Stocks" => { cost => "0.00", market => "0.00", gainLoss => "0.00" },
                        "Assets" => { cost => "0.00", market => "0.00", gainLoss => "0.00" },
                        "Manual" => { gainLoss => "0.00" },
                      );

  my @objectsToProcess = ("Stocks");
  push @objectsToProcess, "Manual" if ($data{displayManualEntries});

  foreach my $objType (@objectsToProcess) #, "Assets"
  {
    # remove the trailing s, if it exists.
    (my $objTypeSingular = $objType) =~ s/s$//;

    # first output the html table and header.
    if ($objType eq "Stocks")
    {
      $output .= <<"END_OF_CODE";
<b>Stocks:</b>
<table border="0" cellspacing="2" cellpadding="0" width="100%">
  <tr>
    <th>&nbsp;</th><th align="left" valign="bottom" nowrap>SubAccount / Lot</th><th align="center" valign="bottom">Purchased</th><th align="right" valign="bottom">Sold</th><th align="right" valign="bottom">Cost&nbsp;&nbsp;</th><th align="right" valign="bottom">Sale<br>Value</th><th align="right" valign="bottom">Shares</th><th align="right" valign="bottom">Gain&nbsp;/<br>Loss&nbsp;&nbsp;</th><th align="right" valign="bottom">LT&nbsp;/<br>ST&nbsp;&nbsp;</th>
  </tr>
  <tr>
    <td align="center" colspan="9"><hr></td>
  </tr>
END_OF_CODE
    }
    elsif ($objType eq "Assets")
    {
      $output .= "<b>Assets</b>\n";
    }
    elsif ($objType eq "Manual")
    {
      $output .= <<"END_OF_CODE";
<b>Manual:</b>
<table border="0" cellspacing="2" cellpadding="0" width="100%">
  <tr>
    <th>&nbsp;</th><th align="left" valign="bottom" nowrap>SubAccount</th><th align="right" valign="bottom">Sold</th><th align="right" valign="bottom">Gain&nbsp;/<br>Loss&nbsp;&nbsp;</th><th align="right" valign="bottom">LT&nbsp;/<br>ST&nbsp;&nbsp;</th>
  </tr>
  <tr>
    <td align="center" colspan="5"><hr></td>
  </tr>
END_OF_CODE
    }

    # now we need to walk over the CGLDisplayObject and display all CGLObjects that are stored in it.
    foreach my $cglObj (@{$cglDisplayObj->{"cgl$objType"}})
    {
      # update the section totals.
      if ($objType =~ /^(Stocks|Assets)$/)
      {
        $sectionTotals{$objType}{cost} += $cglObj->{cost};
        $sectionTotals{$objType}{market} += $cglObj->{market};
      }
      $sectionTotals{$objType}{gainLoss} += $cglObj->{gainLoss};
      # display the state of the CGLObject (+ or -) but only if there are Assets or Stocks to display for it.
      $output .= "  <tr>\n    <td align=\"right\">";
      if ((scalar @{$cglObj->{cglObjects}->{"cgl$objTypeSingular" . "Objects"}} > 0) && $printable eq "false")
      {
        my $image = ($cglObj->state eq "c" ? "plus.gif" : "minus.gif");
        my $phrase = ($cglObj->state eq "c" ? "Expand" : "Contract");
        $output .= "<a href=\"#\" onMouseOver=\"window.status='$phrase'\" onClick=\"javascript: toggle('cgl', '$objType|$cglObj->{type}'); return false;\"><img border=\"0\" src=\"$siteDir/js/Accounting/accounting/images/$image\" alt=\"$phrase\"></a>&nbsp;";
      }
      else
      {
        $output .= "&nbsp;";
      }
      $output .= "</td>\n";

      # now display the CGLObject info
      if ($cglObj->{state} eq "c")
      {
        (my $cglName = ($cglObj->type eq "both" ? "ST & LT Capital GL" : ($cglObj->type eq "ST" ? "Short Term Capital GL" : "Long Term Capital GL"))) =~ s/&/&amp;/g;
        if ($objType eq "Stocks")
        {
          $output .= "    <td align=\"left\" colspan=\"2\" nowrap>$cglName</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->{cost}, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->market, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->gainLoss, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
        }
        elsif ($objType eq "Assets")
        {
        }
        elsif ($objType eq "Manual")
        {
          $output .= "    <td align=\"left\" colspan=\"2\" nowrap>$cglName</td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->gainLoss, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
        }
        $output .= "  </tr>\n";
      }
      else
      {
        (my $cglName = ($cglObj->type eq "both" ? "ST & LT Capital GL" : ($cglObj->type eq "ST" ? "Short Term Capital GL" : "Long Term Capital GL"))) =~ s/&/&amp;/g;
        if ($objType eq "Stocks")
        {
          $output .= "    <td align=\"left\" colspan=\"2\" nowrap>$cglName</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";  # the totals go at the bottom of the list of Stocks and Assets.
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "  </tr>\n";

          # we also have to display all Stocks and their status's here.
          foreach my $stockObj (@{$cglObj->{cglObjects}->{cglStockObjects}})
          {
            $output .= "  <tr>\n";
            $output .= "    <td align=\"right\">&nbsp;</td>\n";
            $output .= "    <td align=\"left\" nowrap>";
            if (scalar @{$stockObj->{cglStockLotObjects}} > 0 && $printable eq "false")
            {
              my $image = ($stockObj->state eq "c" ? "plus.gif" : "minus.gif");
              my $phrase = ($stockObj->state eq "c" ? "Expand" : "Contract");
              $output .= "<a href=\"#\" onMouseOver=\"window.status='$phrase'\" onClick=\"javascript: toggle('sub-acct', '$objType|$cglObj->{type}|$stockObj->{id}|$stockObj->{type}'); return false;\"><img border=\"0\" src=\"$siteDir/js/Accounting/accounting/images/$image\" alt=\"$phrase\"></a>&nbsp;";
            }
            else
            {
              $output .= "&nbsp;&nbsp;&nbsp;&nbsp;";
            }
            (my $stockName = $stockObj->{name}) =~ s/&/&amp;/g;
            $output .= "$stockName - $stockObj->{type}</td>\n";

            # now output the Stock info
            if ($stockObj->{state} eq "c")
            {
              $output .= "    <td align=\"center\">&nbsp;</td>\n";
              $output .= "    <td align=\"center\">&nbsp;</td>\n";
              @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->cost, negativeFormat => $negativeFormat);
              $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
              @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->market, negativeFormat => $negativeFormat);
              $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
              $output .= "    <td align=\"right\">$stockObj->{shares}</td>\n";
              @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->gainLoss, negativeFormat => $negativeFormat);
              $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "  </tr>\n";
            }
            else
            {
              # display the header info and all lots for this stock
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "  </tr>\n";

              # now display all lots for this account/stock pair.
              foreach my $lotObj (@{$stockObj->{cglStockLotObjects}})
              {
                $output .= "  <tr>\n";
                $output .= "    <td>&nbsp;</td>\n";
                my $spaceString = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;";
                $spaceString .= "&nbsp;&nbsp;" if ($printable eq "true");
                $output .= "    <td align=\"left\">$spaceString$lotObj->{refNum}</td>\n";  # we need to make this link somewhere, don't know where yet.
                $output .= "    <td align=\"center\">$lotObj->{purchased}</td>\n";
                $output .= "    <td align=\"right\">$lotObj->{sold}</td>\n";
                @paddingValue = $displayItemObj->displayMoney(amount => $lotObj->cost, negativeFormat => $negativeFormat);
                $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
                @paddingValue = $displayItemObj->displayMoney(amount => $lotObj->market, negativeFormat => $negativeFormat);
                $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
                $output .= "    <td align=\"right\">$lotObj->{shares}</td>\n";
                @paddingValue = $displayItemObj->displayMoney(amount => "$lotObj->{gainLoss}", negativeFormat => $negativeFormat);
                $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
                $output .= "    <td align=\"right\" style=\"padding-right: 1em;\"><b>$lotObj->{type}</b></td>\n";
                $output .= "  </tr>\n";
              }

              # display the totals for the stock after all lots.
              $output .= "  <tr>\n";
              $output .= "    <td>&nbsp;</td>\n";
              my $spaceString = "&nbsp;&nbsp;&nbsp;";
              $spaceString .= "&nbsp;" if ($printable eq "true");
              $output .= "    <td><b>$spaceString$stockName Total - $stockObj->{type}</b></td>\n";
              $output .= "    <td align=\"center\">&nbsp;</td>\n";
              $output .= "    <td align=\"center\">&nbsp;</td>\n";
              @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->cost, negativeFormat => $negativeFormat);
              $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
              @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->market, negativeFormat => $negativeFormat);
              $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
              $output .= "    <td align=\"right\"><b>$stockObj->{shares}</b></td>\n";
              @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->gainLoss, negativeFormat => $negativeFormat);
              $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
              $output .= "    <td align=\"right\">&nbsp;</td>\n";
              $output .= "  </tr>\n";
            }
          }
        }
        elsif ($objType eq "Assets")
        {
          # display the Assets
          $output .= "  <tr>\n    <td colspan=\"10\">&nbsp;</td>\n  </tr>\n" if (scalar @{$cglObj->{cglObjects}->{cglAssetObjects}} > 0);
          foreach my $assetObj (@{$cglObj->{cglObjects}->{cglAssetObjects}})
          {
            $output .= "  <tr>\n";
            $output .= "    <td align=\"right\">&nbsp;</td>\n";
            $output .= "    <td align=\"left\" nowrap>&nbsp;&nbsp;&nbsp;";
            (my $assetName = $assetObj->{name}) =~ s/&/&amp;/g;
            $output .= "$assetName</td>\n";

            # now output the Asset info
            $output .= "    <td align=\"center\">$assetObj->{purchased}</td>\n";
            $output .= "    <td align=\"center\">$assetObj->{sold}</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $assetObj->cost, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $assetObj->market, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            $output .= "    <td align=\"right\">&nbsp;</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $assetObj->gainLoss, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            $output .= "    <td align=\"right\" style=\"padding-right: 1em;\"><b>$assetObj->{type}</b></td>\n";
            $output .= "  </tr>\n";
          }
        }
        elsif ($objType eq "Manual")
        {
          $output .= "    <td align=\"left\" colspan=\"2\" nowrap>$cglName</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";  # the totals go at the bottom of the list of Stocks and Assets.
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "  </tr>\n";
          foreach my $manualObj (@{$cglObj->{cglObjects}->{cglManualObjects}})
          {
            $output .= "  <tr>\n";
            $output .= "    <td align=\"right\">&nbsp;</td>\n";
            $output .= "    <td align=\"left\" nowrap>&nbsp;&nbsp;&nbsp;";
            (my $manualName = $manualObj->{name}) =~ s/&/&amp;/g;
            $output .= "$manualName</td>\n";

            # now output the Manual Entry info
            $output .= "    <td align=\"right\">$manualObj->{sellDate}</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $manualObj->gainLoss, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            $output .= "    <td align=\"right\" style=\"padding-right: 1em;\"><b>$manualObj->{type}</b></td>\n";
            $output .= "  </tr>\n";
          }
        }

        # display the Totals for the Account.
        $output .= "  <tr>\n";
        if ($objType eq "Stocks" || $objType eq "Assets")
        {
          $output .= "    <td>&nbsp;</td>\n";
          my $spaceString = "";
          $output .= "    <td align=\"left\" colspan=\"2\"><b>$spaceString$cglName Total</b></td>\n";
          $output .= "    <td>&nbsp;</td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->cost, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->market, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->gainLoss, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
        }
        elsif ($objType eq "Manual")
        {
          $output .= "    <td>&nbsp;</td>\n";
          my $spaceString = "";
          $output .= "    <td align=\"left\" colspan=\"2\"><b>$spaceString$cglName Total</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $cglObj->gainLoss, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><span style=\"border-top: solid thin;\"><b>" . $paddingValue[1] . "</b></span></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
        }
        $output .= "  </tr>\n";
      }
    }

    # this row will be the calibration row that causes the table to properly proportion the cells.
    $output .= "  <tr>\n";
    if ($objType eq "Stocks" || $objType eq "Assets")
    {
      $output .= "    <td width=\"2%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"23%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"15%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
    }
    elsif ($objType eq "Manual")
    {
      $output .= "    <td width=\"2%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"23%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"25%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"25%\" align=\"right\">&nbsp;</td>\n";
      $output .= "    <td width=\"25%\" align=\"right\">&nbsp;</td>\n";
    }
    $output .= "  </tr>\n";

    if (scalar @{$cglDisplayObj->{"cgl$objType"}} == 0)  # This should never happen!
    {
      $output .= "  <tr>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      if ($objType eq "Stocks" || $objType eq "Assets")
      {
        $output .= "    <td colspan=\"8\" align=\"left\">No CGLObjects returned!</td>\n";
      }
      elsif ($objType eq "Manual")
      {
        $output .= "    <td colspan=\"4\" align=\"left\">No CGLObjects returned!</td>\n";
      }
      $output .= "  </tr>\n";
    }
    else
    {
      # display the totals.
      $output .= "</table>\n";
      $output .= "<br>\n";
      $output .= "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\" width=\"100%\">\n";
      $output .= "  <tr>\n";
      $output .= "    <td align=\"right\" width=\"25%\"><b>$objType Totals:</b></td>\n";
      $output .= "    <td align=\"left\" width=\"75%\">&nbsp;</td>\n";
      $output .= "  </tr>\n";
      $output .= "  <tr>\n";
      $output .= "    <td colspan=\"2\">&nbsp;</td>\n";
      $output .= "  </tr>\n";

      foreach my $totalName (keys %{$sectionTotals{$objType}})
      {
        $sectionTotals{$objType}{$totalName} = sprintf "%.2f", $sectionTotals{$objType}{$totalName};
      }

      if ($objType =~ /^(Stocks|Assets)$/)
      {
        $output .= "  <tr>\n";
        $output .= "    <td align=\"right\" width=\"25%\"><b>Total Cost</b></td>\n";
        @paddingValue = $displayItemObj->displayMoney(amount => $sectionTotals{$objType}{cost}, negativeFormat => $negativeFormat);
        $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;\$" . $paddingValue[1] . "</td>\n";
        $output .= "  </tr>\n";
        $output .= "  <tr>\n";
        $output .= "    <td align=\"right\" width=\"25%\"><b>Total Sale Value</b></td>\n";
        @paddingValue = $displayItemObj->displayMoney(amount => $sectionTotals{$objType}{market}, negativeFormat => $negativeFormat);
        $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;\$" . $paddingValue[1] . "</td>\n";
        $output .= "  </tr>\n";
      }
      $output .= "  <tr>\n";
      $output .= "    <td align=\"right\" width=\"25%\"><b>Total Gain / Loss</b></td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $sectionTotals{$objType}{gainLoss}, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;\$" . $paddingValue[1] . "</td>\n";
      $output .= "  </tr>\n";

      $grandTotals{cost} += $sectionTotals{$objType}{cost};
      $grandTotals{market} += $sectionTotals{$objType}{market};
      $grandTotals{gainLoss} += $sectionTotals{$objType}{gainLoss};
    }
    $output .= "</table>\n";
    $output .= "<hr />\n";
  }  # end of $objType loop.

  # now I need to display the overall Grand Totals.
  foreach my $name (keys %grandTotals)
  {
    $grandTotals{$name} = sprintf("%.2f", $grandTotals{$name});
  }
  # display the grand totals.
  $output .= "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\" width=\"100%\">\n";
  $output .= "  <tr>\n";
  $output .= "    <td align=\"right\" width=\"25%\"><b>Grand Totals:</b></td>\n";
  $output .= "    <td align=\"left\" width=\"75%\">&nbsp;</td>\n";
  $output .= "  </tr>\n";
  $output .= "  <tr>\n";
  $output .= "    <td colspan=\"2\">&nbsp;</td>\n";
  $output .= "  </tr>\n";

  $output .= "  <tr>\n";
  $output .= "    <td align=\"right\" width=\"25%\"><b>Total Cost</b></td>\n";
  @paddingValue = $displayItemObj->displayMoney(amount => $grandTotals{cost}, negativeFormat => $negativeFormat);
  $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;<b>\$" . $paddingValue[1] . "</b></td>\n";
  $output .= "  </tr>\n";
  $output .= "  <tr>\n";
  $output .= "    <td align=\"right\" width=\"25%\"><b>Total Sale Value</b></td>\n";
  @paddingValue = $displayItemObj->displayMoney(amount => $grandTotals{market}, negativeFormat => $negativeFormat);
  $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;<b>\$" . $paddingValue[1] . "</b></td>\n";
  $output .= "  </tr>\n";
  $output .= "  <tr>\n";
  $output .= "    <td align=\"right\" width=\"25%\"><b>Total Gain / Loss</b></td>\n";
  @paddingValue = $displayItemObj->displayMoney(amount => $grandTotals{gainLoss}, negativeFormat => $negativeFormat);
  $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;<b>\$" . $paddingValue[1] . "</b></td>\n";
  $output .= "  </tr>\n";

  $doc->print($output);  # output the CGL data into the HTMLObject.

  $doc->print("</table>\n");  # close the table out.
  $doc->print("</form>\n");  # close the table out.

  if ($printable eq "true")
  {
    # generate the onLoad Code that will trigger a print of the page.
    my $onLoadCode = "window.print();\n";
    $doc->setOnload(code => $onLoadCode);
  }

  return $doc;
}

=back

=cut

1;
__END__

=head1 Exported FUNCTIONS (non-Inline POD)

  HTMLObject run(command)
    Runs the specified command and returns the resulting HTMLObject.


=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

PC & Web Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
