# Portfolio.pm - The Object Class that provides a Portfolio Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: Portfolio.pm,v 1.6 2005/07/22 17:54:56 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Investments::Portfolio;
use strict;
use HTMLObject::Widgets;
use Portal::Accounting::BaseL;
use Portal::Accounting::DB::Main;
use Portal::Accounting::DB::Security;
use Portal::Accounting::DisplayItems;
use Portal::Data::Config;
use Portal::Accounting::Objects::Account;
use Portal::Accounting::Objects::Stock;
use Portal::Accounting::Objects::PortfolioObject;
use Portal::Accounting::Objects::PortfolioAccountObject;
use Portal::Accounting::Objects::PortfolioLotObject;
use Portal::Accounting::Objects::PortfolioStockObject;
use POSIX qw(tmpnam);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.6 $';

# global variables.
# modify to represent the commands this state provides.
my %commands;
%commands = (
  display => "Portfolio",
  execute => "Portfolio",
  update => "portfolio",
  print => "Portfolio",
  executeFile => "Portfolio"
);

=head1 NAME

Portfolio - Object used to build a Portfolio Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Portfolio;
  my $obj = Portal::Accounting::Portfolio->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

Portfolio is a Portfolio class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::Portfolio module.
 See Portal::AppState(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{commands} = \%commands;
  $self->{widgets} = HTMLObject::Widgets->new();
  if ($self->{widgets}->error)
  {
    $self->error($self->{widgets}->errorMessage);
    return $self;
  }

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub c_execute
{
  my $self = shift;
  my $doc = $self->prepDoc(@_);
  # validate input
  my $dateYear = $self->{methods}->getCurrentDate(format => "%Y");
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $invAccounts = $self->{input}->{invAccounts};
  my $stocks = $self->{input}->{stocks};
  my $date = $self->{input}->{date};
  my $negativeFormat = $self->{input}->{negativeFormat};
  my $displayZeroShares = (exists $self->{input}->{displayZeroShares} ? 1 : 0);
  my $portfolioPhrase = $self->{langObj}->map("Portfolio");

  my @invAccounts = split /\0/, $invAccounts;
  my @stocks = split /\0/, $stocks;

  my $errorString = "";
  if (scalar @invAccounts == 0)
  {
    $errorString .= "You must select an Investment Account!<br>\n";
  }
  if ($date !~ /^\d{4}(-|\/)\d{2}(-|\/)\d{2}$/)
  {
    $errorString .= "date = '$date' is invalid!<br>\n";
  }
  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $errorString .= "negativeFormat = '$negativeFormat' is invalid!<br>\n";
  }

  if (length $errorString > 0)
  {
    $doc = $self->displayCustomizationScreen(doc => $doc, error => $errorString);
    if ($self->error)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }
  else
  {
    my %times = ();
    $times{start} = time();
    my $dbObj = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
    if ($dbObj->error)
    {
      $self->error($dbObj->errorMessage);
      return undef;
    }

    # gather the specified data and store in a storable object.

    # first we need to get the info for the specified accounts and sub accounts to pass into the data gathering routine.
    my $account = Portal::Accounting::Objects::Account->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $ledgerCode);
    $account->set("type", 3);
    my @invAccountObjs = $account->selectArrayByType();
    if ($account->error) { $self->error($account->errorMessage); }
    $times{accounts} = time();
    #!!! should we include bonds and Asset types?
    my $stock = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
    my @stockObjs = $stock->selectArrayByType();
    if ($stock->error) { $self->error($stock->errorMessage); }
    $times{subAccounts} = time();
    my $securitiesObj = Portal::Accounting::DB::Security->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
    if ($securitiesObj->error)
    {
      $self->error($securitiesObj->errorMessage);
      return undef;
    }
    my $portfolioObj = $self->getPortfolioData(accounts => \@invAccountObjs, stocks => \@stockObjs, date => $date, ledgerCode => $ledgerCode,
                                       displayZeroShares => $displayZeroShares, year => $dateYear);
    $times{getPortfolio} = time();
    if (!defined $portfolioObj)
    {
      if ($self->{errorCode} == -1)
      {
        $doc = $self->displayCustomizationScreen(doc => $doc, error => "No Portfolio Data returned for the selected Accounts and Stocks!");
      }
      else
      {
        $doc = $self->displayCustomizationScreen(doc => $doc, error => "Portfolio Object not created!");
      }
      return $doc;
    }
    elsif ($portfolioObj->error)
    {
      $doc->print($portfolioObj->errorMessage);
      return $doc;
    }

    my %data = ( portfolioObj => $portfolioObj, date => $date, invAccountObjs => \@invAccountObjs, stockObjs => \@stockObjs, displayZeroShares => $displayZeroShares, times => \%times );
    my $filename = tmpnam();
    $self->{sessionObj}->{store}->{$filename} = \%data;

    # call the display routine giving it the name of the file to work with.
    $doc = $self->displayPortfolio(doc => $doc, filename => $filename, negativeFormat => $negativeFormat);
    if ($self->error)
    {
      $self->prefixError();
      return undef;
    }
  }
  if ($self->error)
  {
    $doc->print($self->errorMessage);
  }
  return $doc;
}

sub c_update
{
  my $self = shift;
  my $doc = $self->prepDoc(@_);
  my $filename = $self->{input}->{filename};
  my $updateWhat = $self->{input}->{updateWhat};
  my $number = $self->{input}->{number};
  my $negativeFormat = $self->{input}->{negativeFormat};

  if (length $filename == 0)
  {
    $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
    return $doc;
  }

  if ($updateWhat !~ /^(account|stock|refresh)$/)
  {
    $self->error("updateWhat = '$updateWhat' is invalid!<br>\n");
    return $doc;
  }
  if (length $number == 0 && $updateWhat ne "refresh")
  {
    $self->error("number = '$number' is invalid!<br>\n");
    return $doc;
  }
  if ($updateWhat eq "stock" && ($number != -1 && $number != -2 && scalar (split /\|/, $number) != 2))
  {
    $self->error("number = '$number' is invalid for Toggling a Stock!<br>\n");
    return $doc;
  }
  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return $doc;
  }

  if ($updateWhat ne "refresh")
  {
    my %data = %{$self->{sessionObj}->{store}->{$filename}};

    my $portfolioObj = $data{portfolioObj};
    if (!defined $portfolioObj || !$portfolioObj->isValid)
    {
      $self->error("stored portfolioObj is invalid or does not exist!<br>\n");
      return $doc;
    }

    # find the account or stock to toggle
    if ($updateWhat eq "account")
    {
      if ($number == -1 || $number == -2)
      {
        for (my $count=0; $count < scalar @{$portfolioObj->{accounts}}; $count++)
        {
          $portfolioObj->{accounts}->[$count]->status($number == -1 ? "o" : "c");
        }
      }
      else  # normal behaviour
      {
        my $account = $number;
        my $found = 0;
        for (my $count=0; $count < scalar @{$portfolioObj->{accounts}} && !$found; $count++)
        {
          if ($portfolioObj->{accounts}[$count]->{accountNum} == $account)
          {
            $portfolioObj->{accounts}->[$count]->status($portfolioObj->{accounts}->[$count]->{status} eq "c" ? "o" : "c");
            $found = 1;
          }
        }

        if (!$found)
        {
          $self->error("Account = '$account' not found!<br>\n");
          return $doc;
        }
      }
    }
    elsif ($updateWhat eq "stock")
    {
      if ($number == -1 || $number == -2)
      {
        # loop through all Accounts (open or closed) and toggle the flag of all stocks to either be open or closed.
        for (my $count=0; $count < scalar @{$portfolioObj->{accounts}}; $count++)
        {
          $portfolioObj->{accounts}->[$count]->status("o");  # force all Accounts to be open when we deal with all Stocks.
          for (my $count2=0; $count2 < scalar @{$portfolioObj->{accounts}->[$count]->{stocks}}; $count2++)
          {
            $portfolioObj->{accounts}->[$count]->{stocks}->[$count2]->status($number == -1 ? "o" : "c");
          }
        }
      }
      else  # normal behaviour
      {
        (my $account, my $stock) = split /\|/, $number;
        my $found = 0;
        for (my $count=0; $count < scalar @{$portfolioObj->{accounts}} && !$found; $count++)
        {
          if ($portfolioObj->{accounts}->[$count]->{accountNum} == $account)
          {
            # now find the Stock associated with this account.
            my $stockFound = 0;
            for (my $count2=0; $count2 < scalar @{$portfolioObj->{accounts}->[$count]->{stocks}} && !$stockFound; $count2++)
            {
              if ($portfolioObj->{accounts}->[$count]->{stocks}->[$count2]->{subAccountNum} == $stock)
              {
                $portfolioObj->{accounts}->[$count]->{stocks}->[$count2]->status($portfolioObj->{accounts}->[$count]->{stocks}->[$count2]->{status} eq "c" ? "o" : "c");
                $stockFound = 1;
              }
            }
            if (!$stockFound)
            {
              $self->error("Account = '$account' found but Stock = '$stock' not found!<br>\n");
              return $doc;
            }
            $found = 1;
          }
        }

        if (!$found)
        {
          $self->error("Account = '$account' not found!<br>\n");
          return $doc;
        }
      }
    }

    $data{portfolioObj} = $portfolioObj;  # make sure it is back in the data hash.
    $self->{sessionObj}->{store}->{$filename} = \%data;
  }

  # redisplay the Portfolio
  $doc = $self->displayPortfolio(doc => $doc, filename => $filename, negativeFormat => $negativeFormat);
  if ($self->error)
  {
    $doc->print($self->errorMessage);
  }
  return $doc;
}

sub c_print
{
  my $self = shift;
  my $doc = $self->prepDoc(@_);
  my $filename = $self->{input}->{filename};
  my $negativeFormat = $self->{input}->{negativeFormat};

  if (length $filename == 0) # || ! -e $filename)
  {
    $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
    return $doc;
  }
  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return $doc;
  }

  $doc = $self->displayPortfolio(doc => $doc, filename => $filename, negativeFormat => $negativeFormat, printable => "true");
  if ($self->error)
  {
    $doc->print($self->errorMessage);
  }
  return $doc;
}

sub c_executeFile
{
  my $self = shift;
  my $dateYear = $self->{methods}->getCurrentDate(format => "%Y");
  my $ledgerCode = $self->{input}->{ledgerCode};
  my %times = ();
  my $filename = $self->{input}->{filename};
  my $date = $self->{input}->{date};
  my $negativeFormat = $self->{input}->{negativeFormat};

  my $doc = $self->prepDoc(@_);
  if ($date !~ /^\d{4}(-|\/)\d{2}(-|\/)\d{2}$/)
  {
    $self->error("date = '$date' is invalid!<br>\n");
    return $doc;
  }
  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return $doc;
  }
  if (length $filename == 0)
  {
    $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
    return $doc;
  }

  my %data = %{$self->{sessionObj}->{store}->{$filename}};

  $times{start} = $times{accounts} = $times{subAccounts} = time();
  my $securitiesObj = Portal::Accounting::DB::Security->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($securitiesObj->error)
  {
    $self->error($securitiesObj->errorMessage);
    return undef;
  }
  my $portfolioObj = $self->getPortfolioData(accounts => $data{invAccountObjs}, stocks => $data{stockObjs}, date => $date, ledgerCode => $ledgerCode,
                                    displayZeroShares => $data{displayZeroShares}, year => $dateYear);
  $times{getPortfolio} = time();
  if (!defined $portfolioObj)
  {
    if ($securitiesObj->error)
    {
      $self->error($securitiesObj->errorMessage);
      return undef;
    }
    if ($securitiesObj->{errorCode} == -1)
    {
      $doc = $self->displayCustomizationScreen(doc => $doc, error => "No Portfolio Data returned for the selected Accounts and Stocks!");
    }
    else
    {
      $doc = $self->displayCustomizationScreen(doc => $doc, error => "Portfolio Object not created!");
    }
    if ($self->error)
    {
      $self->prefixError();
      return undef;
    }
    return $doc;
  }
  else
  {
    if ($portfolioObj->error)
    {
      $self->error($portfolioObj->errorMessage);
      return undef;
    }
  }

  $data{portfolioObj} = $portfolioObj;
  $data{date} = $date;
  $data{times} = \%times;
  $self->{sessionObj}->{store}->{$filename} = \%data;

  # display the Portfolio
  $doc = $self->displayPortfolio(doc => $doc, filename => $filename, negativeFormat => $negativeFormat);
  if ($self->error)
  {
    $self->prefixError();
    return undef;
  }
}


sub c_display
{
  my $self = shift;
  my %args = ( command => $self->{input}->{command}, @_ );
  my $command = $args{command};
  my $doc = $self->prepDoc(@_);
  $doc = $self->displayCustomizationScreen(doc => $doc, firstTime => 1);
  if ($self->error)
  {
    $doc->print($self->errorMessage);
  }
  return $doc;
}

# Helper functions

# displayCustomizationScreen
# takes: doc, error
# optional: firstTime
# returns: HTMLObject::Normal
sub displayCustomizationScreen
{
  my $self = shift;
  my %args = ( doc => undef, error => "", firstTime => 0, @_ );
  my $doc = $args{doc};
  my $error = $args{error};
  my $firstTime = $args{firstTime};
  my $ledger = $self->{input}->{ledger};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $invAccounts = $self->{input}->{invAccounts};
  my $tmpDate = $self->{methods}->getCurrentDate();
  my $dateYear = $self->{methods}->getCurrentDate(format => "%Y");
  my $date = (exists $self->{input}->{date} ? $self->{input}->{date} : "$tmpDate");
  my $stocks = $self->{input}->{stocks};
  my $negativeFormat = (exists $self->{input}->{negativeFormat} ? $self->{input}->{negativeFormat} : 0);
  my $displayZeroShares = (exists $self->{input}->{displayZeroShares} ? 1 : 0);
  my $dbObj = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj});
  if ($dbObj->error)
  {
    $self->error($dbObj->errorMessage);
    return undef;
  }

  if (! defined $doc)
  {
    $doc = $self->prepDoc(@_);
  }

  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return undef;
  }

  my $ledgerURI = $doc->encodeString(string => $ledger);

  my $title = $self->{langObj}->map("portfolioCustomizationTitle");
  my $execute = $self->{langObj}->map("Execute");
  my $investmentAccounts = $self->{langObj}->map("Investment")." ".$self->{langObj}->map("Accounts");
  my $stocksPhrase = $self->{langObj}->map("Stocks");
  my $datePhrase = $self->{langObj}->map("Date");
  my $selectAll = $self->{langObj}->map("Select All");
  my $unSelectAll = $self->{langObj}->map("Un-Select All");
  my $refresh = $self->{langObj}->map("Refresh");
  my $negativeNumberFormatPhrase = $self->{langObj}->map("negativeNumberFormatPhrase");
  my $stockPricesPhrase = $self->{langObj}->map("stockPrices");
  my $investmentJournalPhrase = $self->{langObj}->map("Investment Journal");
  my $displayZeroSharesPhrase = $self->{langObj}->map("displayZeroShares");

  my $type0Checked = ($negativeFormat == 0 ? " checked" : "");
  my $type1Checked = ($negativeFormat == 1 ? " checked" : "");
  my $type2Checked = ($negativeFormat == 2 ? " checked" : "");

  my $zeroSharesSelected = ($displayZeroShares == 1 ? " checked" : "");

  my $displayItemObj = Portal::Accounting::DisplayItems->new(langObj => $self->{langObj});

  $doc->setLinkDecorations;

  # gather all Investment Accounts
  my $account = Portal::Accounting::Objects::Account->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $ledgerCode);
  $account->set('type', 3);
  my @InvAccounts = $account->selectArrayByType;
  if ($account->error)
  {
    $self->error($account->errorMessage);
  }

  # gather all Stocks
  my $stock = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
  my @Stocks = $stock->selectArrayByType;
  if ($stock->error)
  {
    $self->error($stock->errorMessage);
  }

  my %datePicker = $self->{widgets}->generatePickerCode(type => "date", doc => $doc, form => "mainForm",
                          itemName => "date", itemValue => $date, year => $dateYear,
                          phrase => $datePhrase, class => "menu", linkClass => "menu", width => 300, height => 250);
  if ($self->{widgets}->error)
  {
    $self->setError($self->{widgets}->errorMessage);
    return undef;
  }

  my $siteDir = $self->{configObj}->{siteDir};

  # display input form with multiselect boxes for Accounts and Stocks and a Date Selection Widget.
  $doc->setFocus("body");
  $doc->print(<<"END_OF_CODE");
<form name="mainForm" method="post" action="$siteDir/cgi-bin/index.cgi">
  <input type=hidden name="app" value="$self->{app}">
  <input type=hidden name="state" value="$self->{state}">
  <input type=hidden name="command" value="execute">
  <input type=hidden name="ledger" value="$ledger">
  <input type=hidden name="ledgerCode" value="$ledgerCode">
  <table border=0 cellpadding=1 cellspacing=1 class="menu">
    <tr>
      <td align="center" colspan="3">
END_OF_CODE
  $doc->print(%datePicker);
  $doc->print(<<"END_OF_CODE");
      </td>
    </tr>
    <tr>
      <th align="center"><span class="menu">$investmentAccounts</span></th>
      <th align="center"><span class="menu">$stocksPhrase</span></th>
      <th align="center"><span class="menu">$negativeNumberFormatPhrase</span></th>
    </tr>
    <tr>
      <td rowspan="3" align="center">
        <select name="invAccounts" size="10" multiple>
END_OF_CODE

  my @tmpInvAccounts = split /\0/, $invAccounts;
  foreach my $acctObj (@InvAccounts)
  {
    my $selected = (grep(/$acctObj->{number}/, @tmpInvAccounts) > 0 ? " selected" : ($firstTime ? " selected" : ""));
    $doc->print("          <option value=\"".$acctObj->get("id")."\"$selected>".$acctObj->get('name')."</option>\n");
  }

  $doc->print(<<"END_OF_CODE");
        </select>
        <br>
        <input type="button" name="allAccounts" value="$selectAll" onclick="selectAll(document.mainForm.invAccounts)">
        <br>
        <input type="button" name="noAccounts" value="$unSelectAll" onclick="unSelectAll(document.mainForm.invAccounts)">
      </td>
      <td rowspan="3" align="center">
        <select name="stocks" size="10" multiple>
END_OF_CODE

  my @tmpStocks = split /\0/, $stocks;
  foreach my $stockObj (@Stocks)
  {
    my $selected = (grep(/$stockObj->{number}/, @tmpStocks) > 0 ? " selected" : "");
    $doc->print("          <option value=\"".$stockObj->get('id')."\"$selected>".$stockObj->get("name")."</option>\n");
  }

  $doc->print(<<"END_OF_CODE");
        </select>
        <br>
        <input type="button" name="allStocks" value="$selectAll" onclick="selectAll(document.mainForm.stocks)">
        <br>
        <input type="button" name="noStocks" value="$unSelectAll" onclick="unSelectAll(document.mainForm.stocks)">
      </td>
      <td>&nbsp;</td>
    </tr>
    <tr>
      <td align="center" valign="top">
        <table border="0" cellpadding="2" cellspacing="0">
          <tr>
            <td align="center">
              <input type="radio" name="negativeFormat" value="0"$type0Checked>
            </td>
            <td align="left">
              <font color="red">(</font>xxx.xx<font color="red">)</font>
            </td>
          </tr>
          <tr>
            <td align="center">
              <input type="radio" name="negativeFormat" value="1"$type1Checked>
            </td>
            <td align="left">
              <font color="red">-</font>xxx.xx
            </td>
          </tr>
          <tr>
            <td align="center">
              <input type="radio" name="negativeFormat" value="2"$type2Checked>
            </td>
            <td align="left">
              xxx.xx<font color="red">-</font>
            </td>
          </tr>
        </table>
      </td>
    </tr>
    <tr>
      <td><span class="menu">$displayZeroSharesPhrase</span> <input type="checkbox" name="displayZeroShares" value="1"$zeroSharesSelected></td>
    </tr>
  </table>
  <input type="submit" name="execute" value="$execute">
</form>
</center>
END_OF_CODE

  if (length $error > 0)
  {
    $doc->print("<br><br><center><font color=\"red\">Error:&nbsp;</font>$error</center>\n");
  }

  return $doc;
}

# displayPortfolio
# requires: doc, filename, negativeFormat
# optional: printable
# returns: doc or undef if error
# This routine uses the storable object defined by filename to know what data it needs to display.  It then displays the Portfolio from the
# stored data and provides the user with an interface to selectively browse through the displayed data.
sub displayPortfolio
{
  my $startTime = time();
  my $self = shift;
  my %args = ( doc => undef, filename => "", negativeFormat => 0, printable => "false", @_ );
  my $doc = $args{doc};
  my $filename = $args{filename};
  my $printable = $args{printable};
  my $negativeFormat = $args{negativeFormat};
  my $ledger = $self->{input}->{ledger};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $ledgerColor = $self->{sessionObj}->{store}->{ledgers}->{$ledger}->{color};
  my $displayItemObj = Portal::Accounting::DisplayItems->new(langObj => $self->{langObj});
  my $dateYear = $self->{methods}->getCurrentDate(format => "%Y");

  if (! defined $doc)
  {
    $self->error("doc must be defined!<br>\n");
    return $doc;
  }
  if (length $filename == 0)
  {
    $self->error("filename = '$filename' is invalid or does not exist!<br>\n");
    return $doc;
  }
  if ($ledgerCode eq "")
  {
    $self->error("ledgerCode must be defined!<br>\n");
    return $doc;
  }

  if ($printable !~ /^(true|false)$/)
  {
    $self->error("printable = '$printable' is invalid!<br>\n");
    return $doc;
  }
  if ($negativeFormat < 0 || $negativeFormat > 2)
  {
    $self->error("negativeFormat = '$negativeFormat' is invalid!<br>\n");
    return $doc;
  }

  my %data = %{$self->{sessionObj}->{store}->{$filename}};

  my %times = %{$data{times}};
  my $portfolioObj = $data{portfolioObj};
  if (!defined $portfolioObj || !$portfolioObj->isValid)
  {
    $self->error("stored portfolioObj is invalid or does not exist!<br>\n");
    return $doc;
  }
  my $date = $data{date};

  my $phrases = $self->{langObj};

  my $refresh = $phrases->map("refresh");
  my $printPhrase = $phrases->map("print");
  my $portfolioPhrase = $phrases->map("Portfolio");
  my $datePhrase = $self->{langObj}->map("date");

  my $ledgerURI = $doc->encodeString(string => $ledger);
  my $filenameURI = $doc->encodeString(string => $filename);

  my $siteDir = $self->{configObj}->{siteDir};

  my $url = $self->{configObj}->{siteDir} . "/cgi-bin/index.cgi?app=$self->{app}&state=$self->{state}&command=\%s&ledger=%s&ledgerCode=$ledgerCode&filename=%s&negativeFormat=$negativeFormat";
  my $refreshUrl = sprintf($url, "update", $ledgerURI, $filenameURI);
  $refreshUrl .= "&updateWhat=refresh&number=";
  my $printUrl = sprintf($url, "print", $ledgerURI, $filenameURI);
  my $doAnotherUrl = sprintf($url, "display", $ledgerURI, $filenameURI);
  my $investmentJournalURL = $self->{configObj}->{siteDir} . "/cgi-bin/index.cgi?app=$self->{app}&state=InvJournal&command=display&ledgerCode=$ledgerCode&ledger=$ledgerURI";

  $doc->setLinkDecorations;

  my %datePicker = $self->{widgets}->generatePickerCode(type => "date", doc => $doc, form => "doAnother",
                          itemName => "date", itemValue => $date, year => $dateYear,
                          phrase => $datePhrase, width => 300, height => 250);
  if ($self->{widgets}->error)
  {
    $self->setError($self->{widgets}->errorMessage);
    return undef;
  }

  # generate any needed JavaScript functions
  if ($printable eq "false")
  {
    $doc->setFocus("javascript");
    $doc->print(<<"END_OF_CODE");

function toggleAccount(account)
{
  var d = document.portfolio_form;

  d.updateWhat.value = "account";
  d.number.value = account;

  d.submit();
}

function toggleStock(stock)
{
  var d = document.portfolio_form;

  d.updateWhat.value = "stock";
  d.number.value = stock;

  d.submit();
}

END_OF_CODE

  }

  # now display the body.
  $doc->setFocus("body");
  #$doc->print("Account Time: " . ($times{accounts} - $times{start}) . " SubAccount Time: " . ($times{subAccounts} - $times{accounts}) . " getPortfolio Time: " . ($times{getPortfolio} - $times{subAccounts}) . " (SQL: " . ($portfolioObj->{times}->{sql} - $portfolioObj->{times}->{start}) . " Cast: " . ($portfolioObj->{times}->{cast} - $portfolioObj->{times}->{sql}) . " Process: " . ($portfolioObj->{times}->{end} - $portfolioObj->{times}->{cast}) . ") Total Time: <b>" . ($times{getPortfolio} - $times{start}) . "</b> Display Time: #DISPLAYTIME#<br>\n");
  my $actionUrl = $self->{configObj}->{siteDir} . "/cgi-bin/index.cgi";
  if ($printable eq "false")
  {
    $doc->print(<<"END_OF_CODE");
<a href="$refreshUrl" onMouseOver="window.status='$refresh'">$refresh</a>&nbsp;&nbsp;|&nbsp;
<a href="$doAnotherUrl" onMouseOver="window.status='New View'">New View</a>&nbsp;&nbsp;|&nbsp;
<a href="$investmentJournalURL" onMouseOver="window.status='Investment Journal'">Investment Journal</a>&nbsp;&nbsp;|&nbsp;
<a href="$printUrl" onMouseOver="window.status='$printPhrase'">$printPhrase</a>
<hr>
<table border="0" cellpadding="0" cellspacing="0" width="100%">
  <tr>
    <td align="left" width="33%">
      <a href="#" onMouseOver="window.status='Open All Accounts'" onClick="javascript: toggleAccount(-1); return false;">Open</a>&nbsp;/&nbsp;
      <a href="#" onMouseOver="window.status='Close All Accounts'" onClick="javascript: toggleAccount(-2); return false;">Close</a> All Accounts
    </td>
    <td align="center" width="34%">
      <form name="doAnother" method="post" action="$actionUrl">
        <input type="hidden" name="app" value="$self->{app}">
        <input type="hidden" name="state" value="$self->{state}">
        <input type="hidden" name="command" value="executeFile">
        <input type="hidden" name="ledgerCode" value="$ledgerCode">
        <input type="hidden" name="filename" value="$filename">
        <input type="hidden" name="negativeFormat" value="$negativeFormat">
END_OF_CODE
    $doc->print(%datePicker);
    $doc->print(<<"END_OF_CODE");
        <input type="submit" name="execute" value="Go">
      </form>
    </td>
    <td align="right" width="33%">
      <a href="#" onMouseOver="window.status='Open All Stocks'" onClick="javascript: toggleStock(-1); return false;">Open</a>&nbsp;/&nbsp;
      <a href="#" onMouseOver="window.status='Close All Stocks'" onClick="javascript: toggleStock(-2); return false;">Close</a> All Stocks
    </td>
  </tr>
</table>
END_OF_CODE
  }

  my $headerString = ($printable eq "true" ? "Return" : "Refresh");
  $doc->print(<<"END_OF_CODE");
<center>
<a href="$siteDir/cgi-bin/index.cgi?app=$self->{app}&state=$self->{state}&command=update&updateWhat=refresh&number=&ledgerCode=$ledgerCode&ledger=$ledgerURI&filename=$filenameURI&negativeFormat=$negativeFormat" onMouseOver="window.status='$headerString'"><font color="black"><h2>Portfolio for Ledger <font color="$ledgerColor">$ledger</font> - Date = $date</h2></font></a>
</center>
<form name="portfolio_form" method="post" action="$actionUrl">
  <input type="hidden" name="app" value="$self->{app}">
  <input type="hidden" name="state" value="$self->{state}">
  <input type="hidden" name="command" value="update">
  <input type="hidden" name="ledger" value="$ledger">
  <input type="hidden" name="ledgerCode" value="$ledgerCode">
  <input type="hidden" name="filename" value="$filename">
  <input type="hidden" name="negativeFormat" value="$negativeFormat">
  <input type="hidden" name="updateWhat" value="">
  <input type="hidden" name="number" value="">
</form>
<table border="0" cellspacing="2" cellpadding="0" width="100%">
  <tr>
    <td width="2%">&nbsp;</td>
    <td width="23%">&nbsp;</td>
    <td width="10%">&nbsp;</td>
    <td width="10%">&nbsp;</td>
    <td width="15%">&nbsp;</td>
    <td width="10%">&nbsp;</td>
    <td width="10%">&nbsp;</td>
    <td width="10%">&nbsp;</td>
    <td width="10%">&nbsp;</td>
  </tr>
  <tr>
    <th>&nbsp;</th><th align="left" valign="bottom" nowrap>Account / Stock / Lot</th><th align="center" valign="bottom">Buy<br>Date</th><th align="right" valign="bottom">Shares</th><th align="right" valign="bottom">Cost&nbsp;&nbsp;</th><th align="right" valign="bottom">Cost&nbsp;/<br>Share</th><th align="right" valign="bottom">Price&nbsp;/<br>Share</th><th align="right" valign="bottom">Market<br>Value&nbsp;</th><th align="right" valign="bottom">Gain&nbsp;/<br>Loss&nbsp;&nbsp;</th>
  </tr>
  <tr>
    <td align="center" colspan="9"><hr></td>
  </tr>
END_OF_CODE

  my $portfolioObj = $data{portfolioObj};

  my $output = "";
  my @paddingValue;
  my %grandTotals = ( "totalCost" => "0.00", "totalMarketValue" => "0.00", "totalGainLoss" => "0.00" );
  # now we need to walk over the PortfolioObject and display all Accounts and Stocks that are stored in it.
  foreach my $accountObj (@{$portfolioObj->{accounts}})
  {
    # increment the grand totals.
    $grandTotals{totalCost} += $accountObj->totalCost;
    $grandTotals{totalMarketValue} += $accountObj->totalMarketValue;
    $grandTotals{totalGainLoss} += $accountObj->totalGainLoss;

    # display the state of the Account (+ or -) but only if there are stocks to display for it.
    $output .= "  <tr>\n    <td align=\"right\">";
    if (scalar @{$accountObj->{stocks}} > 0 && $printable eq "false")
    {
      my $image = ($accountObj->status eq "c" ? "plus.gif" : "minus.gif");
      my $phrase = ($accountObj->status eq "c" ? "Expand" : "Contract");
      $output .= "<a href=\"#\" onMouseOver=\"window.status='$phrase'\" onClick=\"javascript: toggleAccount($accountObj->{accountNum}); return false;\"><img border=\"0\" src=\"$siteDir/js/Accounting/accounting/images/$image\" alt=\"$phrase\"></a>&nbsp;";
    }
    else
    {
      $output .= "&nbsp;";
    }
    $output .= "</td>\n";

    # now display the Account info
    if ($accountObj->status eq "c")
    {
      (my $accountName = $accountObj->accountName) =~ s/&/&amp;/g;
      $output .= "    <td align=\"left\" colspan=\"2\" nowrap>$accountName ($accountObj->{accountNum})</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $accountObj->{totalCost}, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $accountObj->totalMarketValue, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $accountObj->totalGainLoss, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
      $output .= "  </tr>\n";
    }
    else
    {
      (my $accountName = $accountObj->accountName) =~ s/&/&amp;/g;
      $output .= "    <td align=\"left\" colspan=\"2\" nowrap>$accountName ($accountObj->{accountNum})</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";  # the totals go at the bottom of the list of Stocks.
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "  </tr>\n";

      # we also have to display all Stocks and their status's here.
      foreach my $stockObj (@{$accountObj->{stocks}})
      {
        $output .= "  <tr>\n";
        $output .= "    <td align=\"right\">&nbsp;</td>\n";
        $output .= "    <td align=\"left\" nowrap>";
        if (scalar @{$stockObj->{lots}} > 0 && $printable eq "false")
        {
          my $image = ($stockObj->status eq "c" ? "plus.gif" : "minus.gif");
          my $phrase = ($stockObj->status eq "c" ? "Expand" : "Contract");
          $output .= "<a href=\"#\" onMouseOver=\"window.status='$phrase'\" onClick=\"javascript: toggleStock('$accountObj->{accountNum}|$stockObj->{subAccountNum}'); return false;\"><img border=\"0\" src=\"$siteDir/js/Accounting/accounting/images/$image\" alt=\"$phrase\"></a>&nbsp;";
        }
        else
        {
          $output .= "&nbsp;&nbsp;&nbsp;&nbsp;";
        }
        (my $stockName = $stockObj->{subAccountName}) =~ s/&/&amp;/g;
        $output .= "$stockName</td>\n";

        # now output the Stock info
        if ($stockObj->status eq "c")
        {
          $output .= "    <td align=\"center\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\"><b>$stockObj->{totalShares}</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->totalCost, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->averageCostPerShare, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->newPricePerShare, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->totalMarketValue, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->totalGainLoss, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          $output .= "  </tr>\n";
        }
        else
        {
          # display the header info and all lots for this stock
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "  </tr>\n";

          # now display all lots for this account/stock pair.
          foreach my $lotObj (@{$stockObj->{lots}})
          {
            $output .= "  <tr>\n";
            $output .= "    <td>&nbsp;</td>\n";
            my $spaceString = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;";
            $spaceString .= "&nbsp;&nbsp;" if ($printable eq "true");
            $output .= "    <td align=\"left\">$spaceString$lotObj->{lotRefNum}</td>\n";  # we need to make this link somewhere, don't know where yet.
            $output .= "    <td align=\"center\">$lotObj->{buyDate}</td>\n";
            $output .= "    <td align=\"right\">$lotObj->{shares}</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $lotObj->cost, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $lotObj->costPerShare, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $lotObj->newPricePerShare, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => $lotObj->marketValue, negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            @paddingValue = $displayItemObj->displayMoney(amount => "$lotObj->{gainLoss}", negativeFormat => $negativeFormat);
            $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\">" . $paddingValue[1] . "</td>\n";
            $output .= "  </tr>\n";
          }

          # display the totals for the stock after all lots.
          $output .= "  <tr>\n";
          $output .= "    <td>&nbsp;</td>\n";
          $output .= "    <td>&nbsp;</td>\n";
          $output .= "    <td>&nbsp;</td>\n";
          $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
          $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
          $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
          $output .= "    <td align=\"right\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
          $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
          $output .= "  </tr>\n";

          $output .= "  <tr>\n";
          $output .= "    <td>&nbsp;</td>\n";
          my $spaceString = "&nbsp;&nbsp;&nbsp;";
          $spaceString .= "&nbsp;" if ($printable eq "true");
          $output .= "    <td><b>$spaceString$stockName Total</b></td>\n";
          $output .= "    <td align=\"center\">&nbsp;</td>\n";
          $output .= "    <td align=\"right\"><b>$stockObj->{totalShares}</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->totalCost, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->averageCostPerShare, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->newPricePerShare, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->totalMarketValue, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          @paddingValue = $displayItemObj->displayMoney(amount => $stockObj->totalGainLoss, negativeFormat => $negativeFormat);
          $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
          $output .= "  </tr>\n";
        }
      }

      # display the Totals for the Account.
      $output .= "  <tr>\n";
      $output .= "    <td>&nbsp;</td>\n";
      $output .= "    <td>&nbsp;</td>\n";
      $output .= "    <td>&nbsp;</td>\n";
      $output .= "    <td>&nbsp;</td>\n";
      $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
      $output .= "    <td align=\"right\"><hr align=\"right\" width=\"75%\"></td>\n";
      $output .= "  </tr>\n";

      $output .= "  <tr>\n";
      $output .= "    <td>&nbsp;</td>\n";
      my $spaceString = "";
      $output .= "    <td align=\"left\" colspan=\"2\"><b>$spaceString$accountName Total</b></td>\n";
      $output .= "    <td>&nbsp;</td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $accountObj->totalCost, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      $output .= "    <td align=\"right\">&nbsp;</td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $accountObj->totalMarketValue, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
      @paddingValue = $displayItemObj->displayMoney(amount => $accountObj->totalGainLoss, negativeFormat => $negativeFormat);
      $output .= "    <td align=\"right\" style=\"padding-right: $paddingValue[0];\"><b>" . $paddingValue[1] . "</b></td>\n";
      $output .= "  </tr>\n";
    }
  }

  # this row will be the calibration row that causes the table to properly proportion the cells.
  $output .= "  <tr>\n";
  $output .= "    <td width=\"2%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"23%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"15%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
  $output .= "    <td width=\"10%\" align=\"right\">&nbsp;</td>\n";
  $output .= "  </tr>\n";

  if (scalar @{$portfolioObj->{accounts}} == 0)  # This should never happen!
  {
    $output .= "  <tr>\n";
    $output .= "    <td align=\"right\">&nbsp;</td>\n";
    $output .= "    <td colspan=\"8\" align=\"left\">No Accounts returned!</td>\n";
    $output .= "  </tr>\n";
  }
  else
  {
    # display the grand totals.
    $output .= "</table>\n";
    $output .= "<br>\n";
    $output .= "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\" width=\"100%\">\n";
    $output .= "  <tr>\n";
    $output .= "    <td align=\"right\" width=\"25%\"><b>Grand Totals:</b></td>\n";
    $output .= "    <td align=\"left\" width=\"75%\">&nbsp;</td>\n";
    $output .= "  </tr>\n";
    $output .= "  <tr>\n";
    $output .= "    <td colspan=\"2\">&nbsp;</td>\n";
    $output .= "  </tr>\n";

    $grandTotals{totalCost} = sprintf("%.2f", $grandTotals{totalCost});
    $grandTotals{totalMarketValue} = sprintf("%.2f", $grandTotals{totalMarketValue});
    $grandTotals{totalGainLoss} = sprintf("%.2f", $grandTotals{totalGainLoss});

    $output .= "  <tr>\n";
    $output .= "    <td align=\"right\" width=\"25%\"><b>Total Cost</b></td>\n";
    @paddingValue = $displayItemObj->displayMoney(amount => $grandTotals{totalCost}, negativeFormat => $negativeFormat);
    $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;<b>\$" . $paddingValue[1] . "</b></td>\n";
    $output .= "  </tr>\n";
    $output .= "  <tr>\n";
    $output .= "    <td align=\"right\" width=\"25%\"><b>Total Market Value</b></td>\n";
    @paddingValue = $displayItemObj->displayMoney(amount => $grandTotals{totalMarketValue}, negativeFormat => $negativeFormat);
    $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;<b>\$" . $paddingValue[1] . "</b></td>\n";
    $output .= "  </tr>\n";
    $output .= "  <tr>\n";
    $output .= "    <td align=\"right\" width=\"25%\"><b>Total Gain / Loss</b></td>\n";
    @paddingValue = $displayItemObj->displayMoney(amount => $grandTotals{totalGainLoss}, negativeFormat => $negativeFormat);
    $output .= "    <td align=\"left\" width=\"75%\">&nbsp; = &nbsp;<b>\$" . $paddingValue[1] . "</b></td>\n";
    $output .= "  </tr>\n";
  }

  $doc->print($output);  # output the portfolio data into the HTMLObject.

  $doc->print("</table>\n");  # close the table out.

  if ($printable eq "true")
  {
    # generate the onLoad Code that will trigger a print of the page.
    my $onLoadCode = "window.print();\n";
    $doc->setOnload(code => $onLoadCode);
  }

  my $endTime = time();

  # now substitute into the Document.
  $doc->printTag(tag => "#DISPLAYTIME#", value => ($endTime - $startTime));

  return $doc;
}

# getPortfolioData
# required: accounts, date, ledgerCode
# optional: stocks, displayZeroShares
# returns: PortfolioObject or undef if error.
# This routine takes the array of accounts, stocks and gathers all info for a Portfolio view upto the specified date.
sub getPortfolioData
{
  my $self = shift;
  my %args = ( accounts => undef, stocks => undef, date => "", ledgerCode => "", displayZeroShares => 0, year => "", @_ );
  if ($self->{debug})
  {
    $self->error("getPortfolioData<br>\n");
  }
  my @accounts = @{$args{accounts}};
  my @stocks = @{$args{stocks}};
  my $date = $args{date};
  my $ledgerCode = $args{ledgerCode};
  my $portfolioObj = undef;
  my $displayZeroShares = $args{displayZeroShares};
  my $year = $args{year};
  my %times = ();

  if (!defined @accounts)
  {
    $self->setError(errorString => "accounts must be defined!<br>\n");
    return $portfolioObj;
  }
  my @dateResult = $self->{methods}->fixupDate(date => $date, year => $year);
  if (!$dateResult[0])
  {
    $self->setError(errorString => "date = '$date' is invalid!<br>$dateResult[2]\n");
    return $portfolioObj;
  }
  else
  {
    # make sure the date is "fixed".
    $date = $dateResult[1];
  }
  if (length $ledgerCode == 0)
  {
    $self->setError(errorString => "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return $portfolioObj;
  }

  $times{start} = time();

  # build up the sql string to execute.
  my $sql = "SELECT DISTINCT(a.lot), a.event_date, a.node_id, a.account_num, a.sub_account_num, a.shares, a.cost, (SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = a.sub_account_num) AS sub_account_name, (SELECT MIN(event_date) FROM lot_node_tb WHERE lot = (SELECT buy_lot FROM lot_parent_tb WHERE lot = a.lot)) as buy_date, (SELECT name FROM account_tb WHERE number = a.account_num) AS account_name FROM lot_node_tb AS a WHERE a.ledger_code = '$ledgerCode'";
  $sql .= " AND a.event_date = (SELECT max(event_date) FROM lot_node_tb WHERE lot = a.lot AND event_date <= '$date')";
  $sql .= " AND a.node_id = (SELECT max(node_id) FROM lot_node_tb WHERE lot = a.lot AND event_date = a.event_date)";
  $sql .= " AND a.shares > '0.0'" if (!$displayZeroShares);
  $sql .= " AND a.account_num IN (";
  for (my $i=0; $i < scalar @accounts; $i++)
  {
    $sql .= "," if ($i > 0);
    $sql .= $accounts[$i]->get("id");
  }
  $sql .= ")";
  if (scalar @stocks > 0)
  {
    $sql .= " AND a.sub_account_num IN (";
    for (my $i=0; $i < scalar @stocks; $i++)
    {
      $sql .= "," if ($i > 0);
      $sql .= $stocks[$i]->get("id");
    }
    $sql .= ")";
  }
  $sql .= " ORDER BY account_name, sub_account_name, a.lot";

  # gather the info for the specified data.
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->setError(errorString => "DBIW Error in getPortfolioData<br>\n" . $self->{dbHandle}->errorMessage);
    return $portfolioObj;
  }

  $times{sql} = time();

  my $result = $sth->fetchall_arrayref;  # get all the items from the database.
  my @result = @{$result};  # cast it to an array.

  $times{cast} = time();

  my @accountObjs = ();
  my $startAccount = 1;
  my $prevAccount = "";
  my $startAccountIndex = 0;
  my $endAccountIndex = 0;

  my %accountNames;
  my $stockName = "";

  foreach my $accountObj (@accounts)
  {
    $accountNames{$accountObj->get("id")} = $accountObj->get("name");
  }

  if (scalar @result == 0)
  {
    $self->{errorCode} = -1;  # signal we didn't get any rows back.
    return undef;
  }

  my $lotIndex = 0;
  my $accountIndex = 3;
  my $stockIndex = 4;
  my $dateIndex = 1;
  my $nodeIndex = 2;
  my $sharesIndex = 5;
  my $costIndex = 6;
  my $accountNameIndex = 9;
  my $subAccountNameIndex = 7;
  my $buyDateIndex = 8;

  my $processed_last_account = 0;  # flag to catch the case where the last account only had 1 lot row returned and wasn't being included in result.

  my %currentStockPrices = ();

  # now walk over the result set and for all rows that have the same account, process all stock and lot pairs and produce a single PortfolioAccountObject.
  for (my $i=0; $i < scalar @result; $i++)
  {
    if ($startAccount)
    {
      $startAccount = 0;  # signal we are done just blindly taking the first account we meet.
      $prevAccount = $result[$i][$accountIndex];
      $startAccountIndex = $i;
      $endAccountIndex = $i;
    }
    elsif ($result[$i][$accountIndex] == $prevAccount)
    {
      $endAccountIndex = $i;
    }
    if (($result[$i][$accountIndex] != $prevAccount && $startAccount == 0) || ($i == scalar(@result) - 1))  # now we need to process all entries for the previous account.
    {
      my @stockObjs = ();
      my @lotObjs = ();
      my %accountTotals = ( totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );
      my %stockTotals = ( totalShares => "0.00", totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );
      my $startStocks = 1;
      my $prevStock = "";
      for (my $j=$startAccountIndex; $j <= $endAccountIndex; $j++)
      {
        # create the PortfolioLotObject
        my $shares = $result[$j][$sharesIndex];
        my $cost   = $result[$j][$costIndex];
        my $pricePerShare = "0.00";

        if (not exists $currentStockPrices{$result[$j][$stockIndex]})
        {
          # lookup the current price from the database.
          my @stockPriceInfo = $self->getCurrentStockPrice(stock => $result[$j][$stockIndex], date => $date, year => $year);
          if ($self->error)
          {
            $self->prefixError();
            return $portfolioObj;
          }
          $currentStockPrices{$result[$j][$stockIndex]} = $stockPriceInfo[0];
        }
        $pricePerShare = sprintf("%.2f", $currentStockPrices{$result[$j][$stockIndex]});

        my $marketValue = $shares * $pricePerShare;
        $marketValue = sprintf("%.2f", $marketValue);
        my $gainLoss = $marketValue - $cost;
        $gainLoss = sprintf("%.2f", $gainLoss);
        my $costPerShare = ($shares !~ /^0+(\.0+)?$/ && $shares ne "" ? $cost / $shares : "0.00");#/
        $costPerShare = sprintf("%.2f", $costPerShare);
        $shares =~ s/^(\d+\.\d)(0+)$/$1/;
        $cost = sprintf("%.2f", $cost);

        my $portfolioLotObj = Portal::Accounting::Objects::PortfolioLotObject->new(accountNum => $result[$j][$accountIndex], subAccountNum => $result[$j][$stockIndex],
                              lotRefNum => $result[$j][$lotIndex], date => $result[$j][$dateIndex], shares => $shares, cost => $cost,
                              newPricePerShare => $pricePerShare, marketValue => $marketValue, gainLoss => $gainLoss, costPerShare => $costPerShare,
                              buyDate => $result[$j][$buyDateIndex], langObj => $self->{langObj});
        if ($portfolioLotObj->error)
        {
          $self->setError(errorString => "PortfolioLotObject Error in getPortfolioData<br>\n" . $portfolioLotObj->errorMessage);
          return $portfolioObj;
        }

        # now see what StockObject this goes with.
        if ($result[$j][$stockIndex] == $prevStock || $startStocks)
        {
          $startStocks = 0;
          $prevStock = $result[$j][$stockIndex];
          $lotObjs[++$#lotObjs] = $portfolioLotObj;

          $stockName = $result[$j][$subAccountNameIndex];

          $stockTotals{totalShares} += $shares;
          $stockTotals{totalCost} += $cost;
          $stockTotals{totalMarketValue} += $marketValue;
          $stockTotals{totalGainLoss} += $gainLoss;

          $accountTotals{totalCost} += $cost;
          $accountTotals{totalMarketValue} += $marketValue;
          $accountTotals{totalGainLoss} += $gainLoss;
        }
        if (($result[$j][$stockIndex] != $prevStock && $startStocks == 0) || ($j == $endAccountIndex))  # create this stock and start next one.
        {
          # make sure all figures are 2 decimal points.
          foreach my $total (keys %stockTotals)
          {
            if ($total eq "totalShares")
            {
              $stockTotals{$total} = sprintf("%.6f", $stockTotals{$total});
              $stockTotals{$total} =~ s/^(\d+\.\d)(0+)$/$1/;
            }
            else
            {
              $stockTotals{$total} = sprintf("%.2f", $stockTotals{$total});
            }
          }

          my $averageCostPerShare = ($stockTotals{totalShares} !~ /^0+\.0+$/ && $stockTotals{totalShares} ne "" ? $stockTotals{totalCost} / $stockTotals{totalShares} : "0.00");#/
          $averageCostPerShare = sprintf("%.2f", $averageCostPerShare);

          my @tmpLotObjs = @lotObjs;
          my $portfolioStockObj = Portal::Accounting::Objects::PortfolioStockObject->new(subAccountName => $stockName, subAccountNum => $prevStock,
                                  totalShares => $stockTotals{totalShares}, totalCost => $stockTotals{totalCost}, totalMarketValue => $stockTotals{totalMarketValue},
                                  accountNum => $prevAccount, totalGainLoss => $stockTotals{totalGainLoss}, status => "c", lots => \@tmpLotObjs,
                                  averageCostPerShare => $averageCostPerShare, newPricePerShare => $tmpLotObjs[0]->{newPricePerShare}, langObj => $self->{langObj});
          if ($portfolioStockObj->error)
          {
            $self->setError(errorString =>  "PortfolioStockObject Error in getPortfolioData<br>\n" .$portfolioStockObj->errorMessage);
            return $portfolioObj;
          }
          $stockObjs[++$#stockObjs] = $portfolioStockObj;

          if (($j + 1) <= $endAccountIndex)
          {
            # reset Lot* variables and Stock* totals
            @lotObjs = ();
            $prevStock = $result[$j][$stockIndex];
            $startStocks = 0;

            $stockName = $result[$j][$subAccountNameIndex];

            $lotObjs[++$#lotObjs] = $portfolioLotObj;

            %stockTotals = ( totalShares => "0.00", totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );

            $stockTotals{totalShares} += $shares;
            $stockTotals{totalCost} += $cost;
            $stockTotals{totalMarketValue} += $marketValue;
            $stockTotals{totalGainLoss} += $gainLoss;

            $accountTotals{totalCost} += $cost;
            $accountTotals{totalMarketValue} += $marketValue;
            $accountTotals{totalGainLoss} += $gainLoss;
          }
          elsif (($result[$j][$stockIndex] != $prevStock && $startStocks == 0) && ($j == $endAccountIndex))
          {  # did we just switch stocks and we are the last one in the returned list?
            # we must make the PortfolioStockObj entry!
            @lotObjs = ();
            $prevStock = $result[$j][$stockIndex];
            $startStocks = 0;

            $stockName = $result[$j][$subAccountNameIndex];

            $lotObjs[++$#lotObjs] = $portfolioLotObj;

            %stockTotals = ( totalShares => "0.00", totalCost => "0.00", totalMarketValue => "0.00", totalGainLoss => "0.00" );

            $stockTotals{totalShares} += $shares;
            $stockTotals{totalCost} += $cost;
            $stockTotals{totalMarketValue} += $marketValue;
            $stockTotals{totalGainLoss} += $gainLoss;

            $accountTotals{totalCost} += $cost;
            $accountTotals{totalMarketValue} += $marketValue;
            $accountTotals{totalGainLoss} += $gainLoss;

            # make sure all figures are 2 decimal points.
            foreach my $total (keys %stockTotals)
            {
              if ($total eq "totalShares")
              {
                $stockTotals{$total} = sprintf("%.6f", $stockTotals{$total});
                $stockTotals{$total} =~ s/^(\d+\.\d)(0+)$/$1/;
              }
              else
              {
                $stockTotals{$total} = sprintf("%.2f", $stockTotals{$total});
              }
            }

            my $averageCostPerShare = ($stockTotals{totalShares} !~ /^0+\.0+$/ && $stockTotals{totalShares} ne "" ? $stockTotals{totalCost} / $stockTotals{totalShares} : "0.00");#/
            $averageCostPerShare = sprintf("%.2f", $averageCostPerShare);

            my @tmpLotObjs = @lotObjs;
            my $portfolioStockObj = Portal::Accounting::Objects::PortfolioStockObject->new(subAccountName => $stockName, subAccountNum => $prevStock,
                                    totalShares => $stockTotals{totalShares}, totalCost => $stockTotals{totalCost}, totalMarketValue => $stockTotals{totalMarketValue},
                                    accountNum => $prevAccount, totalGainLoss => $stockTotals{totalGainLoss}, status => "o", lots => \@tmpLotObjs,
                                    averageCostPerShare => $averageCostPerShare, newPricePerShare => $tmpLotObjs[0]->{newPricePerShare}, langObj => $self->{langObj});
            if ($portfolioStockObj->error)
            {
              $self->setError(errorString =>  "PortfolioStockObject Error in getPortfolioData<br>\n" .$portfolioStockObj->errorMessage);
              return $portfolioObj;
            }
            $stockObjs[++$#stockObjs] = $portfolioStockObj;
          }
        }
      }

      # make sure all figures are 2 decimal points.
      foreach my $total (keys %accountTotals)
      {
        $accountTotals{$total} = sprintf("%.2f", $accountTotals{$total});
      }

      # create the PortfolioAccountObject and insert into accountObjs
      my @tmpStockObjs = @stockObjs;
      my $portfolioAccountObj = Portal::Accounting::Objects::PortfolioAccountObject->new(accountName => $accountNames{$prevAccount}, accountNum => $prevAccount,
                                totalCost => $accountTotals{totalCost}, totalMarketValue => $accountTotals{totalMarketValue}, totalGainLoss => $accountTotals{totalGainLoss},
                                status => "o", stocks => \@tmpStockObjs, langObj => $self->{langObj});
      if ($portfolioAccountObj->error)
      {
        $self->setError(errorString => "PortfolioAccountObject Error in getPortfolioData<br>\n" . $portfolioAccountObj->errorMessage);
        return $portfolioObj;
      }
      $accountObjs[++$#accountObjs] = $portfolioAccountObj;

      # reset Account* variables.
      $startAccount = 0;
      $startAccountIndex = $i;
      $endAccountIndex = $i;
      $prevAccount = $result[$i][$accountIndex];
      if ($i > 0 && $result[$i-1][$accountIndex] != $prevAccount && ($i == scalar(@result) - 1) && !$processed_last_account)
      {
        # signal we have already done this.
        $processed_last_account = 1;
        redo;  # process last entry.
      }
    }
  }

  # now build the portfolioObj.
  $portfolioObj = Portal::Accounting::Objects::PortfolioObject->new(accounts => \@accountObjs, langObj => $self->{langObj});
  if ($portfolioObj->error)
  {
    $self->setError(errorString => "PortfolioObject Error in getPortfolioData<br>\n" . $portfolioObj->errorMessage);
    return $portfolioObj;
  }

  $times{end} = time();

  $portfolioObj->{times} = \%times;

  return $portfolioObj;
}

=back

=cut

1;
__END__

=head1 Exported FUNCTIONS (non-Inline POD)

  HTMLObject run(command)
    Runs the specified command and returns the resulting HTMLObject.


=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
