# StockPrices.pm - The Object Class that provides a StockPrices Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: StockPrices.pm,v 1.7 2005/07/22 17:54:56 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Investments::StockPrices;
use strict;
use HTMLObject::Widgets;
use Portal::Accounting::Main;
use Portal::Accounting::DisplayItems;
use Portal::Accounting::DB::Main;
use Portal::Accounting::Objects::Stock;
use Finance::Quote;
use LWP::Simple;
use HTML::TokeParser::Simple;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);
use POSIX qw(tmpnam);

require Exporter;

@ISA = qw(Portal::Accounting::Main Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.7 $';

# global variables.
# modify to represent the commands this state provides.
my %commands = (
 display => "Stock Prices",
 lookupStockInfo => "Stock Prices",
 update => "Stock Prices"
  );

=head1 NAME

StockPrices - Object used to build a StockPrices Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::StockPrices;
  my $obj = Portal::Accounting::StockPrices->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

StockPrices is a StockPrices class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::StockPrices module.
 See Portal::Accounting::Base(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{commands} = \%commands;
  $self->{widgets} = HTMLObject::Widgets->new();
  if ($self->{widgets}->error)
  {
    $self->error($self->{widgets}->errorMessage);
    return $self;
  }

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub c_update
{
  my $self = shift;
  return $self->c_display;
}

sub c_display
{
  my $self = shift;
  my %args = ( command => $self->{input}->{command}, @_ );
  my $command = $args{command};
  my $doc;
  my $ledger;
  my $ledgerCode;
  my $stockObj = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);

  if (!exists $commands{$command})
  {
    $self->setError(sprintf($self->{langObj}->map("commandNotKnown"), $command));
    return undef;
  }

  $doc = $self->prepDoc;
  if ($command eq "display")
  {
    $doc = $self->displayStockPriceScreen(doc => $doc);
    if ($self->error)
    {
      $self->prefixError();
      return undef;
    }
  }
  elsif ($command eq "lookupStockInfo")
  {
    $doc->setTitle($self->{langObj}->map("chooseStockInfo"));
    $doc->setLinkDecorations;

    $doc->setFocus("body");
    $doc->print(<<"END_OF_CODE");
  <table border="0" cellspacing="0" cellpadding="0" class="main">
    <tr>
      <td class="edit">
        <table border="0" cellspacing="1" cellpadding="3">
END_OF_CODE
    my $result = 0;
    if ((length $self->{input}->{subAccountName} > 0) || (length $self->{input}->{ticker} > 0) || (length $self->{input}->{stockName} > 0))
    {
      $result = $self->lookupStockInfo(doc => $doc, info => join('+', $self->{input}->{subAccountName}, $self->{input}->{ticker}, $self->{input}->{stockName}));
      if ($result == 0 || $self->error)
      {
        $doc->print(sprintf($self->{langObj}->map("errorSearchingStockInfo"), $self->{input}->{subAccountName}, $self->{input}->{ticker}, $self->{input}->{stockName}) . "<br>\n" . $self->errorMessage);
      }
    }
    else
    {
      $doc->print("<tr><td bgcolor=\"#ffffff\"><a href=\"\" onClick=\"window.close(); return false;\">" . $self->{langObj}->map("noStockInfoSuppliedReturned") . "</a></td></tr>\n");
    }
    $doc->print(<<"END_OF_CODE");
        </table>
      </td>
    </tr>
  </table>
END_OF_CODE
  }
  elsif ($command eq "update")
  {
    my $subCommand = $self->{input}->{subCommand};
    my $stocks = $self->{input}->{stocks};  # the selected stocks to work with if command is currentPrice or historicalPrice
    my $manualPrice = $self->{input}->{manualPrice};
    my $startDate = $self->{input}->{startDate};
    my $endDate = $self->{input}->{endDate};
    my $interval = $self->{input}->{interval};
    $ledgerCode = $self->{input}->{ledgerCode};
    $ledger = $self->{sessionObj}->{store}->{ledgers}->{$ledgerCode}->name;

    if ($subCommand !~ /^(currentPrice|historicalPrice|manualPrice)$/)
    {
      $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "subCommand = '$subCommand' is invalid!<br>\n");
      if ($self->error)
      {
        $self->prefixError();
        return undef;
      }
      return $doc;
    }

    my @stocks = split /\0/, $stocks;
    my @tickers = ();
    my @stockObjs = ();
    my $tickers = "";

    if (scalar @stocks > 0)
    {
      # lookup the ticker for each specified stock
      foreach my $stock (@stocks)
      {
        my $i = scalar @stockObjs;
        $stockObjs[$i] = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
        if ($stockObjs[$i]->set($stockObj->{id}, $stock)) { $stockObjs[$i]->select; }
        if ($stockObjs[$i]->error)
        {
          $self->error($stockObjs[$i]->errorMessage);
        }
      }
    }
    else
    {
      # get all stocks in the system and create the array of tickers
      @stockObjs = $stockObj->getListOfStocks(marketOnly => 1);
      if ($stockObj->error)
      {
        $self->error($stockObj->errorMessage);
        $doc->print($self->errorMessage);
      }
    }
    foreach my $stockObj (@stockObjs)
    {
      if (!$stockObj->isValid) { $self->error($stockObj->errorMessage); }
      elsif ($stockObj->get("ticker") ne "" && $stockObj->get("market") ne "None")
      {
        $tickers .= ", " if (length $tickers > 0);
        $tickers .= $stockObj->get("ticker");
        $tickers[scalar @tickers] = $stockObj->get("ticker");
      }
    }
    if ($subCommand eq "currentPrice")
    {
      # get the list of stocks to work with (sub account numbers)
      my @stocks = split /\0/, $stocks;
      my @tickers = ();
      my $tickers = "";

      if (scalar @stocks > 0)
      {
        # lookup the ticker for each specified stock
        foreach my $stock (@stocks)
        {
          if ($stock eq "") { next;  }
          my $stockObj = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
          if ($stockObj->error)
          {
            $doc->print($stockObj->errorMessage);
            return $doc;
          }
          if ($stockObj->selectIdByTicker($stock)) { $stockObj->select; }
          if ($stockObj->error)
          {
            $doc = $self->displayStockPriceScreen(doc => $doc, errorString => $stockObj->errorMessage);
            if ($self->error)
            {
              $doc->print($self->errorMessage);
            }
            return $doc;
          }
          if ($stockObj->get("ticker") eq "" || $stockObj->get("market") eq "None") {  next;  }
          $tickers .= ", " if (length $tickers > 0);
          $tickers .= $stockObj->get("ticker");
          $tickers[scalar @tickers] = $stockObj->get("ticker");
        }
      }
      else
      {
        # get all stocks in the system and create the array of tickers
        my @stocks = $stockObj->getListOfStocks(marketOnly => 1);
        if ($stockObj->error)
        {
          $self->error($stockObj->errorMessage);
          return undef;
        }

        foreach my $stockObj (@stocks)
        {
          $tickers .= ", " if (length $tickers > 0);
          $tickers .= $stockObj->get("ticker");
          $tickers[scalar @tickers] = $stockObj->get("ticker");
        }
      }

      # have it lookup the prices and store the results.
      (my $result, my $goodTickers, my $badTickers) = $self->updateStockPrices(tickerList => \@tickers, type => "current");
      if ($self->error)
      {
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => $self->errorMessage);
        return $doc;
      }

      # return to the main screen.
      $doc = $self->displayStockPriceScreen(doc => $doc, statusString => "Current Prices gathered for ticker(s) <b>" . join(", ", @{$goodTickers}) . "</b>" . (@{$badTickers} > 0 ? ";<br><br>Current Prices not gathered for ticker(s) <b>" . join(", ", @{$badTickers}) . "</b>.<br><br>" : ".") . "<br>\n") if ($result);
      $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "Failed to gather Current Prices for ticker(s) <b>$tickers</b>.<br>\n") if (!$result);
      if ($self->error)
      {
        $self->prefixError();
        return undef;
      }
    }
    elsif ($subCommand eq "historicalPrice")
    {
      if ($startDate !~ /^\d{4}[-\\\/]\d{1,2}[-\\\/]\d{1,2}$/)
      {
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "startDate = '$startDate' is invalid!\n");
        if ($self->error)
        {
          $self->prefixError();
          return undef;
        }
        return $doc;
      }
      if ($endDate !~ /^\d{4}[-\\\/]\d{1,2}[-\\\/]\d{1,2}$/)
      {
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "endDate = '$endDate' is invalid!\n");
        if ($self->error)
        {
          $self->prefixError();
          return undef;
        }
        return $doc;
      }
      if ($interval !~ /^(day|week|month)$/)
      {
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "interval = '$interval' is invalid!<br>\n");
        if ($self->error)
        {
          $self->prefixError();
          return undef;
        }
        return $doc;
      }
      # have it lookup the prices and store the results.
      (my $result, my $goodTickers, my $badTickers) = $self->updateStockPrices(tickerList => \@tickers, type => "historical", startDate => $startDate, endDate => $endDate, interval => $interval);
      if ($self->error)
      {
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => $self->errorMessage);
        return $doc;
      }

      # return to the main screen.
      $doc = $self->displayStockPriceScreen(doc => $doc, statusString => "Historical Prices gathered for ticker(s) <b>" . join(", ", @{$goodTickers}) . "</b>" . (@{$badTickers} > 0 ? ";<br><br>Historical Prices not gathered for ticker(s) <b>" . join(", ", @{$badTickers}) . "</b><br><br>" : "") . "; from <b>$startDate</b> to <b>$endDate</b> by <b>$interval</b>.<br>\n") if ($result);
      $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "Failed to gather Historical Prices for ticker(s) <b>$tickers</b>; from <b>$startDate</b> to <b>$endDate</b> by <b>$interval</b>.<br>\n") if (!$result);
      if ($self->error)
      {
        $self->prefixError();
        return undef;
      }
    }
    elsif ($subCommand eq "manualPrice")
    {
      if (not defined $manualPrice || $manualPrice eq "")
      {
        my $manualPriceNeedsToBeSpecifiedPhrase = $self->{langObj}->map("manualPriceNeedsToBeSpecified");
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "$manualPriceNeedsToBeSpecifiedPhrase<br>\n");
        return $doc;
      }
      if (scalar @stockObjs == 0)
      {
        # output an error message stating you have to choose stocks to update
        my $chooseStocksManuallyPhrase = $self->{langObj}->map("chooseStocksManually");
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "$chooseStocksManuallyPhrase<br>\n");
        return $doc;
      }

      if ($startDate !~ /^\d{4}[-\\\/]\d{1,2}[-\\\/]\d{1,2}$/)
      {
        $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "startDate = '$startDate' is invalid!\n");
        if ($self->error)
        {
          $self->prefixError();
          return undef;
        }
        return $doc;
      }

      # have it lookup the prices and store the results.
      my $result = $self->updateManualPrice(stocks => \@stockObjs, price => $manualPrice, date => $startDate);

      # return to the main screen.
      $doc = $self->displayStockPriceScreen(doc => $doc, statusString => "Manual Prices set for stock <b>".@stockObjs[0]->get("name")."</b> for date $startDate.<br>\n") if ($result);
      $doc = $self->displayStockPriceScreen(doc => $doc, errorString => "Failed to set Manual Prices for ticker(s) <b>".@stockObjs[0]->get("name")."</b>.<br>\n") if (!$result);
    }
  }

  if ($self->error)
  {
    $doc->print($self->errorMessage);
  }
  return $doc;
}

# displayStockPriceScreen
# requires: doc
# optional: errorString, statusString
# returns: doc
# summary: This routine displays all stocks in the system and provides an interface
# to get current stock prices or get historical stock prices or manually set prices.
sub displayStockPriceScreen
{
  my $self = shift;
  my %args = ( doc => undef, errorString => "", statusString => "", @_ );
  my $doc = $args{doc};
  my $errorString = $args{errorString};
  my $statusString = $args{statusString};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $ledger = $self->{sessionObj}->{store}->{ledgers}->{$ledgerCode}->name;
  my $startDate = $self->{methods}->getDateOffset(date => "now", offset => "1 year ago", format => "%Y-%m-%d");
  my $tmpDate = $self->{methods}->getCurrentDate();
  my $dateYear = $self->{methods}->getCurrentDate(format => "%Y");

  if (!defined $doc)
  {
    $self->error("doc must be defined!<br>\n");
    return undef;
  }

  # get list of All Stocks in the System with Tickers
  my $stockObj = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
  my @stocks = $stockObj->getListOfStocks(marketOnly => 1);
  if ($stockObj->error)
  {
    $self->error($stockObj->errorMessage);
    return undef;
  }

  my $ledgerURI = $doc->encodeString(string => $ledger);

  my $title = $self->{langObj}->map("stockPriceTitle");
  my $investmentAccounts = $self->{langObj}->map("Investment")." ".$self->{langObj}->map("Accounts");
  my $stocksPhrase = $self->{langObj}->map("Stocks");
  my $manualPricePhrase = $self->{langObj}->map("Manual Price");
  my $dateInfoPhrase = $self->{langObj}->map("dateInfo");
  my $datePhrase = $self->{langObj}->map("date");
  my $selectAll = $self->{langObj}->map("Select All");
  my $unSelectAll = $self->{langObj}->map("Un-Select All");
  my $refresh = $self->{langObj}->map("refresh");
  my $resetPhrase = $self->{langObj}->map("reset");
  my $updateManualStockPricePhrase = $self->{langObj}->map("updateManualStockPrice");
  my $negativeNumberFormatPhrase = $self->{langObj}->map("negativeNumberFormatPhrase");
  my $goPhrase = $self->{langObj}->map("go");
  my $herePhrase = $self->{langObj}->map("here");
  my $noStocksCreatedPhrase = $self->{langObj}->map("noStocksCreated");
  my $noTickeredStocksCreatedPhrase = $self->{langObj}->map("noTickeredStocksCreated");
  my $toCreateStocksPhrase = $self->{langObj}->map("toCreateStocks");
  my $toMaintainStocksPhrase = $self->{langObj}->map("toMaintainStocks");
  my $maintainStocksPhrase = $self->{langObj}->map("maintainStocks");
  my $updateCurrentStockPricePhrase = $self->{langObj}->map("updateCurrentStockPrice");
  my $updateHistoricalStockPricePhrase = $self->{langObj}->map("updateHistoricalStockPrice");
  my $dayPhrase = $self->{langObj}->map("day");
  my $weekPhrase = $self->{langObj}->map("week");
  my $monthPhrase = $self->{langObj}->map("month");
  my $startDatePhrase = $self->{langObj}->map("Start")." ".$self->{langObj}->map("Date");
  my $endDatePhrase = $self->{langObj}->map("End")." ".$self->{langObj}->map("Date");
  my $intervalPhrase = $self->{langObj}->map("interval");
  my $stockPriceInstructionsPhrase = $self->{langObj}->map("stockPriceInstructions");

  my $displayItemObj = Portal::Accounting::DisplayItems->new(langObj => $self->{langObj});
  my $siteDir = $self->{configObj}->{siteDir};

  $doc->setTitle($title);
  $doc->setLinkDecorations;

  $doc->setFocus("body");

  # do some quick validity checks.
  if (scalar @stocks == 0)
  {
    $doc->print("$noStocksCreatedPhrase<br>$goPhrase <a href=\"?app=$self->{app}&state=SubAccounts&command=list&type=70\" onMouseOver=\"window.status='$herePhrase'\">$herePhrase</a> $toCreateStocksPhrase\n");
    return $doc;
  }

  my %tickers = ();  # hash to store subaccount number -> ticker values.
  my %names = ();  # hash to store subaccount number -> name values.

  my %startDatePicker = $self->{widgets}->generatePickerCode(type => "date", doc => $doc, form => "stockPriceForm",
    itemName => "startDate", itemValue => $startDate, year => $dateYear,
    phrase => $startDatePhrase, class => "menu", linkClass => "menu", width => 300, height => 250,
  );
  if ($self->{widgets}->error)
  {
    $self->setError($self->{widgets}->errorMessage);
    return undef;
  }

  my %endDatePicker = $self->{widgets}->generatePickerCode(type => "date", doc => $doc, form => "stockPriceForm",
    itemName => "endDate", itemValue => $tmpDate, year => $dateYear,
    phrase => $endDatePhrase, class => "menu", linkClass => "menu", width => 300, height => 250,
    onClick => "if (document.stockPriceForm.manualPrice.value == '') { document.stockPriceForm.subCommand.options[0].selected=true; }", onChange => "if (document.stockPriceForm.manualPrice.value != '') { document.stockPriceForm.subCommand.options[1].selected=true; }"
  );
  if ($self->{widgets}->error)
  {
    $self->setError($self->{widgets}->errorMessage);
    return undef;
  }

  # generate the JavaScript hashes we will need.
  $doc->setFocus("javascript");
  $doc->print(<<"END_OF_CODE");

var stockInfo = new Array();

function setStock(ticker, name)
{
  this.ticker = ticker;
  this.name = name;
}

function manualStock()
{
  var d = document.stockPriceForm;
  if (d.manualPrice.value != "")
  {
    // update the command to execute to be the Update Manual Price
    d.subCommand.options[1].selected = true;
  }
}

END_OF_CODE

  my $stocksString = "";
  foreach my $stockObj (@stocks)
  {
    my $id = $stockObj->get("id");
    my $ti = $stockObj->get("ticker");
    my $na = $stockObj->get("name");
    $stocksString .= "          <option value=\"$id\">$na ($ti)\n";
    $tickers{$id} = $ti;
    $names{$id} = $na;
    $doc->print("stockInfo[\"$id\"] = new setStock(\"$ti\", \"$na\");\n");
  }

  if (length $errorString > 0)
  {
    $errorString = "<font color=\"red\">Error: </font>" . $errorString . "<br>\n";
  }
  if (length $statusString > 0)
  {
    $statusString = "<font color=\"red\">Status: </font>" . $statusString . "<br>\n";
  }

  $doc->setFocus("body");
  $doc->print(<<"END_OF_CODE");
<br>
$statusString
$errorString
<br>
<form name="stockPriceForm" method="post">
  <input type="hidden" name="app" value="$self->{app}">
  <input type="hidden" name="state" value="$self->{state}">
  <input type="hidden" name="command" value="update">
  <input type="hidden" name="ledger" value="$ledger">
  <input type="hidden" name="ledgerCode" value="$ledgerCode">
  <table border="0" cellpadding="5" cellspacing="0">
    <tr>
      <th align="center"><a href="?app=$self->{app}&state=SubAccounts&command=list&type=70">$stocksPhrase</a></th><th align="center">$manualPricePhrase</th><th align="center">$dateInfoPhrase</th>
    </tr>
    <tr>
      <td align="center" valign="top">
        <select name="stocks" size="10" multiple>
$stocksString
        </select>
        <br>
        <input type="button" name="selectAllStocks" value="$selectAll" onClick="javascript:selectAll(document.stockPriceForm.stocks);">
        <br>
        <input type="button" name="unSelectAllStocks" value="$unSelectAll" onClick="javascript:unSelectAll(document.stockPriceForm.stocks);">
      </td>
      <td align="center" valign="top">
        <input type="text" name="manualPrice" value="" size="10" onChange="manualStock();">
      </td>
      <td align="left" valign="top">
END_OF_CODE
  $doc->print(%startDatePicker);
  $doc->print(<<"END_OF_CODE");
        <br>
END_OF_CODE
  $doc->print(%endDatePicker);
  $doc->print(<<"END_OF_CODE");
        <br>
        $intervalPhrase <select name="interval" onChange="document.stockPriceForm.subCommand.options[0].selected=true">
          <option value="day">$dayPhrase
          <option value="week">$weekPhrase
          <option value="month">$monthPhrase
        </select>
      </td>
    </tr>
  </table>
  <select name="subCommand">
    <option value="historicalPrice">$updateHistoricalStockPricePhrase
    <option value="manualPrice">$updateManualStockPricePhrase
  </select>
  <input type="submit" name="update" value="$goPhrase">&nbsp;<input type="reset" value="$resetPhrase">
  <br>
  $stockPriceInstructionsPhrase
</form>
END_OF_CODE

  return $doc;
}

# lookupStockInfo
# requires: doc, info
# optional:
# returns: 1 on success, 0 on failure
# summary: This routine displays all stocks found by sending the value of info to
# the financial URL and parsing the returned data apart to present a listing of
# rows with the symbol, name, market, and industry of the result of the search
sub c_lookupStockInfo
{
  my $self = shift;
  my %args = ( doc => undef, info => "", @_ );
  my $doc = $args{doc};
  my $info = $args{info};

  my $financeURL = "http://finance.yahoo.com/l?m=US&t=S&s=";

  # output header fields
  $doc->print("<tr bgcolor=\"#cdcdcd\"><th><font size=\"-1\">" . $self->{langObj}->map("Symbol") . "</font></th><th><font size=\"-1\">" . $self->{langObj}->map("Name") . "</font></th><th><font size=\"-1\">" . $self->{langObj}->map("Market") . "</font></th><th><font size=\"-1\">" . $self->{langObj}->map("Industry") . "</font></th></tr>\n");

  my $financeHTML = LWP::Simple::get($financeURL . $info);
  my $htmlString = HTML::TokeParser::Simple->new(\$financeHTML);
  my $numValidTokens = 0;
  my $rowString = "";
  my $ticker = "";
  my $stockName = "";
  my $market = "";
  my $industry = "";
  while (my $token = $htmlString->get_token)
  {
    if ($token->as_is =~ /Add to My Portfolio/)
    {
      my $tdCount = 0;
      my $alreadyOutput = 0;
      my $alreadyTR = 0;
      my $previousValidToken = 0;
      # move token position after the /td that closes 'Add to My Portfolio'
      while (not ($token->is_end_tag('/td')))
      {
        $token = $htmlString->get_token;
      }
      # parse any returned data
      while (not ($token->as_is =~ /View Quotes for All Above Symbols/))
      {
        $token = $htmlString->get_token;
        if ($token->is_end_tag('/td'))
        {
          if ($previousValidToken > 0)
          {
            $numValidTokens++;
            $previousValidToken = 0;
          }
          $tdCount++;
          $alreadyOutput = 0;
        }
        if (($token->is_text) && (not ($token->as_is =~ /View Quotes for All Above Symbols/)) && (not ($token->as_is =~ /^Add$/)) && (not ($token->as_is =~ /^\s$/)))
        {
          if ($alreadyOutput > 0)
          {
            if ($tdCount == 0)
            {
              $ticker .= " ";
            }
            elsif ($tdCount == 1)
            {
              $stockName .= " ";
            }
            elsif ($tdCount == 2)
            {
              $market .= " ";
            }
            elsif ($tdCount == 3)
            {
              $industry .= " ";
            }
          }
          if ($tdCount == 0)
          {
            $ticker .= $token->as_is;
          }
          elsif ($tdCount == 1)
          {
            $stockName .= $token->as_is;
          }
          elsif ($tdCount == 2)
          {
            $market .= $token->as_is;
          }
          elsif ($tdCount == 3)
          {
            $industry .= $token->as_is;
          }
          $alreadyOutput = 1;
          $previousValidToken = 1;
        }
        if ($tdCount == 5)
        {
          $rowString = $doc->td(-content => $doc->a(href => "", onclick => "window.opener.document.sub_acc_form.sub_acct_name.value = '" . $stockName . "'; window.opener.document.sub_acc_form.ticker.value = '" . $ticker . "'; window.opener.document.sub_acc_form.stockName.value = '" . $stockName . "'; window.opener.matchMarket('" . $market . "'); window.close(); return false;", -content => $doc->font(size => "-2", -content => $ticker))) .
                       $doc->td(-content => $doc->font(size => "-2", -content => $stockName)) .
                       $doc->td(-content => $doc->font(size => "-2", -content => $market)) .
                       $doc->td(-content => $doc->font(size => "-2", -content => $industry));
          $doc->print($doc->tr(bgcolor => "#ffffff", -content => $rowString) . "\n");
          $ticker = "";
          $stockName = "";
          $market = "";
          $industry = "";
          $tdCount = 0;
          $alreadyOutput = 0;
        }
      }
    }
  }

  if ($numValidTokens == 0)
  {
    $doc->print("<tr><td colspan=\"4\" bgcolor=\"#ffffff\"><a href=\"\" onClick=\"window.close(); return false;\">" . $self->{langObj}->map("noStockInfoReturned") . "</a></td></tr>\n");
  }

  return 1;
}

=item array updateStockPrices(tickerList => | stocks => , startDate =>, endDate =>, type =>, interval =>)

 Summary: this function retrieves data for the sent in list of stocks from the Internet by calling the StockQuote
           object methods to gather the data; or it takes the manualy prices given
           it then inserts the data into the database in the stock_price_tb table
 Required: tickerList, [startDate, endDate] (if type = historical)
 Optional: type = current/historical (defaults to current price), interval (if type = historical)
  Returns: array where index 0 is bool (1 on success, 0 on failure)
           index 1 is array ref to goodTickers
           index 2 is array ref to badTickers

=cut

sub updateStockPrices
{
  my $self = shift;
  my %args = ( @_ );
  my $errorString = "";
  my $tickerList = $args{'tickerList'};
  my $startDate = $args{'startDate'};
  my $endDate = $args{'endDate'};
  my $type = $args{'type'};
  my $manualPrice = $args{'manualPrice'};
  my $interval = $args{'interval'};
  my @goodTickers = ();
  my @badTickers = ();

  if (not defined $args{'tickerList'})
  {
    $errorString = "tickerList must be specified.";
    $self->error($errorString);
    return (0, \@goodTickers, \@badTickers);
  }
  if (defined $type && $type ne "historical" && $type ne "current" && $type ne "manual")
  {
    $errorString = "Invalid type ($type).  Must be 'historical' or 'current' or 'manual'.";
    $self->error($errorString);
    return (0, \@goodTickers, \@badTickers);
  }
  if (defined $type && $type eq "manual" && ((not defined $args{'manualPrice'} || $manualPrice eq "") && (not defined $args{'startDate'} || $startDate eq "")))
  {
    $errorString = "'manualPrice' and 'startDate' must be specified when type => 'manual'.";
    $self->error($errorString);
    return (0, \@goodTickers, \@badTickers);
  }

  if ((not defined $type) || $type eq "current" || ($type ne "current" && $type ne "historical" && $type ne "manual"))
  {
    # get the prices for the list of stocks
    (my $result, my $currentPrices) = $self->getStockQuotes(tickerList => $tickerList);
    if ($result ne "")
    {
      $errorString = "Looking up current prices failed.<br>" . $result;
      $self->error($errorString);
      return (0, \@goodTickers, \@badTickers);
    }
    my %priceHash = %{$currentPrices};
    # insert the prices into the database
    foreach my $ticker (keys %priceHash)
    {
      if ($ticker eq "") { next;  }
      my $stockObj = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
      if ($stockObj->selectIdByTicker($ticker)) { $stockObj->select; }
      if ($stockObj->error)
      {
        $errorString = "Failed to lookup stock information for ticker ($ticker).<br>" . $stockObj->errorMessage;
        $self->error($errorString);
        return (0, \@goodTickers, \@badTickers);
      }
      if ($stockObj->get("market") eq "None") {  next;  }
      push @goodTickers, $ticker;
      my @tempArray = keys %{$priceHash{$ticker}};
      my $date = $tempArray[0];
      my $sth = $self->{dbHandle}->write(sql => "DELETE FROM stock_price_tb WHERE sub_acct = ? AND date_of = ?", plug => [ $stockObj->get("id"), $date ]);
      if ($self->{dbHandle}->error())
      {
        $errorString = "Problem occurred while deleting the stock prices for sub-account (".$stockObj->get("id")."), ticker ($ticker) on date ($date).<br>" . $self->{dbHandle}->errorMessage;
        $self->error($errorString);
        $self->{dbHandle}->rollback;
        return (0, \@goodTickers, \@badTickers);
      }
      $sth = $self->{dbHandle}->write(sql => "INSERT INTO stock_price_tb (date_of, price, sub_acct) VALUES (?, ?, ?)", plug => [ $date, $priceHash{$ticker}{$date}, $stockObj->get("id") ]);
      if ($self->{dbHandle}->error())
      {
        $errorString = "Problem occurred while inserting the stock price for sub-account (".$stockObj->get("id")."), ticker ($ticker) on date ($date).<br>" . $self->{dbHandle}->errorMessage;
        $self->error($errorString);
        $self->{dbHandle}->rollback;
        return (0, \@goodTickers, \@badTickers);
      }
    }
    foreach my $ticker (@{$tickerList})
    {
      if ($ticker eq "") {  next;  }
      if (not grep /^$ticker$/, @goodTickers)
      {
        push @badTickers, $ticker;
      }
    }
  }
  elsif ($type eq "historical")
  {
    if ($startDate !~ /^\d{4}[-\\\/]\d{1,2}[-\\\/]\d{1,2}$/)
    {
      $self->setError("startDate = '$startDate' is invalid!\n");
      return (0, \@goodTickers, \@badTickers);
    }
    if ($endDate !~ /^\d{4}[-\\\/]\d{1,2}[-\\\/]\d{1,2}$/)
    {
      $self->setError("endDate = '$endDate' is invalid!\n");
      return (0, \@goodTickers, \@badTickers);
    }
    if ($interval !~ /^(d|w|m)/i)
    {
      $interval = "d"; # daily prices
    }
    else
    {
      $interval =~ /^(.)/;
      $interval = $1;
    }

    $startDate =~ /^(\d{4})-(\d{2})-(\d{2})$/;
    my $startMonth = $2;
    my $startDay = $3;
    my $startYear = $1;
    $endDate =~ /^(\d{4})-(\d{2})-(\d{2})$/;
    my $endMonth = $2;
    my $endDay = $3;
    my $endYear = $1;

    # get the prices for the list of stocks
    foreach my $ticker (@{$tickerList})
    {
      if ($ticker eq "") {  next;  }
      my $stockObj = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
      if ($stockObj->selectIdByTicker($ticker)) { $stockObj->select; }
      if ($stockObj->error())
      {
        $errorString = "Failed to lookup stock information for ticker ($ticker).<br>" . $stockObj->errorMessage;
        $self->error($errorString);
        return (0, \@goodTickers, \@badTickers);
      }
      if ($stockObj->get("market") eq "None") {  next;  }
      (my $result, my $historicalPrices) = $self->getHistoricalQuotes(ticker => $ticker, interval => $interval, startDay => $startDay, startMonth => $startMonth, startYear => $startYear, endDay => $endDay, endMonth => $endMonth, endYear => $endYear);
      if ($self->error)
      {
        $errorString = "Looking up historical prices failed on $ticker:$stockObj->{market}:$stockObj->{name}<br>" . $result;
        $self->error($errorString);
        return (0, \@goodTickers, \@badTickers);
      }
      my %priceHash = %{$historicalPrices};
      if (keys %priceHash == 0)
      {
        push @badTickers, $ticker;
      }
      else
      {
        push @goodTickers, $ticker;
      }
      foreach my $date (keys %priceHash)
      {
        my $sth = $self->{dbHandle}->write(sql => "DELETE FROM stock_price_tb WHERE sub_acct = ? AND date_of = ?", plug => [ $stockObj->get("id"), $date ]);
        if ($self->{dbHandle}->error())
        {
          $errorString = "Problem occurred while deleting the stock prices for sub-account (".$stockObj->get("id")."), ticker ($ticker) on date ($date).<br>" . $self->{dbHandle}->errorMessage;
          $self->error($errorString);
          $self->{dbHandle}->rollback;
          return (0, \@goodTickers, \@badTickers);
        }
        $sth = $self->{dbHandle}->write(sql => "INSERT INTO stock_price_tb (date_of, price, sub_acct) VALUES (?, ?, ?)", plug => [ $date, $priceHash{$date}, $stockObj->get("id") ]);
        if ($self->{dbHandle}->error())
        {
          $errorString = "Problem occurred while inserting the stock price for sub-account (".$stockObj->get("id")."), ticker ($ticker) on date ($date).<br>" . $self->{dbHandle}->errorMessage;
          $self->error($errorString);
          $self->{dbHandle}->rollback;
          return (0, \@goodTickers, \@badTickers);
        }
      }
    }
  }

  # everything looks good
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->error())
  {
    $errorString = "Problem committing work.<br>" . $self->{dbHandle}->errorMessage;
    $self->error($errorString);
    $self->{dbHandle}->rollback;
    return (0, \@goodTickers, \@badTickers);
  }

  return (1, \@goodTickers, \@badTickers);
}

sub updateManualPrice
{
  my $self = shift;
  my %args = ( stocks => undef, date => '', price => 0, @_ );
  my $errorString = "";
  my $date = $args{'date'};
  my $manualPrice = $args{'price'};
  if (scalar @{$args{stocks}} <= 0) { $self->error('<p>No stocks sent to manualPrice!</p>\n'); return 0; }

  # insert the prices into the database
  foreach my $stockObj (@{$args{stocks}})
  {
    if (!$stockObj->isValid) { $self->error('Warning! Invalid stock.'.$stockObj->errorMessage); return 0;  }
    my $sth = $self->{dbHandle}->write(sql => "DELETE FROM stock_price_tb WHERE sub_acct = ? AND date_of = ?", plug => [ $stockObj->get("id"), $date ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("Problem occurred while deleting the stock prices for sub-account (".$stockObj->get("id")."), ticker (".$stockObj->get("ticker").") on date ($date).<br>" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return 0;
    }
    $sth = $self->{dbHandle}->write(sql => "INSERT INTO stock_price_tb (date_of, price, sub_acct) VALUES (?, ?, ?)", plug => [ $date, $manualPrice, $stockObj->get("id") ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("Problem occurred while inserting the stock price for sub-account (".$stockObj->get("id")."), ticker (".$stockObj->get('ticker').") on date ($date).<br>" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return 0;
    }
  }
  $self->{dbHandle}->commit;
  if ($self->{dbHandle}->error())
  {
    $errorString = "Problem committing work.<br>" . $self->{dbHandle}->errorMessage;
    $self->error($errorString);
    $self->{dbHandle}->rollback;
    return 0;
  }

  return 1;
}

=item @ getStockQuotes(tickerList =>, url =>)
  Summary:   Use http://yahoo.com/ site to retrieve a comma-separated-values
    file with the data for the stock symbols in the format
    "<ticker>",<last price>,"<date M[M]/D[D]/YYYY>".  If the data
    or the symbol does not exist, "N/A" is returned in the <date> field.
  Requires:  tickerList (a reference to a list of ticker symbols for stocks to look up the last price)
  Optional:  url => alternative url to use.  Defaults to
     http://finance.yahoo.com/d/quotes.csv

  Returns:   array with the first element containing an error message (if any) and the second element containing a multi-dimensional hash with {<ticker>}{<date>}-><last price> on sucess or undef on error
=cut
sub getStockQuotes
{
  my %args = ( tickerList => undef, url => "http://finance.yahoo.com/d/quotes.csv", @_ );
  my $errorString = "StockQuote::getStockQuotes() - Error: ";
  my %stockQuotes = ();

  if (! defined $args{tickerList} || scalar @{$args{tickerList}} == 0)
  {
    return ($errorString . "tickerList empty.  tickerList must be a list of tickers.", undef);
  }
  my @tickerList = @{$args{'tickerList'}};
  # encode the list of stocks to retrieve using '+' as the separator
  my $stockList = join("+", @tickerList);
  # build the URL for retrieving the stock quote information file
  my $URL = "\"$args{url}?s=$stockList&f=sl1d1&e=.csv\"";
  # create a random file in /tmp/...
  my $tempFile = tmpnam();
  # make sure that "wget" exists on the system and is in the right place
  if (not -e "/usr/bin/wget")
  {
    return ($errorString . "/usr/bin/wget does not exist!  Make sure the 'wget' package is installed and in the necessary location.", undef);
  }
  # use "wget" to retrieve the URL
  my $call = "/usr/bin/wget -q -O $tempFile $URL > /dev/null 2>&1";
  my @args = ($call);
  my $result = system(@args);
  if ($result != 0)
  {
    unlink $tempFile;
    return ($errorString . "System call to retrieve stock prices failed. $call", undef);
  }
  # open the temporary file which should contain the data for the stocks
  if (!open(STOCKQUOTES, "<$tempFile"))
  {
    return ($errorString . "Could not open temporary file.", undef);
  }
  # read each line of data, splitting on commas
  while (my $line = <STOCKQUOTES>)
  {
    (my $symbol, my $price, my $date) = split(',', $line, 3);
    chomp $date;
    if ($date !~ /N\/A/) # valid value returned for this row of data in the file
    {
      $symbol =~ s/\"//g; # remove double quotes from symbol and date information
      $date =~ s/\"//g;

      (my $stockTicker) = grep /^$symbol$/i, @tickerList;  # we should probably check here to make sure Yahoo! is not returning to us stock tickers/quotes we didn't ask for
      $stockQuotes{$stockTicker} = ();
      $stockQuotes{$stockTicker}{$date} = $price;
    }
  }
  # close the temporary file
  close(STOCKQUOTES);

  # remove the temporary file
  unlink $tempFile;

#  if (scalar keys %stockQuotes == 0)
#  {
#    return ($errorString . "No stock quotes returned.", undef);
#  }

  return ("", \%stockQuotes);
}

=item @ getHistoricalQuotes(ticker =>, . . .)

  Summary:  Use http://ichart.finance.yahoo.com/ site to retrieve
    a comma-separated-values file with the data for the stock
    symbol in the format <date D[D]-MMM-YY>,<open>,<high>,<low>,<close>,<volume>.
    Date format changes for monthly gatherings.  The close price
    exists for the last day of the month.   <MMM YY>
    If the data or the symbol does not exist, "No data available
    for <ticker>" is returned in the first line of the file.
  Requires: ticker (the ticker symbol to lookup close price for the date range)
           startMonth (the beginning of the date range to lookup closing price information)
           startDay
           startYear
           endMonth (the end of the date range to lookup closing price information)
           endDay
           endYear
  Optional: interval (d = daily, w = weekly, m = monthly;
            defaults to daily in the absence of the argument or in the case
            of an unrecognized value)
  Returns: array with the first element containing an error message
  (if any) and the second element containing a hash with
  {<date>}-><last price> on sucess or undef on error

=cut
sub getHistoricalQuotes
{
  my $self = shift;
  my %args = ( ticker => "", startMonth => undef, startDay => undef, startYear => undef,
    endMonth => undef, endDay => undef, endYear => undef, interval => undef,
    url => "http://ichart.finance.yahoo.com/table.csv",
    @_ );
  my $errorString = "StockQuote::getHistoricalQuotes() - Error: ";
  my $symbol = $args{'ticker'};
  my $startMonth = $args{'startMonth'};
  my $startDay = $args{'startDay'};
  my $startYear = $args{'startYear'};
  my $endMonth = $args{'endMonth'};
  my $endDay = $args{'endDay'};
  my $endYear = $args{'endYear'};
  my $interval = $args{'interval'};
  my %historicalQuotes = ();

  if ($symbol eq "")
  {
    $self->error("ticker empty or not specified.");
    return undef;
  }
  if ((not defined $startMonth) || $startMonth eq "")
  {
    $self->error("startMonth empty or not specified.");
    return undef;
  }
  if ($startMonth !~ /^((0?[1-9])|(1[0-2]))$/)
  {
    # invalid month value
    $self->error("Invalid value for startMonth (" . $startMonth . ").");
    return undef;
  }
  $startMonth--;  # fixup per Yahoo's change
  if ((not defined $startDay) || $startDay eq "")
  {
    $self->error("startDay empty or not specified.");
    return undef;
  }
  if ($startDay !~ /^((0?[1-9])|([12][0-9])|(3[01]))$/)
  {
    # invalid day value
    $self->error("Invalid value for startDay (" . $startDay . ").");
    return undef;
  }
  if ((not defined $startYear) || $startYear eq "")
  {
    $self->error("startYear empty or not specified.");
    return undef;
  }
  if ($startYear !~ /^[0-9]+$/)
  {
    # invalid year value
    $self->error("Invalid value for startYear (" . $startYear . ").");
    return undef;
  }
  if ((not defined $endMonth) || $endMonth eq "")
  {
    $self->error("endMonth empty or not specified.");
    return undef;
  }
  if ($endMonth !~ /^((0?[1-9])|(1[0-2]))$/)
  {
    # invalid month value
    $self->error("Invalid value for endMonth (" . $endMonth . ").");
    return undef;
  }
  $endMonth--;  # fixup per Yahoo's change
  if ((not defined $endDay) || $endDay eq "")
  {
    $self->error("endDay empty or not specified.");
    return undef;
  }
  if ($endDay !~ /^((0?[1-9])|([12][0-9])|(3[01]))$/)
  {
    # invalid day value
    $self->error("Invalid value for endDay (" . $endDay . ").");
    return undef;
  }
  if ((not defined $endYear) || $endYear eq "")
  {
    $self->error("endYear empty or not specified.");
    return undef;
  }
  if ($endYear !~ /^[0-9]+$/)
  {
    # invalid year value
    $self->error("Invalid value for endYear (" . $endYear . ").");
    return undef;
  }
  if ((not defined $interval) || ($interval ne "d" && $interval ne "w" && $interval ne "m"))
  {
    $interval = "d";
  }

  # build the URL for retrieving the stock quote information file
  my $URL = "\"$args{url}?s=$symbol&a=$startMonth&b=$startDay&c=$startYear&d=$endMonth&e=$endDay&f=$endYear&g=$interval&q=q&y=0&z=$symbol&x=.csv\"";
  # create a random file in /tmp/...
  my $tempFile = tmpnam();
  # make sure that "wget" exists on the system and is in the right place
  if (not -e "/usr/bin/wget")
  {
    $self->error("/usr/bin/wget does not exist!  Make sure the 'wget' package is installed and in the necessary location.");
    return undef;
  }
  # use "wget" to retrieve the URL
  my $call = "/usr/bin/wget -q -O $tempFile $URL > /dev/null 2>&1";
  my @args = ($call);
  my $result = system(@args);
  if ($result != 0)
  {
    unlink $tempFile;
    $self->error("System call to retrieve stock prices failed. $call");
    return undef;
  }
  # open the temporary file which should contain the data for the stocks
  if (!open(HISTORICALQUOTES, "<$tempFile"))
  {
    $self->error("Could not open temporary file.");
    return undef;
  }
  # read the first line to determine if there is valid data in the file and in what order it should be processed
  my $line = <HISTORICALQUOTES>;
  my @fields = split(',', $line);
  if ((scalar @fields) <= 1)
  {
    # not enough fields returned
    $self->error("Invalid data returned. (Not enough fields)");
    return undef;
  }
  my %fieldOrder = ();
  for (my $labelCount=0; $labelCount < (scalar @fields); $labelCount++)
  {
    if ($fields[$labelCount] =~ /date|week of/i)
    {
      $fieldOrder{'DATE'} = $labelCount;
    }
    elsif ($fields[$labelCount] =~ /close/i)
    {
      $fieldOrder{'CLOSE'} = $labelCount;
    }
  }
  # read each line of data, splitting on commas
  while ($line = <HISTORICALQUOTES>)
  {
    my @fieldValues = split(',', $line);
    my $date = $fieldValues[$fieldOrder{'DATE'}];
    my $price = $fieldValues[$fieldOrder{'CLOSE'}];

    my $day;
    my $month;
    my $year;
    # massage date format into DD-MMM-YYYY for recognized date formats
    if ($date =~ /^(\w+) (\d+)$/)
    {
      $month = $1;
      $year = $2;

      if ($year =~ /\d{2}/) # if two-digit year, convert it to four-digit year; otherwise, leave it alone
      {
        if (($year < 100) && ($year >= 50))
        {
          $year += 1900; # change year format into 1900 + year using 1950 as the starting point; this code will only work up till the year 2049 for two-digit years
        }
      }

      if ($month =~ /apr/i || $month =~ /jun/i || $month =~ /sep/i || $month =~ /nov/i)
      {
        $day = "30";
      }
      elsif ($month =~ /jan/i || $month =~ /mar/i || $month =~ /may/i || $month =~ /jul/i || $month =~ /aug/i || $month =~ /oct/i || $month =~ /dec/i)
      {
        $day = "31";
      }
      elsif ($month =~ /feb/i)
      {
        # use the date command to determine the ending day of february for the given year
        $day = `/bin/date -d "01-Mar-$year 1 day ago" +%d`;
        chomp $day;
      }
    }
    elsif ($date =~ /^(\d+)-(\w+)-(\d+)$/)
    {
      $day = $1;
      $month = $2;
      $year = $3;

      if ($year =~ /\d{2}/) # if two-digit year, convert it to four-digit year; otherwise, leave it alone
      {
        if (($year < 100) && ($year >= 50))
        {
          $year += 1900; # change year format into 1900 + year using 1950 as the starting point; this code will only work up till the year 2049 for two-digit years
        }
      }
    }
    $date = $day . "-" . $month . "-" . $year;
    $historicalQuotes{$date} = $price;
  }
  # close the temporary file
  close(HISTORICALQUOTES);

  # remove the temporary file
  unlink $tempFile;

#  if (scalar keys %historicalQuotes == 0)
#  {
#    return ($errorString . "No stock quotes returned.", undef);
#  }

  return ("", \%historicalQuotes);
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::Base(3)

=cut
