# WatchList.pm - The Object Class that provides a WatchList Object
# Copyright (c) 2000, 2002 PC & Web Xperience, Inc. http://www.pcxperience.com/
# All rights reserved.  This program is free software; you can redistribute it
# and/or modify it under the same terms of the GPL

=head1 NAME

WatchList - Object used to build a WatchList Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Investements::WatchList;
  my $obj = Portal::App::WatchList->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

=head1

WatchList is a WatchList class.

=cut

package Portal::Accounting::Investments::WatchList;
use strict;
use Portal::Accounting::Ledger;
use Portal::Accounting::DB::Security;
use Portal::Accounting::Objects::Stock;
use Finance::Quote;
use POSIX;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Ledger Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.13 $';

# global variables.
# modify to represent the commands this state provides.
my %commands = ( display => "Watchlists",
  edit => "Edit Watchlist",
  save => "Save Watchlist",
  delete => "Delete Watchlist",
  new => "Create Watchlist",
  statusScreen => "Status",
  view => "Watchlist",
);


=head1 Exported FUNCTIONS

=head1

=head2  scalar new(langObj)

    Creates a new instance of the Portal::Accounting::WatchList
    object.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # assign anything that you need
  $self->{state} = "Investments::WatchList";
  $self->{commands} = \%commands;

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.
  $self->{stockObj} = Portal::Accounting::Objects::Stock->new(dbHandle => $self->{dbHandle}, langObj => $self->{langObj}, debug => 1);
  if ($self->{stockObj}->error)
  {
    $self->prefixError("Error while initializing DB::Main<br>\n".$self->{stockObj}->errorMessage);
    return $self;
  }
  $self->{tickerURL} = "http://finance.yahoo.com/q?s=%s";

  return $self;
}

=head2 bool isValid(void)

    Returns 0 or 1 to indicate if the object is valid.  The error will be
    available via errorMessage().

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # do validation code here.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->prefixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub c_display
{
  my $self = shift;
  my %args = ( message => "", @_ );
  my $doc = $self->prepDoc;
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  $message .= $self->{input}->{message};
  my $ALL = $phrases->map("ALL");
  my $SHORT = $phrases->map("SHORT");
  $doc->setFocus("body");
  my @lists = $self->getWatchlistNames(ledgerCode => $ledgerCode);
  if ($self->error)
  {
    $doc->print("There was an error while attempting to retrieve the watch lists\n".$self->errorMessage);
    return $doc;
  }
  if ($message ne "")
  {
    $doc->print("$message<br>\n");
    $doc->print("<hr>\n");
  }
  $doc->print("<form name=\"mainForm\" method=\"post\">\n");
  $doc->print("  <input type=\"hidden\" name=\"app\" value=\"Accounting\">\n");
  $doc->print("  <input type=\"hidden\" name=\"state\" value=\"Investments::WatchList\">\n");
  $doc->print("  <input type=\"hidden\" name=\"command\" value=\"view\">\n");
  $doc->print("  <input type=\"hidden\" name=\"ledgerCode\" value=\"$ledgerCode\">\n");
  if (@lists > 0)
  {
    $doc->print("<select name=\"groupName\">\n");
    $doc->print("  <option value=\"ALL\">$ALL</option>\n");
    $doc->print("  <option value=\"SHORT\">$SHORT</option>\n");
    foreach my $list (@lists)
    {
      $doc->print("  <option value=\"$list\">$list</option>\n");
    }
    $doc->print("</select>\n");
    $doc->print("<a href=\"\" onClick=\"document.mainForm.submit(); return false;\">".$phrases->map("View")."</a> ");
    $doc->print("<a href=\"?$self->{getInfo}&command=edit&new=0\" onClick=\"document.mainForm.command.value='edit'; document.mainForm.submit(); return false;\" title=\"".$phrases->map("Edit")."\">".$phrases->map("Edit")."</a>");
    $doc->print("&nbsp;");
    $doc->print("<a href=\"?$self->{getInfo}&command=delete\" onClick=\"if (confirm('" . $phrases->map("confirmDeleteWatchList") . "')) { document.mainForm.command.value='delete'; document.mainForm.submit(); } return false;\" title=\"".$phrases->map("Delete")."\">".$phrases->map("Delete")."</a>");
    $doc->print("&nbsp;");
  }
  else
  {
    $doc->print($phrases->map("No Watchlists defined")."<br>\n");
  }
  $doc->print("<a href=\"?app=$self->{app}&state=$self->{state}&command=new&new=1\" onClick=\"document.mainForm.command.value='new'; document.mainForm.submit(); return false;\" title=\"".$phrases->map("New")."\">".$phrases->map("New")."</a>");
  $doc->print("</form>\n");
  return $doc;
}

sub c_new
{
  my $self = shift;
  return $self->c_edit;
}

sub c_edit
{
  my $self = shift;
  my %args = ( message => "", @_ );
  my $command = $self->{input}->{command};
  my $new = 1;
  if ($new eq "") { $new = 1; }
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $groupName = "";
  my $oldGroupName = "";
  my $assignedTickers = "";
  my $short = 0;

  # get list of all stocks in the system
  my @stocks = $self->{stockObj}->getListOfStocks();
  if ($self->{stockObj}->error)
  {
    return $self->c_display(message => $self->{stockObj}->errorMessage);
  }
  if (scalar @stocks == 0)
  {
    return $self->statusScreen(message => $phrases->map("noStocksCreated"));
  }

  if ($command eq "save")
  { #pull params from $self->input bc error caused redisplay
    $groupName = $self->{input}->{groupName};
    if ($groupName eq "") { return $self->c_display(message => $phrases->map("GroupNameCannotBeEmpty"));  }
    $oldGroupName = $self->{input}->{oldGroupName};
    $assignedTickers = $self->{input}->{assignedTickers};
    $new = $self->{input}->{new};
  }
  elsif ($command eq "edit")
  {
    $new = 0;
    $groupName = $self->{input}->{groupName};
    if ($groupName eq "ALL" || $groupName eq "SHORT") { return $self->c_display(message => $phrases->map("cannotEditShortAll"));  }
    $oldGroupName = $self->{input}->{groupName};
    # determine which stocks are assigned to the given groupName
    $short = $self->isShort($groupName);
    my @assignedStocks = $self->getWatchlistArray(ticker => "yes", orderBy => "name", group => $groupName, ledgerCode => $ledgerCode);
    if ($self->error)
    {
      $message = $phrases->map("errorLookingUpStockWatchListInfo") . "<br>" . $message;
    }
    foreach my $assignedStock (@assignedStocks)
    {
      #(my $sTicker, my $sName, my $sSubAcct) = @{$assignedStock};
      if ($assignedTickers ne "")
      {
        $assignedTickers .= ", ";
      }
      #$assignedTickers .= $sTicker;
      $assignedTickers .= $assignedStock->get("ticker");
    }
  }
  my $doc = $self->prepDoc;
  if ($message ne "")
  {
    $doc->print("$message<br>\n");
    $doc->print("<hr>\n");
  }
  # javascript code
  $doc->setFocus("javascript");
  $doc->print(<<"ENDOF");
function assignAllTickers()
{
  var allTickers = '';
  for (var count=0; count < document.editForm.availableTickersSelect.options.length; count++)
  {
    if (document.editForm.availableTickersSelect.options[count].value != '')
    {
      if (count > 0 && allTickers != '')
      {
        allTickers += ', ';
      }
      allTickers += document.editForm.availableTickersSelect.options[count].value;
    }
  }
  document.editForm.assignedTickers.value = allTickers;
}

function moveToAssignedTickers()
{
  if (document.editForm.availableTickersSelect.options[document.editForm.availableTickersSelect.selectedIndex].value != '')
  {
    if (document.editForm.assignedTickers.value != '')
    {
      document.editForm.assignedTickers.value += ', ';
    }
    document.editForm.assignedTickers.value += document.editForm.availableTickersSelect.options[document.editForm.availableTickersSelect.selectedIndex].value;
  }
}

function validateAssignedTickers()
{
  var regEXP = /^(\\s*([\\w\\d\\.]+),\\s*)*(\\s*[\\w\\d\\.]+\\s*)\$/;
  var result = document.editForm.assignedTickers.value.match(regEXP);
  if (result != null)
  {
    return true;
  }
  else
  {
    return false;
  }
}

function validateEditForm()
{
  if (!validateAssignedTickers())
  {
ENDOF
  $doc->print("    alert('" . $phrases->map("enterListOfCommaSeparatedTickers") . "');\n");
  $doc->print(<<"ENDOF");
    return false;
  }

  if (document.editForm.groupName.value.match(/^\\s*\$/))
  {
ENDOF
  $doc->print("    alert('" . sprintf($phrases->map("InvalidWatchListName"), '') . "');\n");
  $doc->print(<<"ENDOF");
    return false;
  }

  return true;
}
ENDOF
  #display text field with name of list above first multi-select
  my $Save = $phrases->map("Save");
  my $assAll = $phrases->map("AssignAllTickers");
  my $unAss = $phrases->map("unAssignAllTickers");
  my $Ass = $phrases->map("Assigned");
  my $Avail = $phrases->map("Available");
  my $include = $phrases->map("Include In Short List");
  my $checked = "";
  if ($short) { $checked = "checked"; }
  $doc->setFocus("body");
  $doc->print(<<"ENDOF");
<form name="editForm" method="post">
  $self->{postInfo}
  <input type="hidden" name="command" value="save">
  <input type="hidden" name="oldGroupName" value="$oldGroupName">
  <input type="hidden" name="new" value="$new">
  <table>
    <tr valign="bottom">
      <td>
        <input type="text" name="groupName" value="$groupName"><br>
        $Avail<br>
        <select name="availableTickersSelect" size="10" onChange="moveToAssignedTickers();">
ENDOF
  foreach my $stock (@stocks)
  {
    $doc->print("    <option value=\"" . $stock->get("ticker") . "\">" . $stock->get("name") . " (" . $stock->get("ticker") . ")\n");
  }
  $doc->print(<<"ENDOF");
        </select>
      </td>
      <td>
        <input type="button" name="assignAllTickersButton" value=">>" title="$assAll" onClick="assignAllTickers();"><br>
        <input type="button" name="unAssignAllTickersButton" value="<<" title="$unAss" onClick="document.editForm.assignedTickers.value = '';">
      </td>
      <td>
        $Ass<br>
        <textarea name="assignedTickers" rows="13" cols="40" wrap="soft">$assignedTickers</textarea>
      </td></tr>
    <tr>
      <td colspan="2">
        <input name="short" type="checkbox" $checked>$include
      </td>
      <td align="right">
        <a href="" onClick="if (validateEditForm()) { document.editForm.submit(); } return false;">$Save</a>
      </td>
    </tr></table>
  </form>
ENDOF
  my $watchlist = $phrases->map("Watchlists");
  $doc->print("<p><a href=\"?$self->getInfo&command=display\">$watchlist</a></p>\n");
  return $doc;
}

sub c_save
{
  my $self = shift;
  my %args = ( message => "", command => "", @_ );
  my $command = $args{command};
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $groupName = $self->{input}->{groupName};
  my $oldGroupName = $self->{input}->{oldGroupName};
  my $assignedTickers = $self->{input}->{assignedTickers};
  my $short = 'false';
  if (exists $self->{input}->{short}) { $short = 'true';  }

  #verify data
  if ($groupName eq "")
  {
    # if error, return edit( message => "foo")
    return $self->c_edit( message => sprintf($phrases->map("InvalidWatchListName") ) ) ;
  }
  if ($assignedTickers eq "")
  {
    return $self->c_edit(message => sprintf($phrases->map("ChooseTickersForWatchList") ) );
  }
  if (not ($assignedTickers =~ /^(\s*([\w\d\.]+),\s*)*(\s*[\w\d\.]+\s*)$/))
  {
    return $self->c_edit(message => $phrases->map("enterListOfCommaSeparatedTickers")) ;
  }
  # if oldCommand was new, assign the groupName to oldGroupName
  if (!$self->{input}->{new})
  {
    my $result = $self->unWatchAllStocks(ledgerCode => $ledgerCode, groupName => $oldGroupName);
    if ($self->error || $result ne "1")
    {
      return $self->c_edit(message => sprintf($phrases->map("UnWatchingAllStocksWhileSavingWatchListFailed"), $oldGroupName) . " " . $self->errorMessage, groupName => $groupName, oldGroupName => $oldGroupName);
    }
  }
  # remove leading and trailing whitespace from each ticker definition
  $assignedTickers =~ s/\s*//g;
  my @assignedTickers = split /,/, $assignedTickers;
  #attempt to put data in Database (delete old, insert new)
  foreach my $stock (@assignedTickers)
  {
    # lookup sub-account number for each ticker
    my $stockObj = Portal::Accounting::Objects::Stock->new(langObj => $self->{langObj}, dbHandle => $self->{dbHandle});
    if ($stockObj->selectIdByTicker($stock))
    {
      $stockObj->select();
    }
    if ($stockObj->error)
    {
      return $self->c_edit(message => $stockObj->errorMessage, groupName => $groupName, oldGroupName => $oldGroupName);
    }
    # watch each stock
    my $result = $self->watchStock(number => $stockObj->id, ledgerCode => $ledgerCode, groupName => $groupName, short => $short);
    if ($self->error)
    {
      return $self->c_edit(message => sprintf($phrases->map("SavingStockForWatchListFailed"), $stock, $groupName) . $self->errorMessage , groupName => $groupName, oldGroupName => $oldGroupName);
    }
  }
  return $self->statusScreen(message => sprintf($phrases->map("SavingWatchListSucceeded"), $groupName));
}

sub c_delete
{
  my $self = shift;
  my %args = ( message => "", command => "", @_ );
  my $command = $args{command};
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $groupName = $self->{input}->{groupName};
  if ($groupName eq "ALL" || $groupName eq "SHORT") { return $self->c_display(message => $phrases->map("cannotDeleteAllShort"));  }

  if ($self->{input}->{groupName} eq "")
  {
    return $self->statusScreen(message => sprintf($phrases->map("InvalidWatchListName"), $groupName));
  }
  #attempt to delete specified list from db
  my $result = $self->unWatchAllStocks(ledgerCode => $ledgerCode, groupName => $groupName);
  if ($self->error || $result ne "1")
  {
    return $self->statusScreen(message => sprintf($phrases->map("DeletionOfWatchListFailed"), $groupName));
  }
  return $self->statusScreen(message => sprintf($phrases->map("DeletionOfWatchListSucceeded"), $groupName));
}

sub statusScreen
{
  my $self = shift;
  my %args = ( message => "", command => "", @_ );
  my $message = $args{message};
  my $phrases = $self->{langObj};
  my $ledgerCode = $self->{input}->{ledgerCode};

  my $doc = HTMLObject::Normal->new;
  $doc->setTitle($commands{statusScreen});
  $doc->setFocus("body");
  $doc->print(<<"ENDOF");
<hr>
$message
<form name="statusForm" method="post">
  $self->{postInfo}
  <input type="hidden" name="command" value="display">
  <input type="hidden" name="message" value="$message">
ENDOF
  $doc->setOnload(code => "document.statusForm.submit();\n");

  return $doc;
}

# getWatchListHTML
# takes: ledgerCode
# returns: HTML <table> output in the event of success; undef if
# something fails
# This routine queries the database for the list of watched stocks.  If
# no stocks exist in the watched stocks table, output "No stocks are
# currently being watched!", but do not return an error condition.  If
# individual but not all stocks fail to be fetched, only return those
# stocks that were successfully fetched, and indicate those that failed
# in a small NOTE at the bottom of the table output.
sub getWatchListHTML
{
  my $self = shift;
  my %args = ( ledgerCode => undef, group => "Default", @_ );
  my $ledgerCode = $args{ledgerCode};
  my $group = $args{group};
  my $htmlString = "";

  # validate input
  if (length $ledgerCode == 0)
  {
    $self->error("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  # get list of watched stocks from database with ticker symbols returned
  # build an array from the watchList hash
  my $array;
  my $watchList;
  ($watchList, $array) = $self->getWatchlistData(ledgerCode => $ledgerCode, group => $group);
  if ($self->error)  {    return undef;  }

  # fetch the information for the ticker symbols using Finance::Quote module
  # hardcode the use of the "usa" method for now
  my $financeQuoteObj = Finance::Quote->new;
  # set the list of required labels that must exist for any stock quotes returned
  $financeQuoteObj->require_labels("name", "last", "net", "p_change", "volume", "date");
  $financeQuoteObj->failover(1);
  my %quotes = $financeQuoteObj->fetch("usa", @{$array});
  # if nothing returned from the fetch, return an error message, but do not set an error condition
  if (not %quotes)
  {
    $htmlString = "<b><i>" . $self->{langObj}->map("noInformationReturnedFromFetch") . "</i></b>\n";
    return $htmlString;
  }

  # run through fetch results and output a row of information for each successfully
  # returned result ordered by company name
  my $failedTickers = "";
   my $date = POSIX::strftime('%m/%d/%Y', localtime);
   (my $localMonth, my $localDay, my $localYear) = split '/', $date, 3;
#  (my $localMonth, my $localDay, my $localYear) = $self->getDate;
  my $colorClass = "";
  my $totalGL = 0;
  my $totalCost = 0;
  my $totalMkt = 0;
  my $classGainLoss = "WLnoGainLossChange";
  my $Company = $self->{langObj}->map("Company");
  my $Ticker = $self->{langObj}->map("Ticker");
  my $TotalShares = $self->{langObj}->map("Shares");
  my $avgcps = $self->{langObj}->map("avgCostPerShare");
  my $Price = $self->{langObj}->map("Price");
  my $Change = $self->{langObj}->map("Change");
  my $PercentChange = $self->{langObj}->map("percentChange");
  my $volume = $self->{langObj}->map("volumeTraded");
  my $mktval = $self->{langObj}->map("Mkt Val");
  my $totalCost = $self->{langObj}->map("Cost");
  my $GainLoss = $self->{langObj}->map("Gain/Loss");
  $htmlString .= <<"ENDOF";
<center>
<a href="?$self->{getInfo}&command=edit&groupName=$group"><span class="WLgroupName">$group</span></a>
<table class="WLquoteTable" width="100\%" border="0" cellspacing="2" cellpadding="2">
  <tr>
    <td class="WLcompanyNameLabel"><b>$Company</b></td>
    <td class="WLtickerSymbolLabel"><b>$Ticker</b></td>
    <td align="right" class="WLtotalSharesLabel"><b>$TotalShares</b></td>
    <td align="right" class="WLaverageCostPerShareLabel"><b>$avgcps</b></td>
    <td align="right" class="WLlastPriceLabel"><b>$Price</b></td>
    <td align="right" class="WLnetChangeLabel"><b>$Change</b></td>
    <td>&nbsp</td>
    <td align="right" class="WLpercentChangeLabel"><b>$PercentChange</b></td>
    <td align="right" class="WLvolumeTradedLabel"><b>$volume</b></td>
    <td align="right" class="WLtotalMarketValueLabel"><b>$mktval</b></td>
    <td align="right" class="WLtotalCostLabel"><b>$totalCost</b></td>
    <td align="right" class="WLgainLossLabel"><b>$GainLoss</b></td>
    <td>&nbsp</td>
  </tr>
ENDOF

  foreach my $name (sort keys %{$watchList})
  {
    my $ticker = $watchList->{$name}->{ticker};
    my $price = "0.00";
    if (defined $watchList->{$name}->{price}) { $price = $watchList->{$name}->{price}; }

    # build a list of failed ticker symbols to output below the table of results
    if ((exists $quotes{$ticker}) && (not $quotes{$ticker, "success"}))
    {
      if ($failedTickers)
      {
        $failedTickers .= ", ";
      }
      $failedTickers .= $ticker;
    }
    else
    {
      $htmlString .= "  <tr>\n";
      # company name
      #my $name = ucfirst(lc($quotes{$ticker, "name"}));
      $htmlString .= "    <td class=\"WLcompanyName\">$name</td>";
      # ticker symbol (in parenthesis) with link to popup window
      my $substitutedTickerURL = sprintf($self->{tickerURL}, $ticker);
      $htmlString .= "<td class=\"WLtickerSymbol\"><a href=\"$substitutedTickerURL\" target=\"tickerPopUpWindow$ticker\">$ticker</a></td>";
      # total shares for stock
      $htmlString .= "<td class=\"WLtotalShares\">" . $self->commify(sprintf("%.0f", $watchList->{$name}->{shares})) . "</td>";
      # average cost per share for stock
      $htmlString .= "<td class=\"WLaverageCostPerShare\">" . $self->commify(sprintf("%.2f", $watchList->{$name}->{cost_per_share})) . "</td>";
      my $net = "0.00";
      my $p = "0";
      my $volume = "0";
      if (defined $quotes{$ticker, "last"})
      {
        $price = $self->commify($quotes{$ticker, "last"});
        $net = $self->commify($quotes{$ticker, "net"});
        $p = $quotes{$ticker, "p_change"};
        $volume = $self->commify(sprintf("%.01f", $quotes{$ticker, "volume"}/1000000) );
      }
      else  {  $price = $self->commify(sprintf("%.02f",$price));  }

      # last traded price
      $htmlString .= "<td class=\"WLlastPrice\">$price</td>";
      # net change (color coded)
      if ($quotes{$ticker, "net"} > 0)
      {
        $htmlString .= "<td class=\"WLnetIncrease\">$net</td>";
      }
      elsif ($quotes{$ticker, "net"} < 0)
      {
        $htmlString .= "<td class=\"WLnetDecrease\">$net</td>";
      }
      else
      {
        $htmlString .= "<td class=\"WLnetChange\">$net</td>";
      }
      # determine if there needs to be an arrow image and in what direction
      if ($quotes{$ticker, "net"} > 0)
      {
        $htmlString .= "<td class=\"WLnetIncreaseImage\"><img src=\"". $self->{imageDir} . "/images/Accounting/stock_increase.gif\" border=\"0\"></img></td>";
      }
      elsif ($quotes{$ticker, "net"} < 0)
      {
        $htmlString .= "<td class=\"WLnetDecreaseImage\"><img src=\"". $self->{imageDir} . "/images/Accounting/stock_decrease.gif\" border=\"0\"></img></td>";
      }
      else
      {
        # empty table cell
        $htmlString .= "<td>&nbsp;</td>";
      }
      # percent change (color coded)
      if ($quotes{$ticker, "p_change"} > 0)
      {
        $htmlString .= "<td class=\"WLpercentIncrease\">$p&#37;</td>";
      }
      elsif ($quotes{$ticker, "p_change"} < 0)
      {
        $htmlString .= "<td class=\"WLpercentDecrease\">$p&#37;</td>";
      }
      else
      {
        $htmlString .= "<td class=\"WLpercentChange\">$p</td>";
      }
      # volume traded
      $htmlString .= "<td class=\"WLvolumeTraded\">" . $volume . "</td>";
      # total market value
      my $marketValue = $watchList->{$name}->{shares} * $price;
      $htmlString .= "<td class=\"WLmarketValue\">" . $self->commify(sprintf("%.0f", $marketValue)) . "</td>";
      $totalMkt += $marketValue;
      # total cost
      my $cost = $watchList->{$name}->{cost};
      $htmlString .= "<td class=\"WLcost\">" . $self->commify(sprintf("%.0f", $cost)) . "</td>";
      $totalCost += $cost;
      # gain/loss
      my $gl = $marketValue - $cost;
      $totalGL += $gl;
      $classGainLoss = "WLnoGainLossChange";
      if ($gl > 0.0)
      {
        $classGainLoss = "WLgain";
      }
      elsif ($gl < 0.0)
      {
        $classGainLoss = "WLloss";
      }
      $htmlString .= "<td class=\"$classGainLoss\">" . $self->commify(sprintf("%.0f", $gl)) . "</td>";
      $htmlString .= "\n  </tr>\n";
    }
  }

  # sub-totals
  $htmlString .= <<"ENDOF";
    <tr>
      <td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td>
      <td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td>
ENDOF
  $htmlString .= "      <td class=\"WLtotalMarketValue\">";
  $htmlString .= $self->commify(sprintf("%.0f", $totalMkt)) . "</td>\n";
  $htmlString .= "      <td class=\"WLtotalCost\">";
  $htmlString .= $self->commify(sprintf("%.0f", $totalCost)) . "</td>\n";
  $classGainLoss = "WLnoGainLossChange";
  if ($totalGL > 0.0)
  {
    $classGainLoss = "WLgain";
  }
  elsif ($totalGL < 0.0)
  {
    $classGainLoss = "WLloss";
  }
  $htmlString .= "      <td class=\"$classGainLoss"."Total\">";
  $htmlString .= $self->commify(sprintf("%.0f", $totalGL)) . "</td>\n";
  $htmlString .= <<"ENDOF";
      </td><td>&nbsp;</td>
    </tr>
  </table>
ENDOF
  # add the list of failed ticker lookups, if any, to the end of the table output
  if ($failedTickers)
  {
    $htmlString .= "<br><br>" . $self->{langObj}->map("tickersFailedToLookup") . ": <b>$failedTickers</b>\n";
  }
  $htmlString .= "</span>\n";

  return $htmlString;
}

sub c_view
{
  my $self = shift;
  my $doc = $self->prepDoc;
  my $group = $self->input->{groupName};
  my $ledgerCode = $self->{input}->{ledgerCode};
  my $edit = $self->langObj->map("Watchlists");
  my %foo;

  $foo{css} = <<"ENDOF";  # output the CSS definitions for the Stock Watch List
        .WLtitle { font-weight: bold; text-decoration: none; }
        .WLgroupName { background-color: #ffffff; font-weight: bold; font-size: x-large; text-decoration: none; }
        .WLoddGroup { background-color: #ffFFff; }
        .WLevenGroup { background-color: #fFEFfF; }
        .WLquoteTable {}
        .WLcompanyNameLabel {}
        .WLtickerSymbolLabel {}
        .WLtotalSharesLabel {}
        .WLaverageCostPerShareLabel {}
        .WLlastPriceLabel {}
        .WLaverageCostPerShareLabel {}
        .WLnetChangeLabel {}
        .WLpercentChangeLabel {}
        .WLvolumeTradedLabel {}
        .WLtotalMarketValueLabel {}
        .WLgainLossLabel {}
        .WLcompanyName { text-transform: capitalize; }
        .WLtickerSymbol {}
        .WLtotalShares { text-align: right; }
        .WLcost { text-align: right; }
        .WLtotalCost { text-align: right; font-weight: bold; }
        .WLaverageCostPerShare { text-align: right; }
        .WLlastPrice { font-weight: bold; text-align: right; }
        .WLnetIncrease { color: #00AA00; text-align: right; }
        .WLnetDecrease { color: #FF0000; text-align: right; }
        .WLnetChange { text-align: right; }
        .WLnetIncreaseImage {}
        .WLnetDecreaseImage {}
        .WLpercentIncrease { color: #00AA00; text-align: right; }
        .WLpercentDecrease { color: #FF0000; text-align: right; }
        .WLpercentChange { text-align: right; }
        .WLvolumeTraded { text-align: right; }
        .WLmarketValue { text-align: right; }
        .WLtotalMarketValue { text-align: right; font-weight: bold; }
        .WLgain { color: #00AA00; text-align: right; }
        .WLloss { color: #FF0000; text-align: right; }
        .WLnoGainLossChange { text-align: right; }
        .WLgainTotal { color: #00AA00; text-align: right; font-weight: bold; }
        .WLlossTotal { color: #FF0000; text-align: right; font-weight: bold; }
        .WLnoGainLossChangeTotal { text-align: right; font-weight: bold; }
        .WLquoteDate {}
ENDOF
  $doc->print(%foo);
  $doc->setFocus("body");
  my $toggle = 0;
  my $class;
  if ($group eq "SHORT")
  {
    my @w = $self->getWatchlistNames(ledgerCode => $ledgerCode, summary => 1);
    foreach my $list (@w)
    {
      if ($toggle) { $class = "WLoddGroup"; }
      else { $class = "WLevenGroup";  }
      $toggle = ! $toggle;
      my $stockHTML = "<span class=\"$class\">".$self->getWatchListHTML(ledgerCode => $ledgerCode, group => $list)."</span>\n";
      if ($self->error)
      {
        $doc->print("<b><i>" . $self->langObj->map("errorLookingUpStockWatchListInfo") . "</i></b><br>\n");
      }
      else
      {
        $doc->print($stockHTML);
      }
    }
  }
  elsif ($group eq "ALL")
  {
    my @w = $self->getWatchlistNames(ledgerCode => $ledgerCode);
    foreach my $list (@w)
    {
      if ($toggle) { $class = "WLoddGroup"; }
      else { $class = "WLevenGroup";  }
      $toggle = ! $toggle;
      my $stockHTML = "<span class=\"$class\">".$self->getWatchListHTML(ledgerCode => $ledgerCode, group => $list)."</span>\n";
      if ($self->error)
      {
        $doc->print("<b><i>" . $self->langObj->map("errorLookingUpStockWatchListInfo") . "</i></b><br>\n");
      }
      else
      {
        $doc->print($stockHTML);
      }
    }
  }
  else
  {
    my $stockHTML = $self->getWatchListHTML(ledgerCode => $ledgerCode, group => $group);
    if ($self->error)
    {
      $doc->print("<b><i>" . $self->langObj->map("errorLookingUpStockWatchListInfo") . "</i></b><br>\n");
    }
    else
    {
      $doc->print($stockHTML);
    }
  }
  $doc->print("<p><a href=\"?$self->{getInfo}&command=display\" title=\"$edit\"><sppan class=\"WLtitle\">$edit</span></a></p>");
  return $doc;
}

=item % getWatchlistData

 required args: ledgerCode
 optional: groupName
 returns: Hash of data that goes with watchlists
 This routine selects all stocks from the stock_watch_tb and pertinent data for the given group

=cut
sub getWatchlistData
{
  my $self = shift;
  my %args = ( ledgerCode => "", group => "Default", @_ );
  my $ledgerCode = $args{ledgerCode};
  if ($ledgerCode eq "" || $ledgerCode =~ /';"/) { $self->error("<p>Invalid ledgerCode in getWatchListData: $ledgerCode</p>");  return undef;  }
  my $group = $args{group};
  if ($group eq "" || $group =~ /';"/) { $self->error("<p>Invalid ledgerCode in getWatchListData: $group</p>");  return undef;  }
  my %data;
  my @tickers;
  my $manual = "";

 #!!!!! this whole SQL needs to be rethought.  It pulls nodes on Teds data that should be zero.

  #my $sql = "SELECT stocks_tb.ticker, stocks_tb.name, group_name, SUM(shares), SUM(cost), SUM(cost)/SUM(shares) FROM lot_node_tb AS ltb, stocks_tb, stock_watch_tb WHERE ltb.ledger_code = '$ledgerCode' AND stock_watch_tb.ledger_code = ltb.ledger_code AND ltb.sub_account_num = stock_watch_tb.sub_acct AND stocks_tb.sub_acct = ltb.sub_account_num AND ltb.node_id = (SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND lot = ltb.lot AND event_date <= DATE('now') AND sub_account_num = stock_watch_tb.sub_acct) AND (SELECT shares FROM lot_node_tb WHERE node_id = ltb.node_id) <> 0 GROUP BY group_name, name, ticker UNION SELECT stocks_tb.ticker, stocks_tb.name, group_name, 0, 0, 0 FROM stocks_tb, stock_watch_tb WHERE stock_watch_tb.ledger_code = '$ledgerCode' AND stock_watch_tb.sub_acct = stocks_tb.sub_acct AND ((SELECT COUNT(lot) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct) = 0 OR (SELECT shares FROM lot_node_tb WHERE node_id = (SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct)) = 0) GROUP BY group_name, name, ticker ORDER BY group_name ASC, name ASC, ticker ASC";
  my $sql = "SELECT stocks_tb.sub_acct AS id, stocks_tb.name, stocks_tb.ticker, SUM(shares), SUM(cost), SUM(cost)/SUM(shares) AS cost_per_share, stocks_tb.market";
  $sql .=   " FROM lot_node_tb AS ltb, stock_watch_tb, stocks_tb";
  $sql .=   " WHERE ltb.ledger_code = '$ledgerCode' AND stock_watch_tb.ledger_code = ltb.ledger_code AND stock_watch_tb.group_name = '$group' ";
  $sql .=   "   AND ltb.sub_account_num = stock_watch_tb.sub_acct AND stocks_tb.sub_acct = ltb.sub_account_num";
  $sql .=   "   AND ltb.node_id = (";
  $sql .=   "     SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND lot = ltb.lot AND event_date <= DATE('now') AND sub_account_num = stock_watch_tb.sub_acct";
  $sql .=   "   ) AND (SELECT shares FROM lot_node_tb WHERE node_id = ltb.node_id) <> 0 ";
  $sql .=   " GROUP BY name, ticker, id, market";
  $sql .=   " UNION SELECT stocks_tb.sub_acct AS id, stocks_tb.name, stocks_tb.ticker, 0, 0, 0, stocks_tb.market FROM stocks_tb, stock_watch_tb";
  $sql .=   " WHERE stock_watch_tb.ledger_code = '$ledgerCode' AND stock_watch_tb.sub_acct = stocks_tb.sub_acct ";
  $sql .=   "   AND stock_watch_tb.group_name = '$group' ";
  $sql .=   "   AND (";
  $sql .=   "     (SELECT COUNT(lot) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND stocks_tb.sub_acct = stock_watch_tb.sub_acct) = 0";
  $sql .=   "     OR ";
  $sql .=   "     (";
  $sql .=   "       SELECT shares FROM lot_node_tb WHERE node_id = (";
  $sql .=   "         SELECT MAX(node_id) FROM lot_node_tb WHERE ledger_code = stock_watch_tb.ledger_code AND sub_account_num = stock_watch_tb.sub_acct";
  $sql .=   "       )";
  $sql .=   "     ) = 0";
  $sql .=   "   ) GROUP BY name, ticker, id, market ORDER BY name ASC";
#   $self->error($sql);
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (my $info = $sth->fetchrow_hashref)
  {
    if ($info->{market} eq "None")
    {
      if ($manual ne "") { $manual .= ","; }
      $manual .= " $info->{id} ";
    }
    else  { push @tickers, $info->{ticker}; }
    $data{$info->{name}} = $info;
#    $self->error("<p>$info->{name}: $info->{ticker}: $info->{market}: $manual</p>\n");
  }
  if ($manual ne "")
  {
    $sql = "SELECT MAX(price) AS price, name FROM stocks_tb, stock_price_tb WHERE stocks_tb.sub_acct = stock_price_tb.sub_acct AND date_of <= DATE('now') AND stocks_tb.sub_acct IN ($manual) GROUP BY name";
    $sth = $self->{dbHandle}->read($sql);
    if ($self->{dbHandle}->error)
    {
      $self->error($self->{dbHandle}->errorMessage);
      return undef;
    }
    while (my $info = $sth->fetchrow_hashref)
    {
      $data{$info->{name}}->{price} = $info->{price};
    }
  }
#   $self->error("<p>$sql</p>\n");
#   return undef;
  return (\%data, \@tickers);
}

=item @Objects::Stock getWatchlistArray

 required args: ledgerCode
 optional:
   orderBy - name or ticker
   group -> defaults to 'Default'
 returns: Array of Stock Objects
 This routine selects all stocks from the stock_watch_tb with the given group

=cut
sub getWatchlistArray
{
  my $self = shift;
  my %args = ( orderBy => "name", ledgerCode => "", group => "Default", @_ );
  my $orderBy = $args{orderBy};
  if ($orderBy ne "ticker") { $orderBy = "name";  } #prevent SQL hacks
  my $ledgerCode = $args{ledgerCode};
  if ($ledgerCode eq "" || $ledgerCode =~ /';"/) { $self->error("<p>Invalid ledgerCode in getWatchListArray: $ledgerCode</p>");  return undef;  }
  my $group = $args{group};
  my @stocks;
  my $sth;
  my @info;
  my $sql = "SELECT stocks_tb.name, stock_watch_tb.sub_acct, stocks_tb.ticker";
    $sql .= " FROM stocks_tb, stock_watch_tb";
    $sql .= " WHERE ledger_code = '$ledgerCode' AND stocks_tb.sub_acct = stock_watch_tb.sub_acct AND group_name = '$group'";
#    $sql .= " AND stocks_tb.market <> 'None'";
    $sql .= " ORDER BY $orderBy ASC";

  # validate input
  if (length $orderBy == 0)
  {
   $self->error("orderBy = '$orderBy' is invalid!<br>\n");
    return undef;
  }

  #get sub account list
  #foreach
    # make new object
    #$obj->select

  # now execute the sql statement
  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    my $obj = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->dbHandle);
    if ($obj->error) { $self->error($obj->errorMessage); return undef;  }
    $obj->set("id", $info[1]);
    if ($obj->select)
    {
      $stocks[++$#stocks] = $obj;
    }
    if ($obj->error) { $self->error($obj->errorMessage); return undef;  }
  }
  return @stocks;
}

=item array[] getWatchlistNames(ledgerCode =>)

  This routine selects all of the watchlists from the stock_watch_tb.

=cut
sub getWatchlistNames
{
  my $self = shift;
  my %args = (ledgerCode => "", summary => 0, @_);
  my $ledger = $args{ledgerCode};
  my $summary = $args{summary};
  if ($summary) { $summary = "WHERE short = true "; }
  else { $summary = "";  }
  my $sql = "SELECT DISTINCT(group_name) FROM stock_watch_tb $summary ORDER BY group_name";
  if ($ledger eq "")
  {
     $self->error("Ledger Code was not defined!<br>\n");
     return undef;
  }
  my @lists = ();

  my $sth = $self->{dbHandle}->read($sql);
  if ($self->{dbHandle}->error())
  {
    $self->error("Error retrieving watch lists!\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  my $index = 0;
  while ((my @info = $sth->fetchrow_array))
  {
    $lists[$index++] = $info[0];
  }
  return @lists;
}

sub isShort
{
  my $self = shift;
  my $group = shift;
  my $sql = "SELECT short FROM stock_watch_tb WHERE group_name = '$group' AND short = true";
  my $sth = $self->{dbHandle}->read($sql);
  if ($self->{dbHandle}->error) { $self->error($self->{dbHandle}->errorMessage);  }
  if (my $info = $sth->fetchrow_hashref)
  {
    return 1;
  }
  return 0;
}

# watchStock
# requires: number, ledgerCode
# optional: groupName (if not specified, defaults to 'Default')
# returns: 1=OK, 0=Error, -1=Stock Sub Account Already exists
# summary: creates the watch this stock entry in the database.
sub watchStock
{
  my $self = shift;
  my %args = ( number => "", ledgerCode => "", groupName => "Default", short => 'false', @_ );
  my $number = $args{number};
  my $ledgerCode = $args{ledgerCode};
  my $groupName = $args{groupName};
  my $short = $args{short};
  my $sth;
  my @info;

  if ($number !~ /^\d+$/)
  {

    $self->error("Stock Sub Account = '$number' is invalid!<br>\n");
    return 0;
  }

  if (length $ledgerCode == 0)
  {

    $self->error("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return 0;
  }

  if (length $groupName == 0)
  {

    $self->error("groupName = '$groupName' is invalid!<br>\n");
    return 0;
  }

  #check for already on the list
  $sth = $self->{dbHandle}->read(sql => "SELECT ledger_code FROM stock_watch_tb WHERE sub_acct = ? AND ledger_code = ? AND group_name = ?",
                                    plug => [  $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    return -1;  # The stock sub account is already being watched!
  }

  my $rc = $self->{dbHandle}->write(sql => "INSERT INTO stock_watch_tb (sub_acct, ledger_code, group_name, short) VALUES (?, ?, ?, ?)",
                                    plug => [ $number, $ledgerCode, $groupName, $short ]);
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return 0;
  }
  $self->{dbHandle}->commit;
  return 1;
}

# unWatchStock
# requires: number, ledgerCode
# optional: groupName (if not specified, defaults to 'Default')
# returns: 1=OK, 0=Error, -1=Stock Sub Account No Longer exists
# summary: deletes the watch this stock entry in the database.
sub unWatchStock
{
  my $self = shift;
  my %args = ( number => "", ledgerCode => "", groupName => "Default", @_ );
  my $number = $args{number};
  my $ledgerCode = $args{ledgerCode};
  my $groupName = $args{groupName};
  my $sth;
  my @info;

  if ($number !~ /^\d+$/)
  {

    $self->error("Stock Sub Account = '$number' is invalid!<br>\n");
    return (0);
  }

  if (length $ledgerCode == 0)
  {

    $self->error("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return (0);
  }

  if (length $groupName == 0)
  {

    $self->error("groupName = '$groupName' is invalid!<br>\n");
    return (0);
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT ledger_code FROM stock_watch_tb WHERE sub_acct = ? AND ledger_code = ? AND group_name = ?",
                                    plug => [  $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->error)
  {

    $self->error($self->{dbHandle}->errorMessage);
    return (0);
  }
  @info = $sth->fetchrow_array;
  if (!defined $info[0])
  {
    return (-1, $info[0]);  # The stock sub account is not being watched!
  }

  my $rc = $self->{dbHandle}->write(sql => "DELETE FROM stock_watch_tb WHERE sub_acct = ? AND ledger_code = ? AND group_name = ?",
                                    plug => [ $number, $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->error)
  {

    $self->error($self->{dbHandle}->errorMessage);
    return (0);
  }
  $self->{dbHandle}->commit;

  return (1);
}

# unWatchAllStocks
# requires: ledgerCode
# optional: groupName (if not specified, defaults to 'Default')
# returns: 1=OK, 0=Error
# summary: deletes all the watch this stock entry in the database for the specified ledger.
sub unWatchAllStocks
{
  my $self = shift;
  my %args = ( ledgerCode => "", groupName => "Default", @_ );
  my $ledgerCode = $args{ledgerCode};
  my $groupName = $args{groupName};

  if (length $ledgerCode == 0)
  {

    $self->error("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return (0);
  }

  if (length $groupName == 0)
  {

    $self->error("groupName = '$groupName' is invalid!<br>\n");
    return (0);
  }

  my $rc = $self->{dbHandle}->write(sql => "DELETE FROM stock_watch_tb WHERE ledger_code = ? AND group_name = ?",
                                    plug => [ $ledgerCode, $groupName ]);
  if ($self->{dbHandle}->error)
  {

    $self->error($self->{dbHandle}->errorMessage);
    return (0);
  }
  $self->{dbHandle}->commit;

  return (1);
}

1;
__END__

=head1

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

JT Moree
PC & Web Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::App(3)

=cut
