# CGLObject.pm - The Object Class that provides a CGLObject Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: CGLObject.pm,v 1.4 2005/08/08 22:18:37 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::CGLObject;
use strict;
use Portal::Accounting::Objects::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.03';

=head1 NAME

CGLObject - Object used to build a CGLObject Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::CGLObject;
  my $obj = Portal::Accounting::Objects::CGLObject->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

CGLObject is a CGLObject class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(cglObjects, type, cost, market, gainLoss, state, name)

 Creates a new instance of the Portal::Accounting::Objects::CGLObject object.
 See Portal::Accounting::Objects::Base(3) for a listing of required arguments.

 requires: cglObjects, type, cost, market, gainLoss, name
 optional: state

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( cglObjects => undef, type => "both", state => "o", cost => "0.00", market => "0.00", gainLoss => "0.00", name => "stocks", @_ );

  if ($self->didErrorOccur)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{cglObjects} = $args{cglObjects};
  $self->{type} = $args{type};
  $self->{state} = $args{state};
  $self->{cost} = $args{cost};
  $self->{market} = $args{market};
  $self->{gainLoss} = $args{gainLoss};
  $self->{name} = $args{name};

  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
    return $self;
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if (!defined $self->{cglObjects})
  {
    $self->missing("cglObjects");
  }
  if (!$self->{cglObjects}->isValid)
  {
    $self->invalid("cglObjects", $self->{cglObjects}->errorMessage);
  }
  if ($self->{type} !~ /^(both|ST|LT)$/)
  {
    $self->invalid("type", $self->{type});
  }
  if ($self->{state} !~ /^(o|c)$/)
  {
    $self->invalid("state", $self->{state});
  }
  if ($self->{cost} !~ /^(-?\d+\.\d{2})$/)
  {
    $self->invalid("cost", $self->{cost});
  }
  if ($self->{market} !~ /^(-?\d+\.\d{2})$/)
  {
    $self->invalid("market", $self->{market});
  }
  if ($self->{gainLoss} !~ /^(-?\d+\.\d{2})$/)
  {
    $self->invalid("gainLoss", $self->{gainLoss});
  }
  if ($self->{name} !~ /^(stocks|assets|manual)$/)
  {
    $self->invalid("name", $self->{name});
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
    }
    $result .= $spaceStr . "<CGLObject>\n";
    # this needs to be fleshed out if we really are going to generate XML for this object.
    foreach my $name (qw{ticker name number type market class})
    {
      if ($name eq "number")
      {
        $result .= $spaceStr . "  <code>$self->{$name}</code>\n";
      }
      else
      {
        $result .= $spaceStr . "  <$name>$self->{$name}</$name>\n";
      }
    }
    $result .= $spaceStr . "</CGLObject>\n";
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::Objects::Base(3)

=cut
