# Investment.pm - The Object Class that provides a Investment Object
#$Id: Investment.pm,v 1.4 2005/08/08 22:18:37 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2003 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Investment;
use strict;
use Portal::Accounting::Objects::Base;
use Portal::Accounting::Objects::Account;
use Portal::Accounting::Objects::Stock;
use Portal::Accounting::Objects::SubEntity;

use vars qw($AUTOLOAD $VERSION @ISA @EXPORT %Elements);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.4 $';

=head1 NAME

InvestmentObject - Investment Object for deriving other objects.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Investment;
  @ISA = qw(Portal::Accounting::Objects::Investment Exporter AutoLoader);

=head1 DESCRIPTION

InvestmentObject is a Investment class for Accounting Objects such as
Objects::Ledger and Objects::Account.  It contains methods for
handling dataInvestment select, insert, update, delete.  Also contains
methods for handling forms which are defined in the child objects.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new( )

    Creates a new instance of Portal::Accounting::Objects::Investment.
    See Portal::Investment(3) for a listing of required arguments.

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (ledgerCode => '', @_);

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }

  # instantiate anything unique to this module
  $self->{objects} = {
    buy  => "Buy",
    sell => "Sell",
    fee  => "Adjustment",
    memo => "Memo",
    acqs => "Merger",
    rtcp => "ReturnOfCapital",
    spof => "SpinOff",
    splt => "StockSplit",
    sdiv => "StockDividend",  #deprecated!!! will remove in the future
    sxfr => "StockTransfer"
  };

  my $chart = Portal::Accounting::Objects::Account->new(langObj => $self->langObj, dbHandle => $self->{dbHandle}, ledgerCode => $args{ledgerCode});
  $chart->set('type', 3);
  $self->{accounts} = $chart->selectOptionsByType();
  if ($chart->error) { $self->error("Failed to get accounts<br>".$chart->errorMessage."\n"); return $self;  }
  if (! defined $self->{accounts}) { $self->{accounts} = {names => [], values => [] }; }

  my $stock = Portal::Accounting::Objects::Stock->new(langObj => $self->langObj, dbHandle => $self->{dbHandle});
  $self->{stocks} = $stock->selectOptions();
  if ($stock->error) { $self->error("Failed to stocks<br>".$stock->errorMessage."\n"); return $self;  }
  if (! defined $self->{stocks}) { $self->{stocks} = {names => [], values => [] }; }

  my $subE = Portal::Accounting::Objects::SubEntity->new(langObj => $self->langObj, dbHandle => $self->{dbHandle}, ledgerCode => $args{ledgerCode});
  $self->{subEntities} = $subE->selectOptions();
  if ($subE->error) { $self->error("Failed to stocks<br>".$subE->errorMessage."\n"); return $self;  }
  if (! defined $self->{subEntities}) { $self->{subEntities} = {names => [], values => [] }; }

  # do anything else you might need to do.
  return $self;
}

=item bool isValid(void)

 Returns 0 for invalid or 1 for valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item string map("invCode")

  This routine takes an investment object type and returns a phrase mapping out of the language module
  It will set the error flag if there is a problem.
  The valid types are
    buy
    sell
    fee
    memo
    acqs
    rtcp
    spof
    splt
    sdiv
    sxfr

=cut
sub map
{
  my $self = shift;
  my $foo = shift;
  if (defined $self->{objects}->{$foo})
  {
    return $self->{langObj}->map($self->{objects}->{$foo});
  }
  else { $self->error("<p>'$foo' is not defined as an object</p>\n"); }
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Base(3)

=cut
