#$Id: JournalEntry.pm,v 1.25 2005/08/16 21:45:22 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::JournalEntry;

use strict;
use Portal::Accounting::Objects::JournalEntryRow;
use Portal::Accounting::Objects::BaseL;
use vars qw($VERSION @ISA @EXPORT $AUTOLOAD);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.25 $';
=head1 NAME

JournalEntry - Object used to retrieve, insert, and manipulate a journal transaction.

=head1 SYNOPSIS

  use JournalEntry;
  my $transobj = JournalEntry->new(langObj => . . .);
  $transobj->set("id","12345");
  $transObj->select();
  if ($transobj->error())
  {
    die $transobj->errorMessage();
  }

  valid data fields can be set with $transobj->set('field', 'value');
    "id" => { -Type => "hidden", -Value => '-1' },
    "journal" => {-Type => "hidden", -Value => "G" },
    "date" => {-Type => "hidden", },
    "payee" => {-Type => "hidden", },
    "remember" => {-Type => "hidden", -Value => "false"},
    "trackingId" => {-Type => "hidden", },
    "origId" => {-Type => "hidden", },
    "ledgerCode" => {-Type => "hidden", },
    "workingYear" => {-Type => "hidden", },
    "lastClose" => {-Type => "hidden", },

=head1 DESCRIPTION

JournalEntry is a class that stores data for an accounting
JournalEntry.  It provides a container for a JournalEntry that will be used
to write to the database or update an existing record in the database.

=head1 Exported FUNCTIONS

=item scalar new(langObj =>, ledger => )

    Creates a new instance of the Portal::Accounting::Objects::JournalEntry
    object.  A new instance is created in an invalid state (i.e., missing
    fields, etc.).  isValid() will be called on the object before using the
    object in a database JournalEntry.

=cut
sub new
{
  my $class = shift;
  my %args = (ledger => undef, debug => 0, @_ );
  my $lo = $args{ledger};
  my $lc = $lo->get('code');
  my $color = $lo->get('color');
  my $self = $class->SUPER::new(ledgerCode => $lc, @_);
  my $l = $self->langObj;
  if ($args{debug} == 1) { $self->debug("<p>debug on</p>\n"); }

  $self->{data}->{main} = {
    "app" =>  {-Type => "hidden", -Value => "Accounting", },
    "state" =>  {-Type => "hidden", -Value => "JournalEntry", },
    "command" =>  {-Type => "hidden", -Value => "edit", },
    "id" => { -Type => "text", -ReadOnly => 1, -Value => '-1',
      -onload => "setEditLink(this.form);",
      },
    "journal" => {-Type => "text", -ReadOnly => 1, -Value => "G" },
    "date" => {-Type => "text", -ReadOnly => 1, },
    "payee" => {-Type => "text", -ReadOnly => 1, },
    "remember" => {-Type => "hidden", -ReadOnly => 1, -Value => "false", -Label => '',},
    "trackingId" => {-Type => "hidden", },
    "origId" => {-Type => "hidden", },
    "ledgerCode" => {-Type => "hidden", -Value => $lc },
    "workingYear" => {-Type => "hidden", },
    "lastClose" => {-Type => "hidden", },
    "windowName" =>  {-Type => "hidden", -DataType => "string", },
  };
  $self->{headers} = <<"ENDOF";
  <table width="100%" cellspacing="0" cellpadding="0" border="1" bgcolor="$color">
    <tr valign="top">
      <th style="width: 5em; text-align: left;">
        <a title="#LANG=Journal#">#LANG=Jnl#</a>
        #LANG=Id#
      </th>
      <th style="width: 1em; text-align: left;">
        <a title='#LANG=Remember#'>#LANG=R#</a>
      </th>
      <th style="width: 7em; text-align: left;">
        #LANG=Date#
      </th>
      <th style="text-align: left;">
        #LANG=Payee# / #LANG=Description#
      </th>
      <th style="width: 2em; text-align: center; ">
        <a title='#LANG=Delete#'>X</a>
      </th>
    </tr>
  </table>
ENDOF
  $self->{template}->{main} = <<"ENDOF";

  <table width="100%" cellspacing="0" cellpadding="0" border="1" bgcolor="$color">
    <tr valign="top">
      <td style="width: 5em;">
        <form #FORMARGS#>
        #FIELD=app#
        #FIELD=state#
        #FIELD=command#
        <a name="link_#LABEL=id#" href="?app=Accounting&state=JournalEntry&command=edit">#FIELD=journal#
        #FIELD=id#</a>
        #FIELD=ledgerCode#
      </td>
      <td style="width: 1em;">
        #FIELD=remember# #LABEL=remember#
      </td>
      <td style="width: 7em;">
        #FIELD=date#
      </td>
      <td>
        #FIELD=payee#&nbsp;
      </td>
      <td style="width: 2em; font-size: tiny; text-align: center;">
        <small><input style="font-size: tiny; border-style:none;" type="button" value="X" name="delete" onclick="if (confirm('#LANG=Delete journal entry#'+' '+this.form.id.value)) { deleteJournalEntry(this.form); }"></small>
        </form>
      </td>
    </tr>
  </table>
ENDOF

  $self->{rowObjects} = [];

  #pull all date information needed from ledgerObject
  $self->set('workingYear', $lo->workingYear());
  $self->set('lastClose', $lo->lastClose);

#   $self->{id} = "id";
#   $self->{INSERTorder}[0] = [ qw( ledger_code date journal_code payee_payer remember tracking_id orig_id ) ] ;  #insert order for db column names
#   $self->{newOnly}[0] = { orig_id => "", ref_num => ''};
#   $self->{newOnly}[1] = { ref_num => ''};
#   $self->{db2data}[0] = { ref_num => 'id', ledger_code => 'ledgerCode', trans_date => 'date', journal_code => 'journal', payee_payer => 'payee' };  #map database columns to ->data hash, not specified assumes same name
#   $self->{data2db}[0] = reverse %{$self->{db2data}[0]};
#   $self->{db2data}[1] = { parent_num => 'id' };
#   $self->{data2db}[1] = reverse %{$self->{db2data}[0]};
#   $self->{SELECT}[0] = "SELECT ##COLUMNS## FROM journal_entry_tb WHERE ref_num = #FIELD=id#";
#   $self->{INSERT}[0] = "INSERT INTO journal_entry_tb (##COLUMNS##) VALUES (##VALUES##)";
#   $self->{UPDATE}[0] = "UPDATE journal_entry_tb SET ##UPDATE## WHERE ref_num = #FIELD=id#";
#   #remember to use single quotes around ##foo## if needed
   $self->{DELETE}[0] = "DELETE FROM journal_entry_tb WHERE ref_num = #FIELD=id#";
   $self->{DELETE}[1] = "DELETE FROM debitcredit_tb WHERE parent_num = #FIELD=id#";
#   #find out ifis this being referenced from other tables before deleting
#   $self->{preDELETE}[0] = "";
#   #what should we output as the value of the thing just deleted?
#   $self->{postDELETE}[0] = "SELET ref_num||':'||payee_payer FROM journal_entry_tb WHERE ref_num = '#FIELD=id#'";

  $self->{journals} = { 'G' => $l->map('General'), 'CSR' => $l->map('Cash Receipts'), 'CP' => $l->map('Cash Purchases'), 'TX' => $l->map('Transfers')};
  return $self;
}

=item scalar populate(id, ledgerCode, journal, date, payee, remember, rowObjects, trackingId, origId)

  Fills in the data quickly for the object
    trackingId = the genSysId value from an XML JournalEntry Import. Can be "".
    origId = the original id value for this entry.  Can be "".

=cut
sub populate
{
  my $self = shift;
  my %args = ( id => -1,
#      ledgerCode => "",  #set at new
      journal => 'G',
      date => "",
      payee => '',
      remember => '',
      rowObjects => undef,
      trackingId => "",
      origId => "",
      rows => 0,
      @_ );

  $self->{data}->{main}->{id}->{-Value}           = $args{id};
#  $self->{data}->{main}->{ledgerCode}->{-Value}   = $args{'ledgerCode'}; #set at new
  $self->{data}->{main}->{journal}->{-Value}  = $args{'journal'};
  $self->{data}->{main}->{date}->{-Value}         = $args{'date'};
  $self->{data}->{main}->{payee}->{-Value}   = $args{'payee'};
  $self->{data}->{main}->{trackingId}->{-Value}   = $args{trackingId};
  $self->{data}->{main}->{origId}->{-Value}       = $args{origId};
  if ($args{'remember'} eq '') { $self->set("remember", "false"); }
  else { $self->set("remember", "true"); }
  if (defined $args{rowObjects})
  {
    foreach my $row (@{$self->{rowObjects}})
    {
      $self->addRow($row);
    }
  }
  else { $self->{rowObjects} = []; }
  if (defined $args{rows} && $args{rows} =~ /^\d+$/)
  {
    for (my $i = 0; $i < $args{rows}; $i++)
    {
      my $subA = -1;
      #create row obj and populate it
      my $obj = Portal::Accounting::Objects::JournalEntryRow->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $self->get("ledgerCode"), debug => $self->{debug});
      if ($obj->error) { $self->error($obj->errorString); }
      if ($args{"subAccount_$i"} ne "") { $subA = $args{"subAccount_$i"}; }
      $obj->populate(
        counter => $i,
        id => $args{'id'},
        date => $args{'date'},
        comment => $args{"comment_$i"},
        account => $args{"account_$i"},
        subAccount => $subA,
        subEntity => $args{"subEntity_$i"},
        dorc => $args{"dorc_$i"},
        amount => $args{"amount_$i"},
        cleared => $args{"cleared_$i"},
      );
        $self->debug("<hr>rows $i: ".$obj->get("date"));
      if ($obj->error) { $self->error($obj->errorMessage); }
      if ($obj->debug) { $self->debug($obj->debugString); }
      push @{$self->{rowObjects}}, $obj;
    }
  }
  else
  {
    $self->error("<p>Invalid rows: ".$self->get("rows")."</p>\n");
  }

  return 1;
}

# isValid
# validates the object in its current state
sub isValid
{
  my $self = shift;
  # Can we call the parent at the bottom to pick up on the error here without repeating code
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  if ($self->get("id") != -1 && $self->get("id") !~ /^\d\d*$/)
  {
    $self->invalid("id", $self->get("id"));
  }
  if ($self->get("ledgerCode") eq "")
  {
    $self->missing( "ledgerCode");
  }
  if (not exists $self->{journals}->{$self->get("journal")})
  {
    $self->invalid( "journal", $self->get("journal"));
  }
  if ($self->get("date") !~ /^\d{8}$/ && $self->get("date") !~ /^\d{4}[\-\\\/]\d{1,2}[\-\\\/]\d{1,2}$/)
  {
    $self->invalid( "date", $self->get("date"));
  }
  if (($self->get("remember") ne "true" and $self->get("remember") ne "false"))
  {
    $self->set("remember","false");
  }
  if (scalar @{ $self->{rowObjects} } < 2)
  {
    $self->prefixError("The JournalEntry rows array contains too few items (" . (scalar @{ $self->{rowObjects} }) . ")!");
  }
  my $sum = 0;
  my $debit_sum = 0;
  my $credit_sum = 0;
  for (my $rowcount=0; $rowcount < scalar @{ $self->{rowObjects} }; $rowcount++)
  {
    if (ref($self->{rowObjects}[$rowcount]) ne "Portal::Accounting::Objects::JournalEntryRow")
    {
      $self->prefixError("Row $rowcount is not a valid JournalEntryRow (" . ref($self->{rowObjects}[$rowcount]) . ")!<br>");
    }
    elsif (!$self->{rowObjects}[$rowcount]->isValid())
    {
      $self->prefixError("JournalEntry->JournalEntryRow->isValid(): Row $rowcount has invalid data! " . $self->{rowObjects}[$rowcount]->errorMessage."<br>");
    }
    else
    {
      my $dorc = $self->{rowObjects}[$rowcount]->get("dorc");
      if ($dorc eq "true")
      {
        $debit_sum += $self->{rowObjects}[$rowcount]->get("amount");
      }
      elsif ($dorc eq "false")
      {
        $credit_sum += $self->{rowObjects}[$rowcount]->get("amount");
      }
      else
      {
        $self->error("<p>Invalid dorc field: $dorc</p>\n");
      }
    }
  }
  $sum = sprintf("%.2f", $debit_sum - $credit_sum);
  if ($sum != 0.00)
  {
    if ($debit_sum == 0.00)
    {
      $self->prefixError( "No debits entered!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
    elsif ($credit_sum == 0.00)
    {
      $self->prefixError("No credits entered!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
    if ($sum < 0.00)
    {
      $self->prefixError("Credits are greater than debits!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
    elsif ($sum > 0.00)
    {
      $self->prefixError("Debits are greater than credits!<br>Debit Sum: " . $debit_sum . "<br>Credit Sum: " . $credit_sum . "<br>Sum: " . $sum);
    }
  }
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item scalar addRow(rowObject)
  OR (accountNum =>, dorc =>, amount =>, [comment =>, subEntity =>, subAccountNum =>, cleared =>])

    Adds a row to the rowObjects array in the JournalEntry.
    Returns the row number of the newly added row or undef on failure.
    takes a rowObject with the proper data or a list of data to create a row.

=cut
sub addRow
{
  my $self = shift;
  my $transrowobj;
  if (@_ == 1)
  {
    #a row object was sent. check it and use it.
    $transrowobj = @_[0];
    if (! ref($transrowobj) eq "Portal::Accounting::Objects::JournalEntryRow") { $self->error("<p>row is not a JournalEntryRow</p>\n"); return undef; }
    $transrowobj->date($self->get("date"));
  }
  else
  {
    # create a new JournalEntryRow with the values passed in
    $transrowobj = Portal::Accounting::Objects::JournalEntryRow->new(langObj => $self->{langObj}, ledgerCode => $self->get("ledgerCode"), dbHandle => $self->{dbHandle});
    $transrowobj->populate(@_, date => $self->get("date"), ledgerCode => $self->get('ledgerCode'));
  }
  if ($transrowobj->error())
  {
    $self->error("Failed to add row!<br>\n" . $transrowobj->errorMessage);
    return undef;
  }
  if ($transrowobj->isValid)
  {
    push (@{$self->{rowObjects}}, $transrowobj);
    return scalar @{$self->{rowObjects}} - 1;
  }
  $self->error($transrowobj->genErrorString("all"));
  return undef;
}

=item scalar deleteRow(row)

    Removes the row specified from the rowObjects array.

=cut
sub deleteRow
{
  my $self = shift;
  my %args = ( @_ );
  my $row = $args{'row'};

  if (not exists $args{'row'})
  {
    $self->prefixError( "Row number value argument missing!");
    return 0;
  }
  if ($row !~ /^[0-9]+$/)
  {
    $self->prefixError( "Invalid row number argument!  The value does not appear to be numeric.");
    return 0;
  }
  if ($row > 0 && $row < scalar @{$self->{rowObjects}})
  {
    # remove the row from the array using splice to "cut" the element out
    splice(@{$self->{rowObjects}}, $row, 1);
  }
  else
  {
    $self->prefixError( "Invalid row number!  That row does not exist.");
    return 0;
  }
  return 1;
}

=item boolean setDate('2004-9-9')

  Returns 0 for failure and sets error
  1 for success

  Will set all dates in rows and in object

=cut
sub setDate
{
  my $self = shift;
  my $date = shift;
  if ($date =~ /^(\d{4})[\-\\\/](\d{1,2})[\-\\\/](\d{1,2})$/)
  {
    $date = $1 ."-". sprintf("%02d", $2) ."-". sprintf("%02d", $3);
  }
  else
  {
    $self->error("Invalid date: $date<br>\n");
    return 0;
  }
  $self->{data}->{main}->{date}->{-Value} = $date;
  foreach my $row (@{$self->{rowObjects}})
  {
    $row->setDate($date);
  }
  return 1;
}

=item scalar insert(commit =>)

    Attempts to insert the contents of the JournalEntryObject
    into the database.
    Returns: the reference number for the new transaction or returns undef and sets error on failure.
    requires:
    optional: commit

=cut
sub insert
{
  my $self = shift;
  my %args = ( commit => 0, @_ );

  # check the validity of the transaction object
  if ($self->isValid())
  {
    if ($self->get('id') == -1)
    {
      # get the nextval(journal_entry_tb_ref_num_seq) of the serial sequence for the table
      my $sth = $self->{dbHandle}->read(sql => "SELECT nextval('journal_entry_tb_ref_num_seq')");
      if ($self->{dbHandle}->error())
      {
        $self->error("Failed to get the nextval() of the journal entry sequence!<br>\n" . $self->{dbHandle}->errorMessage);
        return undef;
      }
      my @info = $sth->fetchrow_array;
      $self->set('id', $info[0]);
    }
    else
    {
      $self->error("<p>Error!  Tried to insert a transaction with id greater than -1: ".$self->get('id').". Perhaps you need to update it instead.</p>\n");
      return 0;
    }
    # insert the JournalEntryObject data into the database
    $self->{dbHandle}->write(
      sql => "INSERT INTO journal_entry_tb (ref_num, journal_code, trans_date, payee_payer, remember, ledger_code, tracking_id, orig_id) values (?, ?, ?, ?, ?, ?, ?, ?)", plug => [
        $self->{data}->{main}->{id}->{-Value},
        $self->{data}->{main}->{journal}->{-Value},
        $self->{data}->{main}->{date}->{-Value},
        $self->{data}->{main}->{payee}->{-Value},
        $self->{data}->{main}->{remember}->{-Value},
        $self->{data}->{main}->{ledgerCode}->{-Value},
        $self->{data}->{main}->{trackingId}->{-Value},
        $self->{data}->{main}->{origId}->{-Value}
      ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("Failed to insert journal entry!<br>\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return undef;
    }
    foreach my $obj (@{ $self->{rowObjects} })
    {
      $obj->set('id', $self->get('id'));
      $obj->insert(commit => 0);
      if ($obj->error) { $self->error($obj->errorMessage); }
    }
    if ($args{'commit'} eq "1") { $self->{dbHandle}->commit; }
  }
  else
  {
    return undef;
  }

  # return the created reference number from the database
  return $self->get('id');
}

=item scalar update(commit =>)

    Attempts to update the contents of the JournalEntry
    to the database.
    requires: JournalEntryObject to update
    optional:  commit
    returns:  0 on failure, positive (1) value on successfully updating the reference number contained in the JournalEntryObject to the database

=cut
sub update
{
  my $self = shift;
  my %args = ( commit => 1, @_ );
  my $refNum = $self->{data}->{main}->{id}->{-Value};
  my $sth;
  if ($refNum == -1)
  {
    $self->error("Invalid object.  Cannot update transaction with id of '-1'!\n");
    return 0;
  }
  # check the validity of the transaction object
  if ($self->isValid() )
  {
    # update the JournalEntryObject data in the database
    $sth = $self->{dbHandle}->write(
      sql => "UPDATE journal_entry_tb SET journal_code = ?, trans_date = ?, payee_payer = ?, remember = ? WHERE ref_num = ?", plug => [
        $self->get("journal"),
        $self->get("date"),
        $self->get("payee"),
        $self->get("remember"),
        $self->get("id") ]
      );
    if ($self->{dbHandle}->error())
    {
      $self->error("Failed to update journal entry!<br>\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return 0;
    }
    if ($sth == 0 )
    {
      # this error occurs in the case that the reference number does not exist in the database; thus, there is nothing to
      # update and the completion of the update process is not necessary or valid
      $self->error("Transaction does not exist for id: ".$refNum."<br>$sth");
      $self->{dbHandle}->rollback;
      return 0;
    }
    # remove all rows from the database for the reference number
    $self->{dbHandle}->write(sql => "DELETE FROM debitcredit_tb WHERE parent_num = ?", plug => [ $refNum ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("Failed to delete row entries!<br>\n" . $self->{dbHandle}->errorMessage);
      $self->{dbHandle}->rollback;
      return 0;
    }
    # insert the rows in the JournalEntryObject->JournalEntryRowObject(s) into the database
    foreach my $obj (@{ $self->{rowObjects} })
    {
      $obj->insert(commit => 0);
      if ($obj->error) { $self->error($obj->errorMessage); }
    }

    if ($args{'commit'} eq "1")
    {
      # finish the transaction
      $self->{dbHandle}->commit;
    }
  }
  else
  {
    return 0;
  }
  return 1;
}

sub insertJERows
{
  my $self = shift;
  my $commit = shift;
  my $refNum = $self->get('id');

  if ($refNum == -1)
  {
      $self->error("Failed to insert row entry because refnum is invalid! '-1'<br>");
      $self->{dbHandle}->rollback;
      return 0;
  }
  for (my $rowcount=0; $rowcount < scalar @{ $self->rowObjects }; $rowcount++)
  {
    my $sql = "INSERT INTO debitcredit_tb (parent_num, trans_date, comment, sub_e_code, account_num, sub_account_num, amount, dorc, cleared, counter, ledger_code, d_amount, c_amount) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, dorcify(?, ?, debit_increases_account_num(?)), dorcify(?, ?, class_action_account_num(?, 'cash')) )";
    my $r = $self->{rowObjects}->[$rowcount];

    $self->{dbHandle}->write(sql => $sql, plug => [
      $refNum,
      $self->get("date"),
      $r->comment,
      $r->subEntity,
      $r->accountNum,
      $r->subAccountNum,
      $r->amount,
      $r->dorc,
      $r->cleared,
      $rowcount,
      $self->{data}->{main}->{ledgerCode}->{-Value},
      $r->amount,
      $r->dorc,
      $r->accountNum
    ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("Failed to insert row entry!<br>\n" . $self->{dbHandle}->errorMessage);
      return 0;
    }
  }
  if ($commit == 1) { $self->{dbHandle}->commit; }
  return 1;
}

=item boolean select( )

    Retrieves the data from a transaction in the database
    for the internal id and places into the object.
    returns # of records selected. 0 means no transaction was found

=cut
sub select
{
  my $self = shift;
  my $refNum = $self->get('id');
  if ($refNum < 0)
  {
    $self->error("Invalid reference number!");
    return 0;
  }

  # retrieve the data for the reference number from the database and place it in the JournalEntryObject
  my @info;
  my $sql = "SELECT journal_code, trans_date, payee_payer, remember, ledger_code, tracking_id, orig_id, (SELECT COUNT(parent_num) FROM debitcredit_tb WHERE parent_num = ref_num) AS rows FROM journal_entry_tb WHERE ref_num = ?";
  my $sth = $self->{dbHandle}->read(sql => $sql, plug => [ $refNum ]);
  if ($self->{dbHandle}->error) { $self->error($self->{dbHandle}->errorMessage);  return undef; }
  if (! (@info = $sth->fetchrow_array))  { $self->warning($self->{langObj}->map("No record found"));  return 0; }
  $self->{data}->{main}->{journal}->{-Value} = $info[0];
  $self->{data}->{main}->{date}->{-Value} = $info[1];
  $self->{data}->{main}->{payee}->{-Value} = $info[2];
  $self->{data}->{main}->{remember}->{-Value} = $info[3];
  $self->{data}->{main}->{ledgerCode}->{-Value} = $info[4];
  $self->{data}->{main}->{trackingId}->{-Value} = $info[5];
  $self->{data}->{main}->{origId}->{-Value} = $info[6];
  my $i;
  for ( $i = 0 ; $i < $info[7]; $i++)
  {
    my $n = Portal::Accounting::Objects::JournalEntryRow->new(langObj => $self->{langObj}, ledgerCode => $self->get("ledgerCode"), dbHandle => $self->{dbHandle});
    $n->set("id", $refNum);
    $n->set("counter", $i);
    if ($self->debug){ $self->debug("<p>id: $refNum<br>counter: $i</p>\n");}
    $n->select();
    if ($n->error()) { $self->error($n->errorMessage); }
    if ($self->debug){ $self->debug($n->debugString);}
    $self->{rowObjects}[$i] = $n;
  }
  if ($info[7] < 2) { $self->error("<p>Invalid object: $refNum!  Must have at least 2 rows!</p>"); }
  return $i;
}

=item string javascript(var)

  This routine will convert this object to a javascript object.
  the variable to use when creating in javascript is optional.
  If not specified 'JournalEntry' is used.

=cut
sub javascript
{
  my $self = shift;
  my $var = shift;
  if (! defined $var) { $var = "JournalEntry"; }
  my $j = "var $var = new Object();\n";
  $j .= << "ENDOF";
  $var.id = '$self->{data}->{main}->{id}->{-Value}';
  $var.date = '$self->{data}->{main}->{date}->{-Value}';
  $var.payee = '$self->{data}->{main}->{payee}->{-Value}';
  $var.remember = '$self->{data}->{main}->{remember}->{-Value}';
  $var.journal = '$self->{data}->{main}->{journal}->{-Value}';
  $var.ledgerCode = '$self->{data}->{main}->{ledgerCode}->{-Value}';
  $var.origId = '$self->{data}->{main}->{origId}->{-Value}';
  $var.trackingId = '$self->{data}->{main}->{trackingId}->{-Value}';
  $var.rows = new Array();
ENDOF
  for (my $rowcount=0; $rowcount < scalar @{ $self->rowObjects }; $rowcount++)
  {
    $j .= $self->{rowObjects}->[$rowcount]->javascript($var);
  }
  $j .= "//END $var\n";
  $j .= <<"ENDOF";
  if (opener)
  {
    if (opener.postRecall) opener.postRecall($var);
    else { alert("function postRecall not found"); }
  }
ENDOF
  return $j;
}

=item string PAYEES(journal =>, style =>, days =>)

  This routine creates a javascript hash using payees as keys
  and transaction id (refNum) as values.  Mostly this is
  for use in a recall searchbox. It will accept an optional
  journal Code (G, CP, CSR, etc) to limit the payees.
  Also the days parameter will limit the search to the most
  recent transactions.  Defaults to 365.

  optional Styles: G, memo

=cut
sub PAYEES
{
  my $self = shift;
  my %args = (days => 365, style => 'G', journal => '', @_);
  my %sql;
  my $l = $self->get("ledgerCode");
  my $r = '';
  my $days = $args{days};
  if ($days !~ /^\d+$/) { $days = 365; }
  my $jW = "";
  my $j = $args{journal};
  if ($j ne "" && exists $self->{journals}->{$j})
  {
    $jW = " AND j.journal_code = '$j'";
  }
  $sql{G} = <<"SELECT";
  SELECT
   j.payee_payer AS payee,
   j.ref_num AS id
  FROM journal_entry_tb AS j
  WHERE j.ledger_code = '$l' AND j.payee_payer <> '' AND j.remember = true AND j.trans_date + $days > NOW() $jW
   UNION SELECT
     j.payee_payer AS payee,
     MAX(j.ref_num) AS id
  FROM journal_entry_tb AS j
  WHERE j.ledger_code = '$l' AND j.payee_payer <> '' AND j.trans_date + $days > NOW() $jW
  GROUP BY payee ORDER BY payee
SELECT
  $sql{memo} = <<"SELECT";
 SELECT
   j.ref_num AS id,
   '*'|| j.payee_payer || '   ('|| (SELECT comment FROM debitcredit_tb WHERE ref_num = parent_num AND counter = 0)||')' AS payee
  FROM journal_entry_tb AS j
  WHERE j.ledger_code = '$l' AND j.payee_payer <> '' AND j.remember = true AND j.trans_date + $days > NOW() $jW
 UNION SELECT
    MAX(j.ref_num) AS id,
    j.payee_payer || '    (' || debitcredit_tb.comment || ')' AS payee
  FROM journal_entry_tb AS j, debitcredit_tb
  WHERE j.payee_payer <> '' AND j.trans_date + $days > NOW() $jW
    AND debitcredit_tb.parent_num =
      (SELECT max(ref_num) FROM journal_entry_tb WHERE payee_payer = j.payee_payer)
    AND debitcredit_tb.counter = 0
  GROUP BY j.payee_payer, debitcredit_tb.comment
 ORDER BY payee
SELECT

  if (not defined $sql{$args{style}}) { $args{style} = "G"; }
  my $sth = $self->{dbHandle}->read($sql{$args{style}});
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage."for code: $j");
    return undef;
  }
  while (my $foo = $sth->fetchrow_hashref)
  {
    $r .= "PAYEES['$foo->{payee}'] = '$foo->{id}';\n";
  }
  return $r."/*\n".$sql{$j}."\n*/\n";
}

sub selectPayeeOptions
{
  my $self = shift;
  if ($self->get('ledgerCode') eq '') { $self->error("<p>Invalid ledger code</p>\n"); return undef; }
  $self->{LIST}[0] = "SELECT DISTINCT(payee_payer), payee_payer FROM journal_entry_tb WHERE ledger_code = '#FIELD=ledgerCode#' ORDER BY payee_payer";
  return $self->SUPER::selectOptions(@_);
}

sub printHTML
{
  my $self = shift;
  my $r = "<table border='1' cellpadding='0' cellspacing='0'><tr>\n";
  $r .= "<td>$self->{data}->{main}->{id}->{-Value}</td>";
  $r .= "<td>$self->{data}->{main}->{remember}->{-Value}</td>";
  $r .= "<td>$self->{data}->{main}->{date}->{-Value}</td>";
  $r .= "<td>$self->{data}->{main}->{payee}->{-Value}</td>";
  $r .= "<td>$self->{data}->{main}->{ledgerCode}->{-Value}</td>";
  $r .= "</tr></table>";
  $r .= "<p>Rows:" . scalar @{$self->{rowObjects}} ."</p>\n<table border='1' cellpadding='0' cellspacing='0'>";
  my $i = 0;
  foreach my $row (@{$self->rowObjects})
  {
    $r .= $row->printHTML();
  }
  return $r;
}

sub isNew
{
  my $self = shift;
  return ($self->get('id') eq '-1');
}

sub printForm
{
  my $self = shift;
  $self->{data}->{main}->{id}->{-Label} = $self->{data}->{main}->{id}->{-Value};
  if ($self->get('remember') eq 'true') { $self->{data}->{main}->{remember}->{-Label} = "+"; }
  my %form = $self->SUPER::printForm(@_, name => 'edit_'.$self->get('id') );
  if (!$self->isNew && !$self->isValid)
  {
    $form{body} .= $self->errorMessage;
  }
  for (my $rowcount=0; $rowcount < scalar @{ $self->{rowObjects} }; $rowcount++)
  {
    my %new = $self->{rowObjects}[$rowcount]->printForm();
    if ($self->{rowObjects}[$rowcount]->error) { $self->error($self->{rowObjects}[$rowcount]->errorMessage); }
    else { $form{body} .= $new{body}; }
  }
  return %form;
}

sub printHeaders
{
  my $self = shift;
  my $r = $self->{headers};
  my $row;
  if (scalar @{$self->{rowObjects}} > 0)
  {
    $row = $self->{rowObjects}[0];
  }
  else
  {
    $row = Portal::Accounting::Objects::JournalEntryRow->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $self->get('ledgerCode'));
  }
  $r .= $row->{headers};
  return $self->langSubstitute($r);
}

sub delete
{
  my $self = shift;
  my %journals = ( 'CP' => '',
    'CSR' => '',
    'G' => '',
    'TX' => '',
  );
  my $sql = "SELECT journal_code FROM journal_entry_tb WHERE ref_num = ".$self->get('id');
  my $sth = $self->{dbHandle}->read($sql);
  if ($self->{dbHandle}->error) { $self->error($self->{dbHandle}->errorMessage); return undef; }
  if (my @info = $sth->fetchrow_array)
  {
    if (! defined $journals{$info[0]}) { $self->error('Cannot delete from journal '.$info[0]); return -1; }
  }
  else { $self->error('no record found while checking for journal_code'); return undef; }
  return $self->SUPER::delete(@_);
}

1;
__END__

=pod

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.

=head1 AUTHOR

James A. Pattie (mailto:james@pcxperience.com), Jason A. Pattie (mailto:pattieja@pcxperience.com)

=head1 SEE ALSO

perl(1), JournalEntry(3), DBIWrapper(3)

=cut
