# JournalEntryRow.pm - Contains the data for a debitcredit_tb table row.
# Created by Jason A. Pattie, 11/07/2000.
#$Id: JournalEntryRow.pm,v 1.12 2005/08/16 21:45:22 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::JournalEntryRow;
use Portal::Accounting::Objects::BaseL;
use strict;
use vars qw($VERSION @ISA @EXPORT $AUTOLOAD);

require Exporter;
@ISA = qw(Portal::Accounting::Objects::BaseL Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.12 $';

=head1 NAME

JournalEntryRow - Object that contains a row of data used to build a JournalEntry in the accounting database.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::JournalEntryRow;
  my $transrowobj = JournalEntryRow->new();
  $transrowobj->populate(date => '12-14-1999',
                             comment => '',
                             subEntity => '',
                             account => 1001,
                             subAccount => -1,
                             dorc => 'true',
                             amount => '2579.63',
                             cleared => 'r');
  if ($transrowobj->error())
  {
    die $transrowobj->errorMessage();
  }

  # this object is generally used to build an array passed to
  # the constructor of a JournalEntryObject
  my @rowObjectArray;
  $rowObjectArray[0] = $transrowobj;
  $rowObjectArray[1] = $transrowobj;
  $rowObjectArray[1]->set('dorc','false');

  # check for validness of row objects values
  # before adding to JournalEntryObject
  if ($rowObjectArray[1]->isValid())
  {
    ...
  }
  else
  {
    die $rowObjectArray[1]->errorMessage();
  }

=head1 DESCRIPTION

JournalEntryRow is used as a container class that is incorporated
into the Objects::JournalEntry class.

=head1 Exported FUNCTIONS

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (ledgerCode => '', debug => 0, dbHandle => undef, @_);
  if ($args{debug} == 1) { $self->debug("<p>row debug on</p>\n"); }
  $self->{dbHandle} = $args{dbHandle};
  $self->{id} = "id";
  $self->{table} = "debitcredit_tb";
  $self->{INSERTorder}[0] = [ qw( parent_num counter ledger_code trans_date account_num sub_account_num sub_e_code comment dorc amount c_amount d_amount ) ] ;  #insert order for db column names
  $self->{newOnly}[0] = { parent_num => '', counter => ''};
  $self->{db2data}[0] = { 'parent_num' => 'id', 'ledger_code' => 'ledgerCode', 'trans_date' => 'date',
    'account_num' => 'account', 'sub_account_num' => 'subAccount', 'sub_e_code' => 'subEntity'
  };  #map database columns to ->data hash, not specified assumes same name
#  $self->{data2db}[0] = reverse %{$self->{db2data}[0]};
  $self->{data2db}[0] = { id => 'parent_num', 'ledgerCode' => 'ledger_code', 'date' => 'trans_date',
    'account' => 'account_num', 'subAccount' => 'sub_account_num', 'subEntity' => 'sub_e_code' };
  $self->{SELECT}[0] = "SELECT ##COLUMNS## FROM debitcredit_tb WHERE parent_num = #FIELD=id# AND counter = #FIELD=counter#";
  $self->{INSERT}[0] = "INSERT INTO debitcredit_tb (##COLUMNS##) VALUES (##VALUES##)";
  $self->{UPDATE}[0] = "UPDATE debitcredit_tb SET ##UPDATE## WHERE parent_num = #FIELD=id# AND counter = #FIELD=counter#";
  #remember to use single quotes around ##foo## if needed
  $self->{DELETE}[0] = "DELETE FROM debitcredit_tb WHERE parent_num = #FIELD=id# AND counter = #FIELD=counter#";

  $self->{data}->{main} = {
    "id" => { -Label => "Id", -Type => "hidden", -Value => -1, -DataType => "int" },
    "ledgerCode" => { -Type => "hidden", -Value => $args{ledgerCode}},
    "amount" => { -Type => "calculator", -DataType => "float",
      -WidgetOptions => { calcButton => 0, undoButton => 0, helpLink => 0 }
    },
    "date" => { -Type => "text", -DataType => "string" },
    "comment" => { -Type => "text", -ReadOnly => 1, -DataType => "string" },
    "dorc" => { -Type => "text", -Value => 'true', -DataType => "string" },
    "subEntity" => { -Type => "text", -ReadOnly => 1, -DataType => "string" },
    "subAccount" => { -Type => "hidden", -ReadOnly => 1, -DataType => "int", -Value => -1 },
    "account" => { -Type => "hidden", -ReadOnly => 1, -DataType => "int" },
    "cleared" => { -Type => "text", -DataType => "string", -Value => "_" },
    "counter" => { -Type => "text", -DataType => "int" },
    "c_amount" => { -Type => "text", -DataType => "float" },
    "d_amount" => { -Type => "text", -DataType => "float" },
    "debit" => { -Label => " ", -Type => "hidden", },
    "credit" => { -Label => " ", -Type => "hidden", },
  };
  $self->{headers} = <<"ENDOF";
  <table width="100%" cellspacing="0" cellpadding="0" border="1">
    <tr>
      <th style="width: 3em; text-align: left;">
        <a title="#LANG=SubEntity#">S</a>
      </th>
      <th style="width: 15em; text-align: left;">
        #LANG=Account#
      </th>
      <th style="width: 11em; text-align: left;">
        #LANG=Sub Account#
      </th>
      <th style="width: 6em; text-align: right;">
        #LANG=Debit#
      </th>
      <th style="width: 6em; text-align: right;">
        #LANG=Credit#
      </th>
      <th style=" text-align: left;">
        &nbsp;&nbsp;#LANG=Comment#
      </th>
    </tr>
  </table>
ENDOF
  $self->{template}->{main} = <<"ENDOF";
  <table width="100%" cellspacing="0" cellpadding="0" border="1">
    <tr valign="top">
      <td style="width: 3em;">
        #FIELD=subEntity#&nbsp;
      </td>
      <td style="width: 15em;">
        #LABEL=account# #FIELD=account#
      </td>
      <td style="width: 11em;">
        #LABEL=subAccount# #FIELD=subAccount#
      </td>
      <td style="width: 6em; text-align: right;">
        #LABEL=debit#
      </td>
      <td style="width: 6em; text-align: right;">
        #LABEL=credit#
      </td>
      <td>
        #FIELD=comment#&nbsp;
      </td>
    </tr>
  </table>
ENDOF
  return $self;
}

=item   scalar isValid(void)

    Runs through the data fields contained in the class to make sure
    they contain valid data.  This function will be called
    automatically when embedded in a JournalEntryObject.

=cut
sub isValid
{
  my $self = shift;
  if (!$self->SUPER::isValid())
  {
    $self->error("skipping rest of isValid");
    return undef;
  }

  if (!defined $self->{dbHandle})
  {
    $self->missing("dbHandle");
  }
  if ($self->get("account") !~ /^[1-5]\d+$/)
  {
    $self->invalid("account", $self->get("account"));
  }
  if ($self->get("subAccount") !~ /^\d+$/ && $self->get("subAccount") != -1)
  {
    $self->invalid("subAccount", $self->get("subAccount"));
  }
  if ($self->get("dorc") ne "true" && $self->get("dorc") ne "false")
  {
    $self->invalid("dorc", $self->get("dorc"));
  }
  my $amount = $self->get("amount");
  if ($amount !~ /^([0-9]*).([0-9]{2})$/) { $self->invalid("amount", $amount); }
  elsif ($amount < 0) { $self->invalid("amount", $amount); }
  if ($self->get("d_amount") !~ /^-?([0-9]*).([0-9]{2})$/)
  {
    $self->invalid("d_amount", $self->get("d_amount"));
  }
  if ($self->get("c_amount") !~ /^-?([0-9]*).([0-9]{2})$/)
  {
    $self->invalid("c_amount", $self->get("c_amount"));
  }
  if ($self->get("cleared") !~ /[_xr]/)
  {
    $self->set("cleared",'_');
  }

  if ($self->{error})
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }
  return 1;
}

=item boolean populate()

  requires:  date, account, dorc, amount
  optional:  comment, subEntity, subAccount, cleared

    Fills in this instance of the Portal::Accounting::Objects::JournalEntryRow
    and validates the values contained in it.

=cut
sub populate
{
  my $self = shift;
  if (!$self->SUPER::populate(@_))
  { $self->prefixError('populate failed!'); return 0; }
  if ($self->get("account") =~ /^1/)
  {
    #!!! what if this is a cash account?  the c_amounts are different?
    if ($self->get("dorc") == 'true')
    {
      $self->{data}->{main}->{c_amount}->{-Value}  -= $self->get("amount");  #increasing asset decreases cash
      $self->{data}->{main}->{d_amount}->{-Value}  = $self->get("amount");  #debits increase assets
    }
    else
    {
      $self->{data}->{main}->{c_amount}->{-Value}  = $self->get("amount"); #decreasing asset like selling which increases cash
      $self->{data}->{main}->{d_amount}->{-Value}  -= $self->get("amount"); #credits decrease assets
    }
  }
  elsif ($self->get("account") =~ /^2/)
  {
    if ($self->get("dorc") == 'true')
    {
      $self->{data}->{main}->{d_amount}->{-Value}  -= $self->get("amount"); #debit decrease liability
      $self->{data}->{main}->{c_amount}->{-Value}  -= $self->get("amount"); #decreasing liability uses cash
    }
    else
    {
      $self->{data}->{main}->{c_amount}->{-Value}  = $self->get("amount"); #decreasing liability uses cash
      $self->{data}->{main}->{c_amount}->{-Value}  = $self->get("amount");  #increasing liability gets cash
    }
  }
  elsif ($self->get("account") =~ /^3/)
  {
    if ($self->get("dorc") == 'true')
    {
      $self->{data}->{main}->{d_amount}->{-Value}  -= $self->get("amount"); #decreasing equity with debit
      $self->{data}->{main}->{c_amount}->{-Value}  -= $self->get("amount"); #decreasing equity uses cash
    }
    else
    {
      $self->{data}->{main}->{d_amount}->{-Value}  = $self->get("amount"); #increasing equity with credit
      $self->{data}->{main}->{c_amount}->{-Value}  = $self->get("amount"); #increasing equity gets cash
    }
  }
  elsif ($self->get("account") =~ /^4/)  #same as 3
  {
    if ($self->get("dorc") == 'true')
    {
      $self->{data}->{main}->{d_amount}->{-Value}  -= $self->get("amount"); #decreasing revenue with debit
      $self->{data}->{main}->{c_amount}->{-Value}  -= $self->get("amount"); #decreasing revenues uses cash
    }
    else
    {
      $self->{data}->{main}->{d_amount}->{-Value}  = $self->get("amount"); #increasing revenue with credit
      $self->{data}->{main}->{c_amount}->{-Value}  = $self->get("amount"); #increasing revenues increases cash
    }
  }
  elsif ($self->get("account") =~ /^5/)  #same as 1
  {
    if ($self->get("dorc") == 'true')
    {
      $self->{data}->{main}->{d_amount}->{-Value}  = $self->get("amount");  #expenses increased by debit
      $self->{data}->{main}->{c_amount}->{-Value}  -= $self->get("amount");  #expenses use cash
    }
    else
    {
      $self->{data}->{main}->{d_amount}->{-Value}  -= $self->get("amount");  #expenses decreased by credit
      $self->{data}->{main}->{c_amount}->{-Value}  = $self->get("amount"); #negative expenses refund cash
    }
  }
  else { $self->prefixError("invalid account") ; return undef; }

  return $self->isValid;
}

sub setDate
{
  my $self = shift;
  my $date = shift;
  if ($date =~ /^(\d{4})[\-\\\/](\d{1,2})[\-\\\/](\d{1,2})$/)
  {
    $date = $1 ."-". sprintf("%02d", $2) ."-". sprintf("%02d", $3);
  }
  else
  {
    $self->error("Invalid date: $date<br>\n");
    return 0;
  }
  $self->{data}->{main}->{date}->{-Value} = $date;
  return 1;
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut
sub printXML
{
  my $self = shift;
  my $result = "";
  my %args = ( fullXML => 0, indent => 2, @_ );
  my $fullXML = $args{fullXML};
  my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
  my $spaceStr = " " x $indent;

  my %mapping = ("comment" => "comment", "accountCode" => "account", "date" => "rowDate",
                 "subAccountCode" => "subAccount", "dorc" => "dorc", "amount" => "amount",
                 "subEntityCode" => "subEntity", "cleared" => "cleared");

  if ($self->isValid())
  {
    if ($fullXML)
    {
      $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
    }
    $result .= $spaceStr . "<line>\n";
    my @tags = qw{dorc accountCode amount subAccountCode comment subEntityCode cleared};
    push @tags, "date" if ($fullXML);
    foreach my $name (@tags)
    {
      $result .= $spaceStr . "  <$name>" . $self->{$mapping{$name}} . "</$name>\n";
    }
    $result .= $spaceStr . "</line>\n";
  }
  else
  {
    $result .= "Data not valid!\n\n";
    $result .= $self->errorMessage();
    die $result;
  }

  return $result;
}

=item string javascript(var)

  var is the name of the js variable to use. defaults to JournalEntry

=cut
sub javascript
{
  my $s = shift;
  my $var = shift;
  if (not defined $var) { $var = "JournalEntry"; }
  my $row = $s->get("counter");
  my $r = "  $var.rows[$row] = new Object();\n";
  $r .= "  $var.rows[$row].account = '".$s->get("account")."';\n";
  $r .= "  $var.rows[$row].subAccount = '".$s->get("subAccount")."';\n";
  $r .= "  $var.rows[$row].dorc = '".$s->get("dorc")."';\n";
  $r .= "  $var.rows[$row].subEntity = '".$s->get("subEntity")."';\n";
  $r .= "  $var.rows[$row].comment = '".$s->get("comment")."';\n";
  $r .= "  $var.rows[$row].amount = '".$s->get("amount")."';\n";
  $r .= "  $var.rows[$row].c_amount = '".$s->get("c_amount")."';\n";
  $r .= "  $var.rows[$row].d_amount = '".$s->get("d_amount")."';\n";
  $r .= "  $var.rows[$row].cleared = '".$s->get("cleared")."';\n";
  return $r;
}

sub printForm
{
  my $self = shift;
  if ($self->get('dorc') eq "true") { $self->{data}->{main}->{debit}->{-Label} = $self->get('amount'); }
  else { $self->{data}->{main}->{credit}->{-Label} = $self->get('amount'); }
  return $self->SUPER::printForm(@_);
}

sub isNew
{
  my $self = shift;
  if ($self->get('id') == -1) { return 1; }
  return 0;
}

sub select
{
  my $self = shift;
  my $r = $self->SUPER::select(@_);
  #convert dorc field
  if ($self->get('dorc') == 0)
  {
    $self->set('dorc', 'false');
  }
  elsif ($self->get('dorc') == 1)
  {
    $self->set('dorc', 'true');
  }
  else
  {
    $self->error('Invalid dorc'.$self->get('dorc'));
  }
  #set account sub account labels to names/codes
  if (!$self->selectNames) { $self->error("<p>could not set account and sub account names</p>\n"); }
  return $r;
}

#used to set label fields for account and sub account to names
sub selectNames
{
  my $self = shift;
  my $asql = "SELECT name, taxable FROM account_tb WHERE number = '".$self->get('account')."'";
  my $result = $self->{dbHandle}->read($asql);
  if ($self->{dbHandle}->error) { $self->error($self->{dbHandle}->errorMessage); }
  elsif (my $info = $result->fetchrow_hashref)
  {
    $self->{data}->{main}->{account}->{-Label} = $info->{name};
    if ($info->{taxable}) { $self->{data}->{main}->{account}->{-Label} .= "*"; }
  }
  my $sub = $self->get('subAccount');
  if ($sub != -1)
  {
    my $ssql = "SELECT sub_acc_name AS name FROM sub_acc_tb WHERE sub_acc_number = '$sub'";
    $result = $self->{dbHandle}->read($ssql);
    if ($self->{dbHandle}->error) { $self->error($self->{dbHandle}->errorMessage); }
    elsif (my $info = $result->fetchrow_hashref)
    {
      $self->{data}->{main}->{subAccount}->{-Label} = $info->{name};
    }
  }
  if ($self->error) { return 0; }
  return 1;
}

1;
__END__

=item   scalar error(errorString)
    Sets error = 1 and $self->{errorString} = errorString.

=cut
=item   scalar error()
    Returns the value of error.

=cut
=item   scalar errorMessage(void)
    Returns the value of errorString.

=cut
=pod

   NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.

=head1 AUTHOR

JT Moree (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), JournalEntryObject(3)

=cut
