# Ledger.pm - The Object Class that provides a Ledger Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: Ledger.pm,v 1.34 2005/08/04 01:55:24 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Ledger;
use strict;
use Portal::Accounting::Objects::Base;
use Date::Calc qw(Add_Delta_Days Today);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.34 $';

=head1 NAME

Ledger - Object used to build a Ledger Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Ledger;
  my $ledgerObj = Portal::Accounting::Objects::Ledger->new()
  $ledgerObj->populate( . . .)
  if ($ledgerObj->error())
  {
    die $ledgerObj->errorMessage();
  }

=head1 DESCRIPTION

Ledger is a Ledger class.  This represents an entry in the
ledger_tb table in the database.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(langObj => )

    Creates a new instance of Portal::Accounting::Objects::Ledger.
    See Portal::Base(3) for a listing of required arguments.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->prefixError();
    return $self;
  }
  my $l = $self->{langObj};

  # instantiate anything unique to this module
  my @date = Today;
  my $cash = $l->map("Cash");
  my $accrual = $l->map("Accrual");
  my $update = $l->map("Update Month Starts")."-->";
  my $Month = $l->map("Month");
  my $StartDate = $l->map("Start Date");
  $self->{data}->{main} = {
    "command" => { -Type => "hidden",  -Value => "save" },
    "app" => { -Type => "hidden", -Value => "Accounting" },
    "state" => { -Type => "hidden", -Value => "Main" },
    "windowName" => { -Type => "hidden", -Value => "", -DataType => 'string', },
    "new" => { -Type => "hidden", -DataType => 'int', -Value => "1" },
    "color" => { -Label => $l->map("Color"), -Type => "color-picker", -Value => "#00FF00",
      -DataType => 'string',
    },
    "name" => { -Label => $l->map("Ledger Name"), -Type => "text", size => 15,
      -DataType => 'string',
    },
    "code" => { -Label => $l->map("Ledger Code"), -Type => "text", size => 15,
      -DataType => 'string',
    },
    "fiscalYear" => { -Label => $l->map("Fiscal Year"), -Type => "text", size => 6,
      -DataType => 'string', onchange => 'javascript: this.form.year.value = this.value;'
    },
    "year" => { -Label => $l->map("Year"), -Type => "text", size => 4,
      -DataType => 'string',
    },
    "month" => { -Label => $l->map("Month"), -Type => "text", size => 2,
      -DataType => 'string', -Value => '01',
    },
    "day" => { -Label => $l->map("Day"), -Type => "text", size => 2,
      -DataType => 'string', -Value => '01',
    },
    "businessType" => { -Label => $l->map("Business Type"), -Type => "select",
      -Options => $self->{formObj}->createSelectOptions($self->getBusinessTypeNames),
      -DataType => "int"
    },
    "basis" => { -Type => "radio", -DataType => 'string', -Label => $l->map("Basis"), -Options => { names => [ $cash, $accrual ], values => [ qw(cash accrual) ] }, -Value => 'cash'
    },
    "start2" => { -Label => "2", -Type => "text", size => 6, -Value => "02-01",
      -DataType => 'string',
    },
    "start3" => { -Label => "3", -Type => "text", size => 6, -Value => "03-01",
      -DataType => 'string',
    },
    "start4" => { -Label => "4", -Type => "text", size => 6, -Value => "04-01",
      -DataType => 'string',
    },
    "start5" => { -Label => "5", -Type => "text", size => 6, -Value => "05-01",
      -DataType => 'string',
    },
    "start6" => { -Label => "6", -Type => "text", size => 6, -Value => "06-01",
      -DataType => 'string',
    },
    "start7" => { -Label => "7", -Type => "text", size => 6, -Value => "07-01",
      -DataType => 'string',
    },
    "start8" => { -Label => "8", -Type => "text", size => 6, -Value => "08-01",
      -DataType => 'string',
    },
    "start9" => { -Label => "9", -Type => "text", size => 6, -Value => "09-01",
      -DataType => 'string',
    },
    "start10" => { -Label => "10", -Type => "text", size => 6, -Value => "10-01",
      -DataType => 'string',
    },
    "start11" => { -Label => "11", -Type => "text", size => 6, -Value => "11-01",
      -DataType => 'string',
    },
    "start12" => { -Label => "12", -Type => "text", size => 6, -Value => "12-01",
      -DataType => 'string',
    },
    "updateSI" => { -Type => "button", -Value => $update, onclick => "updateStartInfo(this.form);" },
  };
  $self->{css}->{main} = <<"ENDOF";
table.editLedger { border: none; }
td.months { text-align: right; padding-left: 4px; }
th.months { text-align: right; padding-left: 4px; }
td.starts {  }
th.starts {  }
ENDOF
  $self->{template}->{main} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIELD=app#
  #FIELD=state#
  #FIELD=command#
  #FIELD=new#
  <input type="hidden" name="ledgerOpen" value="0">
  #FIELD=windowName#
  <table border="1" cellpadding="5" cellspacing="0" width="100%">
    <tr>
      <td align="left" valign="top">
        <p>
        <b>#LABEL=name#</b>&nbsp;#FI=name#
        </p><p>
        <b>#LABEL=code#</b>&nbsp;#FI=code#
        </p><p>
        <b>#LABEL=color#</b>&nbsp;#FI=color#
        </p><p>
        <b>#LABEL=businessType#</b><br>
        #FI=businessType#
        </p><p>
        <b>#LABEL=basis#</b><br> #FI=basis#
        </p>
      </td>
      <td>
        #LABEL=fiscalYear# #FI=fiscalYear#
        Begins on<br>#LABEL=year# #FI=year# #LABEL=month# #FI=month# #LABEL=day# #FI=day#
        <br>#FIELD=updateSI#
        </p><p><table class="editLedger" width="100%">
        <tr><td>
        <table class="editLedger">
          <tr>
            <th class="months">$Month</th>
            <th class="starts">$StartDate</th>
          </tr>
          <tr>
            <td class="months">#LABEL=start2#</td>
            <td class="starts">#FIELD=start2#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start3#</td>
            <td class="starts">#FIELD=start3#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start4#</td>
            <td class="starts">#FIELD=start4#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start5#</td>
            <td class="starts">#FIELD=start5#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start6#</td>
            <td class="starts">#FIELD=start6#</td>
          </tr>
        </table>
        </td><td>
        <table class="editLedger">
          <tr>
            <td class="months">#LABEL=start7#&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td>
            <td class="starts">#FIELD=start7#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start8#</td>
            <td class="starts">#FIELD=start8#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start9#</td>
            <td class="starts">#FIELD=start9#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start10#</td>
            <td class="starts">#FIELD=start10#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start11#</td>
            <td class="starts">#FIELD=start11#</td>
          </tr>
          <tr>
            <td class="months">#LABEL=start12#</td>
            <td class="starts">#FIELD=start12#</td>
          </tr>
        </table>
        </td></tr></table>
      </td>
    </tr>
  </table>
  <input type="submit" name="Submit" value="Submit"></form>
ENDOF
  $self->{javascript}->{main} = <<"ENDOF";
function updateStartInfo(d)
{
  // get the first month and day values
  var month = parseInt(d.month.value);
  var day = parseInt(d.day.value);
  if (isNaN(month) || month < 1 || month > 12) { alert('Invalid month'); return false; }
  if (isNaN(day) || day < 1 || day > 31) { alert('Invalid day'); return false; }
  d.day.value = day;
  d.month.value = month;
  var n=new Number(month);

  // now update the start2-12 values to be month+1-day
  // take care to handle the month+1 > 12 case
  for (var i=2; i<13; i++)
  {
    // alert("i='"+i+"', month='"+n+"', day='"+day+"'");
    if (n + 1 > 12)
    {
      n = 1;
    }
    else
    {
      n++;
    }
    try
    {
      eval("d.start"+i+".value='"+(n.toString().length == 1 ? "0" : "") + n.toString() +"-"+day+"';");
    }
    catch(exception)
    {
      alert(exception);
    }
  }
}
ENDOF
  $self->{profile}->{main} = {
    required => [ qw( name code color basis businessType fiscalYear year month day start2 start3 start4 start5 start6 start7 start8 start9 start10 start11 start12 ) ],
    optional => [  ],
    constraints => {
      name => qr/^[\d\w][\s\d\w]+$/,
      code => qr/^[\d\w]+$/,
      color => qr/^\#[A-Fa-f0-9]{6}$/,
      businessType => qr/^\d$/,
      fiscalYear => qr/^\d{4}$/,
      month => qr/^\d{1,2}$/,
      day => qr/^\d{1,2}$/,
      year => qr/^\d{4}$/,
      start2 => qr/^\d{1,2}-\d{1,2}$/,
      start3 => qr/^\d{1,2}-\d{1,2}$/,
      start4 => qr/^\d{1,2}-\d{1,2}$/,
      start5 => qr/^\d{1,2}-\d{1,2}$/,
      start6 => qr/^\d{1,2}-\d{1,2}$/,
      start7 => qr/^\d{1,2}-\d{1,2}$/,
      start8 => qr/^\d{1,2}-\d{1,2}$/,
      start9 => qr/^\d{1,2}-\d{1,2}$/,
      start10 => qr/^\d{1,2}-\d{1,2}$/,
      start11 => qr/^\d{1,2}-\d{1,2}$/,
      start12 => qr/^\d{1,2}-\d{1,2}$/,
    },
  };

  $self->{table} = 'ledger_tb';
  $self->{id} = 'code';
  $self->{INSERTorder}[0] = [ qw( name code color business_type fy_mo fy_day start2 start3 start4 start5 start6 start7 start8 start9 start10 start11 start12 ) ] ;  #insert order for db column names
  $self->{newOnly}[0] = { code => "", day => "", month => '', year => '', fiscalYear => '', start2 => '', start3 => '', start4 => '', start5 => '', start6 => '', start7 => '', start8 => '', start9 => '', start10 => '', start11 => '', start12 => '' };
  $self->{db2data}[0] = {
   business_type => 'businessType',
   cash_or_accrual => 'basis',
   fy_day => 'day',
   fy_mo => 'month',
   };  #map database columns to ->data hash, not specified assumes same name
  $self->{data2db}[0] = reverse %{$self->{db2data}[0]};

  $self->{SELECT}[0] = "SELECT ##COLUMNS## FROM $self->{table} WHERE $self->{id} = '#FIELD=$self->{id}#'";
  $self->{INSERT}[0] = "INSERT INTO $self->{table} (cash_or_accrual, ##COLUMNS##) VALUES (false, ##VALUES##)";
  $self->{UPDATE}[0] = "UPDATE $self->{table} SET ##UPDATE## WHERE code = '#FIELD=$self->{id}#'";
  #remember to use single quotes around ##foo## if needed
  $self->{DELETE}[0] = "DELETE FROM $self->{table} WHERE code = '#FIELD=$self->{id}#'";
  #what should we output as the value of the thing just deleted?
  $self->{postDELETE}[0] = "SELECT name||', '||code FROM $self->{table} WHERE code = '#FIELD=$self->{id}#' ";
  $self->{LIST}[0] = "SELECT name||' ('||code||')' AS name, code FROM $self->{table} ORDER BY code";

  # do anything else you might need to do.
  $self->{offset} = ();
  return $self;
}

=item boolean populate( . . . )

    new       - 1 for yes, 0 for no
    name      - text
    code      - text
    color     - text
    basis - string (cash, accrual)
    businessType  - integer pertaining to a business type (see business_type_tb)
    fiscalYear -  the first fiscalYear
    year      - the actual year for the first fiscal year
    month     - the first month of the fiscal year
    day       - the first day of the first month
    start2 - start12  - begginning of each month in "MMDD" format

=cut
sub populate
{
  my $self = shift;
  my %input = (fiscalYear => '', year => '', month => '', day => '', workingYear => '', new => 1, @_);
  $self->{knownYears}->{$input{fiscalYear}} = "$input{year}-$input{month}-$input{day}";
  $self->{offset}[1] = $input{year} - $input{fiscalYear};
  if ($input{workingYear} ne '')
  {
    $self->{workingYear} = $input{workingYear};
  }
  else
  {
    $self->{workingYear} = $input{fiscalYear};
  }

  $self->set('new', $input{new});
  if (!$self->isNew)
  {
    $self->{data}->{main}->{updateSI}->{-Type} = "hidden";
  }
  return $self->SUPER::populate(@_);
}

=item bool isNew(void)

  This routine checks the new flag to see if the ledger already exists in the database.
  1 for new, 0 for already exists

=cut
sub isNew
{
  my $self = shift;
  return $self->{data}->{main}->{new}->{-Value};
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

#   if (not defined $self->{knownYears})
#   {
#     $self->missing("knownYears");
#   }
#   else
#   {
#     foreach my $year (keys %{$self->{knownYears}} )
#     {
#       if ($year !~ /^(\d{4})$/) #check the key for a valid integer (year)
#       {
#         $self->invalid('knownYears_'.$year, $year);
#       }
#       elsif ($self->{knownYears}->{$year} !~ m/^(\d{8})$/ && $self->{knownYears}->{$year} !~ s/^(\d{4})([-\/\\])(\d{1,2})([-\/\\])(\d{1,2})$/$1.sprintf("%02d", $3 ).sprintf("%02d",$5)/e)
#       {
#         $self->invalid("knownYears->$year", $self->{knownYears}->{$year});
#       }
#     }
#   }
  if ($self->{offset}[1] !~ m/^(-?\d+)$/)
  {
    $self->invalid('offset1', $self->{offset}[1]);
  }

  my $years = 0;
  for (my $index = 2; $index <= 12; $index++)
  {
    my $d = $self->get("start$index");
    if ($d =~ s/^(\d{2})(\d{2})$/$1-$2/ || $d =~ s/^(\d{1,2})-(\d{1,2})$/sprintf("%02d",$1)."-".sprintf("%02d",$2)/e)
    {
      $self->set("start$index", $d);
      $d =~ s/-//g;
      my $e = $self->start($index - 1);
      $e =~ s/-//g;
      if ($d < $e)
      {
        if ( ++$years > 1 )
        {
          $self->invalid("start".$index, $self->get("start$index : is out of order"));
          $years = 1;
        }
      }
      elsif ($d == $e)
      {
        $self->invalid("start".$index, $self->get("start$index is the same as start".$index - 1));
      }
    }
    else
    {
      $self->invalid("start".$index, $self->get("start$index"));
    }
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item String[] businessTypeName(void)

 Returns the business type name from the proper lanquage
 if it is set in the object or an empty string if not set.
 There is no guarantee that it is set!

=cut

sub businessTypeName
{
  my $self = shift;
  for (my $i; $i < @{$self->{data}->{main}->{businessType}->{-Options}->{values}}; $i++)
  {
    if ($self->{data}->{main}->{businessType}->{-Options}->{values}[$i] == $self->get('businessType'))
    {
      return $self->{data}->{main}->{businessType}->{-Options}->{names}[$i];
    }
  }
  return "";
}

=item string end(month => , fiscalYear|delta => ) or end(<month>)

 This routine returns the ending date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

 delta takes an integer that will be added to the current working
 year and used as the year in the result

=cut
sub end
{
  my $self = shift;
  my $month;
  my $year = 0;
  my %args;
  my $return = "";
  if (scalar @_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => 0, delta => 0, @_);
    if ($args{delta} !~ /^-?\d+$/)
    {
      $self->prefixError($self->{langObj}->map("Invalid") . " delta '". $args{delta} . "'<br>\n");
      return undef;
    }
    $month = $args{month};
    if ($args{delta} != 0)
    {
      $year = $self->workingYear() + $args{delta};
    }
    else
    {
      $year = $args{fiscalYear};
    }
  }
  if ($year !~ m/^(\d)+$/)
  {
     $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("year") . " '". $year . "'<br>\n");
     return undef;
  }
  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
     $self->prefixError($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("month") . " '". $month . "'<br>\n");
     return undef;
  }
  my @date = split '-', $self->start(month => ($month % 12) + 1, fiscalYear => $year  );
  if ($month == 12)
  {
    $date[0]++;
  }
  if (@date != 3)
  {
    $self->prefixError("Invalid split of date string: ".$self->start(month => $month % 12 + 1, fiscalYear => $year));
    return undef;
  }
  @date = Add_Delta_Days(@date, -1);
  if ($year == 0)
  {  return sprintf("%02d",$date[1])."-".sprintf("%02d",$date[2]);  }
  return sprintf("%02d",$date[0])."-".sprintf("%02d",$date[1])."-".sprintf("%02d",$date[2]);
}

=item date[] getQuarterEnds([fiscal year])

 Returns an array of strings representing the ending dates of Quarters
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut
sub getQuarterEnds
{
  my $self = shift;
  my @quarters;
  my $year = 0;
  my @month = qw [ 3 6 9 12 ];
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }

  my $index = -1;
  for my $month (@month)
  {
    if ($year)
    {
      $quarters[++$index] = $self->end(month => $month, fiscalYear => $year);
    }
    else
    {
      $quarters[++$index] = $self->end($month);
    }
  }

  return @quarters;
}

=item date[] getQuarterStarts([fiscal year])

 Returns an array of strings representing the starting dates of Quarters
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getQuarterStarts
{
  my $self = shift;
  my @quarters = ();
  my $year = 0;
  my @month = qw [ 3 6 9 12 ];
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@month)
  {
    $quarters[++$index] = $self->start(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $quarters[$index];
      $quarters[$index] = "$date[1]-$date[2]";
    }
  }

  return @quarters;
}

=item date[] getMonthEnds([fiscal year])

 Returns an array of strings representing the ending dates of months
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getMonthEnds
{
  my $self = shift;
  my @months = qw [ 1 2 3 4 5 6 7 8 9 10 11 12 ];
  my @return;
  my $year = 0;
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@months)
  {
    $return[++$index] = $self->end(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $return[$index];
      $return[$index] = "$date[1]-$date[2]";
    }
  }

  return @return;
}

=item date[] getMonthStarts([fiscal year])

 Returns an array of strings representing the starting dates of months
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getMonthStarts
{
  my $self = shift;
  my @months = qw [ 1 2 3 4 5 6 7 8 9 10 11 12 ];
  my @return;
  my $year = 0;
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@months)
  {
    $return[++$index] = $self->start(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $return[$index];
      $return[$index] = "$date[1]-$date[2]";
    }
  }

  return @return;
}

=item int offset(month)

 This routine calculates the offset for each month to change the
 fiscal year value into a calendar year value.

=cut

sub offset
{
  my $self = shift;
  my $month = shift;

  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
    $self->error("<p>Invalid month sent to offset</p>\n");
    return undef;
  }
  if ($month == 1) { return $self->{offset}[$month]; }
  if (! defined $self->{offset}[$month])
  {
    my $offset = $self->{offset}[1];
    if ($self->get("start2") < sprintf("%02d",$self->get("month"))."-".sprintf("%02d",$self->get("day")))
    {
      $offset++;
    }
    $self->{offset}[2] = $offset;
    for (my $index = 3; $index <= 12; $index++)
    {
      #compare this date+offset with the last one+offset
      my $foo = $index - 1;
      if ($self->get("start$index") < $self->get("start" . $foo))
      {
        $offset++;
      }
      $self->{offset}[$index] = $offset;
    }
  }
  return $self->{offset}[$month];
}

=item string start(month => , fiscalYear =>) or start(<month>)

 This routine returns the starting date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

=cut

sub start
{
  my $self = shift;
  my $month;
  my $year = "";
  my %args;
  my $return = "";
  if (@_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => "", @_);
    $month = $args{month};
    $year = $args{fiscalYear};
  }
  if ($month == 1) {  $return = sprintf("%02d",$self->get("month"))."-".sprintf("%02d",$self->get("day")); }
  else { $return = $self->get("start$month"); }
  if ($return !~ m/^(\d{2})-?(\d{2})$/)
  {
    $self->error("Invalid internal month start: '$return'<br>\n");
    return undef;
  }
  $return = $1."-".$2;
  if ($year eq "")
  {  return $return; }
  elsif ($year =~ m/^\d+$/)
  {  return $year + $self->offset($month) ."-".$return;  }
  $self->prefixError("Invalid year: '".$year."'<br>\n");
  return undef;
}

=item ISO-Date firstDate()

 Returns the first date the books were opened for.

=cut

sub firstDate
{
  my $self = shift;

  # try not to calculate it every time.
  if (!defined $self->{firstDate})
  {
    my @years = keys %{$self->{knownYears}};
    my $start1 = $self->get('month')."-".$self->get('day');

    # we build this up based upon the oldest known year + offset of first month and
    # tack on the first month-day entry.
    my $firstDate = ((sort { $a <=> $b } @years)[0] + $self->offset(1)) . "-" . $start1;

    $self->{firstDate} = $firstDate;
  }

  return $self->{firstDate};
}

=item ISO-Date formatDate(string)

 This routine takes an 8 digit date and returns the it formatted in the ISO standard of YYYY-MM-DD
 if a year is specified.  In either case, we have -'s for seperators.

 If an invalid date is specified, we return "" and set an error.

=cut

sub formatDate
{
  my $self = shift;
  my $date = shift;

  if ($date =~ /^(\d{4})(\d{2})(\d{2})$/)
  {
    $date = "$1-$2-$3";
  }
  elsif ($date =~ /^(\d{2})(\d{2})$/)
  {
    $date = "$1-$2";
  }
  elsif ($date !~ /^((\d{4}-)?\d{2}-\d{2})$/)
  {
    $self->invalid("date", $date);
    $date = "";
    $self->error($self->genErrorString("all"));
  }

  return $date;
}

=item string printXML(fullXML bool, indent int)

 This routine returns an XML snippet for this object for use in such
 things as exporting data

 fullXML - defaults to 0 (false).  If true, we output the xml header tag.
 indent - defaults to 2.  Specifies how many spaces in our top level tag
          needs to be.  If fullXML is true, then we force 0, otherwise we
          use what the user specifies.

=cut

# sub printXML
# {
#   my $self = shift;
#   my $result = "";
#   my %args = ( fullXML => 0, indent => 2, @_ );
#   my $fullXML = $args{fullXML};
#   my $indent = ($fullXML ? 0 : $args{indent});  # start at 0 indentation if we are making a full XML document.
#   my $spaceStr = " " x $indent;
#
#   if ($self->isValid())
#   {
#     if ($fullXML)
#     {
#       $result .= "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
#       $result .= $spaceStr . "<ledger name=\"".$self->get('name')."\" code=\"".$self->get('code')."\">\n";
#       $indent += 2;
#       $spaceStr = " " x $indent;
#     }
#     $result .= $spaceStr . "<ledgerInfo>\n";
#     foreach my $name (qw{color basis businessType businessTypeName})
#     {
#       $result .= $spaceStr . "  <$name>" . ($name eq "firstDate" ? $self->formatDate($name) : ($name eq "businessTypeName" ? $self->businessTypeName() : $name)) . "</$name>\n";
#     }
#     $result .= $spaceStr . "  <knownYears>\n";
#     foreach my $year (keys %{$self->{knownYears}})
#     {
#       $result .= $spaceStr . "    <info year=\"$year\" endDate=\"" . $self->formatDate($self->{knownYears}->{$year}) . "\"/>\n";
#     }
#     $result .= $spaceStr . "  </knownYears>\n";
#     $result .= $spaceStr . "  <dates>\n";
#     for (my $i=1; $i <= 12; $i++)
#     {
#       $result .= $spaceStr . "    <info index=\"$i\" start=\"" . $self->start(month => $i) . "\" offset=\"" . $self->offset($i) . "\"/>\n";
#     }
#     $result .= $spaceStr . "  </dates>\n";
#     $result .= $spaceStr . "</ledgerInfo>\n";
#     if ($fullXML)
#     {
#       $indent -= 2;
#       $spaceStr = " " x $indent;
#       $result .= $spaceStr . "</ledger>\n";
#     }
#   }
#   else
#   {
#     $result .= "Data not valid!\n\n";
#     $result .= $self->errorMessage();
#     die $result;
#   }
#
#   return $result;
# }
#
=item integer workingYear()

  Returns the earliest year on the books that is not closed out.

=cut
sub workingYear
{
  my $self = shift;
  return $self->{workingYear};
}

=item date lastClose()

  Returns the last closed out date.

=cut
sub lastClose
{
  my $self = shift;
  return $self->{lastClose};
}

=item Objects::Ledger[] getLedgers(ledgerCode =>)

    Returns an array of Ledger Objects matching the parameters passed.
    No params means select all ledgers.  Empty array is returned if
    no ledgers exist or match criteria.
    NOTE: Will not error on an invalid ledger.  You must do an
    ->isValid on each returned ledgerObj to decide what to do with it.
    Undef returned on errors.

=cut
sub getLedgers
{
  my $self = shift;
  my $ledgerCode;
  if (@_ == 1)
  {  $ledgerCode = @_[0];  }
  else
  {
    my %args = ( ledgerCode => "", @_ ) ;
    $ledgerCode = $args{ledgerCode};
  }
  my @Ledgers = ();
  my $where = "";

  if ($ledgerCode ne "")
  {
    $where .= " AND ";
    $where .= " ledger_tb.code = '$ledgerCode'";
  }

  my $sql = "SELECT ledger_tb.name, ledger_tb.code, ledger_tb.color";
  $sql .= ", ledger_tb.cash_or_accrual ";
  $sql .= ", business_type_tb.name, ledger_tb.business_type ";
  $sql .= ", offset1 ";
  $sql .= ", start1 ";
  $sql .= ", start2 ";
  $sql .= ", start3 ";
  $sql .= ", start4 ";
  $sql .= ", start5 ";
  $sql .= ", start6 ";
  $sql .= ", start7 ";
  $sql .= ", start8 ";
  $sql .= ", start9 ";
  $sql .= ", start10 ";
  $sql .= ", start11 ";
  $sql .= ", start12 ";
  $sql .= ", working_year(ledger_tb.code) ";
  $sql .= "FROM ledger_tb, business_type_tb ";
  $sql .= "WHERE ledger_tb.business_type = business_type_tb.num ".$where;
  $sql .= "ORDER BY ledger_tb.name, ledger_tb.code";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->prefixError($sql."<br>\n". $self->{dbHandle}->errorMessage);
    return undef;
  }
  while (my @info = $sth->fetchrow_array)
  {
    my $knownYears = $self->getKnownYears($info[1]);
    if ($self->error)
    {
      $self->prefixError(" Sent ledger '$ledgerCode' to function getKnownYears<br>\n");
      return undef;
    }

    my $cashOrAccrual = "";
    if ($info[3] =~ /^t|true|1$/)  #!!!!find DB independent way to do this!
    {  $cashOrAccrual = "cash";}
    elsif ($info[3] =~ /^f|false|0$/)
    { $cashOrAccrual = "accrual";}
    else
    { $cashOrAccrual = "error!";}

    my $start = [ $info[6], $info[7], $info[8], $info[9], $info[10], $info[11], $info[12], $info[13], $info[14], $info[15], $info[16], $info[17], $info[18]];
    my $ledgerObj = Portal::Accounting::Objects::Ledger->new( langObj => $self->{langObj});
    $ledgerObj->setup($self->getBusinessTypeNames);
    $ledgerObj->populate(name => $info[0], code => $info[1], color => $info[2],
                    basis => $cashOrAccrual,
                    businessType => $info[5],
                    knownYears => $knownYears,
                    start => $start,
                    fiscalYear => $info[19],
                    new => 0
                    );

    $Ledgers[++$#Ledgers] = $ledgerObj;
  }

  return @Ledgers;
}

=head2 Hash getBusinessTypeNames()

    This routine selects the business type names from the database and returns them as an hash
    of number = name values where the name is looked up from the Language Object so as to be
    Internationalized.

=cut
sub getBusinessTypeNames
{
  my $self = shift;
  my $index = 0;
  my @info;
  my $sth;
  my %types;

  $sth = $self->{dbHandle}->read(sql => "SELECT name, num FROM business_type_tb ORDER BY num");
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $types{$info[1]} = $self->{langObj}->map($info[0]);
  }

  return \%types;
}

=item integer select

  This routine requires that the code be set first. It will
  load records from the database into this object.

=cut
sub select
{
  my $self = shift;
  my $ledgerCode = $self->get('code');
  if ($ledgerCode eq '') { $self->error("<p>Invalid code in select.</p>\n"); }
  $self->selectKnownYears();
  $self->selectLastClose;
  my $sql = "SELECT name, code, color";
  $sql .= ", cash_or_accrual AS basis";
  $sql .= ", business_type ";
  $sql .= ", offset1 ";
  $sql .= ", start1 ";
  $sql .= ", start2 ";
  $sql .= ", start3 ";
  $sql .= ", start4 ";
  $sql .= ", start5 ";
  $sql .= ", start6 ";
  $sql .= ", start7 ";
  $sql .= ", start8 ";
  $sql .= ", start9 ";
  $sql .= ", start10 ";
  $sql .= ", start11 ";
  $sql .= ", start12 ";
  $sql .= ", working_year('$ledgerCode') AS working_year";
  $sql .= ", fy_mo AS month";
  $sql .= ", fy_day AS day";
  $sql .= ", (SELECT MIN(year) FROM known_years_tb WHERE ledger_code = '$ledgerCode') AS year ";
  $sql .= "FROM ledger_tb ";
  $sql .= "WHERE code = '$ledgerCode' ";
  $sql .= "ORDER BY name, code";
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->error($sql."<br>\n". $self->{dbHandle}->errorMessage);
    return undef;
  }
  if (my $info = $sth->fetchrow_hashref)
  {
    my $cashOrAccrual = "";
    if ($info->{basis} =~ /^t|true|1$/)  #!!!!find DB independent way to do this!
    {  $cashOrAccrual = "cash";}
    elsif ($info->{basis} =~ /^f|false|0$/)
    { $cashOrAccrual = "accrual";}
    else
    { $cashOrAccrual = "error!";}

    $self->populate(name => $info->{name},
      code => $info->{code},
      color => $info->{color},
      basis => $cashOrAccrual,
      businessType => $info->{business_type},
      workingYear => $info->{working_year},
      fiscalYear => $info->{year},
      year => $info->{year} + $info->{offset1},
      month => $info->{month},
      day => $info->{day},
      start2 => $info->{start2},
      start3 => $info->{start3} ,
      start4 => $info->{start4} ,
      start5 => $info->{start5} ,
      start6 => $info->{start6} ,
      start7 => $info->{start7} ,
      start8 => $info->{start8} ,
      start9 => $info->{start9} ,
      start10 => $info->{start10},
      start11 => $info->{start11},
      start12 => $info->{start12},
      new => '0',
    );
  }
  return 1;
}

=head2 %years selectKnownYears()

    Returns hash of fiscal years as keys with YYYY-MM-DD as data for each key

=cut
sub selectKnownYears
{
  my $self = shift;
  my $ledgerCode = $self->get('code');
  my $sth;
  my @info;

  if (length $ledgerCode == 0)
  {
    $self->prefixError("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  my $sql = "SELECT year, (SELECT month_end('$ledgerCode',12,year)) FROM known_years_tb WHERE ledger_code = '$ledgerCode' ORDER BY year";
  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->prefixError($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $self->{knownYears}->{$info[0]} = $info[1];
  }

  return $self->{knownYears};
}

=item date selectLastClose()

  This routine simply returns the last closeout date for this ledger.

=cut
sub selectLastClose
{
  my $self = shift;
  my $ledgerCode = $self->get('code');
  if ($ledgerCode eq "")
  {
    $self->error("<p>Error! no ledgerCode sent to getLastClose!</p>\n");
    return undef;
  }
  my $sth = $self->{dbHandle}->write(sql => "SET datestyle TO 'ISO'");
  $sth = $self->{dbHandle}->read(sql => "SELECT MAX(bal_date) FROM acct_bal_tb WHERE ledger_code = '$ledgerCode'");
  if ($self->{dbHandle}->error)
  {
    $self->prefixError("Error in getLastClose!<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  if (my @info = $sth->fetchrow_array)
  {
    $self->{lastClose} = $info[0];
  }
  return $self->{lastClose};
}

sub insert
{
  my $self = shift;
  my $sql = "SELECT name FROM ledger_tb WHERE code = '".$self->get('code')."'";
  my $sth = $self->dbHandle->read($sql);
  if (my @info = $sth->fetchrow_array)
  {
#    $self->invalid('code');
    $self->error("<p>".$self->langObj->map('Code')." ".$self->langObj->map('already used').":".$self->get('code')."</p>\n");
    return undef;
  }
  $sql = "SELECT name FROM ledger_tb WHERE name = '".$self->get('name')."'";
  $sth = $self->dbHandle->read($sql);
  if (my @info = $sth->fetchrow_array)
  {
#    $self->invalid('name');
    $self->error("<p>".$self->langObj->map('Name')." ".$self->langObj->map('already used').":".$self->get('name')."</p>\n");
    return undef;
  }
  my $basis = 'false';
  if ($self->get('basis') eq 'cash') { $basis = 'true'; }
  $self->{INSERT}[0] = "INSERT INTO $self->{table} (cash_or_accrual, ";
  $self->{INSERT}[0] .= " start1, offset1, offset2, offset3, offset4, offset5, offset6, offset7, offset8, offset9, offset10, offset11, offset12, ";
  $self->{INSERT}[0] .= "##COLUMNS##) ";
  $self->{INSERT}[0] .= "VALUES ($basis, '";
  $self->{INSERT}[0] .=   sprintf("%02d",$self->get('month'))."-".sprintf("%02d",$self->get('day'))."'";
  for (my $i = 1; $i <= 12; $i++)
  {
    $self->{INSERT}[0] .= ", '".$self->offset($i)."'";
  }
  $self->{INSERT}[0] .= ", ##VALUES##)";
  $self->{INSERT}[1] = "INSERT INTO known_years_tb (year, closed, ledger_code) VALUES ('#FIELD=fiscalYear#', false, '#FIELD=code#')";
  return $self->SUPER::insert(@_);
}

sub update
{
  my $self = shift;
  my $sql = "SELECT name FROM ledger_tb WHERE name = '".$self->get('name')."' AND code <> '".$self->get('code')."'";
  my $sth = $self->dbHandle->read($sql);
  if (my @info = $sth->fetchrow_array)
  {
#    $self->invalid('name');
    $self->error("<p>".$self->langObj->map('Name')." ".$self->langObj->map('already used').":".$self->get('name')."</p>\n");
    return undef;
  }
  my $basis = 'false';
  if ($self->get('basis') eq 'cash') { $basis = 'true'; }
  $self->{UPDATE}[0] = "UPDATE $self->{table} SET cash_or_accrual = $basis, ##UPDATE## WHERE code = '#FIELD=$self->{id}#'";
  return $self->SUPER::update(@_);
}

sub lite
{
  my $self = shift;
  my %foo;
  $foo{name} = $self->get('name');
  $foo{code} = $self->get('code');
  $foo{color} = $self->get('color');
  return \%foo;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Base(3)

=cut
