# QIFImport.pm - The Object Class that provides a QIFImport Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: QIFImport.pm,v 1.10 2005/07/12 22:28:05 pcxuser Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.
=head1 NAME

QIFImport - Object used to build a QIFImport Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::QIFImport;
  my $obj = Portal::Accounting::QIFImport->new;
  if ($obj->didErrorOccur())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

QIFImport is a QIFImport class.  The object recognizes specific states
that change between steps of the QIF import process.  The object
generates the display for entering a QIF file to import; viewing the
previous import's log file (if any); mapping classes, categories,
securities, and accounts to pre/non-existing accounting entities.

=head1 Exported FUNCTIONS

=cut

package Portal::Accounting::QIFImport;

use strict;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);
use Portal::Data::Config;
use Portal::Accounting::DisplayItems;
use Portal::Accounting::DB::Main;
use Portal::Accounting::Objects::Account;
use Portal::Accounting::Objects::SubAccount;
use Portal::Accounting::Objects::StockObject;
use Portal::Accounting::DB::Security;
use Portal::Accounting::DB::JournalEntry;
require Exporter;

@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '0.01';

# global variables.
# modify to represent the commands this state provides.
use vars '%commands';
%commands = ( display => "Upload File", review_log => "Review Log", parse_class_list => "Parse File for Class List",
                 parse_category_list => "Parse File for Category List", parse_account_list => "Parse File for Account List",
                 parse_security_list => "Parse File for Security List",
                 build_lists_from_account_transactions => "Build Lists from Account Transactions Lists",
                 build_lists_from_investment_transactions => "Build Lists from Investment Transactions Lists",
                 build_lists_from_prices => "Build Lists from Price Data",
                 ask_entities => "Ask Entities",
                 guess_securities => "Guess Securities", map_securities => "Map Securities",
                 guess_classes => "Guess Classes", map_classes => "Map Classes",
                 guess_categories => "Guess Categories", map_categories => "Map Categories",
                 guess_accounts => "Guess Accounts", map_accounts => "Map Accounts",
                 select_import_accounts => "Select Import Accounts",
                 import_account_transactions => "Import Account Transactions",
                 import_buy_investment_transactions => "Import Buy Investment Transactions",
                 import_investment_transactions => "Import Investment Transactions",
                 import_processed_investment_transaction => "Import Processed Investment Transaction",
                 import_prices => "Import Stock Prices",
                 import_complete => "Import Complete", );

=head1 scalar new(langObj, input, cookieObj, dbHandle, logObj, sessionObj,
             userObj, companyObj, browserType, browserVersion,
             browserMaker)
    Creates a new instance of the Portal::Accounting::QIFImport
    object.

=cut

sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my %args = ( langObj => undef, input => {}, cookieObj => undef, dbHandle => undef, logObj => undef, sessionObj => undef, userObj => undef, companyObj => undef,
               browserType => "HTML", browserVersion => "4.0", browserMaker => "", @_ );
  my $errStr = "Portal::Accounting::QIFImport->new() - Error!<br>\n";

  $self->{langObj} = $args{langObj};
  $self->{input} = $args{input};
  $self->{cookieObj} = $args{cookieObj};
  $self->{dbHandle} = $args{dbHandle};
  $self->{logObj} = $args{logObj};
  $self->{sessionObj} = $args{sessionObj};
  $self->{userObj} = $args{userObj};
  $self->{companyObj} = $args{companyObj};
  $self->{browserType} = $args{browserType};
  $self->{browserVersion} = $args{browserVersion};
  $self->{browserMaker} = $args{browserMaker};
  $self->{inFakePortal} = (exists $args{fakePortal} ? 1 : 0);
  $self->{portalConfig} = Portal::Data::Config->new;

  $self->{QIFInputBuffer} = [];
  $self->{currentLine} = 0;
  $self->{processedClassRecords} = 0;
  $self->{processedCategoryRecords} = 0;
  $self->{processedAccountListRecords} = 0;
  $self->{processedAccountRecords} = 0;
  $self->{processedAccountTransRecords} = 0;
  $self->{processedAccountInvstRecords} = 0;
  $self->{processedSecurityRecords} = 0;
  $self->{importedBuyInvestmentTransactions} = 0;
  $self->{importedAccountTransactions} = 0;
  $self->{importedInvestmentTransactions} = 0;
  $self->{importedPriceTransactions} = 0;
  $self->{checkedAccountTransactions} = 0;
  $self->{checkedInvestmentTransactions} = 0;
  $self->{checkedPriceTransactions} = 0;

  $self->{limitRange} = 10;

  $self->{classListBatchSize} = 100;
  $self->{categoryListBatchSize} = 100;
  $self->{accountListBatchSize} = 100;
  $self->{securityListBatchSize} = 100;

  $self->{accountTransactionBatchSize} = 100;
  $self->{investmentTransactionBatchSize} = 10;

  $self->{buildListBatchSize} = 200;

  $self->{pricesBatchSize} = 250;

  # do validation
  $self->{error} = !$self->isValid;
  if ($self->{error})
  {
    $self->{errorString} = $errStr . $self->{errorString};
  }

  return $self;
}

=head1 bool isValid(void)
    	Returns 0 or 1 to indicate if the object is valid.  The error will
    be available via errorMessage().
=cut
sub isValid
{
  my $self = shift;
  my $error = 0;
  my $errorString = "";
  my $errStr = "Portal::Accounting::QIFImport->isValid() - Error!<br>\n";

  # do validation code here.
  if (not defined $self->{langObj})
  {
    $error = 1;
    $errorString .= "langObj must be defined!<br>\n";
  }
  elsif (!$self->{langObj}->isValid)
  {
    $error = 1;
    $errorString .= $self->{langObj}->errorMessage;
  }
  if (!$self->{inFakePortal})
  {
    if (not defined $self->{userObj})
    {
      $error = 1;
      $errorString .= "userObj must be defined!<br>\n";
    }
    elsif (!$self->{userObj}->isValid)
    {
      $error = 1;
      $errorString .= $self->{userObj}->errorMessage;
    }
    if (not defined $self->{companyObj})
    {
      $error = 1;
      $errorString .= "companyObj must be defined!<br>\n";
    }
    elsif (!$self->{companyObj}->isValid)
    {
      $error = 1;
      $errorString .= $self->{companyObj}->errorMessage;
    }
  }
  if (not defined $self->{logObj})
  {
    $error = 1;
    $errorString .= "logObj must be defined!<br>\n";
  }
  elsif (!$self->{logObj}->isValid)
  {
    $error = 1;
    $errorString .= $self->{logObj}->errorMessage;
  }
  if (not defined $self->{cookieObj})
  {
    $error = 1;
    $errorString .= "cookieObj must be defined!<br>\n";
  }
  if (not defined $self->{dbHandle})
  {
    $error = 1;
    $errorString .= "dbHandle must be defined!<br>\n";
  }
  if (not defined $self->{sessionObj})
  {
    $error = 1;
    $errorString .= "sessionObj must be defined!<br>\n";
  }
  if ($self->{browserType} !~ /^(HTML|HDML|WML)$/)
  {
    $error = 1;
    $errorString .= "browserType = '$self->{browserType}' is invalid!<br>\n";
  }
  if (length $self->{browserVersion} == 0)
  {
    $error = 1;
    $errorString .= "browserType = '$self->{browserType}' is invalid!<br>\n";
  }

  $self->{error} = $error;
  $self->{errorString} = $errStr if $error;
  $self->{errorString} .= $errorString;

  return !$error;
}

sub DESTROY
{
  my $self = shift;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;	# strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

=head1 void setError(errorString)
    Sets error = 1 and $self->{errorString} = errorString.

=cut
sub setError
{
  my $self = shift;
  my %args = ( errorString => "", @_ );
  my $errorString = $args{errorString};

  $self->{errorString} = $errorString;
  $self->{error} = 1;
}

=head1 scalar didErrorOccur(void)
    Returns the value of error.

=cut
sub didErrorOccur
{
  my $self = shift;

  return $self->{error};
}
=head1 scalar errorMessage(void)
    Returns the value of errorString.

=cut
sub errorMessage
{
  my $self = shift;

  return $self->{errorString};
}

=head1 NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

  All other methods are private to the object and are not
  usable outside the internals of the object.


HTMLObject run(command)
	returns: The HTMLObject document to display or undef if an error.
=cut
sub run
{
  my $self = shift;
  my %args = ( command => "", @_ );
  my $command = $args{command};
  my $doc;
  my $errStr = "Portal::Accounting::QIFImport->run() - Error!<br>\n";
  my $input = $self->{input};
  my $display = Portal::Accounting::DisplayItems->new();
  my $ledger = "";
  my $siteDir = $self->{portalConfig}->{siteDir};
  my $site_dir = "$siteDir/js/Accounting/accounting";

  my $app = "Accounting";
  my $state = "QIFImport";
  my $staticInfo = "";
  my $currentAccount = "";
  my $line = "";

  if (!exists $commands{$command})
  {
    $self->{error} = 1;
    $self->{errorString} = $errStr . sprintf($self->langObj->phraseObj->phrases->{commandNotKnown}, $command);
    return undef;
  }

  if (defined $input->{'ledger'})
  {
    $ledger = $input->{'ledger'};
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} =  $errStr . "Error.  Could not get a handle to the ledger in $state.";
    return undef;
  }

  $staticInfo = "<input type=\"hidden\" name=\"app\" value=\"$app\">\n";
  $staticInfo .= "<input type=\"hidden\" name=\"state\" value=\"$state\">\n";
  $staticInfo .= "<input type=\"hidden\" name=\"ledger\" value=\"$ledger\">\n";
  if (exists $input->{'qif_file_name'} && $input->{'qif_file_name'} ne "" && $command ne "display" && $command ne "import_complete")
  {
    $staticInfo .= "<input type=hidden name=\"qif_file_name\" value=\"" . $input->{'qif_file_name'} . "\">\n";
  }

  my $ledgerCode = $self->{sessionObj}->{store}->{ledgers}->{$ledger}->{code};
  $self->{ledgerCode} = $ledgerCode;
  $self->{ledger} = $ledger;
  my $dbObject = Portal::Accounting::DB::Main->new(dbHandle => $self->{dbHandle});
  if ($dbObject->didErrorOccur)
  {
    $self->{error} = 1;
    $self->{errorString} = $errStr . "<br>" . $dbObject->errorMessage;
    return undef;
  }

  $doc = HTMLObject::Normal->new;
  $doc->setLinkDecorations();
  # set the limit range for the object from the environment
  if (exists $input->{'limit_range'} && $input->{'limit_range'} =~ /^[0-9]+$/)
  {
    $self->{limitRange} = $input->{'limit_range'};
  }
  if ($command eq "display")
  {
    if (exists $input->{_FileUploadInfo_}->{serverFileNames}->{'qif_file_name'})
    {
      $input->{_FileUploadInfo_}->{serverFileNames}->{'qif_file_name'} = "";
    }
    $doc->setTitle("QIF File Selection");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Selection</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_file_import\" action=\"$siteDir/cgi-bin/index.cgi\" enctype=\"multipart/form-data\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"parse_class_list\">\n");
    $doc->print("<table border=\"1\" cellspacing=\"0\" cellpadding=\"5\">\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right>Select <b>.QIF</b> file to import:&nbsp;</td>\n");
    $doc->print("    <td align=left><input type=\"file\" name=\"qif_file_name\" size=\"50\"></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right>Number of Entries per Screen:&nbsp;</td>\n");
    $doc->print("    <td align=left><select name=\"limit_range\">\n");
    $doc->print("      <option value=\"10\" selected>10\n");
    $doc->print("      <option value=\"20\">20\n");
    $doc->print("      <option value=\"30\">30\n");
    $doc->print("      <option value=\"40\">40\n");
    $doc->print("      <option value=\"50\">50\n");
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td colspan=2 align=right><input type=\"submit\" value=\"Send file ...\" onClick=\"if (document.qif_file_import.qif_file_name.value == '') { alert('You must enter a filename to upload!'); document.qif_file_import.qif_file_name.focus(); return false; };\">&nbsp;<input type=\"reset\"></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("</table>\n");
    $doc->print("</form>\n");
    $doc->print("<br>\n");
    $doc->print("<form method=\"post\" name=\"qif_log_form\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"review_log\">\n");
    $doc->print("<a href=\"$siteDir/cgi-bin/index.cgi\" onClick=\"javascript: document.qif_log_form.submit(); return false;\">Review previous import log entries.</a><br><br>\n");
    $doc->print("</form>\n");
    $doc->print("</center>\n");
    $doc->print("<form method=post action=\"$site_dir/cgi-bin/prg_config.cgi\" name=\"prg_config_form\">\n");
    $doc->print("<input type=hidden name=\"ledger\" value=\"$ledger\">\n");
    $doc->print("<a href=\"$siteDir/js/Accounting/accounting/cgi-bin/prg_config.cgi\" onClick=\"javascript:document.prg_config_form.submit(); return false;\">Configuration</a>\n");
    $doc->print("</form>\n");
  }
  elsif ($command eq "review_log")
  {
    my $logEntries = 0;
    my $sth = $self->{dbHandle}->read(sql => "SELECT message FROM qif_log_tb");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving QIF log entries!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }
    $doc->setTitle("QIF Import Log Entries");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Review Log Entries</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_review_log\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"display\">\n");
    $doc->print("<input type=\"submit\" value=\"Continue...\">\n");
    $doc->print("<br><br>\n");
    $doc->print("<table border=\"1\" cellspacing=\"0\" cellpadding=\"2\">\n");
    $doc->print("  <tr>\n");
    $doc->print("    <th align=center>QIF Log Entries</th>\n");
    $doc->print("  </tr>\n");
    while (my @info = $sth->fetchrow_array)
    {
      $doc->print("  <tr>\n");
      $doc->print("    <td>" . $info[0] . "</td>\n");
      $doc->print("  </tr>\n");

      $logEntries++;
    }
    if ($logEntries == 0)
    {
      $doc->print("  <tr>\n");
      $doc->print("    <td align=center><i>No log entries were returned from the previous QIF import.</i></td>\n");
      $doc->print("  </tr>\n");
    }
    $doc->print("</table>\n");
    $doc->print("<br>\n");
    $doc->print("<input type=\"submit\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "parse_class_list")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{_FileUploadInfo_}->{serverFileNames}->{'qif_file_name'}) || ($input->{_FileUploadInfo_}->{serverFileNames}->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{_FileUploadInfo_}->{serverFileNames}->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{_FileUploadInfo_}->{serverFileNames}->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # clear out any entries in the QIF database tables
    $self->{dbHandle}->write(sql => "DELETE FROM qif_class_tb");
    $self->{dbHandle}->write(sql => "DELETE FROM qif_category_tb");
    $self->{dbHandle}->write(sql => "DELETE FROM qif_account_tb");
    $self->{dbHandle}->write(sql => "DELETE FROM qif_security_tb");
    $self->{dbHandle}->write(sql => "DELETE FROM qif_log_tb");
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error occurred clearing QIF import tables!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into temporary storage tables in the database

    # reset any variables from a previous class list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'processed_class_records'} && $input->{'processed_class_records'} =~ /\d+/)
    {
      $self->{processedClassRecords} = $input->{'processed_class_records'};
    }

    my $line = "";
    my $parseResult = -1;

    # recognize the single class list for the import
    if ($self->{currentLine} == 0)
    {
      while ((defined $line) && ($parseResult == -1))
      {
        $line = $self->getNextLine(fileHandle => $fileHandle);
        if (defined $line)
        {
          if ($line =~ /!Type:Class/)
          {
            $parseResult = $self->parseClassList(fileHandle => $fileHandle);
            if (not $parseResult)
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Class List! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
          }
        }
      }
    }
    else
    {
      $parseResult = $self->parseClassList(fileHandle => $fileHandle);
      if (not $parseResult)
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Class List! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
        unlink $QIFFileName;
        return undef;
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Class List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Class List</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if ($parseResult ne "2")
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_category_list\">\n");
    }
    elsif ($parseResult eq "2")
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_class_list\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"processed_class_records\" value=\"" . $self->{processedClassRecords} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if ($parseResult ne "2")
    {
      $doc->print("Total");
    }
    elsif ($parseResult eq "2")
    {
      $doc->print("Current");
    }
    $doc->print(" classes processed: " . $self->{processedClassRecords} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "parse_category_list")
  {
    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($input->{'qif_file_name'});
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into temporary storage tables in the database

    # reset any variables from a previous category list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'processed_category_records'} && $input->{'processed_category_records'} =~ /\d+/)
    {
      $self->{processedCategoryRecords} = $input->{'processed_category_records'};
    }

    my $line = "";
    my $parseResult = -1;

    # recognize the single category list for the import
    if ($self->{currentLine} == 0)
    {
      while ((defined $line) && ($parseResult == -1))
      {
        $line = $self->getNextLine(fileHandle => $fileHandle);
        if (defined $line)
        {
          if ($line =~ /!Type:Cat/)
          {
            $parseResult = $self->parseCategoryList(fileHandle => $fileHandle);
            if (not $parseResult)
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Category List! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
          }
        }
      }
    }
    else
    {
      $parseResult = $self->parseCategoryList(fileHandle => $fileHandle);
      if (not $parseResult)
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Category List! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
        unlink $QIFFileName;
        return undef;
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Category List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Category List</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if ($parseResult ne "2")
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_account_list\">\n");
    }
    elsif ($parseResult eq "2")
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_category_list\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"processed_category_records\" value=\"" . $self->{processedCategoryRecords} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if ($parseResult ne "2")
    {
      $doc->print("Total");
    }
    elsif ($parseResult eq "2")
    {
      $doc->print("Current");
    }
    $doc->print(" categories processed: " . $self->{processedCategoryRecords} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "parse_account_list")
  {
    my $tempCounter = 0;
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into temporary storage tables in the database

    # reset any variables from a previous account list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'processed_account_list_records'} && $input->{'processed_account_list_records'} =~ /\d+/)
    {
      $self->{processedAccountListRecords} = $input->{'processed_account_list_records'};
    }
    if (defined $input->{'processed_account_records'} && $input->{'processed_account_records'} =~ /\d+/)
    {
      $self->{processedAccountRecords} = $input->{'processed_account_records'};
    }

    my $line = "";
    my $parseResult = -1;
    my $optionAutoSwitch = 0; # used to keep track of the number of times this export header has been encountered
    if (defined $input->{'option_auto_switch'} && $input->{'option_auto_switch'} =~ /\d+/)
    {
      $optionAutoSwitch = $input->{'option_auto_switch'};
    }
    my $clearAutoSwitch = 0; # used to keep track of the number of times this export header has been encountered
    if (defined $input->{'clear_auto_switch'} && $input->{'clear_auto_switch'} =~ /\d+/)
    {
      $clearAutoSwitch = $input->{'clear_auto_switch'};
    }
    my $currentAccount = "unknown_account"; # used to keep track of the current account to use for account transactions
    if (defined $input->{'current_account'} && $input->{'current_account'} ne "")
    {
      $currentAccount = $input->{'current_account'};
    }
    my $accountTypeFound = 0;
    if (defined $input->{'account_type_found'} && $input->{'account_type_found'} =~ /\d+/)
    {
      $accountTypeFound = $input->{'account_type_found'};
    }
    my $accountType = "";
    if (defined $input->{'account_type'} && $input->{'account_type'} ne "")
    {
      $accountType = $input->{'account_type'};
    }

    # recognize the single account list for the import; also determine if any of the accounts entries used before
    # transaction lists have equivalent entries in the account list; if not, add them to the list
    while ((defined $line) && ($parseResult == -1) && ($tempCounter < $self->{accountListBatchSize}))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if ($line =~ /!Option:AutoSwitch/)
        {
          $optionAutoSwitch++;
        }
        elsif ($line =~ /!Clear:AutoSwitch/)
        {
          $clearAutoSwitch++;
        }
        elsif ($line =~ /!Account/ || $input->{'inside_account_list'} eq "1")
        {
          if (($optionAutoSwitch > 0 && $clearAutoSwitch == 0) || ($input->{'inside_account_list'} eq "1"))
          {
            if ($input->{'inside_account_list'} eq "1")
            {
              $self->putBackLine($line);
              $input->{'inside_account_list'} = "0";
            }
            # inside first accounts section
            $parseResult = $self->parseAccountsList(fileHandle => $fileHandle);
            if (not $parseResult)
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Account List! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
          }
          elsif ($optionAutoSwitch > 0 && $clearAutoSwitch > 0)
          {
            # individual accounts with transaction lists after their definition
            $currentAccount = $self->parseAccountRecord(fileHandle => $fileHandle);
            if ((not defined $currentAccount) || $currentAccount eq "")
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Account record! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
            $tempCounter++;
          }
          else
          {
            # invalid account section found in file
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error: Invalid QIF format found!<br>Account section/record found outside of correct list structure.  line = " . $self->{currentLine} . "<br>\n";
            unlink $QIFFileName;
            return undef;
          }
        }
        elsif ($line =~ /!Type:(Bank|Cash|CCard|Oth.A|Oth.L|Bill|Tax|Invoice|Invst)/)
        {
          $accountTypeFound++;
          $accountType = $1;
        }
      }
    }

    if ((not defined $line) && $currentAccount eq "unknown_account")
    {
      if ($accountTypeFound == 1)
      {
        # insert an entry into the QIF database table for the unknown_account account
        $self->{dbHandle}->write(sql => "INSERT INTO qif_account_tb (name, type, sales_tax, description, credit_card_limit, balance, used) VALUES ('$currentAccount', '$accountType', '', 'Unknown Account', '', '', '1')");
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error inserting '$currentAccount' into database!<br>\n" . $self->{dbHandle}->errorMessage;
          unlink $QIFFileName;
          return undef;
        }
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      unlink $QIFFileName;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Account List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Account List</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_security_list\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_account_list\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"current_account\" value=\"$currentAccount\">\n");
      $doc->print("<input type=hidden name=\"option_auto_switch\" value=\"$optionAutoSwitch\">\n");
      $doc->print("<input type=hidden name=\"clear_auto_switch\" value=\"$clearAutoSwitch\">\n");
      $doc->print("<input type=hidden name=\"account_type_found\" value=\"$accountTypeFound\">\n");
      $doc->print("<input type=hidden name=\"account_type\" value=\"$accountType\">\n");
      $doc->print("<input type=hidden name=\"processed_account_list_records\" value=\"" . $self->{processedAccountListRecords} . "\">\n");
      if ($parseResult eq "2")
      {
        $doc->print("<input type=hidden name=\"inside_account_list\" value=\"1\">\n");
      }
      $doc->print("<input type=hidden name=\"processed_account_records\" value=\"" . $self->{processedAccountRecords} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if ($parseResult ne "2")
    {
      $doc->print("Total");
    }
    elsif ($parseResult eq "2")
    {
      $doc->print("Current");
    }
    $doc->print(" account list entries processed: " . $self->{processedAccountListRecords} . "<br>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" account entries processed: " . $self->{processedAccountRecords} . "<br>\n");
    if ((not defined $line) && $accountTypeFound == 1)
    {
      $doc->print("<b>Added \"unknown_account\" entry</b><br>\n");
    }
    $doc->print("</center>\n");
  }
  elsif ($command eq "parse_security_list")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into temporary storage tables in the database

    # reset any variables from a previous security list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'processed_security_records'} && $input->{'processed_security_records'} =~ /\d+/)
    {
      $self->{processedSecurityRecords} = $input->{'processed_security_records'};
    }

    my $line = "";
    my $tempCounter = 0;
    my $parseResult = -1;

    # recognize the multiple security list entries for the import
    while ((defined $line) && ($tempCounter < $self->{securityListBatchSize}))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if ($line =~ /!Type:Security/)
        {
          $parseResult = $self->parseSecurity(fileHandle => $fileHandle);
          if (not $parseResult)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Security entry! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
            unlink $QIFFileName;
            return undef;
          }
          $tempCounter++;
        }
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Security List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Security List</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"build_lists_from_account_transactions\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"parse_security_list\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"processed_security_records\" value=\"" . $self->{processedSecurityRecords} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" securities processed: " . $self->{processedSecurityRecords} . "<br>\n");
    $doc->print("</center>\n");
  }
  # find problem areas that need to be dealt with (accounts missing, class translations, category translations,
  # transfer doubling, securities mapping, etc.)
  elsif ($command eq "build_lists_from_account_transactions")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file recognizing the lists of account transactions looking for L and S fields to determine if
    # categories/classes/accounts are used and whether they exist in their respective temporary database tables;
    # update the used flag if the entity exists or insert the data into temporary storage tables in the database if it does
    # not exist already and set its used flag to true

    # reset any variables from a previous parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'checked_account_transactions'} && $input->{'checked_account_transactions'} =~ /\d+/)
    {
      $self->{checkedAccountTransactions} = $input->{'checked_account_transactions'};
    }

    my $line = "";
    my $tempCounter = 0;
    my $parseResult = -1;

    # recognize the start of an account list and parse the records to find the necessary information
    while ((defined $line) && ($tempCounter < $self->{buildListBatchSize}))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if (($line =~ /!Type:Bank/) || ($line =~ /!Type:Cash/) || ($line =~ /!Type:CCard/) || ($line =~ /!Type:Oth.A/) ||
            ($line =~ /!Type:Oth.L/) || ($line =~ /!Type:Invoice/) || ($line =~ /!Type:Tax/) || ($line =~ /!Type:Bill/) ||
            ($input->{'inside_account_transaction_list'} eq "1"))
        {
          if ($input->{'inside_account_transaction_list'} eq "1")
          {
            $self->putBackLine($line);
            $input->{'inside_account_transaction_list'} = 0;
          }
          else
          {
            $line = ""; # clear the line variable
          }
          while (defined $line && $line !~ /^\!/ && $tempCounter < $self->{buildListBatchSize})
          {
            $line = $self->getNextLine(fileHandle => $fileHandle);
            if (defined $line)
            {
              if ($line =~ /^[LS](.+?)\r?\n?$/)
              {
                my $tempValue = $1;
                my $category;
                my $category_class;
                my $transfer_account;
                my $transfer_class;
                # determine which of the following cases the L or S entry is
                # category/class or [transfer]/class
                $tempValue =~ /^((\[(.+?)\](\/(.+?))?)|((.+?)(\/(.+?))?))$/;
                $transfer_account = $3;
                $transfer_class = $5;
                $category = $7;
                $category_class = $9;

                $transfer_account =~ s/'/\\'/g;
                $transfer_class =~ s/'/\\'/g;
                $category =~ s/'/\\'/g;
                $category_class =~ s/'/\\'/g;

                if ($category ne "" || $transfer_account ne "")
                {
                  # update the used flag if the entity already exists; otherwise, add a new entity to the database table
                  if ($transfer_account ne "")
                  {
                   my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_account_tb SET used = '1' WHERE name = ?", plug => [ $transfer_account ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF account table for account = '$transfer_account'!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                    if ($updateResult == 0)
                    {
                      # insert "unlisted" account into database QIF list table
                      $self->{dbHandle}->write(sql => "INSERT INTO qif_account_tb (name, used) VALUES (?, '1')", plug => [ $transfer_account ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" account ($transfer_account) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                    }

                    if ($transfer_class ne "")
                    {
                      $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_class_tb SET used = '1' WHERE name = ?", plug => [ $transfer_class ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF class table for class = '$transfer_class'!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                      if ($updateResult == 0)
                      {
                        # insert "unlisted" class into database QIF list table
                        $self->{dbHandle}->write(sql => "INSERT INTO qif_class_tb (name, used) VALUES (?, '1')", plug => [ $transfer_class ]);
                        if ($self->{dbHandle}->didErrorOccur())
                        {
                          $self->{error} = 1;
                          $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" class ($transfer_class) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                          return undef;
                        }
                      }
                    }
                  }
                  elsif ($category ne "")
                  {
                    my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_category_tb SET used = '1' WHERE name = ?", plug => [ $category ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF category table for category = '$category'!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                    if ($updateResult == 0)
                    {
                      # insert "unlisted" category into database QIF list table
                      $self->{dbHandle}->write(sql => "INSERT INTO qif_category_tb (name, used) VALUES (?, '1')", plug => [ $category ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" category ($category) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                    }

                    if ($category_class ne "")
                    {
                      $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_class_tb SET used = '1' WHERE name = ?", plug => [ $category_class ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF class table for class = '$category_class'!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                      if ($updateResult == 0)
                      {
                        # insert "unlisted" class into database QIF list table
                        $self->{dbHandle}->write(sql => "INSERT INTO qif_class_tb (name, used) VALUES (?, '1')", plug => [ $category_class ]);
                        if ($self->{dbHandle}->didErrorOccur())
                        {
                          $self->{error} = 1;
                          $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" class ($category_class) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                          return undef;
                        }
                      }
                    }
                  }

                  # increment the checked/processed count
                  $tempCounter++;
                  $self->{checkedAccountTransactions}++;
                  if ($tempCounter >= $self->{buildListBatchSize})
                  {
                    $parseResult = 2;
                  }
                }
              }
            }
          }
          if ($line =~ /^\!/)
          {
            $self->putBackLine($line);
          }
        }
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Building List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Building List of Used Entities from Account Transactions</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"build_lists_from_investment_transactions\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"build_lists_from_account_transactions\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"checked_account_transactions\" value=\"" . $self->{checkedAccountTransactions} . "\">\n");
      if ($parseResult eq "2")
      {
        $doc->print("<input type=hidden name=\"inside_account_transaction_list\" value=\"1\">\n");
      }
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" account transactions checked: " . $self->{checkedAccountTransactions} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "build_lists_from_investment_transactions")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file recognizing the lists of investment transactions looking for L fields to determine if
    # categories/classes/accounts are used and whether they exist in their respective temporary database tables;
    # update the used flag if the entity exists or insert the data into temporary storage tables in the database if it does
    # not exist already and set its used flag to true

    # reset any variables from a previous parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'checked_investment_transactions'} && $input->{'checked_investment_transactions'} =~ /\d+/)
    {
      $self->{checkedInvestmentTransactions} = $input->{'checked_investment_transactions'};
    }

    my $line = "";
    my $tempCounter = 0;
    my $parseResult = -1;

    # recognize the start of an account list and parse the records to find the necessary information
    while ((defined $line) && ($tempCounter < $self->{buildListBatchSize}))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if (($line =~ /!Type:Invst/) || ($input->{'inside_investment_transaction_list'} eq "1"))
        {
          if ($input->{'inside_investment_transaction_list'} eq "1")
          {
            $self->putBackLine($line);
            $input->{'inside_investment_transaction_list'} = 0;
          }
          else
          {
            $line = ""; # clear the line variable
          }
          while (defined $line && $line !~ /^\!/ && $tempCounter < $self->{buildListBatchSize})
          {
            $line = $self->getNextLine(fileHandle => $fileHandle);
            if (defined $line)
            {
              my $security = "";
              my $category = "";
              my $category_class = "";
              my $transfer_account = "";
              my $transfer_class = "";

              if ($line =~ /^Y(.+?)\r?\n?$/)
              {
                $security = $1;
                $security =~ s/'/\\'/g;
              }
              elsif ($line =~ /^L(.+?)\r?\n?$/)
              {
                my $tempValue = $1;
                # determine which of the following cases the L entry is
                # category/class or [transfer]/class or category/class|[transfer]/class
                $tempValue =~ /^(((.+?)(\/(.+?))?\|\[(.+?)\](\/(.+?))?)|(\[(.+?)\](\/(.+?))?)|((.+?)(\/(.+?))?))$/;
                my $temp1 = $3;
                my $temp2 = $5;
                my $temp3 = $6;
                my $temp4 = $8;
                my $temp5 = $10;
                my $temp6 = $12;
                my $temp7 = $14;
                my $temp8 = $16;

                if ($temp1 ne "" && $temp3 ne "")
                {
                  $category = $temp1;
                  $category_class = $temp2;
                  $transfer_account = $temp3;
                  $transfer_class = $temp4;
                }
                elsif ($temp5 ne "")
                {
                  $transfer_account = $temp5;
                  $transfer_class = $temp6;
                }
                elsif ($temp7 ne "")
                {
                  $category = $temp7;
                  $category_class = $temp8;
                }

                $transfer_account =~ s/'/\\'/g;
                $transfer_class =~ s/'/\\'/g;
                $category =~ s/'/\\'/g;
                $category_class =~ s/'/\\'/g;
              }

              if ($security ne "" || $category ne "" || $transfer_account ne "")
              {
                # update the used flag if the entity already exists; otherwise, add a new entity to the database table
                if ($security ne "")
                {
                  my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_security_tb SET used = '1' WHERE name = ?", plug => [ $security ]);
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF security table for security = '$security'!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  if ($updateResult == 0)
                  {
                    # insert "unlisted" security into database QIF list table
                    $self->{dbHandle}->write(sql => "INSERT INTO qif_security_tb (name, used) VALUES (?, '1')", plug => [ $security ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" security ($security) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                  }
                }
                if ($transfer_account ne "")
                {
                  my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_account_tb SET used = '1' WHERE name = ?", plug => [ $transfer_account ]);
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF account table for account = '$transfer_account'!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  if ($updateResult == 0)
                  {
                    # insert "unlisted" account into database QIF list table
                    $self->{dbHandle}->write(sql => "INSERT INTO qif_account_tb (name, used) VALUES (?, '1')", plug => [ $transfer_account ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" account ($transfer_account) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                  }

                  if ($transfer_class ne "")
                  {
                    $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_class_tb SET used = '1' WHERE name = ?", plug => [ $transfer_class ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF class table for class = '$transfer_class'!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                    if ($updateResult == 0)
                    {
                      # insert "unlisted" class into database QIF list table
                      $self->{dbHandle}->write(sql => "INSERT INTO qif_class_tb (name, used) VALUES (?, '1')", plug => [ $transfer_class ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" class ($transfer_class) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                    }
                  }
                }
                if ($category ne "")
                {
                  my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_category_tb SET used = '1' WHERE name = ?", plug => [ $category ]);
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF category table for category = '$category'!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  if ($updateResult == 0)
                  {
                    # insert "unlisted" category into database QIF list table
                    $self->{dbHandle}->write(sql => "INSERT INTO qif_category_tb (name, used) VALUES (?, '1')", plug => [ $category ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" category ($category) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                  }

                  if ($category_class ne "")
                  {
                    $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_class_tb SET used = '1' WHERE name = ?", plug => [ $category_class ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF class table for class = '$category_class'!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                    if ($updateResult == 0)
                    {
                      # insert "unlisted" class into database QIF list table
                      $self->{dbHandle}->write(sql => "INSERT INTO qif_class_tb (name, used) VALUES (?, '1')", plug => [ $category_class ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" class ($category_class) into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                    }
                  }
                }

                # increment the checked/processed count
                $tempCounter++;
                $self->{checkedInvestmentTransactions}++;
                if ($tempCounter >= $self->{buildListBatchSize})
                {
                  $parseResult = 2;
                }
              }
            }
          }
          if ($line =~ /^\!/)
          {
            $self->putBackLine($line);
          }
        }
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Building List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Building List of Used Entities from Investment Transactions</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"build_lists_from_prices\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"build_lists_from_investment_transactions\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"checked_investment_transactions\" value=\"" . $self->{checkedInvestmentTransactions} . "\">\n");
      if ($parseResult eq "2")
      {
        $doc->print("<input type=hidden name=\"inside_investment_transaction_list\" value=\"1\">\n");
      }
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" investment transactions checked: " . $self->{checkedInvestmentTransactions} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "build_lists_from_prices")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file recognizing the lists of prices looking for ticker symbols to determine if
    # securities are used and whether they exist in their respective temporary database table;
    # update the used flag if the entity exists or insert the data into the temporary storage table in the database if it
    # does not exist already and set its used flag to true

    # reset any variables from a previous parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'checked_price_transactions'} && $input->{'checked_price_transactions'} =~ /\d+/)
    {
      $self->{checkedPriceTransactions} = $input->{'checked_price_transactions'};
    }
    my $currentPriceSymbol = "";
    if (defined $input->{'current_price_symbol'} && $input->{'current_price_symbol'} ne "")
    {
      $currentPriceSymbol = $input->{'current_price_symbol'};
    }

    my $line = "";
    my $tempCounter = 0;
    my $parseResult = -1;

    # recognize the start of an account list and parse the records to find the necessary information
    while ((defined $line) && ($tempCounter < $self->{buildListBatchSize}))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if ($line =~ /!Type:Prices/)
        {
          $line = ""; # clear the line variable
          while (defined $line && $line !~ /^\!/ && $tempCounter < $self->{buildListBatchSize})
          {
            $line = $self->getNextLine(fileHandle => $fileHandle);
            if (defined $line)
            {
              if ($line =~ /^"(.+)",/)
              {
                my $price_symbol = $1;
                $price_symbol =~ s/'/\\'/g;
                $price_symbol = uc($price_symbol);

                if ($price_symbol ne "" && $price_symbol ne $currentPriceSymbol)
                {
                  $currentPriceSymbol = $price_symbol;
                  # update the used flag if the entity already exists; otherwise, add a new entity to the database table
                  my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_security_tb SET used = '1' WHERE symbol = ?", plug => [ $price_symbol ]);
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error occurred updating QIF security table for ticker symbol = '$price_symbol'!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  if ($updateResult == 0)
                  {
                    # insert "unlisted" security into database QIF list table; default to having the name be the same as the ticker
                    $self->{dbHandle}->write(sql => "INSERT INTO qif_security_tb (name, symbol, used) VALUES (?, ?, '1')", plug => [ $price_symbol, $price_symbol ]);
                    if ($self->{dbHandle}->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "Error occurred inserting \"unlisted\" security (symbol = '$price_symbol') into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
                      return undef;
                    }
                  }
                }

                # increment the checked/processed count
                $tempCounter++;
                $self->{checkedPriceTransactions}++;
              }
            }
          }
          if ($line =~ /^\!/)
          {
            $self->putBackLine($line);
          }
        }
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF File Parse for Building List");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF File Parse for Building List of Used Entities from Prices</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_file_parse\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"ask_entities\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"build_lists_from_prices\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"checked_price_transactions\" value=\"" . $self->{checkedPriceTransactions} . "\">\n");
      $doc->print("<input type=hidden name=\"current_price_symbol\" value=\"$currentPriceSymbol\">\n");
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_file_parse.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" prices checked: " . $self->{checkedPriceTransactions} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "ask_entities")
  {
    # ask the user whether to attempt to map all entities, or only those used in the import file
    $doc->setTitle("QIF Ask Entities");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Entities</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_ask_entities\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"guess_securities\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    $doc->print("<table border=1 cellspacing=0 cellpadding=2>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td><input type=radio name=\"use_entities\" value=\"all\">Map all entities</td>\n");
    $doc->print("    <td><input type=radio name=\"use_entities\" value=\"used\" checked>Only map used entities</td>\n");
    $doc->print("  </tr>\n");
    $doc->print("</table>\n");
    $doc->print("<input type=submit name=\"submit_entities\" value=\"Continue...\">&nbsp;<input type=reset>\n");
    $doc->print("</form>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "guess_securities")
  {
    # securities are only mapped to stock sub-accounts (type 70)
    my $sth2;
    if ($input->{'use_entities'} eq "all")
    {
      $sth2 = $self->{dbHandle}->read(sql => "SELECT count(*) FROM qif_security_tb");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the maximum number of securities to map!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth2 = $self->{dbHandle}->read(sql => "SELECT count(*) FROM qif_security_tb WHERE used = '1'");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the maximum number of used securities to map!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    my @info = $sth2->fetchrow_array;
    my $maxSecurities = $info[0];
    my $securityOffset = (exists $input->{'security_offset'}) ? $input->{'security_offset'} : 0;
    # retrieve parsed securities from the temporary database tables
    my $sth;
    if ($input->{'use_entities'} eq "all")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name, symbol FROM qif_security_tb ORDER BY name LIMIT " . $self->{limitRange} . " OFFSET $securityOffset");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Attempted to retrieve all QIF securities from database.<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name, symbol FROM qif_security_tb WHERE used = '1' ORDER BY name LIMIT " . $self->{limitRange} . " OFFSET $securityOffset");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Attempted to retrieve all used QIF securities from database.<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    my @info = ();
    my @securityNames = ();
    my @securitySymbols = ();
    while (@info = $sth->fetchrow_array)
    {
      $securityNames[scalar @securityNames] = $info[0];
      $securitySymbols[scalar @securitySymbols] = $info[1];
    }

    # retrieve all sub-accounts from the database of type Stock
    my @subAccounts = $dbObject->getListOfSubAccounts(type => "70", orderBy => "sub_acc_name");
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Attempted to get a list of sub-accounts!<br>\n" . $dbObject->errorMessage;
      return undef;
    }

    # securities can only be mapped to stock sub-accounts
    $doc->setTitle("QIF Map Securities");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Securities</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_guess_securities\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"");
    if (scalar @securityNames)
    {
      $doc->print("map_securities");
    }
    else
    {
      # if there are no securities to map, skip the security insertion code and go to the next step in the process
      $doc->print("guess_classes");
    }
    $doc->print("\">\n");
    $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    if (scalar @securityNames)
    {
      $doc->print("<input type=hidden name=\"max_securities\" value=\"$maxSecurities\">\n");
      $doc->print("<input type=hidden name=\"security_offset\" value=\"" . $securityOffset . "\">\n");
      $doc->print("<p>Working with securities: <b>" . ($securityOffset + 1) . "</b> - <b>" . (($securityOffset + $self->{limitRange} > $maxSecurities) ? $maxSecurities : ($securityOffset + $self->{limitRange})) . "</b> of <b>" . $maxSecurities . "</b></p>\n");
      $doc->print("<table border=1 cellspacing=0 cellpadding=2>\n");
      $doc->print("  <tr>\n");
      $doc->print("    <th>QIF Security Name</th><th>Existing Sub-Account</th><th>Create Sub-Account</th>\n");
      $doc->print("  </tr>\n");
      for (my $rowCount=0; $rowCount < scalar @securityNames; $rowCount++)
      {
        $doc->print("  <tr>\n");
        $doc->print("    <td align=center><b>$securityNames[$rowCount]</b><br>$securitySymbols[$rowCount]<input type=hidden name=\"security_name$rowCount\" value=\"$securityNames[$rowCount]\"><input type=hidden name=\"security_symbol$rowCount\" value=\"$securitySymbols[$rowCount]\"></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"security$rowCount\" value=\"exist\"");
        # determine if this security name exists in the database already; if it does, select it; if not, put the name of the class in the "new" sub-account field
        my $matchFound = 0;
        for (my $subAccountCount=0; $subAccountCount < scalar @subAccounts; $subAccountCount++)
        {
          if ($securityNames[$rowCount] =~ /$subAccounts[$subAccountCount]->{name}/)
          {
            # match found
            $matchFound = 1;
            $doc->print(" checked");
          }
        }
        $doc->print("><select name=\"security_select$rowCount\" onChange=\"document.qif_guess_securities.security$rowCount" . "[0].checked = true;\">\n");
        for (my $subAccountCount=0; $subAccountCount < scalar @subAccounts; $subAccountCount++)
        {
          $doc->print("      <option value=\"$subAccounts[$subAccountCount]->{number}\"");
          if ($securityNames[$rowCount] =~ /$subAccounts[$subAccountCount]->{name}/)
          {
            $doc->print (" selected");
          }
          $doc->print(">$subAccounts[$subAccountCount]->{name}\n");
        }
        $doc->print("    </select></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"security$rowCount\" value=\"new\"");
        if (not $matchFound)
        {
          $doc->print(" checked");
        }
        $doc->print("><input type=text name=\"security_text$rowCount\" value=\"");
        if (not $matchFound)
        {
          $doc->print("$securityNames[$rowCount]");
        }
        $doc->print("\" onChange=\"document.qif_guess_securities.security$rowCount" . "[1].checked = true;\"></td>\n");
        $doc->print("  </tr>\n");
      }
      $doc->print("</table>\n");
      $doc->print("<input type=submit name=\"submit_securities\" value=\"Continue...\" onClick=\"return validateForm();\">&nbsp;<input type=reset>\n");

      $doc->setFocus("javascript");
      $doc->print("  function validateForm()\n");
      $doc->print("  {\n");
      $doc->print("    var errorString = \"\";\n\n");
      for (my $rowCount=0; $rowCount < scalar @securityNames; $rowCount++)
      {
        $doc->print("    if (document.qif_guess_securities.security" . $rowCount . "[1].checked && document.qif_guess_securities.security_text" . $rowCount . ".value == '')\n");
        $doc->print("    {\n");
        (my $tempSecurityName = $securityNames[$rowCount]) =~ s/'/\\'/g;
        $doc->print("      errorString += '\\nThere must be a non-empty value for the sub-account for mapping \"" . $tempSecurityName . "\".';\n");
        $doc->print("    }\n");
      }
      $doc->print("\n");
      $doc->print("    if (errorString.length > 0)\n");
      $doc->print("    {\n");
      $doc->print("      alert(errorString);\n");
      $doc->print("      return false;\n");
      $doc->print("    }\n");
      $doc->print("    return true;\n");
      $doc->print("  }\n");
    }
    else
    {
      $doc->setFocus("body");
      $doc->print("No securities exist in the database.  This most likely means that no securities existed in the QIF import file to parse.<br><br>\n");
      $doc->print("<input type=submit name=\"submit_securities\" value=\"Continue...\">\n");
    }
    $doc->setFocus("body");
    $doc->print("</form>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "map_securities")
  {
    # securities are only mapped to stock sub-accounts (type 70)
    # get the list of security names from the input fields and their values to map in the QIF database tables
    my @securityNames = ();
    my @securitySymbols = ();
    my @subAccountNumbers = ();
    my @subAccountExistNew = ();
    my $rowCount = 0;
    while (exists $input->{"security_name$rowCount"})
    {
      $securityNames[$rowCount] = $input->{"security_name$rowCount"};
      $securitySymbols[$rowCount] = $input->{"security_symbol$rowCount"};
      if ($input->{"security$rowCount"} eq "exist")
      {
        $subAccountNumbers[$rowCount] = $input->{"security_select$rowCount"};
        $subAccountExistNew[$rowCount] = 1;
      }
      else
      {
        if ($input->{"security_text$rowCount"} eq "")
        {
          $subAccountNumbers[$rowCount] = -1;
          $subAccountExistNew[$rowCount] = 1;
        }
        else
        {
          $subAccountNumbers[$rowCount] = $input->{"security_text$rowCount"};
          $subAccountExistNew[$rowCount] = 0;
        }
      }
      $rowCount++;
    }
    # update security mappings; create any non-existent sub-accounts
    # create sub-accounts of type Stock (70)
    my $subAccountLength = $dbObject->getSubAccountLength;
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the minimum sub-account number length.<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    for ($rowCount=0; $rowCount < scalar @securityNames; $rowCount++)
    {
      if (not $subAccountExistNew[$rowCount])
      {
        # create a new sub-account; if it already exists, check to see if it is of the same type as the sub-account to be created;
        # if it is, simply assign that sub-account without creating a new sub-account
        my $nextSubAccountNumber = $dbObject->getNextAvailableSubAccountNumber(minLength => $subAccountLength);
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Attempted to get the next available sub-account number<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        my $subAccountObj = Portal::Accounting::Objects::SubAccount->new(name => $subAccountNumbers[$rowCount], number => $nextSubAccountNumber, type => 70, minSubAccountLength => $subAccountLength);
        if ($subAccountObj->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error creating new sub-account object.<br>\n" . $subAccountObj->errorMessage;
          return undef;
        }
        my @subAccountCreateResult = $dbObject->createSubAccount(subAccountObj => $subAccountObj);
        if ($subAccountCreateResult[0] == -1)
        {
          # sub-account already exists with the same name and number; map "new" security to existing sub-account if type same
          if ($subAccountCreateResult[1] == 70)
          {
            $subAccountNumbers[$rowCount] = $nextSubAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Sub-account already exists with the same name and number!  Type does not match (Type=" . $subAccountCreateResult[1] . ", Name=" . $subAccountObj->name . ", Number=" . $subAccountObj->number . ").";
            return undef;
          }
        }
        elsif ($subAccountCreateResult[0] == -2)
        {
          # sub-account already exists with the same name; map "new" security to existing sub-account if type same
          if ($subAccountCreateResult[1] == 70)
          {
            $subAccountNumbers[$rowCount] = $subAccountCreateResult[2];
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Sub-account already exists with the same name!  Type does not match (Type=" . $subAccountCreateResult[1] . ", Name=" . $subAccountObj->name . ", Number=" . $subAccountCreateResult[2] . ").";
            return undef;
          }
        }
        elsif ($subAccountCreateResult[0] == -3)
        {
          # sub-account already exists with the same number; map "new" security to existing sub-account if type same
          if ($subAccountCreateResult[1] == 70)
          {
            $subAccountNumbers[$rowCount] = $nextSubAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Sub-account already exists with the same number!  Type does not match (Type=" . $subAccountCreateResult[1] . ", Name=" . $subAccountObj->name . ", Number=" . $subAccountObj->number . ").";
            return undef;
          }
        }
        elsif (not $subAccountCreateResult[0])
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to create a new sub-account!<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        else
        {
          # create a new stock entry
          my $stockObj = Portal::Accounting::Objects::StockObject->new(ticker => uc($securitySymbols[$rowCount]), name => $subAccountNumbers[$rowCount], number => $nextSubAccountNumber, market => 'None');
          if ($stockObj->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Problem creating new StockObject object!<br>\n" . $stockObj->errorMessage;
            return undef;
          }
          my @stockCreateResult = $dbObject->createStock(stockObj => $stockObj);
          # the -1 return case is OK; it means the stock already exists in the exact form that matches the current stock "creation"
          if ($stockCreateResult[0] == -2)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Stock already exists with the same name!  (Ticker=" . $stockCreateResult[1] . ", Name=" . $stockObj->name . ", Number=" . $stockCreateResult[2] . ").";
            return undef;
          }
          elsif ($stockCreateResult[0] == -3)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Stock already exists with the same number!  (Ticker=" . $stockCreateResult[1] . ", Name=" . $stockCreateResult[2] . ", Number=" . $stockObj->number . ").";
            return undef;
          }
          elsif ($stockCreateResult[0] == -4)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Stock already exists with the same ticker!  (Ticker=" . $stockObj->ticker . ", Name=" . $stockCreateResult[1] . ", Number=" . $stockCreateResult[2] . ").";
            return undef;
          }
          elsif (not $stockCreateResult[0])
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Tried to create a new stock!<br>\n" . $dbObject->errorMessage;
            return undef;
          }
          else
          {
            # creation successful; map new sub-account number to subAccountNumber[] array element
            $subAccountNumbers[$rowCount] = $nextSubAccountNumber;
          }
        }
      }
      # update security mapping in QIF table
      my $sth = $self->{dbHandle}->write(sql => "UPDATE qif_security_tb SET map_sub_account = ? WHERE name = ?", plug => [ $subAccountNumbers[$rowCount], $securityNames[$rowCount] ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to update the security mapping! (" . $securityNames[$rowCount] . ")<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF Map Securities");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Securities - Successful</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_map_securities\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"");
    if ($input->{'security_offset'} + $self->{limitRange} >= $input->{'max_securities'})
    {
      $doc->print("guess_classes");
    }
    else
    {
      $doc->print("guess_securities");
    }
    $doc->print("\">\n");
    $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    $doc->print("<input type=hidden name=\"max_securities\" value=\"" . $input->{'max_securities'} . "\">\n");
    $doc->print("<input type=hidden name=\"security_offset\" value=\"" . ($input->{'security_offset'} + $self->{limitRange}) . "\">\n");
    if ($input->{'security_offset'} + $self->{limitRange} >= $input->{'max_securities'})
    {
      $doc->print("<p>Mapping: <b>Complete</b></p>\n");
    }
    else
    {
      $doc->print("<p>Mapped: <b>" . (($input->{'security_offset'} + $self->{limitRange}) > $input->{'max_securities'} ? $input->{'max_securities'} : ($input->{'security_offset'} + $self->{limitRange})) . "</b> of <b>" . $input->{'max_securities'} . "</b></p>\n");
    }
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_map_securities.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_securities\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    for ($rowCount=0; $rowCount < scalar @securityNames; $rowCount++)
    {
      $doc->print("\"" . $securityNames[$rowCount] . "\"&nbsp;=>&nbsp;\"" . $subAccountNumbers[$rowCount] . "\"<br>\n");
    }
    $doc->print("</center>\n");
  }
  elsif ($command eq "guess_classes")
  {
    # retrieve all sub-entity names and codes from the database
    my @subEntities = $dbObject->getArrayOfSubEntities(ledgerCode => $ledgerCode);
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Attempted to get a list of sub-entities!<br>\n" . $dbObject->errorMessage;
      return undef;
    }

    # retrieve all parsed classes from the temporary database tables
    my $sth;
    if ($input->{'use_entities'} eq "all")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name FROM qif_class_tb ORDER BY name");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Attempted to retrieve all QIF classes from database.<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name FROM qif_class_tb WHERE used = '1' ORDER BY name");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Attempted to retrieve all used QIF classes from database.<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    my @info = ();
    my @classNames = ();
    while (@info = $sth->fetchrow_array)
    {
      $classNames[scalar @classNames] = $info[0];
    }

    # classes can only be mapped to sub-entities
    $doc->setTitle("QIF Map Classes");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Classes</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_guess_classes\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"");
    if (scalar @classNames)
    {
      $doc->print("map_classes");
    }
    else
    {
      # if there are no classes to map, skip the class insertion code and go to the next step in the process
      $doc->print("guess_categories");
    }
    $doc->print("\">\n");
    $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    if (scalar @classNames)
    {
      $doc->print("<table border=1 cellspacing=0 cellpadding=2>\n");
      $doc->print("  <tr>\n");
      $doc->print("    <th>QIF Class Name</th><th>Existing Sub-Entity</th><th>Create Sub-Entity</th>\n");
      $doc->print("  </tr>\n");
      for (my $rowCount=0; $rowCount < scalar @classNames; $rowCount++)
      {
        $doc->print("  <tr>\n");
        $doc->print("    <td align=center><b>$classNames[$rowCount]</b><input type=hidden name=\"class_name$rowCount\" value=\"$classNames[$rowCount]\"></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"class$rowCount\" value=\"exist\"");
        # determine if this class name exists in the database already; if it does, select it; if not, put the name of the class in the "new" sub-entity field
        my $matchFound = 0;
        for (my $subEntityCount=0; $subEntityCount < scalar @subEntities; $subEntityCount++)
        {
          if ($classNames[$rowCount] =~ /$subEntities[$subEntityCount]->{name}/)
          {
            # match found
            $matchFound = 1;
            $doc->print(" checked");
          }
        }
        $doc->print("><select name=\"class_select$rowCount\" onChange=\"document.qif_guess_classes.class$rowCount" . "[0].checked = true;\">\n");
        for (my $subEntityCount=0; $subEntityCount < scalar @subEntities; $subEntityCount++)
        {
          $doc->print("      <option value=\"$subEntities[$subEntityCount]->{code}\"");
          if ($classNames[$rowCount] =~ /$subEntities[$subEntityCount]->{name}/)
          {
            $doc->print (" selected");
          }
          $doc->print(">$subEntities[$subEntityCount]->{name}\n");
        }
        $doc->print("    </select></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"class$rowCount\" value=\"new\"");
        if (not $matchFound)
        {
          $doc->print(" checked");
        }
        $doc->print("><input type=text name=\"class_text$rowCount\" value=\"");
        if (not $matchFound)
        {
          $doc->print("$classNames[$rowCount]");
        }
        $doc->print("\" onChange=\"document.qif_guess_classes.class$rowCount" . "[1].checked = true;\"></td>\n");
        $doc->print("  </tr>\n");
      }
      $doc->print("</table>\n");
      $doc->print("<input type=submit name=\"submit_classes\" value=\"Continue...\" onClick=\"return validateForm();\">&nbsp;<input type=reset>\n");

      $doc->setFocus("javascript");
      $doc->print("  function validateForm()\n");
      $doc->print("  {\n");
      $doc->print("    var errorString = \"\";\n\n");
      for (my $rowCount=0; $rowCount < scalar @classNames; $rowCount++)
      {
        $doc->print("    if (document.qif_guess_classes.class" . $rowCount . "[1].checked && document.qif_guess_classes.class_text" . $rowCount . ".value == '')\n");
        $doc->print("    {\n");
        $doc->print("      errorString += '\\nThere must be a non-empty value for the sub-entity for mapping \"" . $classNames[$rowCount] . "\".';\n");
        $doc->print("    }\n");
      }
      $doc->print("\n");
      $doc->print("    if (errorString.length > 0)\n");
      $doc->print("    {\n");
      $doc->print("      alert(errorString);\n");
      $doc->print("      return false;\n");
      $doc->print("    }\n");
      $doc->print("    return true;\n");
      $doc->print("  }\n");
    }
    else
    {
      $doc->setFocus("body");
      $doc->print("No classes were retrieved from the database.  This most likely means that no classes existed in the QIF import file to parse or no classes were used by this import.<br><br>\n");
      $doc->print("<input type=submit name=\"submit_classes\" value=\"Continue...\">\n");
    }
    $doc->setFocus("body");
    $doc->print("</form>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "map_classes")
  {
    # get the list of class names from the input fields and their values to map in the QIF database tables
    my @classNames = ();
    my @subEntityCodes = ();
    my @existNew = ();
    my $rowCount = 0;
    while (exists $input->{"class_name$rowCount"})
    {
      $classNames[$rowCount] = $input->{"class_name$rowCount"};
      if ($input->{"class$rowCount"} eq "exist")
      {
        $subEntityCodes[$rowCount] = $input->{"class_select$rowCount"};
        $existNew[$rowCount] = 1;
      }
      else
      {
        $subEntityCodes[$rowCount] = $input->{"class_text$rowCount"};
        $existNew[$rowCount] = 0;
      }
      $rowCount++;
    }
    # update class mappings; create any non-existent sub-entities
    for ($rowCount=0; $rowCount < scalar @classNames; $rowCount++)
    {
      if (not $existNew[$rowCount])
      {
        # check for the existence of a sub-entity with the same name or code
        my $sth = $self->{dbHandle}->read(sql => "SELECT name, code FROM sub_entity_tb WHERE name = ? OR code = ? AND ledger_code = ?", plug => [ $classNames[$rowCount], $subEntityCodes[$rowCount], $ledgerCode ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to retrieve any sub-entities with the same name or code! (" . $classNames[$rowCount] . ")<br>\n" . $self->{dbHandle}->errorMessage;
          return undef;
        }
        if (my @info = $sth->fetchrow_array)
        {
          # sub-entity already exists with the same name or code; map "new" class to existing sub-entity
          if ($info[0] eq $classNames[$rowCount])
          {
            $subEntityCodes[$rowCount] = $info[1];
          }
        }
        else
        {
          my $sth = $self->{dbHandle}->write(sql => "INSERT INTO sub_entity_tb (name, code, ledger_code) VALUES (?, ?, ?)", plug => [ $classNames[$rowCount], $subEntityCodes[$rowCount], $ledgerCode ]);
          if ($self->{dbHandle}->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Tried to insert a new sub-entity into the database! (" . $classNames[$rowCount] . ")<br>\n" . $self->{dbHandle}->errorMessage;
            return undef;
          }
        }
      }
      # update class mapping in QIF table
      my $sth = $self->{dbHandle}->write(sql => "UPDATE qif_class_tb SET map_sub_entity = ? WHERE name = ?", plug => [ $subEntityCodes[$rowCount], $classNames[$rowCount] ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to update the class mapping! (" . $classNames[$rowCount] . ")<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF Map Classes");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Classes - Successful</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_map_classes\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"guess_categories\">\n");
    $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_map_classes.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_classes\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    for ($rowCount=0; $rowCount < scalar @classNames; $rowCount++)
    {
      $doc->print("\"" . $classNames[$rowCount] . "\" => \"" . $subEntityCodes[$rowCount] . "\"<br>\n");
    }
    $doc->print("</center>\n");
  }
  elsif ($command eq "guess_categories")
  {
    my @categoryNames = ();
    my @categoryDescriptions = ();
    my @taxRelatedFlags = ();
    my @incomeExpenseFlags = ();
    my $maxCategories = 0;
    my $categoryOffset = 0;
    my $sth2;
    if ($input->{'use_entities'} eq "all")
    {
      $sth2 = $self->{dbHandle}->read(sql => "SELECT count(*) FROM qif_category_tb");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the maximum number of categories to map!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth2 = $self->{dbHandle}->read(sql => "SELECT count(*) FROM qif_category_tb WHERE used = '1'");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the maximum number of used categories to map!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    my @info = $sth2->fetchrow_array;
    $maxCategories = $info[0];
    $categoryOffset = (exists $input->{'category_offset'}) ? $input->{'category_offset'} : 0;
    my $sth;
    if ($input->{'use_entities'} eq "all")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name, description, tax_related_item, income_expense_flag FROM qif_category_tb ORDER BY name LIMIT " . $self->{limitRange} . " OFFSET $categoryOffset");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get all the categories from the QIF table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name, description, tax_related_item, income_expense_flag FROM qif_category_tb WHERE used = '1' ORDER BY name LIMIT " . $self->{limitRange} . " OFFSET $categoryOffset");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get all the categories from the QIF table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    if ($maxCategories > 0)
    {
      while (my @info = $sth->fetchrow_array)
      {
        $categoryNames[scalar @categoryNames] = $info[0];
        $categoryDescriptions[scalar @categoryDescriptions] = $info[1];
        $taxRelatedFlags[scalar @taxRelatedFlags] = $info[2];
        $incomeExpenseFlags[scalar @incomeExpenseFlags] = $info[3];
      }
    }
    if (scalar @categoryNames)
    {
      # get a list of all income and expense accounts in the system along with all sub-accounts that can be assigned to them
      # get list of sub-account types for the revenue and expense account types
      my $firstDigit = 4;
      my $sth4 = $self->{dbHandle}->read(sql => "SELECT num FROM acct_type_tb WHERE first_digit = ?", plug => [ $firstDigit ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get a list of account types for first_digit starting with $firstDigit.<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
      my @incomeTaxableAccounts = ();
      my @incomeNonTaxableAccounts = ();
      my @incomeSubAccounts = ();
      my %subAccountTypes = ();
      while (my @info = $sth4->fetchrow_array)
      {
        push @incomeTaxableAccounts, $dbObject->getListOfAccounts(location => "global", type => $info[0], taxable => "true", active => "true", orderBy => "type, name");
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (income taxable) for account type = " . $info[0] . "<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        push @incomeNonTaxableAccounts, $dbObject->getListOfAccounts(location => "global", type => $info[0], taxable => "false", active => "true", orderBy => "type, name");
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (income non-taxable) for account type = " . $info[0] . "<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        my $accountType = $info[0];
        my $sth3 = $self->{dbHandle}->read(sql => "SELECT sub_acct_type FROM acct_type_tb WHERE num = ?", plug => [ $accountType ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to get a list of sub-account types assigned to account type $accountType.<br>\n" . $self->{dbHandle}->errorMessage;
          return undef;
        }
        my @info2 = $sth3->fetchrow_array;
        if (@info2)
        {
          my @subAccountTypes = split /,/, $info2[0];
          for (my $typeCount=0; $typeCount < scalar @subAccountTypes; $typeCount++)
          {
            $subAccountTypes{$subAccountTypes[$typeCount]} = 1;
          }
        }
      }
      foreach my $subAccountType (sort keys %subAccountTypes)
      {
        push @incomeSubAccounts, $dbObject->getListOfSubAccounts(type => $subAccountType, orderBy => "sub_acc_name");
      }

      $firstDigit = 5;
      my $sth4 = $self->{dbHandle}->read(sql => "SELECT num FROM acct_type_tb WHERE first_digit = ?", plug => [ $firstDigit ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get a list of account types for first_digit starting with $firstDigit.<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
      my @expenseTaxableAccounts = ();
      my @expenseNonTaxableAccounts = ();
      my @expenseSubAccounts = ();
      %subAccountTypes = ();
      while (my @info = $sth4->fetchrow_array)
      {
        push @expenseTaxableAccounts, $dbObject->getListOfAccounts(location => "global", type => $info[0], taxable => "true", active => "true", orderBy => "type, name");
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (expense taxable) for account type = " . $info[0] . "<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        push @expenseNonTaxableAccounts, $dbObject->getListOfAccounts(location => "global", type => $info[0], taxable => "false", active => "true", orderBy => "type, name");
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (expense non-taxable) for account type = " . $info[0] . "<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        my $accountType = $info[0];
        my $sth3 = $self->{dbHandle}->read(sql => "SELECT sub_acct_type FROM acct_type_tb WHERE num = ?", plug => [ $accountType ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to get a list of sub-account types assigned to account type $accountType.<br>\n" . $self->{dbHandle}->errorMessage;
          return undef;
        }
        my @info2 = $sth3->fetchrow_array;
        if (@info2)
        {
          my @subAccountTypes = split /,/, $info2[0];
          for (my $typeCount=0; $typeCount < scalar @subAccountTypes; $typeCount++)
          {
            $subAccountTypes{$subAccountTypes[$typeCount]} = 1;
          }
        }
      }
      foreach my $subAccountType (sort keys %subAccountTypes)
      {
        push @expenseSubAccounts, $dbObject->getListOfSubAccounts(type => $subAccountType, orderBy => "sub_acc_name");
      }

      $doc->setFocus("javascript");
      $doc->print("var incomeTaxableAccounts = new Array();\n");
      $doc->print("var incomeNonTaxableAccounts = new Array();\n");
      $doc->print("var expenseTaxableAccounts = new Array();\n");
      $doc->print("var expenseNonTaxableAccounts = new Array();\n");
      $doc->print("var incomeSubAccounts = new Array();\n");
      $doc->print("var expenseSubAccounts = new Array();\n");
      $doc->print("\n");
      for (my $incomeCount=0; $incomeCount < scalar @incomeTaxableAccounts; $incomeCount++)
      {
        $doc->print("incomeTaxableAccounts[$incomeCount] = new Array();\n");
        (my $tempAccount = $incomeTaxableAccounts[$incomeCount]->name) =~ s/\'/\\'/g;
        $doc->print("incomeTaxableAccounts[$incomeCount][0] = \"" . $tempAccount . " ~\";\n");
        $doc->print("incomeTaxableAccounts[$incomeCount][1] = \"" . $incomeTaxableAccounts[$incomeCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $incomeCount=0; $incomeCount < scalar @incomeNonTaxableAccounts; $incomeCount++)
      {
        $doc->print("incomeNonTaxableAccounts[$incomeCount] = new Array();\n");
        (my $tempAccount = $incomeNonTaxableAccounts[$incomeCount]->name) =~ s/\'/\\'/g;
        $doc->print("incomeNonTaxableAccounts[$incomeCount][0] = \"" . $tempAccount . "\";\n");
        $doc->print("incomeNonTaxableAccounts[$incomeCount][1] = \"" . $incomeNonTaxableAccounts[$incomeCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $expenseCount=0; $expenseCount < scalar @expenseTaxableAccounts; $expenseCount++)
      {
        $doc->print("expenseTaxableAccounts[$expenseCount] = new Array();\n");
        (my $tempAccount = $expenseTaxableAccounts[$expenseCount]->name) =~ s/\'/\\'/g;
        $doc->print("expenseTaxableAccounts[$expenseCount][0] = \"" . $tempAccount . " ~\";\n");
        $doc->print("expenseTaxableAccounts[$expenseCount][1] = \"" . $expenseTaxableAccounts[$expenseCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $expenseCount=0; $expenseCount < scalar @expenseNonTaxableAccounts; $expenseCount++)
      {
        $doc->print("expenseNonTaxableAccounts[$expenseCount] = new Array();\n");
        (my $tempAccount = $expenseNonTaxableAccounts[$expenseCount]->name) =~ s/\'/\\'/g;
        $doc->print("expenseNonTaxableAccounts[$expenseCount][0] = \"" . $tempAccount . "\";\n");
        $doc->print("expenseNonTaxableAccounts[$expenseCount][1] = \"" . $expenseNonTaxableAccounts[$expenseCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $incomeCount=0; $incomeCount < scalar @incomeSubAccounts; $incomeCount++)
      {
        $doc->print("incomeSubAccounts[$incomeCount] = new Array();\n");
        (my $tempSubAccount = $incomeSubAccounts[$incomeCount]->name) =~ s/\'/\\'/g;
        $doc->print("incomeSubAccounts[$incomeCount][0] = \"" . $tempSubAccount . "\";\n");
        $doc->print("incomeSubAccounts[$incomeCount][1] = \"" . $incomeSubAccounts[$incomeCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $expenseCount=0; $expenseCount < scalar @expenseSubAccounts; $expenseCount++)
      {
        $doc->print("expenseSubAccounts[$expenseCount] = new Array();\n");
        (my $tempSubAccount = $expenseSubAccounts[$expenseCount]->name) =~ s/\'/\\'/g;
        $doc->print("expenseSubAccounts[$expenseCount][0] = \"" . $tempSubAccount . "\";\n");
        $doc->print("expenseSubAccounts[$expenseCount][1] = \"" . $expenseSubAccounts[$expenseCount]->number . "\";\n");
      }
      $doc->print("\n");
      $doc->print("function validateForm()\n");
      $doc->print("{\n");
      $doc->print("  var errorString = \"\";\n");
      $doc->print("\n");
      for (my $rowCount=0; $rowCount < scalar @categoryNames; $rowCount++)
      {
        $doc->print("    if (document.qif_guess_categories.category_account" . $rowCount . "[1].checked && document.qif_guess_categories.category_account_text" . $rowCount . ".value == '')\n");
        $doc->print("    {\n");
        $doc->print("      errorString += '\\nThere must be a non-empty value for the account for mapping \"" . $categoryNames[$rowCount] . "\".';\n");
        $doc->print("    }\n");
        $doc->print("    if (document.qif_guess_categories.category_sub_account" . $rowCount . "[1].checked && document.qif_guess_categories.category_sub_account_text" . $rowCount . ".value == '')\n");
        $doc->print("    {\n");
        $doc->print("      errorString += '\\nThere must be a non-empty value for the sub-account for mapping \"" . $categoryNames[$rowCount] . "\".';\n");
        $doc->print("    }\n");
      }
      $doc->print("\n");
      $doc->print("  if (errorString != \"\")\n");
      $doc->print("  {\n");
      $doc->print("    alert(errorString);\n");
      $doc->print("    return false;\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  return true;\n");
      $doc->print("}\n");

      # only use the last two "fields" of data in the category definition to guess the account/sub-account for the category
      my $onLoadCode = "var numCategoryNames = " . (scalar @categoryNames) . ";\n";
      $onLoadCode .= "\n";
      $onLoadCode .= "for (var rowCount=0; rowCount < numCategoryNames; rowCount++)\n";
      $onLoadCode .= "{\n";
      $onLoadCode .= "  eval(\"var taxable = document.qif_guess_categories.taxable\" + rowCount + \".value\");\n";
      $onLoadCode .= "  eval(\"var incomeExpense = document.qif_guess_categories.income_expense\" + rowCount + \".value\");\n";
      $onLoadCode .= "  var tempAccountsArray;\n";
      $onLoadCode .= "  var tempSubAccountsArray;\n";
      $onLoadCode .= "  eval(\"document.qif_guess_categories.category_account_select\" + rowCount + \".options.length = 0;\");\n";
      $onLoadCode .= "  eval(\"document.qif_guess_categories.category_sub_account_select\" + rowCount + \".options.length = 0;\");\n";
      # set the values of the select drop-downs based on the type and taxability of the category
      $onLoadCode .= "  if (taxable == 1 && incomeExpense == 1)\n";
      $onLoadCode .= "  {\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < incomeTaxableAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_account_select\" + rowCount + \".options[tempCount] = new Option('\" + incomeTaxableAccounts[tempCount][0] + \"', '\" + incomeTaxableAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < incomeSubAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_sub_account_select\" + rowCount + \".options[tempCount] = new Option('\" + incomeSubAccounts[tempCount][0] + \"', '\" + incomeSubAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    tempAccountsArray = incomeTaxableAccounts;\n";
      $onLoadCode .= "    tempSubAccountsArray = incomeSubAccounts;\n";
      $onLoadCode .= "  }\n";
      $onLoadCode .= "  else if (taxable == 1 && incomeExpense == 0)\n";
      $onLoadCode .= "  {\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < expenseTaxableAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_account_select\" + rowCount + \".options[tempCount] = new Option('\" + expenseTaxableAccounts[tempCount][0] + \"', '\" + expenseTaxableAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < expenseSubAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_sub_account_select\" + rowCount + \".options[tempCount] = new Option('\" + expenseSubAccounts[tempCount][0] + \"', '\" + expenseSubAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    tempAccountsArray = expenseTaxableAccounts;\n";
      $onLoadCode .= "    tempSubAccountsArray = expenseSubAccounts;\n";
      $onLoadCode .= "  }\n";
      $onLoadCode .= "  else if (taxable == 0 && incomeExpense == 1)\n";
      $onLoadCode .= "  {\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < incomeNonTaxableAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_account_select\" + rowCount + \".options[tempCount] = new Option('\" + incomeNonTaxableAccounts[tempCount][0] + \"', '\" + incomeNonTaxableAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < incomeSubAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_sub_account_select\" + rowCount + \".options[tempCount] = new Option('\" + incomeSubAccounts[tempCount][0] + \"', '\" + incomeSubAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    tempAccountsArray = incomeNonTaxableAccounts;\n";
      $onLoadCode .= "    tempSubAccountsArray = incomeSubAccounts;\n";
      $onLoadCode .= "  }\n";
      $onLoadCode .= "  else if (taxable == 0 && incomeExpense == 0)\n";
      $onLoadCode .= "  {\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < expenseNonTaxableAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_account_select\" + rowCount + \".options[tempCount] = new Option('\" + expenseNonTaxableAccounts[tempCount][0] + \"', '\" + expenseNonTaxableAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    for (var tempCount=0; tempCount < expenseSubAccounts.length; tempCount++)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      eval(\"document.qif_guess_categories.category_sub_account_select\" + rowCount + \".options[tempCount] = new Option('\" + expenseSubAccounts[tempCount][0] + \"', '\" + expenseSubAccounts[tempCount][1] + \"');\");\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    tempAccountsArray = expenseNonTaxableAccounts;\n";
      $onLoadCode .= "    tempSubAccountsArray = expenseSubAccounts;\n";
      $onLoadCode .= "  }\n";
      # split the category array on ':' characters
      $onLoadCode .= "  var categoryArray = eval(\"document.qif_guess_categories.category_name\" + rowCount + \".value.split(':')\");\n";
      $onLoadCode .= "  if (categoryArray.length > 0)\n";
      $onLoadCode .= "  {\n";
      $onLoadCode .= "    var tempAccount = \"\";\n";
      $onLoadCode .= "    var tempSubAccount = \"\";\n";
      $onLoadCode .= "    if (categoryArray.length > 1)\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      tempSubAccount = categoryArray[categoryArray.length - 1];\n";
      $onLoadCode .= "      tempAccount = categoryArray[categoryArray.length - 2];\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    else\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      tempAccount = categoryArray[categoryArray.length - 1];\n";
      $onLoadCode .= "    }\n";
      # attempt to find a match for the last two elements of the category name
      $onLoadCode .= "    if (tempAccount != '')\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      var matchFound = -1;\n";
      $onLoadCode .= "      var accountRegExp = new RegExp(\"\" + tempAccount);\n";
      $onLoadCode .= "      for (var count=0; count < tempAccountsArray.length; count++)\n";
      $onLoadCode .= "      {\n";
      $onLoadCode .= "        if (tempAccountsArray[count][0].match(accountRegExp))\n";
      $onLoadCode .= "        {\n";
      $onLoadCode .= "          matchFound = count;\n";
      $onLoadCode .= "        }\n";
      $onLoadCode .= "      }\n";
      $onLoadCode .= "      if (matchFound > -1)\n";
      $onLoadCode .= "      {\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_account_select\" + rowCount + \".options[\" + matchFound + \"].selected = true;\");\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_account\" + rowCount + \"[0].checked = true;\");\n";
      $onLoadCode .= "      }\n";
      $onLoadCode .= "      else\n";
      $onLoadCode .= "      {\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_account_text\" + rowCount + \".value = tempAccount;\");\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_account\" + rowCount + \"[1].checked = true;\");\n";
      $onLoadCode .= "      }\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "    if (tempSubAccount != '')\n";
      $onLoadCode .= "    {\n";
      $onLoadCode .= "      var matchFound = -1;\n";
      $onLoadCode .= "      var subAccountRegExp = new RegExp(\"\" + tempSubAccount);\n";
      $onLoadCode .= "      for (var count=0; count < tempSubAccountsArray.length; count++)\n";
      $onLoadCode .= "      {\n";
      $onLoadCode .= "        if (tempSubAccountsArray[count][0].match(subAccountRegExp))\n";
      $onLoadCode .= "        {\n";
      $onLoadCode .= "          matchFound = count;\n";
      $onLoadCode .= "        }\n";
      $onLoadCode .= "      }\n";
      $onLoadCode .= "      if (matchFound > -1)\n";
      $onLoadCode .= "      {\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_sub_account_select\" + rowCount + \".options[\" + matchFound + \"].selected = true;\");\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_sub_account\" + rowCount + \"[0].checked = true;\");\n";
      $onLoadCode .= "      }\n";
      $onLoadCode .= "      else\n";
      $onLoadCode .= "      {\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_sub_account_text\" + rowCount + \".value = tempSubAccount;\");\n";
      $onLoadCode .= "        eval(\"document.qif_guess_categories.category_sub_account\" + rowCount + \"[1].checked = true;\");\n";
      $onLoadCode .= "      }\n";
      $onLoadCode .= "    }\n";
      $onLoadCode .= "  }\n";
      $onLoadCode .= "}\n";
      $doc->setOnload(code => $onLoadCode);

      $doc->setTitle("QIF Map Categories");
      $doc->setFocus("body");
      $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
      $doc->print("<center>\n");
      $doc->print("<h1>QIF Map Categories</h1><br>\n");
      $doc->print("<form method=\"post\" name=\"qif_guess_categories\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
      $doc->print("$staticInfo");
      $doc->print("<input type=hidden name=\"command\" value=\"map_categories\">\n");
      $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
      $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
      $doc->print("<input type=hidden name=\"max_categories\" value=\"$maxCategories\">\n");
      $doc->print("<input type=hidden name=\"category_offset\" value=\"" . $categoryOffset . "\">\n");
      $doc->print("<p>Working with categories: <b>" . ($categoryOffset + 1) . "</b> - <b>" . (($categoryOffset + $self->{limitRange} > $maxCategories) ? $maxCategories : ($categoryOffset + $self->{limitRange})) . "</b> of <b>" . $maxCategories . "</b></p>\n");
      $doc->print("<table border=1 cellspacing=0 cellpadding=2>\n");
      $doc->print("  <tr>\n");
      $doc->print("    <th align=center>QIF Category Name</th><th align=center>Existing Account</th><th align=center>Create Account</th><th align=center>Existing Sub-account</th><th align=center>Create Sub-account</th>\n");
      $doc->print("  </tr>\n");
      for (my $rowCount=0; $rowCount < scalar @categoryNames; $rowCount++)
      {
        $doc->print("  <tr>\n");
        $doc->print("    <td><b>$categoryNames[$rowCount]</b><br>$categoryDescriptions[$rowCount]\n");
        $doc->print("      <input type=hidden name=\"category_name$rowCount\" value=\"$categoryNames[$rowCount]\">\n");
        $doc->print("      <input type=hidden name=\"taxable$rowCount\" value=\"$taxRelatedFlags[$rowCount]\">\n");
        $doc->print("      <input type=hidden name=\"income_expense$rowCount\" value=\"$incomeExpenseFlags[$rowCount]\">\n");
        $doc->print("    </td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"category_account$rowCount\" value=\"exist\"><select name=\"category_account_select$rowCount\" onChange=\"document.qif_guess_categories.category_account$rowCount" . "[0].checked = true;\">\n");
        $doc->print("        <option value=\"-1\">--------------------\n");
        $doc->print("      </select></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"category_account$rowCount\" value=\"new\"><input type=text name=\"category_account_text$rowCount\" value=\"\" onChange=\"document.qif_guess_categories.category_account$rowCount" . "[1].checked = true;\"></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"category_sub_account$rowCount\" value=\"exist\"><select name=\"category_sub_account_select$rowCount\" onChange=\"document.qif_guess_categories.category_sub_account$rowCount" . "[0].checked = true;\">\n");
        $doc->print("        <option value=\"-1\">---------------\n");
        $doc->print("      </select></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"category_sub_account$rowCount\" value=\"new\"><input type=text name=\"category_sub_account_text$rowCount\" value=\"\" onChange=\"document.qif_guess_categories.category_sub_account$rowCount" . "[1].checked = true;\"></td>\n");
        $doc->print("  </tr>\n");
      }
      $doc->print("</table>\n");
      $doc->print("<input type=submit name=\"submit_categories\" value=\"Continue...\" onClick=\"return validateForm();\">&nbsp;<input type=button value=\"Reset\" onClick=\"document.qif_guess_categories.reset(); doOnLoad();\">\n");
      $doc->print("</form>\n");
      $doc->print("</center>\n");
    }
    else
    {
      # if there are no categories to map, skip the category insertion code and go to the next step in the process
      $doc->setTitle("QIF Map Categories");
      $doc->setFocus("body");
      $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
      $doc->print("<center>\n");
      $doc->print("<h1>QIF Map Categories</h1><br>\n");
      $doc->print("<form method=\"post\" name=\"qif_guess_categories\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
      $doc->print("$staticInfo");
      $doc->print("<input type=hidden name=\"command\" value=\"guess_accounts\">\n");
      $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
      $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
      $doc->print("No categories were found mapped to transaction entries.<br><br>\n");
      $doc->print("<input type=submit name=\"submit_categories\" value=\"Continue...\">\n");
      $doc->print("</form>\n");
      $doc->print("</center>\n");
    }
  }
  elsif ($command eq "map_categories")
  {
    # get the list of category names from the input fields and their values to map in the QIF database tables
    my @categoryNames = ();
    my @taxable = ();
    my @incomeExpense = ();
    my @accountNumbers = ();
    my @subAccountNumbers = ();
    my @accountExistNew = ();
    my @subAccountExistNew = ();
    my $rowCount = 0;
    while (exists $input->{"category_name$rowCount"})
    {
      $categoryNames[$rowCount] = $input->{"category_name$rowCount"};
      if ($input->{"taxable$rowCount"})
      {
        $taxable[$rowCount] = "true";
      }
      else
      {
        $taxable[$rowCount] = "false";
      }
      if ($input->{"income_expense$rowCount"})
      {
        $incomeExpense[$rowCount] = "9"; # Other Revenue account type
      }
      else
      {
        $incomeExpense[$rowCount] = "13"; # Other Expenses account type
      }
      if ($input->{"category_account$rowCount"} eq "exist")
      {
        $accountNumbers[$rowCount] = $input->{"category_account_select$rowCount"};
        $accountExistNew[$rowCount] = 1;
      }
      else
      {
        $accountNumbers[$rowCount] = $input->{"category_account_text$rowCount"};
        $accountExistNew[$rowCount] = 0;
      }
      if ($input->{"category_sub_account$rowCount"} eq "exist")
      {
        $subAccountNumbers[$rowCount] = $input->{"category_sub_account_select$rowCount"};
        if ($subAccountNumbers[$rowCount] eq "")
        {
          $subAccountNumbers[$rowCount] = -1;
        }
        $subAccountExistNew[$rowCount] = 1;
      }
      else
      {
        if ($input->{"category_sub_account_text$rowCount"} eq "")
        {
          $subAccountNumbers[$rowCount] = -1;
          $subAccountExistNew[$rowCount] = 1;
        }
        else
        {
          $subAccountNumbers[$rowCount] = $input->{"category_sub_account_text$rowCount"};
          $subAccountExistNew[$rowCount] = 0;
        }
      }
      $rowCount++;
    }
    # update category mappings; create any non-existent accounts/sub-accounts
    # based on the "income" or "expense" flag, create an appropriate account of that type
    # create sub-accounts of type Normal (50)
    my $accountLength = $dbObject->getAccountLength;
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the minimum account number length.<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    my $subAccountLength = $dbObject->getSubAccountLength;
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the minimum sub-account number length.<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    for ($rowCount=0; $rowCount < scalar @categoryNames; $rowCount++)
    {
      if (not $accountExistNew[$rowCount])
      {
        # create a new account; if it already exists, check to see if it is of the same type as the account to be created;
        # if it is, simply assign that account without creating a new account
        my $sth = $self->{dbHandle}->read(sql => "SELECT first_digit FROM acct_type_tb WHERE num = ?", plug => [ $incomeExpense[$rowCount] ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the first digit for account type (" . $incomeExpense[$rowCount] . ").<br>\n" . $self->{dbHandle}->errorMessage;
          return undef;
        }
        my @info = $sth->fetchrow_array;
        if ((not defined @info) || (not defined $info[0]) || $info[0] !~ /^\d+$/)
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>First digit ($info[0]) invalid when attempting to retrieve for $incomeExpense[$rowCount] account type.\n";
          return undef;
        }
        my $nextAccountNumber = $dbObject->getNextAvailableAccountNumber(firstDigit => $info[0], minLength => $accountLength);
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Attempted to get the next available account number<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        my $accountObj = Portal::Accounting::Objects::Account->new(name => $accountNumbers[$rowCount], number => $nextAccountNumber, type => $incomeExpense[$rowCount], taxable => $taxable[$rowCount], minAccountLength => $accountLength);
        if ($accountObj->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error creating new account object.<br>\n" . $accountObj->errorMessage;
          return undef;
        }
        my @accountCreateResult = $dbObject->createAccount(accountObj => $accountObj);
        if ($accountCreateResult[0] == -1)
        {
          # account already exists with the same name and number; map "new" category to existing account if type and taxable same
          if ($accountCreateResult[1] == $incomeExpense[$rowCount] && $accountCreateResult[2] eq ($taxable[$rowCount] eq "true" ? 1 : 0))
          {
            $accountNumbers[$rowCount] = $nextAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Account already exists with the same name and number!  Type and taxable do not match (Type=" . $accountCreateResult[1] . ", Taxable=" . $accountCreateResult[2] . ", Account Name=" . $accountObj->name . ", Account Number=" . $accountObj->number . ").";
            return undef;
          }
        }
        elsif ($accountCreateResult[0] == -2)
        {
          # account already exists with the same name; map "new" category to existing account if type and taxable same
          if ($accountCreateResult[1] == $incomeExpense[$rowCount] && $accountCreateResult[2] eq ($taxable[$rowCount] eq "true" ? 1 : 0))
          {
            $accountNumbers[$rowCount] = $accountCreateResult[3];
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Account already exists with the same name!  Type and taxable do not match (Type=" . $accountCreateResult[1] . ", Taxable=" . $accountCreateResult[2] . ", Account Name=" . $accountObj->name . ", Account Number=" . $accountCreateResult[3] . ").";
            return undef;
          }
        }
        elsif ($accountCreateResult[0] == -3)
        {
          # account already exists with the same number; map "new" category to existing account if type same
          if ($accountCreateResult[1] == $incomeExpense[$rowCount] && $accountCreateResult[2] eq ($taxable[$rowCount] eq "true" ? 1 : 0))
          {
            $accountNumbers[$rowCount] = $nextAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Account already exists with the same number!  Type and taxable do not match (Type=" . $accountCreateResult[1] . ", Taxable=" . $accountCreateResult[2] . ", Account Name=" . $accountObj->name . ", Account Number=" . $accountCreateResult[3] . ").";
            return undef;
          }
        }
        elsif (not $accountCreateResult[0])
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to create a new account!<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        else
        {
          # creation successful; map new account number to accountNumbers[] array element
          $accountNumbers[$rowCount] = $nextAccountNumber;
          # assign the newly created account to the ledger being worked with
          if ($dbObject->assignAccountToLedger(accountObj => $accountObj, ledgerCode => $ledgerCode) == 0)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Attempted to assign the newly created account (" . $accountObj->name . " - " . $accountObj->number . ") to the ledger ($ledgerCode)!<br>\n" . $dbObject->errorMessage;
            return undef;
          }
        }
      }
      if (not $subAccountExistNew[$rowCount])
      {
        # create a new sub-account; if it already exists, check to see if it is of the same type as the sub-account to be created;
        # if it is, simply assign that sub-account without creating a new sub-account
        my $nextSubAccountNumber = $dbObject->getNextAvailableSubAccountNumber(minLength => $subAccountLength);
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Attempted to get the next available sub-account number<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        my $subAccountObj = Portal::Accounting::Objects::SubAccount->new(name => $subAccountNumbers[$rowCount], number => $nextSubAccountNumber, type => 50, minSubAccountLength => $subAccountLength);
        if ($subAccountObj->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error creating new sub-account object.<br>\n" . $subAccountObj->errorMessage;
          return undef;
        }
        my @subAccountCreateResult = $dbObject->createSubAccount(subAccountObj => $subAccountObj);
        if ($subAccountCreateResult[0] == -1)
        {
          # sub-account already exists with the same name and number; map "new" category to existing sub-account if type same
          if ($subAccountCreateResult[1] == 50)
          {
            $subAccountNumbers[$rowCount] = $nextSubAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Sub-account already exists with the same name and number!  Type does not match (Type=" . $subAccountCreateResult[1] . ", Name=" . $subAccountObj->name . ", Number=" . $subAccountObj->number . ").";
            return undef;
          }
        }
        elsif ($subAccountCreateResult[0] == -2)
        {
          # sub-account already exists with the same name; map "new" category to existing sub-account if type same
          if ($subAccountCreateResult[1] == 50)
          {
            $subAccountNumbers[$rowCount] = $subAccountCreateResult[2];
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Sub-account already exists with the same name!  Type does not match (Type=" . $subAccountCreateResult[1] . ", Name=" . $subAccountObj->name . ", Number=" . $subAccountCreateResult[2] . ").";
            return undef;
          }
        }
        elsif ($subAccountCreateResult[0] == -3)
        {
          # sub-account already exists with the same number; map "new" category to existing sub-account if type same
          if ($subAccountCreateResult[1] == 50)
          {
            $subAccountNumbers[$rowCount] = $nextSubAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Sub-account already exists with the same number!  Type does not match (Type=" . $subAccountCreateResult[1] . ", Name=" . $subAccountObj->name . ", Number=" . $subAccountObj->number . ").";
            return undef;
          }
        }
        elsif (not $subAccountCreateResult[0])
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to create a new sub-account!<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        else
        {
          # creation successful; map new sub-account number to subAccountNumber[] array element
          $subAccountNumbers[$rowCount] = $nextSubAccountNumber;
        }
      }
      # update category mapping in QIF table
      my $sth = $self->{dbHandle}->write(sql => "UPDATE qif_category_tb SET map_account = ?, map_sub_account = ? WHERE name = ?", plug => [ $accountNumbers[$rowCount], $subAccountNumbers[$rowCount], $categoryNames[$rowCount] ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to update the category mapping! (" . $categoryNames[$rowCount] . ")<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF Map Categories");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Categories - Successful</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_map_categories\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"");
    if ($input->{'category_offset'} + $self->{limitRange} >= $input->{'max_categories'})
    {
      $doc->print("guess_accounts");
    }
    else
    {
      $doc->print("guess_categories");
    }
    $doc->print("\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
    $doc->print("<input type=hidden name=\"max_categories\" value=\"" . $input->{'max_categories'} . "\">\n");
    $doc->print("<input type=hidden name=\"category_offset\" value=\"" . ($input->{'category_offset'} + $self->{limitRange}) . "\">\n");
    if ($input->{'category_offset'} + $self->{limitRange} >= $input->{'max_categories'})
    {
      $doc->print("<p>Mapping: <b>Complete</b></p>\n");
    }
    else
    {
      $doc->print("<p>Mapped: <b>" . (($input->{'category_offset'} + $self->{limitRange}) > $input->{'max_categories'} ? $input->{'max_categories'} : ($input->{'category_offset'} + $self->{limitRange})) . "</b> of <b>" . $input->{'max_categories'} . "</b></p>\n");
    }
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_map_categories.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_categories\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    for ($rowCount=0; $rowCount < scalar @categoryNames; $rowCount++)
    {
      $doc->print("\"" . $categoryNames[$rowCount] . "\"&nbsp;=>&nbsp;\"" . $accountNumbers[$rowCount] . "\",&nbsp;\"" . $subAccountNumbers[$rowCount] . "\"<br>\n");
    }
    $doc->print("</center>\n");
  }
  elsif ($command eq "guess_accounts")
  {
    my @accountNames = ();
    my @accountDescriptions = ();
    my @accountTypes = ();
    my $maxAccounts = 0;
    my $accountOffset = 0;

    my $sth2;
    if ($input->{'use_entities'} eq "all")
    {
      $sth2 = $self->{dbHandle}->read(sql => "SELECT count(*) FROM qif_account_tb");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the maximum number of accounts to map!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth2 = $self->{dbHandle}->read(sql => "SELECT count(*) FROM qif_account_tb WHERE used = '1'");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the maximum number of used accounts to map!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    my @info = $sth2->fetchrow_array;
    $maxAccounts = $info[0];
    $accountOffset = (exists $input->{'account_offset'}) ? $input->{'account_offset'} : 0;
    my $sth;
    if ($input->{'use_entities'} eq "all")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name, description, type FROM qif_account_tb ORDER BY name LIMIT " . $self->{limitRange} . " OFFSET $accountOffset");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get all the accounts from the QIF table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    elsif ($input->{'use_entities'} eq "used")
    {
      $sth = $self->{dbHandle}->read(sql => "SELECT name, description, type FROM qif_account_tb WHERE used = '1' ORDER BY name LIMIT " . $self->{limitRange} . " OFFSET $accountOffset");
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to get all the used accounts from the QIF table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    while (my @info = $sth->fetchrow_array)
    {
      $accountNames[scalar @accountNames] = $info[0];
      $accountDescriptions[scalar @accountDescriptions] = $info[1];
      $accountTypes[scalar @accountTypes] = $info[2];
    }
    if (scalar @accountNames)
    {
      # get a list of all accounts of account types in the system that can be used to map QIF accounts
      my $accountType = "3"; # investment
      my @investmentAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (investment)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "0"; # cash/bank account
      my @cashAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (cash/bank account)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "5"; # current liability
      my @currentLiabilityAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (current liability)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "6"; # long term liability
      my @longTermLiabilityAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (long term liability)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "1"; # current asset
      my @currentAssetAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (current asset)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "2"; # long term asset
      my @longTermAssetAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (long term asset)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "111"; # account receivable
      my @accountReceivableAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (account receivable)<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      $accountType = "15"; # income tax
      my @incomeTaxAccounts = $dbObject->getListOfAccounts(location => "global", type => "$accountType", active => "true", orderBy => "type, name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (income tax)<br>\n" . $dbObject->errorMessage;
        return undef;
      }

      $doc->setFocus("javascript");
      $doc->print("var investmentAccounts = new Array();\n");
      $doc->print("var cashAccounts = new Array();\n");
      $doc->print("var currentLiabilityAccounts = new Array();\n");
      $doc->print("var longTermLiabilityAccounts = new Array();\n");
      $doc->print("var currentAssetAccounts = new Array();\n");
      $doc->print("var longTermAssetAccounts = new Array();\n");
      $doc->print("var accountReceivableAccounts = new Array();\n");
      $doc->print("var incomeTaxAccounts = new Array();\n");
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @investmentAccounts; $tempCount++)
      {
        $doc->print("investmentAccounts[$tempCount] = new Array();\n");
        $doc->print("investmentAccounts[$tempCount][0] = \"" . $investmentAccounts[$tempCount]->name . "\";\n");
        $doc->print("investmentAccounts[$tempCount][1] = \"" . $investmentAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @cashAccounts; $tempCount++)
      {
        $doc->print("cashAccounts[$tempCount] = new Array();\n");
        $doc->print("cashAccounts[$tempCount][0] = \"" . $cashAccounts[$tempCount]->name . "\";\n");
        $doc->print("cashAccounts[$tempCount][1] = \"" . $cashAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @currentLiabilityAccounts; $tempCount++)
      {
        $doc->print("currentLiabilityAccounts[$tempCount] = new Array();\n");
        $doc->print("currentLiabilityAccounts[$tempCount][0] = \"" . $currentLiabilityAccounts[$tempCount]->name . "\";\n");
        $doc->print("currentLiabilityAccounts[$tempCount][1] = \"" . $currentLiabilityAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @longTermLiabilityAccounts; $tempCount++)
      {
        $doc->print("longTermLiabilityAccounts[$tempCount] = new Array();\n");
        $doc->print("longTermLiabilityAccounts[$tempCount][0] = \"" . $longTermLiabilityAccounts[$tempCount]->name . "\";\n");
        $doc->print("longTermLiabilityAccounts[$tempCount][1] = \"" . $longTermLiabilityAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @currentAssetAccounts; $tempCount++)
      {
        $doc->print("currentAssetAccounts[$tempCount] = new Array();\n");
        $doc->print("currentAssetAccounts[$tempCount][0] = \"" . $currentAssetAccounts[$tempCount]->name . "\";\n");
        $doc->print("currentAssetAccounts[$tempCount][1] = \"" . $currentAssetAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @longTermAssetAccounts; $tempCount++)
      {
        $doc->print("longTermAssetAccounts[$tempCount] = new Array();\n");
        $doc->print("longTermAssetAccounts[$tempCount][0] = \"" . $longTermAssetAccounts[$tempCount]->name . "\";\n");
        $doc->print("longTermAssetAccounts[$tempCount][1] = \"" . $longTermAssetAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @accountReceivableAccounts; $tempCount++)
      {
        $doc->print("accountReceivableAccounts[$tempCount] = new Array();\n");
        $doc->print("accountReceivableAccounts[$tempCount][0] = \"" . $accountReceivableAccounts[$tempCount]->name . "\";\n");
        $doc->print("accountReceivableAccounts[$tempCount][1] = \"" . $accountReceivableAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      for (my $tempCount=0; $tempCount < scalar @incomeTaxAccounts; $tempCount++)
      {
        $doc->print("incomeTaxAccounts[$tempCount] = new Array();\n");
        $doc->print("incomeTaxAccounts[$tempCount][0] = \"" . $incomeTaxAccounts[$tempCount]->name . "\";\n");
        $doc->print("incomeTaxAccounts[$tempCount][1] = \"" . $incomeTaxAccounts[$tempCount]->number . "\";\n");
      }
      $doc->print("\n");
      $doc->print("function validateForm()\n");
      $doc->print("{\n");
      $doc->print("  var errorString = \"\";\n");
      $doc->print("\n");
      for (my $rowCount=0; $rowCount < scalar @accountNames; $rowCount++)
      {
        $doc->print("    if (document.qif_guess_accounts.account" . $rowCount . "[1].checked && (document.qif_guess_accounts.account_text" . $rowCount . ".value == '' || document.qif_guess_accounts.account_text" . $rowCount . ".value == 'unknown_account'))\n");
        $doc->print("    {\n");
        $doc->print("      errorString += '\\nThere must be a non-empty (and non unknown_account) value for the account for mapping \"" . $accountNames[$rowCount] . "\".';\n");
        $doc->print("    }\n");
      }
      $doc->print("\n");
      $doc->print("  if (errorString != \"\")\n");
      $doc->print("  {\n");
      $doc->print("    alert(errorString);\n");
      $doc->print("    return false;\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  return true;\n");
      $doc->print("}\n");
      $doc->print("\n");
      $doc->print("function fillAccountSelect(rowCount)\n");
      $doc->print("{\n");
      $doc->print("  eval(\"var type = document.qif_guess_accounts.account_type\" + rowCount + \".value\");\n");
      $doc->print("  eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options.length = 0;\");\n");
      # set the values of the select drop-downs based on the type and taxability of the category
      $doc->print("  if (type == 'Bank' || type == 'Cash')\n"); # cash/bank account
      $doc->print("  {\n");
      $doc->print("    for (var tempCount=0; tempCount < cashAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[tempCount] = new Option('\" + cashAccounts[tempCount][0] + \"', '\" + cashAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else if (type == 'CCard')\n"); # current liability
      $doc->print("  {\n");
      $doc->print("    for (var tempCount=0; tempCount < currentLiabilityAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[tempCount] = new Option('\" + currentLiabilityAccounts[tempCount][0] + \"', '\" + currentLiabilityAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else if (type == 'Invst' || type == 'Port' || type == 'Mutual' || type == '401(k)')\n"); # investment
      $doc->print("  {\n");
      $doc->print("    for (var tempCount=0; tempCount < investmentAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[tempCount] = new Option('\" + investmentAccounts[tempCount][0] + \"', '\" + investmentAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else if (type == 'Invoice' || type == 'Bill')\n"); # current asset
      $doc->print("  {\n");
      $doc->print("    for (var tempCount=0; tempCount < currentAssetAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[tempCount] = new Option('\" + currentAssetAccounts[tempCount][0] + \"', '\" + currentAssetAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else if (type == 'Tax')\n"); # income tax
      $doc->print("  {\n");
      $doc->print("    for (var tempCount=0; tempCount < incomeTaxAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[tempCount] = new Option('\" + incomeTaxAccounts[tempCount][0] + \"', '\" + incomeTaxAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else if (type.search(/Oth.A/) > -1)\n"); # other asset
      $doc->print("  {\n");
      $doc->print("    var optionCount = 0;\n");
      $doc->print("    for (var tempCount=0; tempCount < currentAssetAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + currentAssetAccounts[tempCount][0] + \"', '\" + currentAssetAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < longTermAssetAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + longTermAssetAccounts[tempCount][0] + \"', '\" + longTermAssetAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < accountReceivableAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + accountReceivableAccounts[tempCount][0] + \"', '\" + accountReceivableAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else if (type.search(/Oth.L/) > -1)\n"); # other liability
      $doc->print("  {\n");
      $doc->print("    var optionCount = 0;\n");
      $doc->print("    for (var tempCount=0; tempCount < currentLiabilityAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + currentLiabilityAccounts[tempCount][0] + \"', '\" + currentLiabilityAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < longTermLiabilityAccounts.length; tempCount++)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + longTermLiabilityAccounts[tempCount][0] + \"', '\" + longTermLiabilityAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  else\n"); # unknown type; display all known accounts that can be mapped
      $doc->print("  {\n");
      $doc->print("    var optionCount = 0;\n");
      $doc->print("    for (var tempCount=0; tempCount < investmentAccounts.length; tempCount++)\n"); # investment
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + investmentAccounts[tempCount][0] + \"', '\" + investmentAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < cashAccounts.length; tempCount++)\n"); # cash/bank account
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + cashAccounts[tempCount][0] + \"', '\" + cashAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < currentLiabilityAccounts.length; tempCount++)\n"); # current liability
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + currentLiabilityAccounts[tempCount][0] + \"', '\" + currentLiabilityAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < longTermLiabilityAccounts.length; tempCount++)\n"); # long term liability
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + longTermLiabilityAccounts[tempCount][0] + \"', '\" + longTermLiabilityAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < currentAssetAccounts.length; tempCount++)\n"); # current asset
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + currentAssetAccounts[tempCount][0] + \"', '\" + currentAssetAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < longTermAssetAccounts.length; tempCount++)\n"); # long term asset
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + longTermAssetAccounts[tempCount][0] + \"', '\" + longTermAssetAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("    for (var tempCount=0; tempCount < incomeTaxAccounts.length; tempCount++)\n"); # income tax
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[optionCount++] = new Option('\" + incomeTaxAccounts[tempCount][0] + \"', '\" + incomeTaxAccounts[tempCount][1] + \"');\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("  var tempAccount = eval(\"document.qif_guess_accounts.account_name\" + rowCount + \".value\");\n");
      # attempt to find a match for the account name
      $doc->print("  if (tempAccount != '')\n");
      $doc->print("  {\n");
      $doc->print("    var matchFound = -1;\n");
      $doc->print("    var accountRegExp = new RegExp(\"\" + tempAccount);\n");
      $doc->print("    for (var count=0; count < eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options.length\"); count++)\n");
      $doc->print("    {\n");
      $doc->print("      if (eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[count].text.match(accountRegExp)\"))\n");
      $doc->print("      {\n");
      $doc->print("        matchFound = count;\n");
      $doc->print("      }\n");
      $doc->print("    }\n");
      $doc->print("    if (matchFound > -1)\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_select\" + rowCount + \".options[\" + matchFound + \"].selected = true;\");\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account\" + rowCount + \"[0].checked = true;\");\n");
      $doc->print("    }\n");
      $doc->print("    else\n");
      $doc->print("    {\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account_text\" + rowCount + \".value = tempAccount;\");\n");
      $doc->print("      eval(\"document.qif_guess_accounts.account\" + rowCount + \"[1].checked = true;\");\n");
      $doc->print("    }\n");
      $doc->print("  }\n");
      $doc->print("}\n");

      # guess the account for the QIF account
      my $onLoadCode = "var numAccountNames = " . (scalar @accountNames) . ";\n";
      $onLoadCode .= "\n";
      $onLoadCode .= "for (var rowCount=0; rowCount < numAccountNames; rowCount++)\n";
      $onLoadCode .= "{\n";
      $onLoadCode .= "  fillAccountSelect(rowCount);\n";
      $onLoadCode .= "}\n";
      $doc->setOnload(code => $onLoadCode);

      $doc->setTitle("QIF Map Accounts");
      $doc->setFocus("body");
      $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
      $doc->print("<center>\n");
      $doc->print("<h1>QIF Map Accounts</h1><br>\n");
      $doc->print("<form method=\"post\" name=\"qif_guess_accounts\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
      $doc->print("$staticInfo");
      $doc->print("<input type=hidden name=\"command\" value=\"map_accounts\">\n");
      $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
      $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
      $doc->print("<input type=hidden name=\"max_accounts\" value=\"$maxAccounts\">\n");
      $doc->print("<input type=hidden name=\"account_offset\" value=\"" . $accountOffset . "\">\n");
      $doc->print("<p>Working with accounts: <b>" . ($accountOffset + 1) . "</b> - <b>" . (($accountOffset + $self->{limitRange} > $maxAccounts) ? $maxAccounts : ($accountOffset + $self->{limitRange})) . "</b> of <b>" . $maxAccounts . "</b></p>\n");
      $doc->print("<table border=1 cellspacing=0 cellpadding=2>\n");
      $doc->print("  <tr>\n");
      $doc->print("    <th align=center>QIF Account Name</th><th align=center>Existing Account</th><th align=center>Create Account</th><th align=center>LT</th>\n");
      $doc->print("  </tr>\n");
      for (my $rowCount=0; $rowCount < scalar @accountNames; $rowCount++)
      {
        $doc->print("  <tr>\n");
        $doc->print("    <td><b>$accountNames[$rowCount]</b>");
        if ($accountTypes[$rowCount] eq "")
        {
          $doc->print("&nbsp;(<select name=\"account_type_select$rowCount\" onChange=\"document.qif_guess_accounts.account_type$rowCount.value = document.qif_guess_accounts.account_type_select$rowCount" . "[document.qif_guess_accounts.account_type_select$rowCount.selectedIndex" . "].value; fillAccountSelect($rowCount);\">\n");
          $doc->print("        <option value=\"Bank\" selected>Bank\n");
          $doc->print("        <option value=\"Cash\">Cash\n");
          $doc->print("        <option value=\"CCard\">CCard\n");
          $doc->print("        <option value=\"Invst\">Invst\n");
          $doc->print("        <option value=\"Oth A\">Oth A\n");
          $doc->print("        <option value=\"Oth L\">Oth L\n");
          $doc->print("        <option value=\"Invoice\">Invoice\n");
          $doc->print("        <option value=\"Tax\">Tax\n");
          $doc->print("        <option value=\"Bill\">Bill\n");
          $doc->print("      </select>)");
        }
        $doc->print("<br>$accountDescriptions[$rowCount]\n");
        $doc->print("      <input type=hidden name=\"account_name$rowCount\" value=\"$accountNames[$rowCount]\">\n");
        $doc->print("      <input type=hidden name=\"account_type$rowCount\" value=\"$accountTypes[$rowCount]\">\n");
        $doc->print("    </td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"account$rowCount\" value=\"exist\"><select name=\"account_select$rowCount\" onChange=\"document.qif_guess_accounts.account$rowCount" . "[0].checked = true;\">\n");
        $doc->print("        <option value=\"-1\">--------------------\n");
        $doc->print("      </select></td>\n");
        $doc->print("    <td align=center nowrap><input type=radio name=\"account$rowCount\" value=\"new\">");
        $doc->print("<input type=text name=\"account_text$rowCount\" value=\"\" onChange=\"document.qif_guess_accounts.account$rowCount" . "[1].checked = true;\"></td>");
        if ($accountTypes[$rowCount] =~ /Oth/ || $accountTypes[$rowCount] eq "")
        {
          $doc->print("    <td align=center nowrap><input type=checkbox name=\"account_lt$rowCount\"></td>\n");
        }
        else
        {
          $doc->print("    <td>&nbsp;</td>\n");
        }
        $doc->print("  </tr>\n");
      }
      $doc->print("</table>\n");
      $doc->print("<input type=submit value=\"Continue...\" onClick=\"return validateForm();\">&nbsp;<input type=button value=\"Reset\" onClick=\"document.qif_guess_accounts.reset(); doOnLoad();\">\n");
      $doc->print("</form>\n");
      $doc->print("</center>\n");
    }
    else
    {
      # if there are no accounts in the database to map, skip the account insertion code and go to the next step in the process
      $doc->setTitle("QIF Map Accounts");
      $doc->setFocus("body");
      $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
      $doc->print("<center>\n");
      $doc->print("<h1>QIF Map Accounts</h1><br>\n");
      $doc->print("No accounts exist in the database.  This most likely means that no accounts existed in the QIF import file to parse.<br><br>\n");
      $doc->print("Since no accounts exist (not even the unknown_account account), the QIF input file data cannot be inserted into the database!<br><br>\n");
      $doc->print("</center>\n");
      $doc->print("<form method=post action=\"$site_dir/cgi-bin/prg_config.cgi\" name=\"prg_config_form\">\n");
      $doc->print("<input type=hidden name=\"ledger\" value=\"$ledger\">\n");
      $doc->print("<a href=\"$siteDir/js/Accounting/accounting/cgi-bin/prg_config.cgi\" onClick=\"javascript:document.prg_config_form.submit(); return false;\">Configuration</a>\n");
    }
  }
  elsif ($command eq "map_accounts")
  {
    # get the list of account names from the input fields and their values to map in the QIF database tables
    my @accountNames = ();
    my @accountTypes = ();
    my @taxable = ();
    my @accountNumbers = ();
    my @accountExistNew = ();
    my $rowCount = 0;
    while (exists $input->{"account_name$rowCount"})
    {
      $accountNames[$rowCount] = $input->{"account_name$rowCount"};
      if ($input->{"account_type$rowCount"} eq "Bank" || $input->{"account_type$rowCount"} eq "Cash")
      {
        $accountTypes[$rowCount] = 0; # cash/bank account
        $taxable[$rowCount] = "false";
      }
      elsif ($input->{"account_type$rowCount"} eq "Invst" || $input->{"account_type$rowCount"} eq "Port" || $input->{"account_type$rowCount"} eq "Mutual" || $input->{"account_type$rowCount"} eq "401(k)")
      {
        $accountTypes[$rowCount] = 3; # investment
        $taxable[$rowCount] = "false";
      }
      elsif ($input->{"account_type$rowCount"} eq "CCard")
      {
        $accountTypes[$rowCount] = 5; # current liability
        $taxable[$rowCount] = "false";
      }
      elsif ($input->{"account_type$rowCount"} eq "Invoice" || $input->{"account_type$rowCount"} eq "Bill")
      {
        $accountTypes[$rowCount] = 1; # current asset
        $taxable[$rowCount] = "false";
      }
      elsif ($input->{"account_type$rowCount"} eq "Tax")
      {
        $accountTypes[$rowCount] = 15; # income tax
        $taxable[$rowCount] = "true";
      }
      elsif ($input->{"account_type$rowCount"} =~ /Oth.A/)
      {
        if (exists $input->{"account_lt$rowCount"})
        {
          $accountTypes[$rowCount] = 2; # long term asset
        }
        else
        {
          $accountTypes[$rowCount] = 1; # current asset
        }
        $taxable[$rowCount] = "false";
      }
      elsif ($input->{"account_type$rowCount"} =~ /Oth.L/)
      {
        if (exists $input->{"account_lt$rowCount"})
        {
          $accountTypes[$rowCount] = 6; # long term liability
        }
        else
        {
          $accountTypes[$rowCount] = 5; # current liability
        }
        $taxable[$rowCount] = "false";
      }
      if ($input->{"account$rowCount"} eq "exist")
      {
        $accountNumbers[$rowCount] = $input->{"account_select$rowCount"};
        $accountExistNew[$rowCount] = 1;
      }
      else
      {
        $accountNumbers[$rowCount] = $input->{"account_text$rowCount"};
        $accountExistNew[$rowCount] = 0;
      }
      $rowCount++;
    }
    # update account mappings; create any non-existent accounts
    # based on the type, create an appropriate account of that type
    my $accountLength = $dbObject->getAccountLength;
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the minimum account number length.<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    for ($rowCount=0; $rowCount < scalar @accountNames; $rowCount++)
    {
      if (not $accountExistNew[$rowCount])
      {
        # create a new account; if it already exists, check to see if it is of the same type as the account to be created;
        # if it is, simply assign that account without creating a new account
        my $sth = $self->{dbHandle}->read(sql => "SELECT first_digit FROM acct_type_tb WHERE num = ?", plug => [ $accountTypes[$rowCount] ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to get the first digit for account type (" . $accountTypes[$rowCount] . ").<br>\n" . $self->{dbHandle}->errorMessage;
          return undef;
        }
        my @info = $sth->fetchrow_array;
        if ((not defined @info) || (not defined $info[0]) || $info[0] !~ /^\d+$/)
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>First digit ($info[0]) invalid when attempting to retrieve for $accountTypes[$rowCount] account type.\n";
          return undef;
        }
        my $nextAccountNumber = $dbObject->getNextAvailableAccountNumber(firstDigit => $info[0], minLength => $accountLength);
        if ($dbObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Attempted to get the next available account number<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        my $accountObj = Portal::Accounting::Objects::Account->new(name => $accountNumbers[$rowCount], number => $nextAccountNumber, type => $accountTypes[$rowCount], taxable => $taxable[$rowCount], minAccountLength => $accountLength);
        if ($accountObj->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error creating new account object.<br>\n" . $accountObj->errorMessage;
          return undef;
        }
        my @accountCreateResult = $dbObject->createAccount(accountObj => $accountObj);
        if ($accountCreateResult[0] == -1)
        {
          # account already exists with the same name and number; map "new" category to existing account if type same
          if ($accountCreateResult[1] == $accountTypes[$rowCount])
          {
            $accountNumbers[$rowCount] = $nextAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Account already exists with the same name and number!  Type does not match (Type=" . $accountCreateResult[1] . ", Taxable=" . $accountCreateResult[2] . ", Account Name=" . $accountObj->name . ", Account Number=" . $accountObj->number . ").";
            return undef;
          }
        }
        elsif ($accountCreateResult[0] == -2)
        {
          # account already exists with the same name; map "new" category to existing account if type same
          if ($accountCreateResult[1] == $accountTypes[$rowCount])
          {
            $accountNumbers[$rowCount] = $accountCreateResult[3];
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Account already exists with the same name!  Type does not match (Type=" . $accountCreateResult[1] . ", Taxable=" . $accountCreateResult[2] . ", Account Name=" . $accountObj->name . ", Account Number=" . $accountCreateResult[3] . ").";
            return undef;
          }
        }
        elsif ($accountCreateResult[0] == -3)
        {
          # account already exists with the same number; map "new" category to existing account if type same
          if ($accountCreateResult[1] == $accountTypes[$rowCount])
          {
            $accountNumbers[$rowCount] = $nextAccountNumber;
          }
          else
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Account already exists with the same number!  Type does not match (Type=" . $accountCreateResult[1] . ", Taxable=" . $accountCreateResult[2] . ", Account Name=" . $accountObj->name . ", Account Number=" . $accountCreateResult[3] . ").";
            return undef;
          }
        }
        elsif (not $accountCreateResult[0])
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Tried to create a new account!<br>\n" . $dbObject->errorMessage;
          return undef;
        }
        else
        {
          # creation successful; map new account number to accountNumbers[] array element
          $accountNumbers[$rowCount] = $nextAccountNumber;
          # assign the newly created account to the ledger being worked with
          if ($dbObject->assignAccountToLedger(accountObj => $accountObj, ledgerCode => $ledgerCode) == 0)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Attempted to assign the newly created account (" . $accountObj->name . " - " . $accountObj->number . ") to the ledger ($ledgerCode)!<br>\n" . $dbObject->errorMessage;
            return undef;
          }
        }
      }
      # update account mapping in QIF table
      my $sth = $self->{dbHandle}->write(sql => "UPDATE qif_account_tb SET map_account = ? WHERE name = ?", plug => [ $accountNumbers[$rowCount], $accountNames[$rowCount] ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Tried to update the account mapping! (" . $accountNames[$rowCount] . ")<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }

    # everything looks good
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF Map Accounts");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Accounts - Successful</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_map_accounts\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"");
    if ($input->{'account_offset'} + $self->{limitRange} >= $input->{'max_accounts'})
    {
      $doc->print("select_import_accounts");
    }
    else
    {
      $doc->print("guess_accounts");
    }
    $doc->print("\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    $doc->print("<input type=hidden name=\"use_entities\" value=\"" . $input->{'use_entities'} . "\">\n");
    $doc->print("<input type=hidden name=\"max_accounts\" value=\"" . $input->{'max_accounts'} . "\">\n");
    $doc->print("<input type=hidden name=\"account_offset\" value=\"" . ($input->{'account_offset'} + $self->{limitRange}) . "\">\n");
    if ($input->{'account_offset'} + $self->{limitRange} >= $input->{'max_accounts'})
    {
      $doc->print("<p>Mapping: <b>Complete</b></p>\n");
    }
    else
    {
      $doc->print("<p>Mapped: <b>" . (($input->{'account_offset'} + $self->{limitRange}) > $input->{'max_accounts'} ? $input->{'max_accounts'} : ($input->{'account_offset'} + $self->{limitRange})) . "</b> of <b>" . $input->{'max_accounts'} . "</b></p>\n");
    }
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_map_accounts.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_accounts\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    for ($rowCount=0; $rowCount < scalar @accountNames; $rowCount++)
    {
      $doc->print("\"" . $accountNames[$rowCount] . "\"&nbsp;=>&nbsp;\"" . $accountNumbers[$rowCount] . "\"<br>\n");
    }
    $doc->print("</center>\n");
  }
  elsif ($command eq "select_import_accounts")
  {
    my @equityAccounts = $dbObject->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => "7", active => "true", orderBy => "type, name");
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (Equity)!<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    my @dividendRevenueAccounts = $dbObject->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => "11", active => "true", orderBy => "type, name");
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (Dividend Revenue)!<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    my @interestRevenueAccounts = $dbObject->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => "12", active => "true", orderBy => "type, name");
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (Interest Revenue)!<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    my @otherRevenueAccounts = $dbObject->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => "9", active => "true", orderBy => "type, name");
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (Other Revenue)!<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    my @otherExpenseAccounts = $dbObject->getListOfAccounts(location => "local", ledgerCode => $ledgerCode, type => "13", active => "true", orderBy => "type, name");
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of accounts (Other Expense)!<br>\n" . $dbObject->errorMessage;
      return undef;
    }
    my @subEntities = $dbObject->getArrayOfSubEntities(ledgerCode => $ledgerCode);
    if ($dbObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving list of sub-entities!<br>\n" . $dbObject->errorMessage;
      return undef;
    }

    $doc->setTitle("QIF Select Import Accounts");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Select Import Accounts</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_select_accounts\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"import_account_transactions\">\n");
    $doc->print("<p>Choose from the following accounts to be used for importing QIF data:</p>\n");
    $doc->print("<table border=\"1\" cellspacing=\"0\" cellpadding=\"2\">\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right><b>Equity</b> Account</td>\n");
    $doc->print("    <td align=left><select name=\"equity_account\">\n");
    for (my $count=0; $count < scalar @equityAccounts; $count++)
    {
      $doc->print("      <option value=\"" . $equityAccounts[$count]->number . "\">" . $equityAccounts[$count]->name . " (" . $equityAccounts[$count]->number . ")");
      if ($equityAccounts[$count]->taxable eq "true")
      {
        $doc->print(" ~");
      }
      $doc->print("\n");
    }
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right><b>Dividend Revenue</b> Account</td>\n");
    $doc->print("    <td align=left><select name=\"dividend_revenue_account\">\n");
    for (my $count=0; $count < scalar @dividendRevenueAccounts; $count++)
    {
      $doc->print("      <option value=\"" . $dividendRevenueAccounts[$count]->number . "\">" . $dividendRevenueAccounts[$count]->name . " (" . $dividendRevenueAccounts[$count]->number . ")");
      if ($dividendRevenueAccounts[$count]->taxable eq "true")
      {
        $doc->print(" ~");
      }
      $doc->print("\n");
    }
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right><b>Interest Revenue</b> Account</td>\n");
    $doc->print("    <td align=left><select name=\"interest_revenue_account\">\n");
    for (my $count=0; $count < scalar @interestRevenueAccounts; $count++)
    {
      $doc->print("      <option value=\"" . $interestRevenueAccounts[$count]->number . "\">" . $interestRevenueAccounts[$count]->name . " (" . $interestRevenueAccounts[$count]->number . ")");
      if ($interestRevenueAccounts[$count]->taxable eq "true")
      {
        $doc->print(" ~");
      }
      $doc->print("\n");
    }
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right><b>Other Revenue</b> Account</td>\n");
    $doc->print("    <td align=left><select name=\"other_revenue_account\">\n");
    for (my $count=0; $count < scalar @otherRevenueAccounts; $count++)
    {
      $doc->print("      <option value=\"" . $otherRevenueAccounts[$count]->number . "\">" . $otherRevenueAccounts[$count]->name . " (" . $otherRevenueAccounts[$count]->number . ")");
      if ($otherRevenueAccounts[$count]->taxable eq "true")
      {
        $doc->print(" ~");
      }
      $doc->print("\n");
    }
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right><b>Other Expense</b> Account</td>\n");
    $doc->print("    <td align=left><select name=\"other_expense_account\">\n");
    for (my $count=0; $count < scalar @otherExpenseAccounts; $count++)
    {
      $doc->print("      <option value=\"" . $otherExpenseAccounts[$count]->number . "\">" . $otherExpenseAccounts[$count]->name . " (" . $otherExpenseAccounts[$count]->number . ")");
      if ($otherExpenseAccounts[$count]->taxable eq "true")
      {
        $doc->print(" ~");
      }
      $doc->print("\n");
    }
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("  <tr>\n");
    $doc->print("    <td align=right>Import <b>Sub-Entity</b></td>\n");
    $doc->print("    <td align=left><select name=\"import_sub_entity\">\n");
    $doc->print("      <option value=\"\">--\n");
    for (my $count=0; $count < scalar @subEntities; $count++)
    {
      $doc->print("      <option value=\"" . $subEntities[$count]->{code} . "\">" . $subEntities[$count]->{name} . "\n");
    }
    $doc->print("    </select></td>\n");
    $doc->print("  </tr>\n");
    $doc->print("</table>\n");
    $doc->print("<input type=submit name=\"submit_accounts\" value=\"Continue...\" onClick=\"return validateForm();\">\n");
    $doc->print("</form>\n");
    $doc->print("</center>\n");

    $doc->setFocus("javascript");
    $doc->print("function validateForm()\n");
    $doc->print("{\n");
    $doc->print("  var errorString = \"\";\n");
    $doc->print("\n");
    $doc->print("  if (document.qif_select_accounts.equity_account.selectedIndex == -1)\n");
    $doc->print("  {\n");
    $doc->print("    errorString += \"You must select an Equity account!\\nIf there are no equity accounts in the list,\\nyou will need to create and/or assign\\nan equity account to the ledger\\nin which you are working.\";\n");
    $doc->print("  }\n");
    $doc->print("  if (document.qif_select_accounts.dividend_revenue_account.selectedIndex == -1)\n");
    $doc->print("  {\n");
    $doc->print("    errorString += \"You must select a Dividend Revenue account!\\nIf there are no dividend revenue accounts in the list,\\nyou will need to create and/or assign\\na dividend revenue account to the ledger\\nin which you are working.\";\n");
    $doc->print("  }\n");
    $doc->print("  if (document.qif_select_accounts.interest_revenue_account.selectedIndex == -1)\n");
    $doc->print("  {\n");
    $doc->print("    errorString += \"You must select an Interest Revenue account!\\nIf there are no interest revenue accounts in the list,\\nyou will need to create and/or assign\\nan interest revenue account to the ledger\\nin which you are working.\";\n");
    $doc->print("  }\n");
    $doc->print("  if (document.qif_select_accounts.other_revenue_account.selectedIndex == -1)\n");
    $doc->print("  {\n");
    $doc->print("    errorString += \"You must select an Other Revenue account!\\nIf there are no other revenue accounts in the list,\\nyou will need to create and/or assign\\nan other revenue account to the ledger\\nin which you are working.\";\n");
    $doc->print("  }\n");
    $doc->print("  if (document.qif_select_accounts.other_expense_account.selectedIndex == -1)\n");
    $doc->print("  {\n");
    $doc->print("    errorString += \"You must select an Other Expense account!\\nIf there are no other expense accounts in the list,\\nyou will need to create and/or assign\\nan other expense account to the ledger\\nin which you are working.\";\n");
    $doc->print("  }\n");
    $doc->print("  if (document.qif_select_accounts.import_sub_entity.selectedIndex == -1)\n");
    $doc->print("  {\n");
    $doc->print("    errorString += \"You must select an Sub-Entity!\\nIf there are no sub-entities in the list,\\nyou will need to create and/or assign\\na sub-entity to the ledger\\nin which you are working.\";\n");
    $doc->print("  }\n");
    $doc->print("\n");
    $doc->print("  if (errorString != \"\")\n");
    $doc->print("  {\n");
    $doc->print("    alert(errorString);\n");
    $doc->print("    return false;\n");
    $doc->print("  }\n");
    $doc->print("\n");
    $doc->print("  return true;\n");
    $doc->print("}\n");
  }
  elsif ($command eq "import_buy_investment_transactions")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # prepare for the (likely) event that we will run across Buy, BuyX, or ShrsIn records to import
    # create Security object
    my $secDBObject = Portal::Accounting::DB::Security->new(dbiw => $self->{dbHandle});
    if ($secDBObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error creating new Security database object!<br>\n" . $secDBObject->errorMessage;
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into tables in the database

    # reset any variables from a previous security list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'imported_buy_investment_transactions'} && $input->{'imported_buy_investment_transactions'} =~ /\d+/)
    {
      $self->{importedBuyInvestmentTransactions} = $input->{'imported_buy_investment_transactions'};
    }

    my $line = "";
    my $optionAutoSwitch = 0; # used to keep track of the number of times this export header has been encountered
    if (defined $input->{'option_auto_switch'} && $input->{'option_auto_switch'} =~ /\d+/)
    {
      $optionAutoSwitch = $input->{'option_auto_switch'};
    }
    my $clearAutoSwitch = 0; # used to keep track of the number of times this export header has been encountered
    if (defined $input->{'clear_auto_switch'} && $input->{'clear_auto_switch'} =~ /\d+/)
    {
      $clearAutoSwitch = $input->{'clear_auto_switch'};
    }
    my $currentAccount = "unknown_account"; # used to keep track of the current account to use for account transactions
    if (defined $input->{'current_account'} && $input->{'current_account'} ne "")
    {
      $currentAccount = $input->{'current_account'};
    }
    my $tempCounter = 0;
    my @parseResult = ();
    $parseResult[0] = "2"; # set the default return value to 'no valid records found'

    # import all Buy, BuyX, and ShrsIn transactions first in order to have all lots in the database for future insertions of events
    while (defined $line && $tempCounter < $self->{investmentTransactionBatchSize})
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if ($line =~ /!Option:AutoSwitch/)
        {
          $optionAutoSwitch++;
        }
        elsif ($line =~ /!Clear:AutoSwitch/)
        {
          $clearAutoSwitch++;
        }
        elsif ($line =~ /!Account/)
        {
          if ($optionAutoSwitch > 0 && $clearAutoSwitch > 0)
          {
            # individual accounts with transaction lists after their definition
            $currentAccount = $self->parseAccountRecord(fileHandle => $fileHandle);
            if ($currentAccount eq "")
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Account record! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
          }
        }
        elsif ($line =~ /!Type:Invst/ || $input->{'inside_investment_transaction_list'} eq "1")
        {
          if ($input->{'inside_investment_transaction_list'} eq "1")
          {
            $self->putBackLine($line);
            $input->{'inside_investment_transaction_list'} = 0;
          }
          else
          {
            $line = "";
          }
          while (defined $line && $line !~ /^\!/ && $tempCounter < $self->{investmentTransactionBatchSize})
          {
            @parseResult = $self->parseInvestmentTransaction(fileHandle => $fileHandle, importBuys => "1");
            if ($parseResult[0] eq "0")
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Investment Transaction! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
            if ($parseResult[0] eq "1")
            {
              $tempCounter++;
              $self->{importedBuyInvestmentTransactions}++;

              my %recordValues = %{ $parseResult[1] };

              my $transDate = $recordValues{'date'};
              my $action = $recordValues{'action'};
              my $security = $recordValues{'security'};
              my $price = $recordValues{'price'};
              $price = sprintf("%.6f", $price);
              my $quantity = $recordValues{'quantity'};
              my $commission = $recordValues{'commission'};
              $commission = sprintf("%.2f", $commission);
              my $transferAccount = $recordValues{'transfer_account'};
              my $transferClass = $recordValues{'transfer_class'};
              my $amount = $recordValues{'transaction_amount'};
              my $transferAmount = $recordValues{'transfer_amount'};
              my $qifAccount = $currentAccount;

              my $refNum;

              my $mapSubEntity = "";
              my $sth;
              my @info;
              # lookup the mapping for the class
              if ($transferClass ne "")
              {
                $sth = $self->{dbHandle}->read(sql => "SELECT map_sub_entity FROM qif_class_tb WHERE name = ?", plug => [ $transferClass ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for class ($transferClass)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for class ($transferClass)!<br>Make sure you have exported all Classes that are used in the investment data.";
                  return undef;
                }
                $mapSubEntity = $info[0];
              }
              else
              {
                # assign the default sub-entity mapping to the one chosen by the user
                $mapSubEntity = $input->{'import_sub_entity'};
              }
              # lookup the mapping for the QIF account associated with the event
              $sth = $self->{dbHandle}->read(sql => "SELECT map_account FROM qif_account_tb WHERE name = ?", plug => [ $qifAccount ]);
              if ($self->{dbHandle}->didErrorOccur())
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for QIF account ($qifAccount)!<br>\n" . $self->{dbHandle}->errorMessage;
                return undef;
              }
              @info = $sth->fetchrow_array;
              if (not defined $info[0])
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for QIF account ($qifAccount)!<br>Make sure you have exported all Accounts that are used in the investment data.";
                return undef;
              }
              my $mapQIFAccount = $info[0];
              # lookup the mapping for the QIF security associated with the event
              $sth = $self->{dbHandle}->read(sql => "SELECT map_sub_account FROM qif_security_tb WHERE name = ?", plug => [ $security ]);
              if ($self->{dbHandle}->didErrorOccur())
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for QIF security ($security)!<br>\n" . $self->{dbHandle}->errorMessage;
                return undef;
              }
              @info = $sth->fetchrow_array;
              if (not defined $info[0])
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for QIF security ($security)!<br>Make sure you have exported all Securities that are used in the investment data.";
                return undef;
              }
              my $mapQIFSubAccount = $info[0];

              if ($action eq "BuyX")
              {
                # lookup the mapping for the transfer account
                $sth = $self->{dbHandle}->read(sql => "SELECT map_account FROM qif_account_tb WHERE name = ?", plug => [ $transferAccount ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for transfer account ($transferAccount)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for transfer account ($transferAccount)!<br>Make sure you have exported all Accounts that are used in the investment data.";
                  return undef;
                }
                my $mapTransferAccount = $info[0];
                # do the cash transfer
                my $toAccount;
                my $fromAccount;
                if ($transferAmount < 0)
                {
                  $toAccount = $mapTransferAccount;
                  $fromAccount = $mapQIFAccount;
                }
                else
                {
                  $toAccount = $mapQIFAccount;
                  $fromAccount = $mapTransferAccount;
                }
                # beautify transferAmount
                $transferAmount = sprintf("%.2f", abs($transferAmount));
                $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                if ($self->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer!<br>\n" . $self->errorMessage;
                  return undef;
                }
              }
              elsif ($action eq "ShrsIn")
              {
                my $equityAccount = $input->{'equity_account'};
                if ($equityAccount !~ /^[0-9]+$/)
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Invalid equity account ($equityAccount)!";
                  return undef;
                }
                # create equity from nothing!
                my $toAccount;
                my $fromAccount;
                if ($amount < 0)
                {
                  $toAccount = $mapQIFAccount;
                  $fromAccount = $equityAccount;
                }
                else
                {
                  $toAccount = $equityAccount;
                  $fromAccount = $mapQIFAccount;
                }
                # beautify amount
                $amount = sprintf("%.2f", abs($amount));
                $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$amount");
                if ($self->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for creation of equity!<br>\n" . $self->errorMessage;
                  return undef;
                }
              }

              # create SecurityEventObject object
              my $secObj = Portal::Accounting::Objects::SecurityEventObject->new(eventCode => "10", calcDate => $transDate, transDate => $transDate,
                           subEntity => $mapSubEntity, accountNum => $mapQIFAccount, subAccountNum => $mapQIFSubAccount,
                           price => $price, shares => $quantity, commission => $commission, ledgerCode => $ledgerCode);
              # insert a buy security event
              my $eventRef = $secDBObject->insertBuy(secobj => $secObj);
              if ($secDBObject->didErrorOccur())
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Error inserting buy security event!<br>\n" . $secDBObject->errorMessage;
                return undef;
              }

              if ($action eq "ShrsIn")
              {
                # log ShrsIn entry
                $sth = $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES (?)", plug => [ "ShrsIn event: (Journal RefNum = $refNum, Security EventRef = $eventRef) D=$transDate, S=$security, P=$price, Q=$quantity, A=$amount" ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error logging ShrsIn transaction!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                # commit database transaction
                $self->{dbHandle}->commit;
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
              }
            }

            $line = $self->getNextLine(fileHandle => $fileHandle);
            if (defined $line)
            {
              $self->putBackLine($line);
            }
          }
        }
      }
    }

    $doc->setTitle("QIF Map Buy/BuyX/ShrsIn Transactions");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Map Buy/BuyX/ShrsIn Transactions</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_map_buy_transactions\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"import_investment_transactions\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"import_buy_investment_transactions\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"current_account\" value=\"$currentAccount\">\n");
      $doc->print("<input type=hidden name=\"option_auto_switch\" value=\"$optionAutoSwitch\">\n");
      $doc->print("<input type=hidden name=\"clear_auto_switch\" value=\"$clearAutoSwitch\">\n");
      if ($parseResult[0] eq "1" && $tempCounter >= $self->{investmentTransactionBatchSize})
      {
        $doc->print("<input type=hidden name=\"inside_investment_transaction_list\" value=\"1\">\n");
      }
      $doc->print("<input type=hidden name=\"imported_buy_investment_transactions\" value=\"" . $self->{importedBuyInvestmentTransactions} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"equity_account\" value=\"" . $input->{'equity_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"dividend_revenue_account\" value=\"" . $input->{'dividend_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"interest_revenue_account\" value=\"" . $input->{'interest_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"other_revenue_account\" value=\"" . $input->{'other_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"other_expense_account\" value=\"" . $input->{'other_expense_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"import_sub_entity\" value=\"" . $input->{'import_sub_entity'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"1\">\n");
    if ((not defined $line) && $self->{importedBuyInvestmentTransactions} eq "0")
    {
      $doc->print("<p>No Buy, BuyX, or ShrsIn transactions were found in the database!  This will make it very difficult to continue with the import of investment data.</p>\n");
    }
    else
    {
      # set the page to be submitted by JavaScript
      $doc->setFocus("javascript");
      my $onLoadCode = "document.qif_map_buy_transactions.submit();\n";
      $doc->setOnload(code => $onLoadCode);
      $doc->setFocus("body");
      $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    }
    $doc->print("<input type=submit name=\"submit_transactions\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" Buy/BuyX/ShrsIn investment transaction records processed: " . $self->{importedBuyInvestmentTransactions} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "import_account_transactions")
  {
    $currentAccount = (defined $input->{'current_account'}) ? $input->{'current_account'} : "";
    my $doc2 = $self->importAccountTransactions(currentAccount => $currentAccount, fileName => $input->{'qif_file_name'});
    if (not defined $doc2)
    {
      $self->{error} = 1;
      $self->{errorString} = "Could not import Account Transactions" . $self->{errorString};
      #unlink($input->{qif_file_name});
      return undef;
    }
    return $doc2;
  }
  elsif ($command eq "import_investment_transactions")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # prepare for the (likely) event that we will run across investment records to import
    # create Security object
    my $secDBObject = Portal::Accounting::DB::Security->new(dbiw => $self->{dbHandle});
    if ($secDBObject->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error creating new Security database object!<br>\n" . $secDBObject->errorMessage;
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into tables in the database

    # reset any variables from a previous security list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'imported_investment_transactions'} && $input->{'imported_investment_transactions'} =~ /\d+/)
    {
      $self->{importedInvestmentTransactions} = $input->{'imported_investment_transactions'};
    }

    my $line = "";
    my $optionAutoSwitch = 0; # used to keep track of the number of times this export header has been encountered
    if (defined $input->{'option_auto_switch'} && $input->{'option_auto_switch'} =~ /\d+/)
    {
      $optionAutoSwitch = $input->{'option_auto_switch'};
    }
    my $clearAutoSwitch = 0; # used to keep track of the number of times this export header has been encountered
    if (defined $input->{'clear_auto_switch'} && $input->{'clear_auto_switch'} =~ /\d+/)
    {
      $clearAutoSwitch = $input->{'clear_auto_switch'};
    }
    my $currentAccount = "unknown_account"; # used to keep track of the current account to use for account transactions
    if (defined $input->{'current_account'} && $input->{'current_account'} ne "")
    {
      $currentAccount = $input->{'current_account'};
    }
    my $tempCounter = 0;
    my @parseResult = ();
    $parseResult[0] = "2"; # set the default return value to 'no valid records found'

    my $transDate;
    my $action;
    my $security;
    my $price;
    my $quantity;
    my $description;
    my $memo;
    my $commission;
    my $category;
    my $categoryClass;
    my $transferAccount;
    my $transferClass;
    my $amount;
    my $transferAmount;
    my $qifAccount;

    my $autoSubmit = 1;
    my $numLots = 0;
    my $duplicateStkSplit = 0;
    my $duplicateRtrnCap = 0;

    my $mapCategorySubEntity = "";
    my $mapCategorySubEntity = "";
    my $mapTransferSubEntity = "";
    my $mapQIFAccount;
    my $mapQIFSubAccount = "-1";
    my $mapCategoryAccount = "";
    my $mapCategorySubAccount = "-1";
    my $mapTransferAccount = "";

    # import all other investment transactions
    while ((defined $line) && ($tempCounter < $self->{investmentTransactionBatchSize}) && ($autoSubmit eq "1"))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if ($line =~ /!Option:AutoSwitch/)
        {
          $optionAutoSwitch++;
        }
        elsif ($line =~ /!Clear:AutoSwitch/)
        {
          $clearAutoSwitch++;
        }
        elsif ($line =~ /!Account/)
        {
          if ($optionAutoSwitch > 0 && $clearAutoSwitch > 0)
          {
            # individual accounts with transaction lists after their definition
            $currentAccount = $self->parseAccountRecord(fileHandle => $fileHandle);
            if ($currentAccount eq "")
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Account record! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
          }
        }
        elsif ($line =~ /!Type:Invst/ || $input->{'inside_investment_transaction_list'} eq "1")
        {
          if ($input->{'inside_investment_transaction_list'} eq "1")
          {
            $self->putBackLine($line);
            $input->{'inside_investment_transaction_list'} = 0;
          }
          else
          {
            $line = "";
          }
          while ((defined $line) && ($line !~ /^\!/) && ($tempCounter < $self->{investmentTransactionBatchSize}) && ($autoSubmit eq "1"))
          {
            @parseResult = $self->parseInvestmentTransaction(fileHandle => $fileHandle);
            if ($parseResult[0] eq "0")
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Investment Transaction! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
              unlink $QIFFileName;
              return undef;
            }
            if ($parseResult[0] eq "1")
            {
              $tempCounter++;
              $self->{importedInvestmentTransactions}++;

              my %recordValues = %{ $parseResult[1] };

              $transDate = $recordValues{'date'};
              $action = $recordValues{'action'};
              $security = $recordValues{'security'};
              $price = $recordValues{'price'};
              $price = sprintf("%.6f", $price);
              $quantity = $recordValues{'quantity'};
              $description = $recordValues{'text_reminder'};
              $memo = $recordValues{'memo'};
              $commission = $recordValues{'commission'};
              $commission = sprintf("%.2f", $commission);
              $category = $recordValues{'category'};
              $categoryClass = $recordValues{'category_class'};
              $transferAccount = $recordValues{'transfer_account'};
              $transferClass = $recordValues{'transfer_class'};
              $amount = $recordValues{'transaction_amount'};
              $transferAmount = $recordValues{'transfer_amount'};
              $qifAccount = $currentAccount;

              $numLots = 0;
              $duplicateStkSplit = 0;
              $duplicateRtrnCap = 0;

              $mapCategorySubEntity = "";

              my $sth;
              my @info;
              # lookup the mapping for the class
              if ($categoryClass ne "")
              {
                $sth = $self->{dbHandle}->read(sql => "SELECT map_sub_entity FROM qif_class_tb WHERE name = ?", plug => [ $categoryClass ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for class ($categoryClass)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for class ($categoryClass)!<br>Make sure you have exported all Classes that are used in the investment data.";
                  return undef;
                }
                $mapCategorySubEntity = $info[0];
              }
              $mapTransferSubEntity = "";
              # lookup the mapping for the class
              if ($transferClass ne "")
              {
                $sth = $self->{dbHandle}->read(sql => "SELECT map_sub_entity FROM qif_class_tb WHERE name = ?", plug => [ $transferClass ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for class ($transferClass)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for class ($transferClass)!<br>Make sure you have exported all Classes that are used in the investment data.";
                  return undef;
                }
                $mapTransferSubEntity = $info[0];
              }
              # lookup the mapping for the QIF account associated with the event
              $sth = $self->{dbHandle}->read(sql => "SELECT map_account FROM qif_account_tb WHERE name = ?", plug => [ $qifAccount ]);
              if ($self->{dbHandle}->didErrorOccur())
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for QIF account ($qifAccount)!<br>\n" . $self->{dbHandle}->errorMessage;
                return undef;
              }
              @info = $sth->fetchrow_array;
              if (not defined $info[0])
              {
                $self->{error} = 1;
                $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for QIF account ($qifAccount)!<br>Make sure you have exported all Accounts that are used in the investment data.";
                return undef;
              }
              $mapQIFAccount = $info[0];
              # lookup the mapping for the QIF security associated with the event
              $mapQIFSubAccount = "-1";
              if ($security ne "")
              {
                $sth = $self->{dbHandle}->read(sql => "SELECT map_sub_account FROM qif_security_tb WHERE name = ?", plug => [ $security ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for QIF security ($security)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for QIF security ($security)!<br>Make sure you have exported all Securities that are used in the investment data.";
                  return undef;
                }
                $mapQIFSubAccount = $info[0];
              }
              $mapCategoryAccount = "";
              $mapCategorySubAccount = "-1";
              # lookup the mapping for the category account and sub-account (if any)
              if ($category ne "")
              {
                $sth = $self->{dbHandle}->read(sql => "SELECT map_account, map_sub_account FROM qif_category_tb WHERE name = ?", plug => [ $category ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for category ($category)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for category ($category)!<br>Make sure you have exported all Categories that are used in the investment data.";
                  return undef;
                }
                $mapCategoryAccount = $info[0];
                $mapCategorySubAccount = $info[1];
              }
              $mapTransferAccount = "";
              # lookup the mapping for the transfer account
              if ($transferAccount ne "")
              {
                $sth = $self->{dbHandle}->read(sql => "SELECT map_account FROM qif_account_tb WHERE name = ?", plug => [ $transferAccount ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving mapping for transfer account ($transferAccount)!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (not defined $info[0])
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Mapping not found for transfer account ($transferAccount)!<br>Make sure you have exported all Accounts that are used in the investment data.";
                  return undef;
                }
                $mapTransferAccount = $info[0];
              }

              # import all transactions that do not require user intervention (for choosing lots, for example)
              if ($action eq "Reminder")
              {
                $autoSubmit = 0;
              }
              elsif ($action eq "XIn")
              {
                # do the cash transfer
                my $toAccount;
                my $fromAccount;
                if ($transferAmount < 0)
                {
                  $toAccount = $mapTransferAccount;
                  $fromAccount = $mapQIFAccount;
                }
                else
                {
                  $toAccount = $mapQIFAccount;
                  $fromAccount = $mapTransferAccount;
                }
                # beautify transferAmount
                $transferAmount = sprintf("%.2f", abs($transferAmount));
                my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                if ($self->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                  return undef;
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "XOut")
              {
                # do the cash transfer
                my $toAccount;
                my $fromAccount;
                if ($transferAmount < 0)
                {
                  $toAccount = $mapQIFAccount;
                  $fromAccount = $mapTransferAccount;
                }
                else
                {
                  $toAccount = $mapTransferAccount;
                  $fromAccount = $mapQIFAccount;
                }
                # beautify transferAmount
                $transferAmount = sprintf("%.2f", abs($transferAmount));
                my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                if ($self->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                  return undef;
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "StkSplit")
              {
                # ratio is stored as a number that is multiplied by 10
                my $newShares = $quantity / 10; #/
                my $oldShares = 1;

                # check for duplicate entries before blindly inserting a new record
                # if duplicate exists, skip transaction
                my $sth = $self->{dbHandle}->read(sql => "SELECT event_code, calc_date, s, os, ns FROM sec_event_tb WHERE event_code = ? AND calc_date = ? AND s = ? AND os = '$oldShares' AND ns = '$newShares'", plug => [ "90", $transDate, $mapQIFSubAccount ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error checking for duplicate stock split security event!<br>n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                my @infoTemp = $sth->fetchrow_array;
                if (defined @infoTemp || defined $infoTemp[0])
                {
                  # duplicate entry found
                  $duplicateStkSplit = 1;
                  # log duplicate entry
                  $sth = $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES (?)", plug => [ "Duplicate $action event: D=$transDate, S=$security, Q=$quantity, A=$amount" ]);
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error logging duplicate $action transaction!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  # commit database transaction
                  $self->{dbHandle}->commit;
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                }
                else
                {
                  # create SecurityEventObject object
                  my $secObj = Portal::Accounting::Objects::SecurityEventObject->new(eventCode => "90",
                               calcDate => $transDate, transDate => $transDate,  subAccountNum => $mapQIFSubAccount,
                               oldShares => $oldShares, newShares => $newShares, ledgerCode => $ledgerCode);
                  # insert a stock split security event
                  my $eventRef = $secDBObject->insertStockSplit(secobj => $secObj);
                  if ($secDBObject->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting stock split security event!<br>\n" . $secDBObject->errorMessage;
                    return undef;
                  }
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "RtrnCap" || $action eq "RtrnCapX")
              {
                # check for duplicate entries before blindly inserting a new record
                # if duplicate exists, skip transaction
                my $sth = $self->{dbHandle}->read(sql => "SELECT event_code, calc_date, s FROM sec_event_tb WHERE event_code = ? AND calc_date = ? AND s = ? AND cc = '$amount'", plug => [ "50", $transDate, $mapQIFSubAccount ]);
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error checking for duplicate return of capital security event!<br>n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                my @infoTemp = $sth->fetchrow_array;
                if (defined @infoTemp || defined $infoTemp[0])
                {
                  # duplicate entry found
                  $duplicateRtrnCap = 1;
                  # log duplicate entry
                  $sth = $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES (?)", plug => [ "Duplicate $action event: D=$transDate, S=$security, A=$amount" ]);
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error logging duplicate $action transaction!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  # commit database transaction
                  $self->{dbHandle}->commit;
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                }
                else
                {
                  # create SecurityEventObject object
                  my $secObj = Portal::Accounting::Objects::SecurityEventObject->new(eventCode => "50",
                               calcDate => $transDate, transDate => $transDate,  subAccountNum => $mapQIFSubAccount,
                               amount => $amount, ledgerCode => $ledgerCode);
                  # insert a return of capital security event
                  my $eventRef = $secDBObject->insertReturnOfCapital(secobj => $secObj);
                  if ($secDBObject->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting return of capital security event!<br>\n" . $secDBObject->errorMessage;
                    return undef;
                  }
                }

                if ($action eq "RtrnCapX")
                {
                  # do the cash transfer
                  my $toAccount;
                  my $fromAccount;
                  if ($transferAmount < 0)
                  {
                    $toAccount = $mapQIFAccount;
                    $fromAccount = $mapTransferAccount;
                  }
                  else
                  {
                    $toAccount = $mapTransferAccount;
                    $fromAccount = $mapQIFAccount;
                  }
                  # beautify transferAmount
                  $transferAmount = sprintf("%.2f", abs($transferAmount));
                  my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                  if ($self->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                    return undef;
                  }
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "Div" || $action eq "DivX")
              {
                # dividend revenue account, security sub-account

                # for positive amount
                # debit investment account, -1 sub-account
                # credit dividend revenue account, security sub-account

                # for negative amount, swap above debit and credit

                # do the journal entry
                my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
                if ($transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                # beautify amount
                my $tempAmount = sprintf("%.2f", abs($amount));
                my @transRowObject = [];
                if ($amount < 0)
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $input->{'dividend_revenue_account'}, subAccountNum => $mapQIFSubAccount, dorc => "false", amount => "$tempAmount");
                }
                else
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $input->{'dividend_revenue_account'}, subAccountNum => $mapQIFSubAccount, dorc => "true", amount => "$tempAmount");
                }
                my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                                   journalCode => "G", # hardcoded general journal code
                                                                                   transDate => $transDate,
                                                                                   rowObjects => \@transRowObject);
                if ($transobj->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
                  return undef;
                }

                # insert the transaction into the database
                my $refNum = $transaction->insert(obj => $transobj, commit => 1);
                if ($refNum == -1 || $transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                if ($action eq "DivX")
                {
                  # do the cash transfer
                  my $toAccount;
                  my $fromAccount;
                  if ($transferAmount < 0)
                  {
                    $toAccount = $mapQIFAccount;
                    $fromAccount = $mapTransferAccount;
                  }
                  else
                  {
                    $toAccount = $mapTransferAccount;
                    $fromAccount = $mapQIFAccount;
                  }
                  # beautify transferAmount
                  $transferAmount = sprintf("%.2f", abs($transferAmount));
                  my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                  if ($self->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                    return undef;
                  }
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "IntInc" || $action eq "IntIncX")
              {
                # interest revenue account, security sub-account

                # for positive amount
                # debit investment account, -1 sub-account
                # credit interest revenue account, security sub-account

                # for negative amount, swap above debit and credit

                # do the journal entry
                my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
                if ($transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                # beautify amount
                my $tempAmount = sprintf("%.2f", abs($amount));
                my @transRowObject = [];
                if ($amount < 0)
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $input->{'interest_revenue_account'}, subAccountNum => $mapQIFSubAccount, dorc => "false", amount => "$tempAmount");
                }
                else
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $input->{'interest_revenue_account'}, subAccountNum => $mapQIFSubAccount, dorc => "true", amount => "$tempAmount");
                }
                my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                                   journalCode => "G", # hardcoded general journal code
                                                                                   transDate => $transDate,
                                                                                   rowObjects => \@transRowObject);
                if ($transobj->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
                  return undef;
                }

                # insert the transaction into the database
                my $refNum = $transaction->insert(obj => $transobj, commit => 1);
                if ($refNum == -1 || $transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                if ($action eq "IntIncX")
                {
                  # do the cash transfer
                  my $toAccount;
                  my $fromAccount;
                  if ($transferAmount < 0)
                  {
                    $toAccount = $mapQIFAccount;
                    $fromAccount = $mapTransferAccount;
                  }
                  else
                  {
                    $toAccount = $mapTransferAccount;
                    $fromAccount = $mapQIFAccount;
                  }
                  # beautify transferAmount
                  $transferAmount = sprintf("%.2f", abs($transferAmount));
                  my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                  if ($self->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                    return undef;
                  }
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "MiscInc" || $action eq "MiscIncX")
              {
                # miscellaneous income, security sub-account

                # for positive amount
                # debit investment account, -1 sub-account
                # credit category account, category sub-account
                # or credit other revenue account, -1 sub-account
                my $revenueAccount = ($mapCategoryAccount eq "") ? $input->{'other_revenue_account'} : $mapCategoryAccount;
                my $revenueSubAccount = ($mapCategorySubAccount ne "-1") ? $mapCategorySubAccount : "-1";

                # for negative amount, swap above debit and credit

                # do the journal entry
                my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
                if ($transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                # beautify amount
                my $tempAmount = sprintf("%.2f", abs($amount));
                my @transRowObject = [];
                if ($amount < 0)
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $revenueAccount, subAccountNum => $revenueSubAccount, dorc => "false", amount => "$tempAmount");
                }
                else
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $revenueAccount, subAccountNum => $revenueSubAccount, dorc => "true", amount => "$tempAmount");
                }
                my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                                   journalCode => "G", # hardcoded general journal code
                                                                                   transDate => $transDate,
                                                                                   rowObjects => \@transRowObject);
                if ($transobj->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
                  return undef;
                }

                # insert the transaction into the database
                my $refNum = $transaction->insert(obj => $transobj, commit => 1);
                if ($refNum == -1 || $transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                if ($action eq "MiscIncX")
                {
                  # do the cash transfer
                  my $toAccount;
                  my $fromAccount;
                  if ($transferAmount < 0)
                  {
                    $toAccount = $mapQIFAccount;
                    $fromAccount = $mapTransferAccount;
                  }
                  else
                  {
                    $toAccount = $mapTransferAccount;
                    $fromAccount = $mapQIFAccount;
                  }
                  # beautify transferAmount
                  $transferAmount = sprintf("%.2f", abs($transferAmount));
                  my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                  if ($self->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                    return undef;
                  }
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "MiscExp" || $action eq "MiscExpX")
              {
                # miscellaneous expense, security sub-account

                # for positive amount
                # debit investment account, -1 sub-account
                # credit category account, category sub-account
                # or credit other expense account, -1 sub-account
                my $revenueAccount = ($mapCategoryAccount eq "") ? $input->{'other_expense_account'} : $mapCategoryAccount;
                my $revenueSubAccount = ($mapCategorySubAccount ne "-1") ? $mapCategorySubAccount : "-1";

                # for negative amount, swap above debit and credit

                # do the journal entry
                my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
                if ($transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                # beautify amount
                my $tempAmount = sprintf("%.2f", abs($amount));
                my @transRowObject = [];
                if ($amount < 0)
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $revenueAccount, subAccountNum => $revenueSubAccount, dorc => "false", amount => "$tempAmount");
                }
                else
                {
                  $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                  $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapCategorySubEntity, accountNum => $revenueAccount, subAccountNum => $revenueSubAccount, dorc => "true", amount => "$tempAmount");
                }
                my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                                   journalCode => "G", # hardcoded general journal code
                                                                                   transDate => $transDate,
                                                                                   rowObjects => \@transRowObject);
                if ($transobj->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
                  return undef;
                }

                # insert the transaction into the database
                my $refNum = $transaction->insert(obj => $transobj, commit => 1);
                if ($refNum == -1 || $transaction->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
                  return undef;
                }

                if ($action eq "MiscExpX")
                {
                  # do the cash transfer
                  my $toAccount;
                  my $fromAccount;
                  if ($transferAmount < 0)
                  {
                    $toAccount = $mapQIFAccount;
                    $fromAccount = $mapTransferAccount;
                  }
                  else
                  {
                    $toAccount = $mapTransferAccount;
                    $fromAccount = $mapQIFAccount;
                  }
                  # beautify transferAmount
                  $transferAmount = sprintf("%.2f", abs($transferAmount));
                  my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                  if ($self->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                    return undef;
                  }
                }
                $autoSubmit = 1;
              }
              elsif ($action eq "Sell" || $action eq "SellX" || $action eq "ShrsOut")
              {
                # retrieve the number of lots for the security
                $sth = $self->{dbHandle}->read(sql => "SELECT count(DISTINCT lot) FROM lot_node_tb WHERE sub_account_num = '$mapQIFSubAccount'");
                if ($self->{dbHandle}->didErrorOccur())
                {
                  $self->{error} = 1;
                  $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving number of lots for security (" . $mapQIFSubAccount . ")!<br>\n" . $self->{dbHandle}->errorMessage;
                  return undef;
                }
                @info = $sth->fetchrow_array;
                if (defined @info && $info[0] =~ /^[0-9]+$/)
                {
                  $numLots = $info[0];
                }
                if ($numLots == 1)
                {
                  # attempt to automate the process for selling this (partial) lot
                  $sth = $self->{dbHandle}->read(sql => "SELECT DISTINCT lot FROM lot_node_tb WHERE sub_account_num = '$mapQIFSubAccount'");
                  if ($self->{dbHandle}->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving lot for security (" . $mapQIFSubAccount . ")!<br>\n" . $self->{dbHandle}->errorMessage;
                    return undef;
                  }
                  @info = $sth->fetchrow_array;
                  my $lotNum;
                  if (defined @info && $info[0] =~ /^[0-9]+$/)
                  {
                    $lotNum = $info[0];
                  }

                  # retrieve the balance for the lot
                  my $lotobj = $secDBObject->getLotBalance(eventDate => $transDate, lot => $lotNum, ledgerCode => $ledgerCode);
                  if ($secDBObject->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error getting the balance for lot (" . $lotNum . ")!<br>\n" . $secDBObject->errorMessage;
                    return undef;
                  }
                  # check to make sure the number of shares being sold does not exceed the total balance of shares in the lot
                  if ($quantity <= $lotobj->shares)
                  {
                    my $secObj = undef;
                    #take data and put in event object to give to database
                    $secObj = Portal::Accounting::Objects::SecurityEventObject->new(masterEventRef => "-1",
                      eventCode => "20", calcDate => $transDate, transDate => $transDate,
                      subEntity => "$mapTransferSubEntity", accountNum => $mapQIFAccount, subAccountNum => $mapQIFSubAccount, lot => $lotNum,
                      price => $price, shares => $quantity, commission => $commission, ledgerCode => $ledgerCode);
                   # counter =>   used if placing this event in between other events on same day for same subaccount
                    if ($secObj->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error Creating Object Sell: $quantity shares of Lot: $lotNum<br>\n" . $secObj->errorMessage;
                      return undef;
                    }
                    #give to database
                    my $eventRef = $secDBObject->insertSell(secobj => $secObj);
                    if ($secDBObject->didErrorOccur())
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error Inserting Sell: $quantity shares of Lot: $lotNum<br>\n" . $secDBObject->errorMessage;
                      return undef;
                    }
                    if (not defined $eventRef)
                    {
                      $self->{error} = 1;
                      $self->{errorString} = $errStr . "command = '$command'<br>Error getting event reference for Sell: $quantity shares of Lot: $lotNum<br>\n" . $secDBObject->errorMessage;
                      return undef;
                    }

                    if ($action eq "SellX")
                    {
                      # insert cash transfer
                      my $toAccount;
                      my $fromAccount;
                      if ($transferAmount < 0)
                      {
                        $toAccount = $mapQIFAccount;
                        $fromAccount = $mapTransferAccount;
                      }
                      else
                      {
                        $toAccount = $mapTransferAccount;
                        $fromAccount = $mapQIFAccount;
                      }
                      # beautify transferAmount
                      $transferAmount = sprintf("%.2f", abs($transferAmount));
                      my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                      if ($self->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                        return undef;
                      }
                    }

                    if ($action eq "ShrsOut")
                    {
                      # annihilate equity into nothing
                      my $equityAccount = $input->{'equity_account'};
                      if ($equityAccount !~ /^[0-9]+$/)
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Invalid equity account ($equityAccount)!";
                        return undef;
                      }
                      # create new transaction
                      my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
                      if ($transaction->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
                        return undef;
                      }

                      # beautify amount
                      my $tempAmount = sprintf("%.2f", abs($amount));
                      my @transRowObject = [];
                      if ($amount < 0)
                      {
                        $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                        $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $equityAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                      }
                      else
                      {
                        $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                        $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $equityAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                      }
                      my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                                         journalCode => "G", # hardcoded general journal code
                                                                                         transDate => $transDate,
                                                                                         rowObjects => \@transRowObject);
                      if ($transobj->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
                        return undef;
                      }

                      # insert the transaction into the database
                      my $refNum = $transaction->insert(obj => $transobj, commit => 1);
                      if ($refNum == -1 || $transaction->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
                        return undef;
                      }

                      # log ShrsOut entry
                      my $sth = $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES (?)", plug => [ "ShrsOut event: (Journal RefNum = $refNum, Security EventRef = $eventRef) D=$transDate, S=$mapQIFSubAccount, P=$price, Q=$quantity, A=$amount" ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error logging ShrsOut transaction!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                      # commit database transaction
                      $self->{dbHandle}->commit;
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                    }

                    $autoSubmit = 1;
                  }
                  else
                  {
                    # number of shares being sold exceeds the total number of shares in the lot
                    $numLots = 0;
                    $autoSubmit = 0;  # this will cause the code to pass to the screen that will allow the user to skip the transaction
                  }
                }
                elsif ($numLots > 1)
                {
                  # determine whether the number of shares being sold equals the total number of shares in all the lots of the security
                  my @lotsArray = $secDBObject->getLotBalanceEntriesFor(eventDate => $transDate, subAccountNum => $mapQIFSubAccount, ledgerCode => $ledgerCode);
                  if ($secDBObject->didErrorOccur())
                  {
                    $self->{error} = 1;
                    $self->{errorString} = $errStr . "command = '$command'<br>Error retrieving all lot entries for security (" . $mapQIFSubAccount . ")!<br>\n" . $secDBObject->errorMessage;
                    return undef;
                  }
                  my $lotCount = 0;
                  my $sharesSum = 0.00;
                  for ($lotCount=0; $lotCount < scalar @lotsArray; $lotCount++)
                  {
                    $sharesSum += $lotsArray[$lotCount]->shares;
                  }
                  if ($sharesSum == $quantity)
                  {
                    my $commissionPerShare = sprintf("%.2f", ($commission / $quantity)); #/
                    my $leftoverCommission = sprintf("%.2f", ($commission - ($commissionPerShare * $quantity)));
                    # run through each of the lots to be sold and insert a sell event into the database with the appropriate data
                    my $masterRef = -1;
                    my @eventRefArray = ();
                    # automatically sell all lots of the security
                    for ($lotCount=0; $lotCount < scalar @lotsArray; $lotCount++)
                    {
                      my $secObj = undef;
                      #take data and put in event object to give to database
                      if ($masterRef == -1)
                      {
                        $secObj = Portal::Accounting::Objects::SecurityEventObject->new(masterEventRef => $masterRef,
                          eventCode => "20", calcDate => $transDate, transDate => $transDate,
                          subEntity => "$mapTransferSubEntity", accountNum => $mapQIFAccount, subAccountNum => $mapQIFSubAccount, lot => $lotsArray[$lotCount]->lot,
                          price => $price, shares => $lotsArray[$lotCount]->shares,
                          commission => sprintf("%.2f", ($commissionPerShare * $lotsArray[$lotCount]->shares + $leftoverCommission) ),
                          ledgerCode => $ledgerCode);
                     # counter =>   used if placing this event in between other events on same day for same subaccount
                      }
                      else
                      {
                        $secObj = Portal::Accounting::Objects::SecurityEventObject->new(masterEventRef => $masterRef,
                          eventCode => "20", calcDate => $transDate, transDate => $transDate,
                          subEntity => "$mapTransferSubEntity", accountNum => $mapQIFAccount, subAccountNum => $mapQIFSubAccount, lot => $lotsArray[$lotCount]->lot,
                          price => $price, shares => $lotsArray[$lotCount]->shares,
                          commission => sprintf("%.2f", ($commissionPerShare * $lotsArray[$lotCount]->shares) ),
                          ledgerCode => $ledgerCode);
                     # counter =>   used if placing this event in between other events on same day for same subaccount
                      }
                      if ($secObj->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error Creating Object Sell: " . $lotsArray[$lotCount]->shares . " shares of Lot: " . $lotsArray[$lotCount]->lot . "<br>\n" . $secObj->errorMessage;
                        return undef;
                      }
                      #give to database
                      my $eventRef;
                      $eventRef = $secDBObject->insertSell(secobj => $secObj);
                      if ($secDBObject->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error Inserting Sell: " . $lotsArray[$lotCount]->shares . " shares of Lot: " . $lotsArray[$lotCount]->lot . "<br>\n" . $secDBObject->errorMessage;
                        return undef;
                      }
                      if (not defined $eventRef)
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error getting event reference for Sell: " . $lotsArray[$lotCount]->shares . " shares of Lot: " . $lotsArray[$lotCount]->lot . "<br>\n" . $secDBObject->errorMessage;
                        return undef;
                      }
                      elsif ($masterRef == -1)
                      {
                        $masterRef = $eventRef;
                      }
                      $eventRefArray[scalar @eventRefArray] = $eventRef;
                    }

                    if ($action eq "SellX")
                    {
                      # insert cash transfer
                      my $toAccount;
                      my $fromAccount;
                      if ($transferAmount < 0)
                      {
                        $toAccount = $mapQIFAccount;
                        $fromAccount = $mapTransferAccount;
                      }
                      else
                      {
                        $toAccount = $mapTransferAccount;
                        $fromAccount = $mapQIFAccount;
                      }
                      # beautify transferAmount
                      $transferAmount = sprintf("%.2f", abs($transferAmount));
                      my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $transDate, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
                      if ($self->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
                        return undef;
                      }
                    }

                    if ($action eq "ShrsOut")
                    {
                      # annihilate equity into nothing
                      my $equityAccount = $input->{'equity_account'};
                      if ($equityAccount !~ /^[0-9]+$/)
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Invalid equity account ($equityAccount)!";
                        return undef;
                      }
                      # create new transaction
                      my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
                      if ($transaction->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
                        return undef;
                      }

                      # beautify amount
                      my $tempAmount = sprintf("%.2f", abs($amount));
                      my @transRowObject = [];
                      if ($amount < 0)
                      {
                        $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                        $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $equityAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                      }
                      else
                      {
                        $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $mapQIFAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
                        $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $transDate, subEntity => $mapTransferSubEntity, accountNum => $equityAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
                      }
                      my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                                         journalCode => "G", # hardcoded general journal code
                                                                                         transDate => $transDate,
                                                                                         rowObjects => \@transRowObject);
                      if ($transobj->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
                        return undef;
                      }

                      # insert the transaction into the database
                      my $refNum = $transaction->insert(obj => $transobj, commit => 1);
                      if ($refNum == -1 || $transaction->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
                        return undef;
                      }

                      # log ShrsOut entry
                      my $sth = $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES (?)", plug => [ "ShrsOut event: (Journal RefNum = $refNum, Security EventRef = " . join(",", @eventRefArray) . ") D=$transDate, S=$mapQIFSubAccount, P=$price, Q=$quantity, A=$amount" ]);
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error logging ShrsOut transaction!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                      # commit database transaction
                      $self->{dbHandle}->commit;
                      if ($self->{dbHandle}->didErrorOccur())
                      {
                        $self->{error} = 1;
                        $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
                        return undef;
                      }
                    }

                    $autoSubmit = 1;
                  }
                  else
                  {
                    $autoSubmit = 0;
                  }
                }
              }
            }

            $line = $self->getNextLine(fileHandle => $fileHandle);
            if (defined $line)
            {
              $self->putBackLine($line);
            }
          }
        }
      }
    }

    $doc->setTitle("QIF Map Import Investment Transaction");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Import Investment Transaction</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"inputform\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if ((not defined $line) && $autoSubmit eq "1") # finished processing all investment transactions
    {
      $doc->print("<input type=hidden name=\"command\" value=\"import_prices\">\n");
    }
    else
    {
      if ((($action eq "Reminder" || $action eq "Sell" || $action eq "SellX" || $action eq "ShrsOut") && $autoSubmit ne "1"))
      {
        $doc->print("<input type=hidden name=\"command\" value=\"import_processed_investment_transaction\">\n");
      }
      else
      {
        $doc->print("<input type=hidden name=\"command\" value=\"import_investment_transactions\">\n");
      }
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"current_account\" value=\"$currentAccount\">\n");
      $doc->print("<input type=hidden name=\"option_auto_switch\" value=\"$optionAutoSwitch\">\n");
      $doc->print("<input type=hidden name=\"clear_auto_switch\" value=\"$clearAutoSwitch\">\n");
      if ($parseResult[0] eq "1" && ($tempCounter >= $self->{investmentTransactionBatchSize} || $autoSubmit ne "1"))
      {
        $doc->print("<input type=hidden name=\"inside_investment_transaction_list\" value=\"1\">\n");
      }
      $doc->print("<input type=hidden name=\"imported_investment_transactions\" value=\"" . $self->{importedInvestmentTransactions} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"equity_account\" value=\"" . $input->{'equity_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"dividend_revenue_account\" value=\"" . $input->{'dividend_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"interest_revenue_account\" value=\"" . $input->{'interest_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"other_revenue_account\" value=\"" . $input->{'other_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"other_expense_account\" value=\"" . $input->{'other_expense_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"import_sub_entity\" value=\"" . $input->{'import_sub_entity'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"1\">\n");
    if ((($action eq "Sell" || $action eq "SellX" || $action eq "ShrsOut") && $autoSubmit != 1))
    {
      $doc->print("<input type=hidden name=\"qif_action\" value=\"$action\">\n");
      $doc->print("<input type=hidden name=\"qif_trans_date\" value=\"$transDate\">\n");
      $doc->print("<input type=hidden name=\"qif_security\" value=\"$mapQIFSubAccount\">\n");
      $doc->print("<input type=hidden name=\"qif_price\" value=\"$price\">\n");
      $doc->print("<input type=hidden name=\"qif_quantity\" value=\"$quantity\">\n");
      $doc->print("<input type=hidden name=\"qif_description\" value=\"$description\">\n");
      $doc->print("<input type=hidden name=\"qif_memo\" value=\"$memo\">\n");
      $doc->print("<input type=hidden name=\"qif_commission\" value=\"$commission\">\n");
      $doc->print("<input type=hidden name=\"qif_category_account\" value=\"$mapCategoryAccount\">\n");
      $doc->print("<input type=hidden name=\"qif_category_sub_account\" value=\"$mapCategorySubAccount\">\n");
      $doc->print("<input type=hidden name=\"qif_category_class\" value=\"$mapCategorySubEntity\">\n");
      $doc->print("<input type=hidden name=\"qif_transfer_account\" value=\"$mapTransferAccount\">\n");
      $doc->print("<input type=hidden name=\"qif_transfer_class\" value=\"$mapTransferSubEntity\">\n");
      $doc->print("<input type=hidden name=\"qif_transfer_amount\" value=\"$transferAmount\">\n");
      $doc->print("<input type=hidden name=\"qif_account\" value=\"$mapQIFAccount\">\n");
      $doc->print("<input type=hidden name=\"qif_amount\" value=\"$amount\">\n");

      $doc->print("Action = <b>$action</b><br>\n");
      $doc->print("Date = <b>$transDate</b><br>\n");
      $doc->print("Security = <b>$security</b><br>\n");
      $doc->print("Price = <b>$price</b><br>\n");
      $doc->print("Quantity = <b>$quantity</b><br>\n");
      $doc->print("Commission = <b>$commission</b><br>\n");
      $doc->print("Amount = <b>$amount</b><br>\n");

      my $button = "<input type=\"button\" onClick=\"window.open('index.cgi?app=Accounting&state=InvJournal&command=sharesselect";
      $button .= "&ledger=$ledger&account=$mapQIFAccount&subAccount=$mapQIFSubAccount&date=$transDate";
      $button .= "', 'sharesselection', 'toolbars=no,location=no,status=no,width=550,height=550,resizable,scrollbars');\" value=\"Select Lots\"><br>\n";
      $doc->print($button);
      $doc->print("<input type=\"hidden\" name=\"lots\" value=\"\">\n");
      $doc->print("<input type=\"hidden\" name=\"shares\" value=\"0\">\n");
      $doc->print("<input type=\"hidden\" name=\"skip_transaction\" value=\"0\">\n");

      $doc->print("<input type=submit name=\"skip_submit\" value=\"Skip ($action)\" onClick=\"return validateSkip();\"><input type=submit name=\"submit_transactions\" value=\"Continue...\" onClick=\"return validateForm();\">\n");
      $doc->print("</form>\n");

      $doc->setFocus("javascript");
      $doc->print("function validateSkip()\n");
      $doc->print("{\n");
      $doc->print("  var confirmMessage = \"Skipping transactions may cause data to be lost or become invalid.\\nDo you really want to skip inserting this transaction?\";\n");
      $doc->print("\n");
      $doc->print("  if (confirm(confirmMessage))\n");
      $doc->print("  {\n");
      $doc->print("    document.inputform.skip_transaction.value = '1';\n");
      $doc->print("    return true;\n");
      $doc->print("  }\n");
      $doc->print("  else\n");
      $doc->print("  {\n");
      $doc->print("    return false;\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  return false;\n");
      $doc->print("}\n");
      $doc->print("\n");
      $doc->print("function validateForm()\n");
      $doc->print("{\n");
      $doc->print("  var errorString = \"\";\n");
      $doc->print("\n");
      $doc->print("  if (document.inputform.lots.value == '')\n");
      $doc->print("  {\n");
      $doc->print("    errorString += \"Lots field is empty!  No shares were selected.\\nYou must select the number of shares from existing lots\\nas necessary to total the number of shares\\nin the transaction.\";\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  if (document.inputform.shares.value != $quantity)\n");
      $doc->print("  {\n");
      $doc->print("    errorString += \"The number of shares selected (\" + document.inputform.shares.value + \")\\ndoes not equal the necessary amount of shares ($quantity)!\";\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  if (errorString != \"\")\n");
      $doc->print("  {\n");
      $doc->print("    alert(errorString);\n");
      $doc->print("    return false;\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  return true;\n");
      $doc->print("}\n");
      $doc->setFocus("body");
    }
    elsif ($action eq "Reminder")
    {
      $doc->print("<input type=hidden name=\"qif_action\" value=\"$action\">\n");
      $doc->print("<input type=hidden name=\"qif_trans_date\" value=\"$transDate\">\n");
      $doc->print("<input type=hidden name=\"qif_description\" value=\"$description\">\n");
      $doc->print("<input type=hidden name=\"qif_memo\" value=\"$memo\">\n");
      $doc->print("<input type=hidden name=\"qif_account\" value=\"$mapQIFAccount\">\n");
      $doc->print("Memo: '$memo'<br>");

      $doc->print("Select the Security to which to map the Memo: <select name=\"security_select\">\n");
      # retrieve a list of all securities to choose from
      my @securityList = $dbObject->getListOfSubAccounts(type => "70", orderBy => "sub_acc_name");
      if ($dbObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error obtaining list of security sub-accounts!<br>\n" . $dbObject->errorMessage;
        return undef;
      }
      for (my $listCount=0; $listCount < scalar @securityList; $listCount++)
      {
        $doc->print("  <option value=\"" . $securityList[$listCount]->number . "\">" . $securityList[$listCount]->name . "\n");
      }
      $doc->print("</select>\n");

      $doc->print("<input type=\"hidden\" name=\"skip_transaction\" value=\"0\">\n");

      $doc->print("<input type=submit name=\"skip_submit\" value=\"Skip ($action)\" onClick=\"return validateSkip();\"><input type=submit name=\"submit_transactions\" value=\"Continue...\" onClick=\"return validateForm();\">\n");
      $doc->print("</form>\n");

      $doc->setFocus("javascript");
      $doc->print("function validateSkip()\n");
      $doc->print("{\n");
      $doc->print("  var confirmMessage = \"Skipping transactions may cause data to be lost or become invalid.\\nDo you really want to skip inserting this transaction?\";\n");
      $doc->print("\n");
      $doc->print("  if (confirm(confirmMessage))\n");
      $doc->print("  {\n");
      $doc->print("    document.inputform.skip_transaction.value = '1';\n");
      $doc->print("    return true;\n");
      $doc->print("  }\n");
      $doc->print("  else\n");
      $doc->print("  {\n");
      $doc->print("    return false;\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  return false;\n");
      $doc->print("}\n");
      $doc->print("\n");
      $doc->print("function validateForm()\n");
      $doc->print("{\n");
      $doc->print("  var errorString = \"\";\n");
      $doc->print("\n");
      $doc->print("  if (document.inputform.security_select.options.selectedIndex == -1)\n");
      $doc->print("  {\n");
      $doc->print("    errorString += \"No securities were selected!\\nYou must select the security with which\\nto map the QIF reminder to a memo.\";\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  if (errorString != \"\")\n");
      $doc->print("  {\n");
      $doc->print("    alert(errorString);\n");
      $doc->print("    return false;\n");
      $doc->print("  }\n");
      $doc->print("\n");
      $doc->print("  return true;\n");
      $doc->print("}\n");
      $doc->setFocus("body");
    }
    else
    {
      # set the page to be submitted by JavaScript
      $doc->setFocus("javascript");
      my $onLoadCode = "document.inputform.submit();\n";
      $doc->setOnload(code => $onLoadCode);
      $doc->setFocus("body");
      $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
      $doc->print("<input type=submit name=\"submit_transactions\" value=\"Continue...\">\n");
      $doc->print("</form>\n");
      $doc->print("<br><br>\n");
      if ((not defined $line) && $autoSubmit eq "1")
      {
        $doc->print("Total");
      }
      else
      {
        $doc->print("Current");
      }
      $doc->print(" investment transaction records processed: " . $self->{importedInvestmentTransactions} . "\n");
    }
    $doc->print("</center>\n");
  }
  elsif ($command eq "import_processed_investment_transaction")
  {
    my $skipTransaction = $input->{'skip_transaction'};
    my $action = $input->{'qif_action'};
    my $maxTransactions = $input->{'max_transactions'};
    my $transactionOffset = $input->{'transaction_offset'};

    my $date = $input->{'qif_trans_date'};
    my $lots = $input->{'lots'};
    my $shares = $input->{'shares'};
    my $price = $input->{'qif_price'};
    my $commission = $input->{'qif_commission'};
    my $mapTransferSubEntity = $input->{'qif_transfer_class'};
    my $account = $input->{'qif_account'};
    my $subAccount = $input->{'qif_security'};

    if ($skipTransaction eq "0")
    {
      # create Security object
      my $securityDBObject = Portal::Accounting::DB::Security->new(dbiw => $self->{dbHandle});
      if ($securityDBObject->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errStr . "command = '$command'<br>Error creating new Security database object!<br>\n" . $securityDBObject->errorMessage;
        return undef;
      }

      if ($action eq "Reminder")
      {
        my $description = $input->{'qif_description'};
        my $memo = $input->{'qif_memo'};
        my $security = $input->{'security_select'};
        # create SecurityEventObject object
        my $secObj = Portal::Accounting::Objects::SecurityEventObject->new(eventCode => "memo",
                     calcDate => $date, transDate => $date, memoText => "$description [$memo]",
                     subAccountNum => $security, ledgerCode => $ledgerCode);
        # insert a memo security event
        my $eventRef = $securityDBObject->insertMemo(secobj => $secObj);
        if ($securityDBObject->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errStr . "command = '$command'<br>Error inserting memo security event!<br>\n" . $securityDBObject->errorMessage;
          return undef;
        }
      }
      elsif ($action eq "Sell" || $action eq "SellX" || $action eq "ShrsOut")
      {
        # insert the security event for the Sell, SellX, or ShrsOut
        my @lotArray = split(/\|/, $lots);
        my ($tempLot, $tempShares, $tempNode);
        my $commissionPerShare = sprintf("%.2f", ($commission / $shares)); #/
        my $leftoverCommission = sprintf("%.2f", ($commission - ($commissionPerShare * $shares)));
        # run through each of the lots to be sold and insert a sell event into the database with the appropriate data
        my $masterRef = -1;
        my @eventRefArray = ();
        foreach my $lotIndex (@lotArray)
        {
          ($tempLot, $tempShares) = split(/\=/, $lotIndex);
          if ($tempLot eq "" || $tempShares eq "")
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Invalid lot format! ($lotIndex)";
            return undef;
          }
          else
          {
            $tempNode = $securityDBObject->getLotBalance(lot => $tempLot, eventDate => $date, ledgerCode => $ledgerCode);
            if ($securityDBObject->didErrorOccur)
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Balance for Lot: $tempLot, $date failed.<br>\n" . $securityDBObject->errorMessage;
              return undef;
            }
            #check to make sure the shares chosen are <= balance
            if ($tempShares > $tempNode->{shares})
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Number of shares sold exceeds number of shares in the future! (L=$tempLot, S=$tempShares)";
              return undef;
            }

            my $secObj = undef;
            #take data and put in event object to give to database
            if ($masterRef == -1)
            {
              $secObj = Portal::Accounting::Objects::SecurityEventObject->new(masterEventRef => $masterRef,
                eventCode => "20", calcDate => $date, transDate => $date,
                subEntity => "$mapTransferSubEntity", accountNum => $account, subAccountNum => $subAccount, lot => $tempLot,
                price => $price, shares => $tempShares,
                commission => sprintf("%.2f", ($commissionPerShare * $tempShares + $leftoverCommission) ),
                ledgerCode => $ledgerCode);
           # counter =>   used if placing this event in between other events on same day for same subaccount
            }
            else
            {
              $secObj = Portal::Accounting::Objects::SecurityEventObject->new(masterEventRef => $masterRef,
                eventCode => "20", calcDate => $date, transDate => $date,
                subEntity => "$mapTransferSubEntity", accountNum => $account, subAccountNum => $subAccount, lot => $tempLot,
                price => $price, shares => $tempShares,
                commission => sprintf("%.2f", ($commissionPerShare * $tempShares) ),
                ledgerCode => $ledgerCode);
           # counter =>   used if placing this event in between other events on same day for same subaccount
            }
            if ($secObj->didErrorOccur())
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error Creating Object Sell: $tempShares shares of Lot: $tempLot<br>\n" . $secObj->errorMessage;
              return undef;
            }
            #give to database
            my $eventRef;
            $eventRef = $securityDBObject->insertSell(secobj => $secObj);
            if ($securityDBObject->didErrorOccur())
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error Inserting Sell: $tempShares shares of Lot: $tempLot<br>\n" . $securityDBObject->errorMessage;
              return undef;
            }
            if (not defined $eventRef)
            {
              $self->{error} = 1;
              $self->{errorString} = $errStr . "command = '$command'<br>Error getting event reference for Sell: $tempShares shares of Lot: $tempLot<br>\n" . $securityDBObject->errorMessage;
              return undef;
            }
            elsif ($masterRef == -1)
            {
              $masterRef = $eventRef;
            }
            $eventRefArray[scalar @eventRefArray] = $eventRef;
          }
        }#endfor

        if ($action eq "ShrsOut")
        {
          # annihilate equity into nothing
          my $equityAccount = $input->{'equity_account'};
          if ($equityAccount !~ /^[0-9]+$/)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Invalid equity account ($equityAccount)!";
            return undef;
          }
          # create new transaction
          my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
          if ($transaction->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error instantiating Transaction object for action ($action)!<br>\n" . $transaction->errorMessage;
            return undef;
          }

          # beautify amount
          my $tempAmount = sprintf("%.2f", abs($input->{'qif_amount'}));
          my @transRowObject = [];
          if ($input->{'qif_amount'} < 0)
          {
            $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $mapTransferSubEntity, accountNum => $account, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
            $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $mapTransferSubEntity, accountNum => $equityAccount, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
          }
          else
          {
            $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $mapTransferSubEntity, accountNum => $account, subAccountNum => "-1", dorc => "false", amount => "$tempAmount");
            $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $mapTransferSubEntity, accountNum => $equityAccount, subAccountNum => "-1", dorc => "true", amount => "$tempAmount");
          }
          my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                             journalCode => "G", # hardcoded general journal code
                                                                             transDate => $date,
                                                                             rowObjects => \@transRowObject);
          if ($transobj->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Failed to create new transaction object for action ($action)!<br>\n" . $transobj->errorMessage;
            return undef;
          }

          # insert the transaction into the database
          my $refNum = $transaction->insert(obj => $transobj, commit => 1);
          if ($refNum == -1 || $transaction->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error inserting journal transaction for action ($action)!<br>\n" . $transaction->errorMessage;
            return undef;
          }

          # log ShrsOut entry
          my $sth = $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES (?)", plug => [ "ShrsOut event: (Journal RefNum = $refNum, Security EventRef = " . join(",", @eventRefArray) . ") D=$date, S=$subAccount, P=$price, Q=" . $input->{'qif_quantity'} . ", A=" . $input->{'qif_amount'} ]);
          if ($self->{dbHandle}->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error logging ShrsOut transaction!<br>\n" . $self->{dbHandle}->errorMessage;
            return undef;
          }
          # commit database transaction
          $self->{dbHandle}->commit;
          if ($self->{dbHandle}->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error committing work!<br>\n" . $self->{dbHandle}->errorMessage;
            return undef;
          }
        }

        if ($action eq "SellX")
        {
          # insert cash transfer
          my $toAccount;
          my $fromAccount;
          my $transferAmount = $input->{'qif_transfer_amount'};
          if ($transferAmount < 0)
          {
            $toAccount = $input->{'qif_account'};
            $fromAccount = $input->{'qif_transfer_account'};
          }
          else
          {
            $toAccount = $input->{'qif_transfer_account'};
            $fromAccount = $input->{'qif_account'};
          }
          # beautify transferAmount
          $transferAmount = sprintf("%.2f", abs($transferAmount));
          my $refNum = $self->insertCashTransfer(ledgerCode => $ledgerCode, date => $date, subEntity => $mapTransferSubEntity, accountFrom => $fromAccount, accountTo => $toAccount, amount => "$transferAmount");
          if ($self->didErrorOccur())
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error inserting cash transfer for ($action) transaction!<br>\n" . $self->errorMessage;
            return undef;
          }
        }
      }
    }

    $doc->setTitle("QIF Map Import Investment Transaction");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Import Investment Transaction</h1><br>\n");
    $doc->print("<form method=\"post\" name=\"qif_map_processed_transactions\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"import_investment_transactions\">\n");
    $doc->print("<input type=hidden name=\"current_line\" value=\"" . $input->{'current_line'} . "\">\n");
    $doc->print("<input type=hidden name=\"current_account\" value=\"" . $input->{'current_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"option_auto_switch\" value=\"" . $input->{'option_auto_switch'} . "\">\n");
    $doc->print("<input type=hidden name=\"clear_auto_switch\" value=\"" . $input->{'clear_auto_switch'} . "\">\n");
    $doc->print("<input type=hidden name=\"inside_investment_transaction_list\" value=\"" . $input->{'inside_investment_transaction_list'} . "\">\n");
    $doc->print("<input type=hidden name=\"imported_investment_transactions\" value=\"" . $input->{'imported_investment_transactions'} . "\">\n");

    $doc->print("<input type=hidden name=\"equity_account\" value=\"" . $input->{'equity_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"dividend_revenue_account\" value=\"" . $input->{'dividend_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"interest_revenue_account\" value=\"" . $input->{'interest_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"other_revenue_account\" value=\"" . $input->{'other_revenue_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"other_expense_account\" value=\"" . $input->{'other_expense_account'} . "\">\n");
    $doc->print("<input type=hidden name=\"import_sub_entity\" value=\"" . $input->{'import_sub_entity'} . "\">\n");
    $doc->print("<input type=hidden name=\"limit_range\" value=\"1\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_map_processed_transactions.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_transactions\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if ($skipTransaction eq "0")
    {
      $doc->print("Processed");
    }
    else
    {
      $doc->print("Skipped");
    }
    $doc->print(" QIF <b>$action</b> Event:<br>\n");
    $doc->print("Date = <b>$date</b><br>\n");
    $doc->print("Security = <b>$subAccount</b><br>\n");
    if ($action eq "Sell" || $action eq "SellX" || $action eq "ShrsOut")
    {
      $doc->print("Price = <b>$price</b><br>\n");
      $doc->print("Quantity = <b>" . $input->{'qif_quantity'} . "</b><br>\n");
      $doc->print("Commission = <b>$commission</b><br>\n");
      $doc->print("Amount = <b>" . $input->{'qif_amount'} . "</b><br>\n");
      if ($action eq "SellX")
      {
        $doc->print("Transfer Account = <b>" . $input->{'qif_transfer_account'} . "</b><br>\n");
        $doc->print("Transfer Class = <b>" . $input->{'qif_transfer_class'} . "</b><br>\n");
        $doc->print("Transfer Amount = <b>" . $input->{'qif_transfer_amount'} . "</b><br>\n");
      }
    }
    elsif ($action eq "Reminder")
    {
      $doc->print("Description = <b>" . $input->{'qif_description'} . "</b><br>\n");
      $doc->print("Memo = <b>" . $input->{'qif_memo'} . "</b><br>\n");
    }
    $doc->print("QIF Account = <b>$account</b>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "import_prices")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    my $fileHandle = openFile($QIFFileName);
    if (not defined $fileHandle)
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>Error opening QIF file!<br>The filename (" . $QIFFileName . ") does not exist on the server!";
      return undef;
    }

    # parse the data file, recognizing the different sections as they occur; call helper functions to read the data for the
    # recognized section and insert it into tables in the database

    # reset any variables from a previous security list parse and continue parsing
    if (defined $input->{'current_line'} && $input->{'current_line'} =~ /\d+/)
    {
      $self->{currentLine} = $input->{'current_line'};
      if (!$self->seekFile(lines => $self->{currentLine}, fileHandle =>$fileHandle))
      {
        $self->{errorString} .= "Could not open file in: $command<br>\n";
      }
    }
    else
    {
      $self->{currentLine} = 0;
    }
    if (defined $input->{'imported_price_transactions'} && $input->{'imported_price_transactions'} =~ /\d+/)
    {
      $self->{importedPriceTransactions} = $input->{'imported_price_transactions'};
    }

    my $line = "";
    my $tempCounter = 0;
    my $parseResult = -1;

    # recognize the multiple security list entries for the import
    while ((defined $line) && ($tempCounter < $self->{pricesBatchSize}))
    {
      $line = $self->getNextLine(fileHandle => $fileHandle);
      if (defined $line)
      {
        if ($line =~ /!Type:Prices/)
        {
          $parseResult = $self->parsePrice(fileHandle => $fileHandle);
          if (not $parseResult)
          {
            $self->{error} = 1;
            $self->{errorString} = $errStr . "command = '$command'<br>Error parsing QIF Price entry! line = " . $self->{currentLine} . "<br>\n" . $self->errorMessage;
            unlink $QIFFileName;
            return undef;
          }
          # record processing of this record
          $self->{importedPriceTransactions}++;
          $tempCounter++;
        }
      }
    }

    $doc->setTitle("QIF Import Prices");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Import Prices</h1><br>\n");
    # everything appears to have worked and parsed successfully
    $doc->print("<form method=\"post\" name=\"qif_import_prices\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    if (not defined $line)
    {
      $doc->print("<input type=hidden name=\"command\" value=\"import_complete\">\n");
    }
    else
    {
      $doc->print("<input type=hidden name=\"command\" value=\"import_prices\">\n");
      $doc->print("<input type=hidden name=\"current_line\" value=\"" . $self->{currentLine} . "\">\n");
      $doc->print("<input type=hidden name=\"imported_price_transactions\" value=\"" . $self->{importedPriceTransactions} . "\">\n");
    }
    $doc->print("<input type=hidden name=\"limit_range\" value=\"" . $self->{limitRange} . "\">\n");
    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_import_prices.submit();\n";
    $doc->setOnload(code => $onLoadCode);
    $doc->setFocus("body");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font><br><br>\n");
    $doc->print("<input type=submit name=\"submit_parse\" value=\"Continue...\">\n");
    $doc->print("</form>\n");
    $doc->print("<br><br>\n");
    if (not defined $line)
    {
      $doc->print("Total");
    }
    else
    {
      $doc->print("Current");
    }
    $doc->print(" prices imported: " . $self->{importedPriceTransactions} . "<br>\n");
    $doc->print("</center>\n");
  }
  elsif ($command eq "import_complete")
  {
    # get the filename of the QIF file to parse
    if ((not exists $input->{'qif_file_name'}) || ($input->{'qif_file_name'} eq ""))
    {
      $self->{error} = 1;
      $self->{errorString} = $errStr . "command = '$command'<br>The filename to parse QIF import data was not passed to the server!<br>\n" . $input->{'qif_file_name'};
      return undef;
    }

    my $QIFFileName = $input->{'qif_file_name'};
    # remove the QIF import file
    unlink $QIFFileName;

    $doc->setTitle("QIF Import Transactions");
    $doc->setFocus("body");
    $display->displayMenu(doc=>$doc, style=>"banner",ledger=>$ledger);
    $doc->print("<center>\n");
    $doc->print("<h1>QIF Import Transactions - Successfully Completed</h1><br>\n");
    $doc->print("<p>All data parsed from the QIF import file has been inserted into the system successfully.  Returning to the file input screen.</p>\n");
    $doc->print("<font size=\"-1\">(If this form does not submit automatically, click the 'Continue...' button.)</font>\n");
    $doc->print("<form method=post name=\"qif_import_complete\" action=\"$siteDir/cgi-bin/index.cgi\">\n");
    $doc->print("$staticInfo");
    $doc->print("<input type=hidden name=\"command\" value=\"display\">\n");
    $doc->print("<input type=submit value=\"Continue...\">\n");
    $doc->print("<form>\n");
    $doc->print("</center>\n");

    # set the page to be submitted by JavaScript
    $doc->setFocus("javascript");
    my $onLoadCode = "document.qif_import_complete.submit();\n";
    $doc->setOnload(code => $onLoadCode);
  }

  return $doc;
}
#= NOTE:  All data fields are accessible by specifying the object
#         and pointing to the data member to be modified on the
#         left-hand side of the assignment.
#         Ex.  $obj->variable($newValue); or $value = $obj->variable;
#=cut
#=head2 Internal functions
#	All other methods are private to the object and are not
#	usable outside the internals of the object. These are
#	methods to help break down the different pieces of the QIF file
#
#	string getNextLine(fileHandle)
#		shifts the next available line off the QIFInputBuffer array
#		and calls readNextLine; if no values are in the array,
#		readNextLine is called again to attempt to fill the array
#		with the next line of data; if there are still no values
#		returned from readNextLine, the data file is assumed to be
#		exhausted
sub getNextLine
{
  my $self = shift;
  my %args = (fileHandle => undef, @_);
  my $fileHandle = undef;
  if (defined $args{fileHandle})
  {
    $fileHandle = $args{fileHandle};
  }
  else
  {
    return undef;
  }
  my $line = shift @{$self->{QIFInputBuffer}};
  my $tempLine = $self->readNextLine(fileHandle=>$fileHandle);
  if (not defined $line)
  {
    if (defined $tempLine)
    {
      $line = shift @{$self->{QIFInputBuffer}};
      $self->{currentLine}++;
      $self->readNextLine(fileHandle=>$fileHandle);
    }
  }
  else
  {
    $self->{currentLine}++;
  }

  return $line;
}

#void readNextLine(fileHandle)
#	reads the next line from the submitted QIF import file and
#	places the data into a FIFO
sub readNextLine
{
  my $self = shift;
  my %args = (fileHandle => undef, @_);
  my $fileHandle;
  if (defined $args{fileHandle})
  {
    $fileHandle = $args{fileHandle};
  }
  else
  {
    return undef;
  }

  my $line = <$fileHandle>;
  # 'chomp' the end-of-line markers
  chomp $line;
  if (defined $line)
  {
    push @{$self->{QIFInputBuffer}}, $line;
  }

  return $line;
}

#= putBackLine
#    summary: 'unshift's the given argument back into the FIFO
#    to be obtained by the next call to getNextLine
#
#=cut
sub putBackLine
{
  my $self = shift;
  my $line = shift;
  if (defined $line)
  {
    unshift @{$self->{QIFInputBuffer}}, $line;
  }
  $self->{currentLine}--;
}

#= name: parseClassList
#
#=cut
sub parseClassList
{
  my $self = shift;
  my %args = ( @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parseClassList(): ";
  my $line = "";
  my $tempCounter = 0;

  my $class_name = "";
  my $class_description = "";

  while ((defined $line) && $line !~ /^\!/ && ($tempCounter < $self->{classListBatchSize}))
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line =~ /^N(.+?)\r?\n?$/)
    {
      $class_name = $1;
      $class_name =~ s/'/\\'/g;
    }
    elsif ($line =~ /^D(.+?)\r?\n?$/)
    {
      $class_description = $1;
      $class_description =~ s/'/\\'/g;
    }
    elsif ($line =~ /\^/)
    {
      # check to make sure we at least have a name to use for inserting data into the database
      if ($class_name ne "")
      {
        # insert record into temporary table that stores a list of classes
        $self->{dbHandle}->write(sql => "INSERT INTO qif_class_tb (name, description) VALUES (?, ?)", plug => [ $class_name, $class_description ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error occurred inserting class into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
          return 0;
        }
        # successfully entered data into QIF temporary table; clear variables for next record input
        $class_name = "";
        $class_description = "";
        # record processing of this record
        $self->{processedClassRecords}++;
        $tempCounter++;
      }
    }
  }

  if (defined $line)
  {
    $self->putBackLine($line);
  }

  if ($tempCounter >= $self->{classListBatchSize})
  {
    return 2;
  }

  return 1;
}

#    name: parseCategoryList
sub parseCategoryList
{
  my $self = shift;
  my %args = ( @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parseCategoryList(): ";
  my $line = "";
  my $tempCounter = 0;

  my $category_name = "";
  my $category_description = "";
  my $category_tax_flag = "false";
  my $category_tax_num = "";
  my $category_income_expense_flag = "";

  while ((defined $line) && ($line !~ /^\!/) && ($tempCounter < $self->{categoryListBatchSize}))
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line =~ /^N(.+?)\r?\n?$/)
    {
      $category_name = $1;
      $category_name =~ s/'/\\'/g;
    }
    elsif ($line =~ /^D(.+?)\r?\n?$/)
    {
      $category_description = $1;
      $category_description =~ s/'/\\'/g;
    }
    elsif ($line =~ /^T/)
    {
      $category_tax_flag = "true";
    }
    elsif ($line =~ /^R(.+?)\r?\n?$/)
    {
      $category_tax_num = $1;
    }
    elsif ($line =~ /^I/)
    {
      $category_income_expense_flag = "true";
    }
    elsif ($line =~ /^E/)
    {
      $category_income_expense_flag = "false";
    }
    elsif ($line =~ /\^/)
    {
      # check to make sure we at least have a name and income/expense flag to use for inserting data into the database
      if ($category_name ne "" && $category_income_expense_flag ne "")
      {
        # insert record into temporary table that stores a list of categories
        $self->{dbHandle}->write(sql => "INSERT INTO qif_category_tb (name, description, tax_related_item, tax_ref_num, income_expense_flag) VALUES (?, ?, ?, ?, ?)", plug => [ $category_name, $category_description, $category_tax_flag, $category_tax_num, $category_income_expense_flag ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error occurred inserting category into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
          return 0;
        }
        # successfully entered data into QIF temporary table; clear variables for next record input
        $category_name = "";
        $category_description = "";
        $category_tax_flag = "false";
        $category_tax_num = "";
        $category_income_expense_flag = "";
        # record processing of this record
        $self->{processedCategoryRecords}++;
        $tempCounter++;
      }
    }
  }

  if (defined $line)
  {
    $self->putBackLine($line);
  }

  if ($tempCounter >= $self->{categoryListBatchSize})
  {
    return 2;
  }

  return 1;
}

#    name: parseSecurity
sub parseSecurity
{
  my $self = shift;
  my %args = ( @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parseSecurity(): ";
  my $line = "";

  my $security_name = "";
  my $security_symbol = "";
  my $security_type = "";
  my $security_goal = "";

  while ((defined $line) && $line !~ /\^/)
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line =~ /^N(.+?)\r?\n?$/)
    {
      $security_name = $1;
      $security_name =~ s/'/\\'/g;
    }
    elsif ($line =~ /^S(.+?)\r?\n?$/)
    {
      $security_symbol = $1;
      $security_symbol =~ s/'/\\'/g;
      $security_symbol = uc($security_symbol);
    }
    elsif ($line =~ /^T(.+?)\r?\n?$/)
    {
      $security_type = $1;
    }
    elsif ($line =~ /^G(.+?)\r?\n?$/)
    {
      $security_goal = $1;
    }
  }

  # check to make sure we at least have a name to use for inserting data into the database
  if ($security_name ne "")
  {
    # insert record into temporary table that stores a list of securities
    $self->{dbHandle}->write(sql => "INSERT INTO qif_security_tb (name, symbol, type, goal) VALUES (?, ?, ?, ?)", plug => [ $security_name, $security_symbol, $security_type, $security_goal ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error occurred inserting security into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
    $self->{processedSecurityRecords}++;
  }

  return 1;
}

#    name: parsePrice
sub parsePrice
{
  my $self = shift;
  my %args = ( @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parsePrice(): ";
  my $line = "";

  my $price_symbol = "";
  my $price_amount = "";
  my $price_date = "";

  while ((defined $line) && $line !~ /\^/)
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line !~ /\^/)
    {
      $line =~ /^"(.+)",/;
      $price_symbol = $1;
      $price_symbol =~ s/'/\\'/g;
      $price_symbol = uc($price_symbol);

      if ($line =~ /",(\d+),"/)
      {
        $price_amount = $1;
      }
      elsif ($line =~ /",(\d+)[^\d]+?(\d+)\/(\d+),"/)
      {
        my $coefficient = $1;
        my $numerand = $2;
        my $dividend = $3;

        $price_amount = $coefficient + ($numerand / $dividend); #/
      }
      elsif ($line =~ /",(\d+)\/(\d+),"/)
      {
        my $numerand = $1;
        my $dividend = $2;

        $price_amount = $numerand / $dividend; #/
      }
      elsif ($line =~ /",(\d+\.\d+),"/)
      {
        $price_amount = $1;
      }

      $line =~ /,"(..)\/(..)([\/'])(..)"\r?\n?$/;
      my $month = $1;
      my $day = $2;
      my $separator = $3;
      my $year = $4;

      $month =~ s/[^\d]//g;
      $day =~ s/[^\d]//g;
      $year =~ s/[^\d]//g;
      # translate year into YYYY value
      if ($separator eq "/")
      {
        $year += 1900;
      }
      elsif ($separator eq "'")
      {
        $year += 2000;
      }
      if ($month < 10)
      {
        $month = "0" . (int $month);
      }
      if ($day < 10)
      {
        $day = "0" . (int $day);
      }
      $price_date = $month . "/" . $day . "/" . $year;
    }
  }

  # check to make sure we have a name, amount, and date to use for inserting data into the database
  if ($price_symbol ne "" && $price_amount ne "" && $price_date ne "")
  {
    # for each price, determine the mapped sub-account for the ticker symbol (if possible)
    my $sth = $self->{dbHandle}->read(sql => "SELECT map_sub_account FROM qif_security_tb WHERE symbol = ?", plug => [ $price_symbol ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error retrieving sub-account for ticker ($price_symbol)!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
    my @info = $sth->fetchrow_array;
    if (defined @info && $info[0] =~ /^[0-9]+$/) # don't process price if no mapped sub-account
    {
      # delete all records for the current day for the current ticker from the stock_price_tb table
      my $sth2 = $self->{dbHandle}->write(sql => "DELETE FROM stock_price_tb WHERE date_of = ? AND sub_acct = ?", plug => [ $price_date, $info[0] ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error deleting stock price entries for sub-account (" . $info[0] . ") for ticker ($price_symbol) on date ($price_date)!<br>\n" . $self->{dbHandle}->errorMessage;
        $self->{dbHandle}->rollback;
        return 0;
      }
      # insert the new price record for the ticker
      $sth2 = $self->{dbHandle}->write(sql => "INSERT INTO stock_price_tb (date_of, price, sub_acct) VALUES (?, ?, ?, ?)", plug => [ $price_date, $price_amount, $info[0] ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error inserting stock price entry for sub-account (" . $info[0] . ") for ticker ($price_symbol) on date ($price_date) for amount ($price_amount)!<br>\n" . $self->{dbHandle}->errorMessage;
        $self->{dbHandle}->rollback;
        return 0;
      }
    }
    # everything went ok
    $self->{dbHandle}->commit;
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error committing prices to database!<br>\n" . $self->{dbHandle}->errorMessage;
      return 0;
    }
  }

  return 1;
}

#    name: parseAccountsList
sub parseAccountsList
{
  my $self = shift;
  my %args = ( @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parseAccountsList(): ";
  my $line = "";
  my $tempCounter = 0;

  my $account_name = "";
  my $account_type = "";
  my $sales_tax = "";
  my $description = "";
  my $credit_card_limit = "";
  my $balance = "";

  while ((defined $line) && $line !~ /^\!/ && ($tempCounter < $self->{accountListBatchSize}))
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line =~ /^N(.+?)\r?\n?$/)
    {
      $account_name = $1;
      $account_name =~ s/'/\\'/g;
    }
    elsif ($line =~ /^T(.+?)\r?\n?$/)
    {
      $account_type = $1;
    }
    elsif ($line =~ /^R(.+?)\r?\n?$/)
    {
      $sales_tax = $1;
    }
    elsif ($line =~ /^D(.+?)\r?\n?$/)
    {
      $description = $1;
      $description =~ s/'/\\'/g;
    }
    elsif ($line =~ /^L(.+?)\r?\n?$/)
    {
      $credit_card_limit = $1;
    }
    elsif ($line =~ /^\$(.+?)\r?\n?$/)
    {
      $balance = $1;
    }
    elsif ($line =~ /\^/)
    {
      # check to make sure we at least have a name and type to use for inserting data into the database
      if ($account_name ne "" && $account_type ne "")
      {
        # insert record into temporary table that stores a list of accounts
        $self->{dbHandle}->write(sql => "INSERT INTO qif_account_tb (name, type, sales_tax, description, credit_card_limit, balance) VALUES (?, ?, ?, ?, ?, ?)", plug => [ $account_name, $account_type, $sales_tax, $description, $credit_card_limit, $balance ]);
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error occurred inserting account into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
          return 0;
        }
        # successfully entered data into QIF temporary table; clear variables for next record input
        $account_name = "";
        $account_type = "";
        $sales_tax = "";
        $description = "";
        $credit_card_limit = "";
        $balance = "";
        # record processing of this record
        $self->{processedAccountListRecords}++;
        $tempCounter++;
      }
    }
  }

  if (defined $line)
  {
    $self->putBackLine($line);
  }

  if ($tempCounter >= $self->{accountListBatchSize})
  {
    return 2;
  }

  return 1;
}

#description: after coming to a !Account line in the file
#     this routing will parse out the account information
#required:  fileHandle => a handle to the input file
sub parseAccountRecord
{
  my $self = shift;
  my %args = ( @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parseAccountRecord(): ";
  my $line = "";

  my $account_name = "";
  my $description = "";
  my $account_type = "";
  my $sales_tax = "";

  while ((defined $line) && $line !~ /\^/)
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line =~ /^N(.+?)\r?\n?$/)
    {
      $account_name = $1;
      $account_name =~ s/'/\\'/g;
    }
    elsif ($line =~ /^D(.+?)\r?\n?$/)
    {
      $description = $1;
      $description =~ s/'/\\'/g;
    }
    elsif ($line =~ /^T(.+?)\r?\n?$/)
    {
      $account_type = $1;
    }
    elsif ($line =~ /^R(.+?)\r?\n?$/)
    {
      $sales_tax = $1;
    }
  }

  # check to make sure we at least have a name and type to use for inserting data into the database
  if ($account_name ne "" && $account_type ne "")
  {
    # compare the current account with accounts in the current account list; if it does not exist, enter it into the database
    # update the current account's used flag; based on whether the value (the number of updated rows) returned is greater than
    # 0 or not determines whether the account is already in the temporary table or not and if the account needs to be added
    my $updateResult = $self->{dbHandle}->write(sql => "UPDATE qif_account_tb SET used = '1' WHERE name = ?", plug => [ $account_name ]);
    if ($self->{dbHandle}->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} = $errorHeader . "Error occurred trying to update account in temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
      return undef;
    }
    if ($updateResult == 0)
    {
      # insert "unlisted" account into database QIF list table
      $self->{dbHandle}->write(sql => "INSERT INTO qif_account_tb (name, type, sales_tax, description, used) VALUES (?, ?, ?, ?, '1')", plug => [ $account_name, $account_type, $sales_tax, $description ]);
      if ($self->{dbHandle}->didErrorOccur())
      {
        $self->{error} = 1;
        $self->{errorString} = $errorHeader . "Error occurred inserting \"unlisted\" account into temporary QIF database table!<br>\n" . $self->{dbHandle}->errorMessage;
        return undef;
      }
    }
    # record processing of this record
    $self->{processedAccountRecords}++;
  }
  else
  {
    $account_name = "";
  }

  # return the account name for the following list of transactions
  return $account_name;
}

#=scalar parseInvestmentTransaction(importBuys)
#   takes: optional command-line argument (importBuy) to tell the parse routine whether to search for buy, buyx, or shrsin
#          events or everything else
# returns: array with two elements; the first element contains the return code to allow the calling routine to determine
#          the outcome of the parse; the second element contains a reference to a hash with all the parsed values; there is
#          no differentiation between a blank value ("") that was parsed and a value that is blank ("") because it did not
#          exist in the file record
#          the return codes are as follows: 0 - error occurred parsing the file or inserting record into database
#                                           1 - parse successful and valid record is being returned
#                                           2 - parse successful but no valid records or no records at all were found
sub parseInvestmentTransaction
{
  my $self = shift;
  my %args = ( importBuys => "0", @_ );
  my $fileHandle = $args{'fileHandle'};
  my $errorHeader = "QIFImport->parseInvestmentTransaction(): ";
  my @parseResult = ();
  my %recordValues = ();

  my $importBuys = $args{'importBuys'};
  my $line = "";

  my $skipTransaction = 0;
  my $foundValidRecord = 0;

  my $date = "";
  my $action = "";
  my $security = "";
  my $price = "";
  my $quantity = "";
  my $cleared = "";
  my $text_reminder = "";
  my $memo = "";
  my $commission = "";
  my $category = "";
  my $category_class = "";
  my $transfer_account = "";
  my $transfer_class = "";
  my $transaction_amount = "";
  my $transfer_amount = "";

  # set the default return value to be that no valid records were found ("2")
  $parseResult[0] = "2";

  while ((defined $line) && $line !~ /^\!/ && !$foundValidRecord)
  {
    $line = $self->getNextLine(fileHandle => $fileHandle);

    if ($line =~ /^D(\d+)\/([^\d]*?\d+)([\/'])([^\d]*?\d+)\r?\n?$/)
    {
      my $month = $1;
      my $day = $2;
      my $year = $4;
      if ($3 eq "/")
      {
        $year += 1900;
      }
      elsif ($3 eq "'")
      {
        $year += 2000;
      }
      if ($month < 10)
      {
        $month = "0" . (int $month);
      }
      if ($day < 10)
      {
        $day = "0" . (int $day);
      }
      $date = $month . "/" . $day . "/" . $year;
    }
    elsif ($line =~ /^N(.+?)\r?\n?$/)
    {
      $action = $1;
    }
    elsif ($line =~ /^Y(.+?)\r?\n?$/)
    {
      $security = $1;
      $security =~ s/'/\\'/g;
    }
    elsif ($line =~ /^I(.+?)\r?\n?$/)
    {
      $price = $1;
      # remove all commas from numeric values;
      $price =~ s/,//g;
    }
    elsif ($line =~ /^Q(.+?)\r?\n?$/)
    {
      $quantity = $1;
      $quantity =~ s/,//g;
    }
    elsif ($line =~ /^C(.+?)\r?\n?$/)
    {
      $cleared = $1;
    }
    elsif ($line =~ /^P(.+?)\r?\n?$/)
    {
      $text_reminder = $1;
      $text_reminder =~ s/'/\\'/g;
    }
    elsif ($line =~ /^M(.+?)\r?\n?$/)
    {
      $memo = $1;
      $memo =~ s/'/\\'/g;
    }
    elsif ($line =~ /^O(.+?)\r?\n?$/)
    {
      $commission = $1;
      $commission =~ s/,//g;
    }
    elsif ($line =~ /^L(.+?)\r?\n?$/)
    {
      my $tempValue = $1;
      if ($action ne "")
      {
        if ($action eq "MiscIncX" || $action eq "MiscExpX")
        {
          # category/class|[transfer]/class
          $tempValue =~ /^(.+?)(\/(.+?))?\|\[(.+?)\](\/(.+?))?$/;
          $category = $1;
          $category_class = $3;
          $transfer_account = $4;
          $transfer_class = $6;
          $category =~ s/'/\\'/g;
          $category_class =~ s/'/\\'/g;
          $transfer_account =~ s/'/\\'/g;
          $transfer_class =~ s/'/\\'/g;
        }
        elsif ($action eq "MiscInc" || $action eq "MiscExp")
        {
          # category/class only
          $tempValue =~ /^(.+?)(\/(.+?))?$/;
          $category = $1;
          $category_class = $3;
          $category =~ s/'/\\'/g;
          $category_class =~ s/'/\\'/g;
        }
        else
        {
          # [transfer]/class only
          $tempValue =~ /^\[(.+?)\](\/(.+?))?$/;
          $transfer_account = $1;
          $transfer_class = $3;
          $transfer_account =~ s/'/\\'/g;
          $transfer_class =~ s/'/\\'/g;
        }
      }
      else
      {
        # default to assigning the transfer/class of the transaction since the majority of actions use this case
        $tempValue =~ /^\[(.+?)\](\/(.+?))?$/;
        $transfer_account = $1;
        $transfer_class = $3;
        $transfer_account =~ s/'/\\'/g;
        $transfer_class =~ s/'/\\'/g;
      }
    }
    elsif ($line =~ /^T(.+?)\r?\n?$/)
    {
      $transaction_amount = $1;
      $transaction_amount =~ s/,//g;
    }
    elsif ($line =~ /^\$(.+?)\r?\n?$/)
    {
      $transfer_amount = $1;
      $transfer_amount =~ s/,//g;
    }
    elsif ($line =~ /\^/)
    {
      # check to make sure we at least have a date and action to use for inserting data into the database
      if ($date ne "" && $action ne "")
      {
        # check to make sure we have valid data for the given action
        if ($action eq "Buy")
        {
          # minimum needed for Buy in addition to date
          if ($security eq "" || $transaction_amount eq "" || $price eq "" || $quantity eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "BuyX")
        {
          if ($security eq "" || $transaction_amount eq "" || $price eq "" || $quantity eq "" || $transfer_amount eq "" || $transfer_account eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "Sell")
        {
          if ($security eq "" || $transaction_amount eq "" || $price eq "" || $quantity eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "SellX")
        {
          if ($security eq "" || $transaction_amount eq "" || $price eq "" || $quantity eq "" || $transfer_amount eq "" || $transfer_account eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "ShrsIn")
        {
          if ($security eq "" || $quantity eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "ShrsOut")
        {
          if ($security eq "" || $quantity eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "XIn")
        {
          if ($transaction_amount eq "" || $transfer_account eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "XOut")
        {
          if ($transaction_amount eq "" || $transfer_account eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "RtrnCap")
        {
          if ($security eq "" || $transaction_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "RtrnCapX")
        {
          if ($security eq "" || $transaction_amount eq "" || $transfer_account eq "" || $transfer_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "StkSplit")
        {
          if ($security eq "" || $quantity eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "Reminder")
        {
          if ($text_reminder eq "" && $memo eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "Div")
        {
          if ($security eq "" || $transaction_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "DivX")
        {
          if ($security eq "" || $transaction_amount eq "" || $transfer_account eq "" || $transfer_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "IntInc")
        {
          if ($security eq "" || $transaction_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "IntIncX")
        {
          if ($security eq "" || $transaction_amount eq "" || $transfer_account eq "" || $transfer_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "MiscInc")
        {
          if ($transaction_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "MiscIncX")
        {
          if ($transaction_amount eq "" || $transfer_account eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "MiscExp")
        {
          if ($transaction_amount eq "")
          {
            $skipTransaction = 1;
          }
        }
        if ($action eq "MiscExpX")
        {
          if ($transaction_amount eq "" || $transfer_account eq "")
          {
            $skipTransaction = 1;
          }
        }
        # if the commission is negative, recalculate the price based on the total amount divided by the number of shares
        # and reset the commission to zero
        if ($commission ne "" && $commission < 0.0)
        {
          if ($transaction_amount ne "" && $quantity ne "" && $quantity != 0)
          {
            $price = $transaction_amount / $quantity; #/
            $commission = "0.00";
          }
          else
          {
            $skipTransaction = 1;
          }
        }
        # determine based on the importBuys variable whether to skip importing the record or not
        if ($importBuys eq "0")
        {
          if ($action eq "Buy" || $action eq "BuyX" || $action eq "ShrsIn")
          {
            $skipTransaction = 1;
          }
        }
        else
        {
          if ($action eq "Buy" || $action eq "BuyX" || $action eq "ShrsIn")
          {
            if ($skipTransaction ne "1")
            {
              $skipTransaction = 0;
            }
          }
          else
          {
            $skipTransaction = 1;
          }
        }
        if (not $skipTransaction)
        {
          # record processing of this record
          $self->{processedAccountInvstRecords}++;
          $foundValidRecord = 1;

          $parseResult[0] = 1;

          $recordValues{'date'} = $date;
          $recordValues{'action'} = $action;
          $recordValues{'security'} = $security;
          $recordValues{'price'} = $price;
          $recordValues{'quantity'} = $quantity;
          $recordValues{'cleared'} = $cleared;
          $recordValues{'text_reminder'} = $text_reminder;
          $recordValues{'memo'} = $memo;
          $recordValues{'commission'} = $commission;
          $recordValues{'category'} = $category;
          $recordValues{'category_class'} = $category_class;
          $recordValues{'transfer_account'} = $transfer_account;
          $recordValues{'transfer_class'} = $transfer_class;
          $recordValues{'transaction_amount'} = $transaction_amount;
          $recordValues{'transfer_amount'} = $transfer_amount;

          $parseResult[1] = \%recordValues;
        }
        else
        {
          if (($importBuys eq "0" && $action ne "Buy" && $action ne "BuyX" && $action ne "ShrsIn") || ($importBuys eq "1" && ($action eq "Buy" || $action eq "BuyX" || $action eq "ShrsIn")))
          {
            # record skipping of this record
            $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES ('Invalid $action event (line #" . $self->{currentLine} . "): D=$date, S=$security, P=$price, Q=$quantity, A=$transaction_amount, C=$commission, L=$category\/$category_class\|[$transfer_account]\/$transfer_class, \$=$transfer_amount, M=$text_reminder [$memo]')");
            if ($self->{dbHandle}->didErrorOccur())
            {
              $self->{error} = 1;
              $self->{errorString} = $errorHeader . "Error inserting log record for skipping investment transaction record!<br>\n" . $self->{dbHandle}->errorMessage;
              $parseResult[0] = 0;
              return @parseResult;
            }
          }
        }

        if (!$foundValidRecord)
        {
          # clear variables for next record input
          $date = "";
          $action = "";
          $security = "";
          $price = "";
          $quantity = "";
          $cleared = "";
          $text_reminder = "";
          $memo = "";
          $commission = "";
          $category = "";
          $category_class = "";
          $transfer_account = "";
          $transfer_class = "";
          $transaction_amount = "";
          $transfer_amount = "";

          $skipTransaction = 0;
        }
      }
      else
      {
        # record skipping of this record
        $self->{dbHandle}->write(sql => "INSERT INTO qif_log_tb (message) VALUES ('No date or action!  Invalid $action event (line #" . $self->{currentLine} . "): D=$date, S=$security, P=$price, Q=$quantity, A=$transaction_amount, C=$commission, L=$category\/$category_class\|[$transfer_account]\/$transfer_class, \$=$transfer_amount, M=$text_reminder [$memo]')");
        if ($self->{dbHandle}->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} = $errorHeader . "Error inserting log record for skipping invalid investment transaction record!<br>\n" . $self->{dbHandle}->errorMessage;
          $parseResult[0] = 0;
          return @parseResult;
        }
      }
    }
  }

  if (defined $line && $line !~ /\^/)
  {
    $self->putBackLine($line);
  }

  return @parseResult;
}

#=scalar seekFile(lines, fileHandle)
#	summary: this function seeks into the QIF import file specified the
#	number of lines specified.  Returns 1 for success, 0 for failure
#
#=cut
sub seekFile
{
  my $self = shift;
  my %args = (lines => 0, @_);
  my $fileHandle;
  if (defined $args{fileHandle} && length $args{fileHandle} > 0)
  {
    $fileHandle = $args{fileHandle};
  }
  else
  {
    return 0;
  }
  # seek back into the file currentLine lines
  for (my $lineCount=0; $lineCount < $args{'lines'}; $lineCount++)
  {
    <$fileHandle>;
  }
  $self->{currentLine} = $args{'lines'};
  return 1;
}

# insertCashTransfer
# requires: ledgerCode, date, subEntity, accountFrom, accountTo, and amount
# optional: subAccountTo, subAccountFrom
# summary: inserts a (general) journal entry for the amount based on the debit_increases flag for the transferred from account
# returns: 1 on success, 0 on failure
sub insertCashTransfer
{
  my $self = shift;
  my $errorHeader = "QIFImport->insertCashTransfer(): ";
  my %args = (ledgerCode => undef, date => undef, subEntity => "", accountFrom => undef, accountTo => undef, subAccountTo => -1, subAccountFrom => -1, amount => undef, @_);
  my $ledgerCode = $args{'ledgerCode'};
  my $date = $args{'date'};
  my $subEntity = $args{'subEntity'};
  my $accountFrom = $args{'accountFrom'};
  my $accountTo = $args{'accountTo'};
  my $subAccountTo = $args{'subAccountTo'};
  my $subAccountFrom = $args{'subAccountFrom'};
  my $amount = $args{'amount'};

  if ((not defined $args{'ledgerCode'}) || $args{'ledgerCode'} eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Ledger code value missing or invalid!  The ledger code is a required value.";
    return 0;
  }
  if ((not defined $args{'date'}) || $args{'date'} eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Date value missing or invalid!  The date is a required value.";
    return 0;
  }
  if ((not defined $args{'accountFrom'}) || $args{'accountFrom'} !~ /^[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "accountFrom value missing or invalid!  The account to transfer the amount from is required.";
    return 0;
  }
  if ((not defined $args{'accountTo'}) || $args{'accountTo'} !~ /^[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "accountTo value missing or invalid!  The account to transfer the amount to is required.";
    return 0;
  }
  if ((not defined $args{'amount'}) || $args{'amount'} !~ /^-?[0-9]+(\.[0-9]+)$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Amount value missing or invalid!  The amount is a required value.";
    return 0;
  }
  elsif ($args{'amount'} =~ /^-/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Amount cannot be negative!";
    return 0;
  }
  if ($subAccountTo !~ /^-?[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "subAccountTo is invalid ($subAccountTo)!";
    return 0;
  }
  if ($subAccountFrom !~ /^-?[0-9]+$/)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "subAccountFrom is invalid ($subAccountFrom)!";
    return 0;
  }

  # determine the account type for the accountFrom account
  my $sth = $self->{dbHandle}->read(sql => "SELECT type FROM account_tb WHERE number = ?", plug => [ $accountFrom ]);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error looking up type for accountFrom (" . $accountFrom . ")!<br>\n" . $self->{dbHandle}->errorMessage;
    return 0;
  }
  my @info = $sth->fetchrow_array;
  my $accountType;
  if (defined $info[0])
  {
    $accountType = $info[0];
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving type for accountFrom (" . $accountFrom . ")!";
    return 0;
  }
  # determine the debit_increases flag for the accountFrom account type
  $sth = $self->{dbHandle}->read(sql => "SELECT debit_increases FROM acct_type_tb WHERE num = ?", plug => [ $accountType ]);
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error looking up debit_increases flag for accountFrom (" . $accountFrom . ")!<br>\n" . $self->{dbHandle}->errorMessage;
    return 0;
  }
  my @info = $sth->fetchrow_array;
  my $debitIncreases;
  if (defined @info[0])
  {
    $debitIncreases = $info[0];
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error retrieving debit_increases for accountType (" . $accountType . ")!";
    return 0;
  }

  # build a new Transaction object
  my $transaction = Portal::Accounting::DB::JournalEntry->new( dbiw => $self->{dbHandle} );
  if ($transaction->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error instantiating Transaction object!<br>\n" . $transaction->errorMessage;
    return 0;
  }

  my @transRowObject = [];
  if ($debitIncreases)
  {
    $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $subEntity, accountNum => $accountFrom, subAccountNum => $subAccountFrom, dorc => "false", amount => "$amount");
    $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $subEntity, accountNum => $accountTo, subAccountNum => $subAccountTo, dorc => "true", amount => "$amount");
  }
  else
  {
    $transRowObject[0] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $subEntity, accountNum => $accountFrom, subAccountNum => $subAccountFrom, dorc => "true", amount => "$amount");
    $transRowObject[1] = Portal::Accounting::Objects::JournalEntryRow->new(rowDate => $date, subEntity => $subEntity, accountNum => $accountTo, subAccountNum => $subAccountTo, dorc => "false", amount => "$amount");
  }
  my $transobj = Portal::Accounting::Objects::JournalEntry->new(ledgerCode => $ledgerCode,
                                                                     journalCode => "G", # hardcoded general journal code
                                                                     transDate => $date,
                                                                     memo => "Transfer from $accountFrom to $accountTo",
                                                                     rowObjects => \@transRowObject);
  if ($transobj->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Failed to create new transaction object!<br>\n" . $transobj->errorMessage;
    return 0;
  }

  # insert the transaction into the database
  my $refNum = $transaction->insert(obj => $transobj, commit => 1);
  if ($refNum == -1 || $transaction->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} = $errorHeader . "Error inserting journal transaction for transfer!<br>\n" . $transaction->errorMessage;
    return 0;
  }

  return $refNum;
}

#=head1 HTMLObject importAccountTransactions(currentAccount, fileName)
#
#=cut
#getNextLine
#while !EOF and batchCounter <= batchSize #stop at EOF or when batch is full
  #if this line an !Option:AutoSwitch
    #auto++
    #if auto == 1
      #read until !Clear:AutoSwitch
  #elsif nextLIne starts with !Type
    #if Bank, LIab....
      #if currentAccount eq ""
        #currentAccount = lookupAccount(name => "unknown_account");
    #else
      #thisis an undesired section. How do we skip it?
      #read until another !
      #put line Back
  #elsif nextLine eq !Account
    #currentAccount = parseRecord.
  #else
    #process transactions?
    #putbackLine
    #if nextLine !~ /^\!/ #line starts with !
      #if currentAccount ne "" #everything is oK
        #TransactinObject parseAccountTransaction(getRecord)
        #insertObject(JournalEntry)
        #batchCounter++;
  #getNextLine
sub importAccountTransactions()
{
  my $self = shift;
  my %args = (fileName => undef, @_);
  my $errorStr = "QIFImport->importAccountTransactions(): ";
  my $autoSwitch = 0;
  my $batchSize = $self->{'accountTransactionBatchSize'};
  my $batchCounter = 0;
  my $command = "parse_account_transactions";
  my $currentAccount;
  my $ledger;
  my $nextLine = "";
  my $transObject;
  my $app = "Accounting";
  my $state = "QIFImport";
  my $staticInfo = "";
  my $debugCounter = 0;

  if (defined $args{'currentAccount'})
  {
    $currentAccount = $args{'currentAccount'};
  }
  else
  {
    $currentAccount = "";
  }
  if (not defined $self->{input}->{'ledger'})
  {
    $self->{error} = 1;
    $self->{errorString} .= $errorStr . "Could not get ledger.<br>\n";
    return undef;
  }
  else
  {
    $ledger = $self->{input}->{ledger};
  }

  $staticInfo = "<input type=\"hidden\" name=\"app\" value=\"$app\">\n";
  $staticInfo .= "<input type=\"hidden\" name=\"state\" value=\"$state\">\n";
  $staticInfo .= "<input type=\"hidden\" name=\"ledger\" value=\"$ledger\">\n";
#  $staticInfo .= "<input type=\"hidden\" name=\"ledgerCode\" value=\"$ledgerCode\">\n";
  if (exists $self->{input}->{'qif_file_name'} && $self->{input}->{'qif_file_name'} ne "")
  {
    $staticInfo .= "<input type=hidden name=\"qif_file_name\" value=\"" . $self->{input}->{'qif_file_name'} . "\">\n";
  }

  if (exists $self->{input}->{'current_line'} && $self->{input}->{'current_line'} =~ /^\d+$/)
  {
    $self->{currentLine} = $self->{input}->{'current_line'};
  }
  if (exists $self->{input}->{'auto_switch'} && $self->{input}->{'auto_switch'} =~ /^\d+$/)
  {
    $autoSwitch = $self->{input}->{'auto_switch'};
  }
  if (exists $self->{input}->{'imported_account_transactions'} && $self->{input}->{'imported_account_transactions'} =~ /^\d+$/)
  {
    $self->{importedAccountTransactions} = $self->{input}->{'imported_account_transactions'};
  }

  my $doc = HTMLObject::Normal->new();
  if (not defined $doc)
  {
    $self->setError(errorString => "Error creating HTMLObject document.<br>\n");
    return undef;
  }

  if ((not defined $args{fileName}) || $args{fileName} eq "")
  {
    $self->{error} = 1;
    $self->{errorString} = $errorStr . "command = '$command'<br>The filename to parse QIF import data was not passed in!<br>\n" . $args{fileName};
    return undef;
  }
  my $fileHandle = openFile($args{fileName});
  if (not defined $fileHandle)
  {
    $self->{error} = 1;
    $self->{errorString} = $errorStr . "command = '$command'<br>The file could not be opened!<br>\n" . $args{fileName};
    return undef;
  }

#jump to line in file
  if (!$self->seekFile(fileHandle=>$fileHandle, lines => $self->{currentLine}))
  {
    $self->{error} = 1;
    $self->{errorString} = $errorStr . "command = '$command'<br>The seek method could not seek into the file: " . $args{fileName} . "<br>\n";
    unlink($args{fileName});
    return undef;
  }
  # parse the data file, recognizing the different sections as they occur; call helper functions to read the data
  my @tempArray;

  # create the Database Transaction object once to make the processing go faster and more efficiently
  my $dbTransObject = Portal::Accounting::DB::JournalEntry->new(dbiw => $self->dbHandle);
  if ((not defined $dbTransObject) || $dbTransObject->didErrorOccur())
  {
    $self->setError(errorString => $self->{errorString} . "<br>Failed on creating JournalEntry Database object when sent to 'parseAccountTransaction' in subroutine '$command'.");
    return undef;
  }

  $nextLine = $self->getNextLine(fileHandle => $fileHandle);
  while (defined $nextLine && $batchCounter < $batchSize) #stop at EOF or when batch is full
  {
    if ($nextLine =~ /^\!Option:AutoSwitch/)
    {
      $autoSwitch++;
      if ($autoSwitch == 1)
      {
        while (defined $nextLine && $nextLine !~ /^\!Clear:AutoSwitch/)
        { #read until !Clear:AutoSwitch
          $nextLine = $self->getNextLine(fileHandle => $fileHandle);
        }
      }
    }
    elsif ($nextLine =~ /^\!Type/)
    {
      if ($nextLine =~ /^\!Type:(Bank|Cash|CCard|Oth.A|Oth.L|Bill|Tax|Invoice)/)
      {
        if ($currentAccount eq "") #current account is not set so set it to be ready for next section
        {
          $currentAccount = $self->lookupAccount(name => "unknown_account");
          if ($currentAccount eq "")
          {
            $self->{error} = 1;
            $self->{errorString} .= $errorStr . "Error. Unknown account lookup failed.<br>\n";
          }
        }
      }
      else
      { #thisis an undesired section. How do we skip it?
        $nextLine = $self->getNextLine(fileHandle => $fileHandle);
        while (defined $nextLine && $nextLine !~ /^\!/)
        { #not start of a new section so get next line
          $nextLine = $self->getNextLine(fileHandle => $fileHandle);
        }#when reaching new section, break out of loop
        if (defined $nextLine)
        {
          $self->putBackLine($nextLine);
        }
      }
    }
    elsif ($nextLine =~ /^\!Account/)
    {
      $currentAccount = $self->parseAccountRecord(fileHandle => $fileHandle);
    }
    else
    { #process transactions
      $self->putBackLine($nextLine); #to be ready for next action to getRecord
      if ($nextLine !~ /^\!/) #line starts with !
      {
        if ($currentAccount ne "") #everything is oK
        { #start processing the next lines as transactions
          @tempArray = $self->getRecord(fileHandle => $fileHandle);
          if (scalar @tempArray == 0)
          {
            $self->setError(errorString => $self->{errorString} . "<br>Failed on getRecord in subroutine '$command'.");
            return undef;
          }
          else
          { #record was successful. parse it into a JournalEntry
            $transObject = $self->parseAccountTransaction(lineArray => \@tempArray, currentAccount => $currentAccount);
            if (not defined $transObject)
            {
              $self->setError(errorString => $self->{errorString} . "<br>Failed on creatinng JournalEntry object when sent to 'parseAccountTransaction' in subroutine '$command' on line '" . $self->{currentLine} . "'.");
              return undef;
            }
            else
            {#send JournalEntry to DB::JournalEntry
              #!!!!!!!!
              my $refNum = $dbTransObject->insert(obj => $transObject, commit => 1);
              if ($refNum == -1)
              {
                $self->setError(errorString => $self->{errorString} . "<br>Failed to insert JournalEntry Database object into Database' in subroutine '$command'. line = '" . $self->{currentLine} . "'<br>\n" . $dbTransObject->errorMessage);
                return undef;
              }
            }
            $batchCounter++;
            $self->{importedAccountTransactions}++;
          }
        }
      }
    }#end else
    $nextLine = $self->getNextLine(fileHandle => $fileHandle);
  }#end while

  $doc->setTitle("QIF File Parse Account Transactions");
  $doc->setFocus("body");
  $doc->print("<center>\n");
  $doc->print("<h1>QIF File Parse</h1><br>\n");
    # everything appears to have worked and parsed successfully
  if (not defined $nextLine)
  {
    $doc->print("Total");
  }
  else
  {
    $doc->print("Current");
  }
  $doc->print(" account transactions processed: " . $self->{importedAccountTransactions} . "<br>\n");
  my $formInfo;
  if (not defined $nextLine)
  { #must have reached end of file
    $formInfo  = "<p>Finished importing Account Transactions.  Please click Continue if your browser does not autosubmit.</p>\n";
    $formInfo .= "<input type=hidden name=\"command\" value=\"import_buy_investment_transactions\">\n";
  }
  else
  {
    $formInfo  = "<p>Not finished with the Transactions.  Please click Continue if your browser does not autosubmit.</p>\n";
    $formInfo .= "<input type=hidden name=\"command\" value=\"import_account_transactions\">\n";
    $formInfo .= "<input type=hidden name=\"current_line\" value=\"" . ($self->{currentLine} - 1) . "\">\n";
    $formInfo .= "<input type=hidden name=\"auto_switch\" value=\"$autoSwitch\">\n";
    $formInfo .= "<input type=hidden name=\"current_account\" value=\"$currentAccount\">\n";
    $formInfo .= "<input type=hidden name=\"imported_account_transactions\" value=\"" . $self->{importedAccountTransactions} . "\">\n";
  }
  $doc->print("<form method=\"post\" name=\"mainForm\" action=\"" . $self->{portalConfig}->{siteDir} . "/cgi-bin/index.cgi\">\n");
  $doc->print("$staticInfo");
  $doc->print($formInfo);
  $doc->print("<input type=hidden name=\"equity_account\" value=\"" . $self->{input}->{'equity_account'} . "\">\n");
  $doc->print("<input type=hidden name=\"dividend_revenue_account\" value=\"" . $self->{input}->{'dividend_revenue_account'} . "\">\n");
  $doc->print("<input type=hidden name=\"interest_revenue_account\" value=\"" . $self->{input}->{'interest_revenue_account'} . "\">\n");
  $doc->print("<input type=hidden name=\"other_revenue_account\" value=\"" . $self->{input}->{'other_revenue_account'} . "\">\n");
  $doc->print("<input type=hidden name=\"other_expense_account\" value=\"" . $self->{input}->{'other_expense_account'} . "\">\n");
  $doc->print("<input type=hidden name=\"import_sub_entity\" value=\"" . $self->{input}->{'import_sub_entity'} . "\">\n");
  $doc->print("<input type=hidden name=\"limit_range\" value=\"1\">\n");
  $doc->print("<input type=\"submit\" value=\"Continue...\">\n");
  $doc->print("</form>\n");
  $doc->print("</center>\n");
  $doc->setFocus("javascript");
  my $onLoadCode = "document.mainForm.submit();\n";
  $doc->setOnload(code => $onLoadCode);
  return $doc;
}

#array getRecord(fileHandle)
#		arguments:
#		returns:	array of lines from the file,
#	This method grabs from the next line in the file to a ^
#	returns an array of the lines in the record
sub getRecord
{
  my $self = shift;
  my %args = (@_);
  my $fileHandle = $args{fileHandle};
  my $errorHeader = "QIFImport->getRecord(): ";
  my $lineText = "";
  my @lineArray;
  my $index = 0;

  $lineText = $self->getNextLine(fileHandle => $fileHandle);
  while ((defined $lineText) && $lineText !~ /\^/)
  {
    $lineArray[$index++] = $lineText;
    $lineText = $self->getNextLine(fileHandle => $fileHandle);
  }

  return @lineArray;
}

#=head1 JournalEntry parseAccountTransaction(lineArray=>, currentAccount)
#	required:	array of lines containing transaction from QIF file
#	returns:	undef for error, otherwise a valid JournalEntry
#
#=cut
sub parseAccountTransaction
{
  #define vars
  my $self = shift;
  my %args = ( @_ );
  #instantiate needed objects
  my @lineArray = undef;
  my @rowArray = ();
  my $account = "";
  my $amount = "";
  my $cleared = '_';
  my $codeChar = "";
  my $currentAccount = "";
  my $date = "";
  my $debitIncreases = 0;
  my $dorc = "";
  my $errors = "";
  my $lineAccount = "";
  my $lineDorc = "";
  my $lineSubAccount = "";
  my $lineAmount = 0;
  my $lineMemo = "";
  my $memo = "";
  my $number = "-1";
  my $payee = "";
  my $row = 0;
  my $emptySplitCategory = 0;
  my $subAccount = "";
  my $subEntity = "";
  my $tempLine = "";

  if (not defined $args{lineArray})
  {
    $self->{error} = 1;
    $self->{errorString} = "<br>lineArray not defined!";
    return undef;
  }
  else
  {
    @lineArray = @{ $args{lineArray} };
  }
  if ((not defined $args{currentAccount}) || length $args{currentAccount} == 0)
  {
    $self->{error} = 1;
    $self->{errorString} = "<br>currentAccount not defined or empty!  '" . $args{currentAccount} . "'";
    return undef;
  }
  else
  {
    $currentAccount = $args{currentAccount};
  }
  my $mainAccount = $self->lookupAccount(name => $currentAccount); #change name into number
  if ((not defined $mainAccount) || $mainAccount eq "")
  {
    $self->{error} = 1;
    $self->{errorString} .= "Error! No account returned from lookupAccount when looking up mapped account number for currentAccount = '$currentAccount'.<br>\n";
    return undef;
  }
  $currentAccount = $mainAccount;

  #look for data in arrays
  for (my $counter = 0; $counter < scalar @lineArray; $counter++)
  {
    $tempLine = $lineArray[$counter];
    $tempLine =~ /^(.)(.+?)\r?\n?$/;
    $codeChar = $1;
    $tempLine = $2;
    if (length $tempLine > 0)
    {
      #date
      if ($codeChar eq 'D')
      {
        if ($tempLine =~ /^(\d+)\/([^\d]*?\d+)([\/'])([^\d]*?\d+)$/)
        {
          my $month = $1;
          my $day = $2;
          my $year = $4;
          if ($3 eq "/")
          {
            $year += 1900;
          }
          elsif ($3 eq "'")
          {
            $year += 2000;
          }
          if ($month < 10)
          {
            $month = "0" . (int $month);
          }
          if ($day < 10)
          {
            $day = "0" . (int $day);
          }
          $date = $month . "/" . $day . "/" . $year;
        }
        elsif ($tempLine =~ /^(\d+)\/(\d+)\/(\d+)$/)
        {
          $date = $tempLine;
        }
      }
      #payee
      elsif ($codeChar eq 'P')
      {
        $payee = $tempLine;
      }
      #check/dep #
      elsif ($codeChar eq 'N')
      {
        $number = $tempLine;
        if ($number !~ /^(\d+)$/)
        {
          $number = "-1";
        }
      }
      #memo
      elsif ($codeChar eq 'M')
      {
        $memo = chomp($tempLine);
      }
      #cleared
      elsif ($codeChar eq 'C')
      {
        if ($tempLine eq "R")
        {
          $cleared = 'r';
        }
        elsif ($tempLine eq "X")
        {
          $cleared = 'x';
        }
      }
      elsif ($codeChar eq 'L')
      {
        if ($emptySplitCategory ne "1")
        {
          $emptySplitCategory = "2";
        }
        ($account, $subEntity) = split /\//, $tempLine;
        #look for classes/sub entities code
        if (length $subEntity > 0)
        {
          $subEntity = lookupSubEntity(name => $subEntity);
          if (length $subEntity == 0)
          {#subEntity lookup failed
            $self->{errorstring} .= "JournalEntry has invalid sub entity\n";
          }
        }
        #account or category?
        if ($account =~ /^\[(.+?)\]$/) #name is surrounded by [ ]
        { #this is a tranfer
          $account = $1;
          if ($account eq "")
          {
            $self->{error} = 1;
            $self->{errorString} .= "Account transfer has NO account in it!<br>\n";
            return undef;
          }
          $account = $self->lookupAccount(name => $account); #change name into number
          if ((not defined $account) || $account eq "")
          {
            $self->{error} = 1;
            $self->{errorString} .= "Error! No account returned from lookupAccount.<br>\n";
            return undef;
          }
          $subAccount = -1;
        }
        else
        { #lookup categories to get Account, sub Account mappings to numbers
          ($account, $subAccount) = $self->lookupCategory(name => $account);
          if ($account eq "")
          {
            $self->{error} = 1;
            $self->setError(errorString => $self->{errorString} . "<br>Transaction has bad acccount mapping\n");
            return undef;
          }
          if ($subAccount eq "")
          {
            $self->{error} = 1;
            $self->setError(errorString => $self->{errorString} . "<br>Transaction has bad sub acccount mapping\n");
            return undef;
          }
        }
        if ($account eq $currentAccount)
        {#this is opening balance, change this account to equity account mapped by user
          $account = $self->{input}->{'equity_account'};
        }
        #will create row in rowArray later
      }
      elsif ($codeChar eq 'U')
      {#not sure what U is for, duplicate of T?
      }
      elsif ($codeChar eq 'T')   #amount
      {
        #debit or credit
        $amount = $tempLine;
        $amount =~ s/,//g;
        $amount =~ s/\$//g;
      }
      #address
      elsif ($codeChar eq 'A')
      { #skip this.
      }
      elsif ($codeChar eq 'S') #found split, loop over rows
      {
        #set the split flag
        $emptySplitCategory = 1;
        #split line into class/ and category
        ($lineAccount, $subEntity) = split /\//, $tempLine;
        #look for classes/sub entities code
        if (length $subEntity > 0)
        {
          $subEntity = lookupSubEntity(name => $subEntity);
          if (length $subEntity == 0)
          {#subEntity lookup failed
            $self->{errorstring} .= "Transaction has invalid sub entity\n";
          }
        }
        #account or category?
        if ($lineAccount =~ /^\[(.+?)\]$/) #name is surrounded by [ ]
        { #this is a transfer into an account
          $lineAccount = $1;
          if ($lineAccount eq "")
          {
            $self->{error} = 1;
            $self->{errorString} .= "Account transfer has NO account in it!<br>\n";
            return undef;
          }
          $lineAccount = $self->lookupAccount(name => $lineAccount); #change name into number
          if ((not defined $lineAccount) || $lineAccount eq "")
          {
            $self->{error} = 1;
            $self->{errorString} .= "Error! No account returned from lookupAccount.<br>\n";
            return undef;
          }
          $lineSubAccount = -1; #no sub account given
          if ($lineAccount eq $currentAccount)
          {#this is opening balance, change this account to equity account mapped by user
            $lineAccount = $self->{input}->{'equity_account'};
          }
        }
        else
        { #lookup categories to get Account, sub Account mappings to numbers
          ($lineAccount, $lineSubAccount) = $self->lookupCategory(name => $lineAccount);
          if ($lineAccount eq "")
          {
            $self->{error} = 1;
            $self->setError(errorString => $self->{errorString} . "<br>Transaction has bad acccount mapping\n");
            return undef;
          }
          if ($lineSubAccount eq "")
          {
            $self->{error} = 1;
            $self->setError(errorString => $self->{errorString} . "<br>Transaction has bad sub acccount mapping\n");
            return undef;
          }
        }
        #get next line from array which should be E, %, or $
        $tempLine = $lineArray[++$counter];
        $tempLine =~ /^(.)(.+?)\r?\n?$/;
        $codeChar = $1; #strip first digit out into code
        $tempLine = $2; #data goes back into here
        if ($codeChar eq 'E')
        {
          $lineMemo = $tempLine; #set the memo to the data
          $tempLine = $lineArray[++$counter]; #grab the next line ($ or %)
          $tempLine =~ /^(.)(.+?)\r?\n?$/;
          $codeChar = $1; #strip first digit out into code
          $tempLine = $2; #data goes back into here
        }
        if ($codeChar eq '$')
        { #set amount for this line
          $lineAmount = $tempLine;
          $lineAmount =~ s/,//g;
        }
        elsif ($codeChar eq '%')
        { # % means a percentage so multiply by total to get amount for this line
          #hopefully amount (U or T) has been reached before getting into this loop!!
          $lineAmount = eval($tempLine . " * " . $amount);
        }
        $debitIncreases = 0;
        if ($currentAccount =~ /^1/)
        { #  if mapped account first_digit is 1, debit increases true
          $debitIncreases = 0;
        }
        elsif ($currentAccount =~ /^([2-5])/)
        { #  if mapped account first_digit is 2,3,4, 5 debit increases false
          $debitIncreases = 1;
        }
        else
        {
          $self->{error} = 1;
          $self->{errorString} .= "Invalid First digit on account: $lineAccount<br>\n";
          return undef;
        }
        if ($lineAmount < 0)
        {#  Now if number is negative switch debitincreases for this row
          $lineAmount = sprintf("%.2f", abs($lineAmount));
          $debitIncreases = !$debitIncreases;
        }
        $lineDorc = ($debitIncreases?"true":"false");
        #create row Object
        $rowArray[$row] = Portal::Accounting::Objects::JournalEntryRow->new(
            rowDate => $date,
            memo => $lineMemo,
            subEntity => $subEntity,
            accountNum => $lineAccount,
            subAccountNum => $lineSubAccount,
            dorc => $lineDorc,
            amount => $lineAmount,
            cleared => $cleared
            );
        if (not defined $rowArray[$row])
        {
          $self->{error} = 1;
          $self->{errorString} .= "No row created from method JournalEntryRow->new()";
          return undef;
        }
        if ($rowArray[$row]->didErrorOccur())
        {
          $self->{error} = 1;
          $self->{errorString} .= $rowArray[$row]->errorMessage() . "<br><b>1</b><br>currentAccount = '$currentAccount'<br>lineAccount = '$lineAccount'<br>lineSubAccount = '$lineSubAccount'\n";
          return undef;
        }
        else
        {
          $row++;
        }
      }
    }#end if >0
  }#endfor
  #create the Row for the Account we are in bank/ liability etc.
    #  if mapped account first_digit is 1, debit increases true
  $debitIncreases = 0;
  if ($currentAccount =~ /^1/)
  {
    $debitIncreases = 1;
  }
  elsif ($currentAccount =~ /^([2-5])/)
  {  #  if mapped account first_digit is 2,3,4, 5 debit increases false
    $debitIncreases = 0;
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} .= "Invalid First digit on account: $currentAccount<br>\n";
    return undef;
  }
  if ($amount < 0)
  {#  Now if number is negative switch debit/credit for that row
    $amount = sprintf("%.2f", abs($amount));
    $debitIncreases = !$debitIncreases;
  }
  $dorc = ($debitIncreases?"true":"false");

  $rowArray[$row] = Portal::Accounting::Objects::JournalEntryRow->new(
            rowDate => $date,
            memo => $memo,
            subEntity => $self->{input}->{'import_sub_entity'},
            accountNum => $currentAccount,
            subAccountNum => $number,
            dorc => $dorc,
            amount => $amount,
            cleared => $cleared
            );
  if (not defined $rowArray[$row])
  {
    $self->{error} = 1;
    $self->{errorString} .= "No row created from method JournalEntryRow->new()";
    return undef;
  }
  if ($rowArray[$row]->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} .= $rowArray[$row]->errorMessage() . "<br><b>2</b><br>currentAccount = '$currentAccount'<br>lineAccount = '$lineAccount'<br>lineSubAccount = '$lineSubAccount'\n";
    return undef;
  }
  else
  {
    $row++;
  }
  if ($emptySplitCategory eq "2")
  {#create the L row
    #reverse dorc flag
    if ($dorc eq "false")
    {
      $dorc = "true";
    }
    elsif ($dorc eq "true")
    {
      $dorc = "false";
    }
    $rowArray[$row] = Portal::Accounting::Objects::JournalEntryRow->new(
            rowDate => $date,
            memo => $memo,
            subEntity => $subEntity,
            accountNum => $account,
            subAccountNum => $subAccount,
            dorc => $dorc,
            amount => $amount,
            cleared => $cleared
            );
    if (not defined $rowArray[$row])
    {
      $self->{error} = 1;
      $self->{errorString} .= "No row created from method JournalEntryRow->new()";
      return undef;
    }
    if ($rowArray[$row]->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} .= $rowArray[$row]->errorMessage();
      return undef;
    }
    else
    {
      $row++;
    }
  }
  elsif ($emptySplitCategory eq "0")
  {
    #reverse dorc flag
    if ($dorc eq "false")
    {
      $dorc = "true";
    }
    elsif ($dorc eq "true")
    {
      $dorc = "false";
    }
    $rowArray[$row] = Portal::Accounting::Objects::JournalEntryRow->new(
            rowDate => $date,
            memo => $memo,
            subEntity => $self->{input}->{'import_sub_entity'},
            accountNum => $self->{input}->{'equity_account'},
            subAccountNum => "-1",
            dorc => $dorc,
            amount => $amount,
            cleared => "_"
            );
    if (not defined $rowArray[$row])
    {
      $self->{error} = 1;
      $self->{errorString} .= "No row created from method JournalEntryRow->new()";
      return undef;
    }
    if ($rowArray[$row]->didErrorOccur())
    {
      $self->{error} = 1;
      $self->{errorString} .= $rowArray[$row]->errorMessage();
      return undef;
    }
    else
    {
      $row++;
    }
  }
  #instantiate a JournalEntry and pass in data
  my $transObject = Portal::Accounting::Objects::JournalEntry->new(
      ledgerCode => $self->{ledgerCode},
      journalCode => "G",
      transDate => $date,
      payeePayer => $payee,
      memo => $memo,
      rowObjects => \@rowArray
      );
  if (not defined $transObject)
  {
    $self->{error} = 1;
    $self->{errorString} .= "No Transaction object was defined: JournalEntry->new()";
    return undef;
  }
  if ($transObject->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} .= $transObject->errorMessage();
    return undef;
  }

  return $transObject;
}

#takes the argument from the S or L line and looks
#up the proper account or category mapping
# returns and array of account, subAccount numbers
#or undef
sub lookupCategory()
{
  my $self = shift;
  my %args = ( name => '', @_ );
  my $account = undef;    #account number
  my $name = '';          #original name and account name
  my $subName = '';       #sub account name if found
  my $subAccount = undef; #sub account number
  if (length $args{name} == 0)
  {
    return (undef, undef);
  }
  else
  {
    $name = $args{name};
  }
  #lookup account, subaccount from Category LIST.
  my $result = $self->{dbHandle}->read(sql => "SELECT map_account, map_sub_account FROM qif_category_tb WHERE name = '$name'");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} .= "Error! attempted to read from qif_category_tb.<br>\n";
    return undef;
  }
  my @info = $result->fetchrow_array;
  if (defined @info && defined $info[0] && defined $info[1])
  {
    $account = $info[0];
    $subAccount = $info[1];
  }
  else
  {
    $self->{errorString} = "Error! No mapping returned from qif_category_tb for '$name'<br>\n";
    $self->{error} = 1;
    return (undef, undef);
  }
  return ($account, $subAccount);
}

#takes tha argument from the S or L line and looks
#up the proper account mapping
# returns and account number or undef
sub lookupAccount()
{
  my $self = shift;
  my %args = (name => '', @_ );
  my $account = undef;
  my $name = '';

  if ($args{name} eq '')
  {
    return undef;
  }
  else
  {
    $name = $args{name};
  }

  #use database qif_mapping table to lookup account number for account name given
  my $result = $self->{dbHandle}->read(sql => "SELECT map_account FROM qif_account_tb WHERE name = '$name'");
  if ($self->{dbHandle}->didErrorOccur())
  {
    $self->{error} = 1;
    $self->{errorString} .= "Error! attempt to read from qif_account_tb FAILED.<br>\n";
    return undef;
  }
  my @info = $result->fetchrow_array;
  if (defined @info && defined $info[0])
  {
    $account = $info[0];
  }
  else
  {
    $self->{error} = 1;
    $self->{errorString} = "Error! No mapping returned from qif_account_tb for '$name'<br>\n";
    return undef;
  }
  return $account;
}

#=FILE openFile(string)
#	returns a file handle to file whose name was passed in
#
#=cut
sub openFile
{
  my $fname = shift;

  if (length $fname == 0)
  {
    return undef;
  }
  if (! -e $fname)
  {
    return undef;
  }
  open(F, "+< $fname") or return undef;
  return *F;
}

1;
__END__

=head1 AUTHOR

Jason A. Pattie, JT Moree
PC & Web Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3)

=cut
